/** \file client.c
 * \brief TINE Client-side API Routines
 *      
 * Below are the principal client-side API routines which allow the user to obtain data from
 * elements in the control system.  Data acquistion can either be synchronous (e.g. ExecLink()) 
 * where execution is blocked until a call completes, or asynchronous (e.g. AttachLink()) where
 * the results of a call are given to a callback routine upon completion.  Data can also be 
 * 'monitored' asynchronously either at a supplied polling rate or upon data change.  
 *
 */
#ifndef NO_DOXYGEN
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include "srvcore.h"
#include "swaplib.h"
#include "globals.h"
#include "dbtools.h"
#include "srvip.h"
#include "toolkit.h"
#include "ttycmd.h"
#if defined(IPXNW) || defined(IPXBSD)
#  include "srvipx.h"
#endif
#if defined(MAKE_TINE32_DLL) && !defined(NO_REMOTE_PIPE)
#  include "shmalloc.h"
#endif

void wcSingleCallback(int id,int cc);
void wcGroupCallback(int id,int cc);
void CloseLink(int i);
void doIPXConsumer(void);
int initIPClient(unsigned short port);
void doUDPConsumer(void);
void DataDump(char *txt,char *data,int siz,short fmt,short dir);
int addDependentLink(int linkId);

short lastLnkErrSrc = 0;
int ExecLinkSemaphore = 0;
int AttachLinkSemaphore = 0;
int recvNetGlobalSemaphore = 0;
short lastLnkErr = 0;
int redirectedEntrySlot = -1;
int allowCommonLinks = TRUE;
int gLastGlobalDataSize = 0;
int useMultiThreadedCycler = FALSE;
int useConnectedSockets = FALSE;
TINE_EXPORT char gLastStatusString[STATUS_SIZE];

TINE_EXPORT void UseMultiThreadedCycler(int value)
{
  useMultiThreadedCycler = value;
}
int gUseSynchronousNotification = FALSE;
TINE_EXPORT void UseSynchronousNotification(int value)
{
  gUseSynchronousNotification = value;
}

/**
 * \internal
 * \brief Number of entries in conTbl
 * 
 * \see conTbl
 * \note Incremented only in IncrementConnections() 
 */
int nConnectionTableEntries = 0;
/**
 * \internal
 * \brief Number of entries in wcTable
 * 
 * \see wcTable
 * \note Incremented only in getWildCardTableId() 
 */
int nWildCardTableEntries = 0;
UINT32 nConnectionTimeouts = 0,nArrivals = 0;
int SystemLocalDataTimeSynchronization = FALSE;
int SystemSupressHeartbeatNotification = TRUE;
char gFecAddrBuffer[30],fecprot[4];

/**
 * \internal
 * \brief Client connection table
 */
ConnTableEntry **conTbl = NULL;
          
GroupTableEntry *grpTable = NULL;

wildCardTableEntry **wcTable = NULL;

extern ConnTableEntry connNameServer;
#ifndef clnWorkArea
/**
 * \internal 
 * \brief Buffer for incoming and outgoing data.
 * \todo Overwritten with a macro (DOS  hack) 
 */
BYTE *clnWorkArea = NULL;
/**
 * \internal
 * \brief Size of clnWorkArea in bytes.
 * \todo Overwritten with a macro (DOS hack)
 */
UINT32 clnWorkAreaSize = MIN_WORK_SIZE;
          
#endif
BYTE *clientbuf;

char *modetoa[] = { "CANCEL","SINGLE","REFRESH","POLL","REGISTER","RECEIVE","EVENT","INVALID" };

GlobalListStruct **glbTbl;
int GlobalHeartbeat = GLOBALHEARTBEAT;
/* 
 * General Purpose Routines:
 */
char *gStatusSourceString[2] = { "LCL", "RMT" };
void GetErrorString(int cc,char *str)
{
  if (cc < numErr) strncpy(str,erlst[cc],32);
  else sprintf(str,"error %d",cc);
}
TINE_EXPORT int GetConnectionTable(ConTblInfo *tbl,int *tblSize)
{
  int k, n = 0;
  if (tbl == NULL || tblSize == NULL) return argument_list_error;
  for (k=0; k<nConnectionTableEntries; k++)
  {
    if (conTbl[k]->mode == CM_CANCEL) continue;
    if (conTbl[k]->mode == CM_REGISTER) continue;
    if (n >= *tblSize) continue;
    memcpy(&tbl[n],conTbl[k],CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE+DEVICE_NAME_SIZE+PROPERTY_NAME_SIZE); 
    memset(tbl[n].eqm,0,EQM_NAME_SIZE);
    strncpy(tbl[n].eqm,SrvTbl[conTbl[k]->srvIdx].EqmName,EQM_NAME_SHORTSIZE);
    tbl[n].mode = conTbl[k]->mode;
    tbl[n].access = conTbl[k]->access;
    tbl[n].formatIn = conTbl[k]->formatIn;
    tbl[n].formatOut = conTbl[k]->formatOut;
    tbl[n].sizeIn = conTbl[k]->sizeIn;
    tbl[n].sizeOut = conTbl[k]->sizeOut;
    memcpy(tbl[n].strTagIn,conTbl[k]->strTagIn,TAG_NAME_SIZE);
    memcpy(tbl[n].strTagOut,conTbl[k]->strTagOut,TAG_NAME_SIZE);
    tbl[n].pollRate = conTbl[k]->pollRate;
    tbl[n].heartbeat = conTbl[k]->heartbeat;
    tbl[n].starttime = conTbl[k]->starttime;
    tbl[n].dtimestamp = conTbl[k]->dtimestamp;
    tbl[n].dtimestampMSEC = conTbl[k]->dtimestampMSEC;
    tbl[n].counter = conTbl[k]->counter;
    tbl[n].linkStatus = conTbl[k]->linkStatus;
    n++;
  }
  *tblSize = n;
  return 0;
}
int outputConnectionList(void)
{
  int k,baseMode;
  char Buffer[256],linkstatus[5];
  ConnTableEntry *c;

  memset(Buffer,' ',80);
  sprintf(&Buffer[0]," LINK"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[24],"Contracts\n");
  ttyoutput(Buffer);

  for (k=0; k<nConnectionTableEntries; k++)
  {
    memset(Buffer,' ',80); c = conTbl[k];
    sprintf(&Buffer[0]," %s (%d)",FecTbl[c->fecIdx].Name,c->tineProtocol);
    Buffer[strlen(Buffer)] = ' ';
    baseMode = BASEMODE(c->mode);
    strcpy(linkstatus, baseMode < CM_SINGLE ? "-" : c->tocounter ? "DOWN" : "UP");
    sprintf(&Buffer[24],"[%d] %s %s %s %d msec (%s) - %s\n",
      k,c->expName,c->devName,c->prpName,(int)c->pollRate,modetoa[baseMode],linkstatus);
    ttyoutput(Buffer);
  }
  return 0;
}
int outputConnectionStats(int k)
{
  int baseMode;
  char Buffer[256],linkstatus[5];
  ConnTableEntry *c;

  if (k < 0 || k >= nConnectionTableEntries)
  {
    sprintf(Buffer,"connection %d is not available",k);
    ttyoutput(Buffer);
    sprintf(Buffer,"available connections are %d to %d",k,nConnectionTableEntries-1);
    ttyoutput(Buffer);
    return 0;
  }
  memset(Buffer,' ',80);
  sprintf(&Buffer[0]," Front End"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[24],"Contract");
  ttyoutput(Buffer);

  memset(Buffer,' ',80); c = conTbl[k];
  sprintf(&Buffer[0]," %s (%d)",FecTbl[c->fecIdx].Name,c->tineProtocol);
  Buffer[strlen(Buffer)] = ' ';
  baseMode = BASEMODE(c->mode);
  strcpy(linkstatus, baseMode < CM_SINGLE ? "-" : c->tocounter ? "DOWN" : "UP");
  sprintf(&Buffer[24],"[%d] %s %s %s %d msec (%s) - %s",
    k,c->expName,c->devName,c->prpName,(int)c->pollRate,modetoa[baseMode],linkstatus);
  ttyoutput(Buffer);

  ttyoutput("\n");
  sprintf(Buffer,"%15s%15s%15s%12s%8s%11s","Link Starttime","Last Arrived","Last Timestamp","Is Pending","Counter","Timeouts");
  ttyoutput(Buffer);
# ifdef OS_16BIT
  sprintf(Buffer,"%15ld%15ld%15ld%12s%8d%9d",
    c->starttime,c->time,c->dtimestamp,c->notifyPending ? "YES" : "NO",c->counter,c->tocounter);
# else
  sprintf(Buffer,"%15d%15d%15d%12s%8d%9d",
    c->starttime,c->time,c->dtimestamp,c->notifyPending ? "YES" : "NO",c->counter,c->tocounter);
# endif
  ttyoutput(Buffer);

  ttyoutput("\n");
  sprintf(Buffer,"%15s%15s%15s%12s%14s","Link Heartbeat","Array Size","Num Blocks","Is Grouped","Callback ID");
  ttyoutput(Buffer);
# ifdef OS_16BIT
  sprintf(Buffer,"%15d%15ld%15d%12s%12lu",
    (int)c->heartbeat,c->sizeOut,c->numblks,c->grouped ? "YES" : "NO",c->callbackID);
# else
  sprintf(Buffer,"%15d%15d%15d%12s%12u",
    (int)c->heartbeat,c->sizeOut,c->numblks,c->grouped ? "YES" : "NO",c->callbackID);
# endif
  ttyoutput(Buffer);

  return 0;
}

int dumpConnectionList(void)
{
  return outputConnectionList();
}

int dumpConnectionStats(int i)
{
  return outputConnectionStats(i);
}

int initClientTransport(int port)
{
  int cc = 0;
# if !defined(DOS)
  if ((cc=initControlStructs()) != 0) 
  { /* log the error by don't stop initializing */
    feclog("cannot initialize control structs : %s",erlst[cc]);
    cc = 0;
  }
# endif
  if ((cc=initIPClient((UINT16)port)) != 0 && cc != tcp_not_supported) return cc;
# if defined(IPXBSD) || defined(IPXNW)
  if ((cc=initIPXClient(port)) != 0 && cc != ipx_not_supported) return cc;
# endif
# if !defined(DOS)
  if ((cc=initClnWorkArea(clnWorkAreaSize)) != 0) return cc;
# endif
# if defined(MAKE_TINE32_DLL) && !defined(NO_REMOTE_PIPE)
  if (CreateShmWindow()) feclog("Cannot create shm window!");
# endif
  return cc;
}
/**
 * \internal
 * \brief Initializes the TINE client.
 *
 * initClient has an internal flag which ensures that it is
 * called only once.
 *
 * This function
 * <ul>
 * <li>Finds the user name and stores it in gUserName</li>
 * <li>Calls initFecTable()</li>
 * <li> calls initTransport() </li>
 * </ul>
 * \see gUserName
 * 
 */
int initClient(void)
{
  static int initialized = 0;
  int cc;
  char *ptr = NULL;

  if (initialized) return 0;
  if ((ptr=getenv("TINE_TRANSPORT")) != NULL)
  {
    if (!strnicmp(ptr,"TCP",3)) useConnectedSockets = TRUE;
  }
  InitSystemMutexSet();
  if (WaitForSystemMutex(hSystemClientMutex,-1) != 0) 
  {
    feclog("cannot initialize client resources : wait for client mutex failed!");
    return semaphore_error;
  }
  getSrvIPAddr();
  if (startuptime == 0) startuptime = time(NULL);
  whoami(gUserName);
  /* alloc room for the principal client connection table */
  if ((conTbl=(ConnTableEntry **)SystemCalloc(MaxNumConnections,sizeof(ConnTableEntry *))) == NULL)
    return out_of_client_memory;
# ifndef SMALL_TINE_LIB
  /* alloc room for the wildcard connection table */
  if ((wcTable=(wildCardTableEntry **)SystemCalloc(MaxNumConnections,sizeof(wildCardTableEntry *))) == NULL)
    return out_of_client_memory;
# endif
  if ((cc=initFecTable()) != 0) return cc; /* first call initializes the Fec Table */
  if ((cc=initClientTransport(0)) != 0) return cc;
  if (clnWorkArea == NULL)
  {
    feclog("Client Work Space not allocated !!");
    return un_allocated;
  }
  initialized = 1;
  ReleaseSystemMutex(hSystemClientMutex);
  return 0;
}
TINE_EXPORT void CloseClients(int closesockets)
{
  int n;

  /* For packing efficiency do this in two steps: */
  for (n=0; n<nConnectionTableEntries; n++) 
  {
    if (conTbl == NULL) break; /* how ? */
    if (conTbl[n] == NULL) continue; /* how ? */
    if (conTbl[n]->tineProtocol > 3) CloseLink(n);
  }
  for (n=0; n<nConnectionTableEntries; n++)
  {
    if (conTbl == NULL) break; /* how ? */
    CheckSubscriptions(FALSE);
    if (ReturnToClientCycle)
    {
      if (ReturnToClientCycle%2) hardDelay(20);
      ReturnToClientCycle = 0;
      n--;
    }
  }
  if (closesockets) ShutdownClients(TRUE);
}
GroupTableEntry *findGroup(size_t id)
{
  GroupTableEntry *gt;

  for (gt=grpTable; gt && gt->GrpID != id; gt=gt->next);

  if (!gt)
  {
    if ((gt=(GroupTableEntry *)SystemCalloc(1,sizeof(GroupTableEntry))) == NULL) return NULL;
    gt->GrpID = id;
    gt->next = grpTable;
    grpTable = gt;
  }
  return gt;
}
int isNameServerRequest = FALSE;
int DataTagging = FALSE;
int lastOutputDataLink = -1;
REDIRECTED_LINK *RedirectedLinkList = NULL;
REDIRECTED_LINK *appendRedirectedLinkList(char *con,char *srv,char *prp,char *dev,char *dstctx,char *dstsrv,char *dstdev,char *dstprp)
{
  REDIRECTED_LINK *r;
  char *bpstr, *mpstr;
  if ((r=(REDIRECTED_LINK *)SystemCalloc(1,sizeof(REDIRECTED_LINK))) == NULL) return NULL;
  strncpy(r->srcContext,con,CONTEXT_NAME_SIZE);
  strncpy(r->srcServerName,srv,EXPORT_NAME_SIZE);
  strncpy(r->srcProperty,prp,PROPERTY_NAME_SIZE);
  strncpy(r->srcDeviceName,dev,DEVICE_NAME_SIZE);
  strncpy(r->dstContext,dstctx,CONTEXT_NAME_SIZE);
  strncpy(r->dstServerName,dstsrv,EXPORT_NAME_SIZE);
  strncpy(r->dstDeviceName,dstdev,DEVICE_NAME_SIZE);
  if (strlen(dstprp) > 0) 
  {
    strncpy(r->dstProperty,dstprp,PROPERTY_NAME_SIZE);
    if (strlen(dstprp) <= PROPERTY_NAME_SHORTSIZE && 
        !IsMetaProperty(dstprp,NULL,NULL) && 
        IsMetaProperty(prp,&bpstr,&mpstr))
    {
      strcat(r->dstProperty,".");
      strncat(r->dstProperty,mpstr,15);
    }
  }
  else 
  {
    strncpy(r->dstProperty,prp,PROPERTY_NAME_SIZE);
  }
  r->nxt = RedirectedLinkList;
  if (RedirectedLinkList != NULL) RedirectedLinkList->prv = r;
  RedirectedLinkList = r;
  if (tineDebug)
  {
    dbglog("append /%s/%s/%s[%s] -> /%s/%s/%s[%s] to redirection table",
      con,srv,dev,prp,con,dstsrv,dev,r->dstProperty);
  }
  return r;
}

/**
 * \internal
 * \brief Removes a redirected link from a doubly linked list
 * \param r List Element
 * \param cc unused
 * \todo Remove unused parameter
 */
int removeRedirectedLink(REDIRECTED_LINK *r,int cc)
{
  if (tineDebug)
  {
    dbglog("removing /%s/%s/%s/%s from redirection table",
      r->dstContext,r->dstServerName,r->dstProperty,r->dstDeviceName);
  }
  if (r->prv != NULL) r->prv->nxt = r->nxt;
  else if (r == RedirectedLinkList) RedirectedLinkList = r->nxt;
  SystemFree(r);
  return 0;
}

/**
 * \internal
 * \brief Get data from redirected linked list
 *
 * \param con EqmContext
 * \param srv Server Name 
 * \param prp Property
 * \param dev Device name
 *
 * This function looks for an entry in the global Redirection table
 * whose corresponding fields match the four parameters.
 * If an entry is found, the parameters prp and srv are
 * overwritten with the entry's values.
 * 
 * \return -1 if an entry was found. 0 if no entry was found.
 * 
 * \todo String lengths are hardcoded
 * \todo const.
 * \see RedirectedLinkList
 * 
 */
int redirectLink(char *con,char *srv,char *prp,char *dev)
{
  REDIRECTED_LINK *r=NULL;

  if (srv == NULL || prp == NULL) return 0;
  for (r=RedirectedLinkList; r != NULL; r = r->nxt)
  {
    if (con != NULL && strnicmp(r->srcContext,con,CONTEXT_NAME_SIZE)) continue;
    if (strnicmp(r->srcServerName,srv,EXPORT_NAME_SIZE)) continue;
    if (strnicmp(r->srcProperty,prp,PROPERTY_NAME_SIZE)) continue;
    if (dev != NULL && strnicmp(r->srcDeviceName,dev,DEVICE_NAME_SIZE)) continue;
    strncpy(con,r->dstContext,CONTEXT_NAME_SIZE);
    strncpy(dev,r->dstDeviceName,DEVICE_NAME_SIZE);
    strncpy(srv,r->dstServerName,EXPORT_NAME_SIZE);
    strncpy(prp,r->dstProperty,PROPERTY_NAME_SIZE);
    if (tineDebug)
    {
      dbglog("redirecting /%s/%s/%s/%s -> /%s/%s/%s/%s from redirection table",
        r->srcContext,r->srcServerName,r->srcProperty,dev,
        r->dstContext,r->dstServerName,r->dstProperty,r->dstDeviceName);
    }
    return -1;
  }
  return 0;
}

void markTCPConnectionAsIdle(int slot)
{
  int i;
  if (conTbl[slot]->TCPsocket > 0)       /* mark the connection as idle ? */
  {
    for (i=0; i<nConnectionTableEntries; i++)    /* run thru the list again */
    {
      if (i != slot && conTbl[i]->TCPsocket == conTbl[slot]->TCPsocket) break;
    }
    if (i == nConnectionTableEntries)                    /* didn't find it */
    {                /* the last entry holding the socket is being re-used */
      AddIdleConnection(conTbl[slot]->TCPsocket,conTbl[slot]->fecIdx);
    }
  }
  return;
}

int ConTblRdrCmp(ConnTableEntry *cte,char *cxt,char *grp,char *dev,char *prp)
{
  REDIRECTED_LINK *r = cte->rdrLink;
  if (r == NULL) return -1;
  if (cte->mode == CM_CANCEL) return -1;
  if (cte->rdrID <= 0) return -1;
  if (strnicmp(r->srcContext,cxt,CONTEXT_NAME_SIZE)) return -1;
  if (strnicmp(r->srcServerName,grp,EXPORT_NAME_SIZE)) return -1;
  if (strnicmp(r->srcProperty,prp,PROPERTY_NAME_SIZE)) return -1;
  if (strnicmp(r->srcDeviceName,dev,DEVICE_NAME_SIZE)) return -1;
  return 0;
}
int assignConTblMemory(int i)
{
  if (i < 0) return invalid_index;
  if (i >= MaxNumConnections) return resources_exhausted;
  if ((conTbl[i]=(ConnTableEntry *)SystemCalloc(1,sizeof(ConnTableEntry))) == NULL) return out_of_client_memory;
  return 0;
}

void prepNameServerEntry(DTYPE *dout,DTYPE *din,char *devcxt,char *devgrp,char *devnam,char *devprp) 
{
  conTbl[0] = &connNameServer;
  conTbl[0]->tineProtocol = gTineServicesProtocol;
  conTbl[0]->inetProtocol = UDP;
  strcpy(conTbl[0]->devName,devnam);
  strcpy(conTbl[0]->prpName,devprp);
  if (dout != NULL) 
  {
   conTbl[0]->sizeOut = dout->dArrayLength;
   conTbl[0]->formatOut = dout->dFormat;
   conTbl[0]->dataOut = (char *)dout->data.vptr;
  } 
  else 
  {
   conTbl[0]->sizeOut = 0;
   conTbl[0]->formatOut = CF_NULL;
   conTbl[0]->dataOut = NULL;
  }
  if (din != NULL) 
  {
   conTbl[0]->sizeIn = din->dArrayLength;
   conTbl[0]->formatIn = din->dFormat;
   conTbl[0]->dataIn = (char *)din->data.vptr;
   conTbl[0]->bytesTotal = conTbl[0]->sizeIn*fmtsizeof(conTbl[0]->formatIn);
   conTbl[0]->datamtu = UDP_BUFFER_SIZE-SUBSCRIPTION_SIZE-HEADERSTRUCT_SIZE;
   if (conTbl[0]->bytesTotal < conTbl[0]->datamtu) 
   {
     conTbl[0]->numblksIn = 1;
     conTbl[0]->bytesRemaining = conTbl[0]->bytesTotal;
     conTbl[0]->nextDataSize = conTbl[0]->bytesTotal;
   } 
   else 
   { 
     conTbl[0]->bytesTotal = 0; 
   }
  } 
  else 
  {
   conTbl[0]->sizeIn = 0;
   conTbl[0]->formatIn = CF_NULL;
   conTbl[0]->dataIn = NULL;
  }
  conTbl[0]->notifyPending = 1;
  if (conTbl[0]->bufOut) { SystemFree(conTbl[0]->bufOut); conTbl[0]->bufOut = NULL; }
  if (conTbl[0]->bufOutBlks) { SystemFree(conTbl[0]->bufOutBlks); conTbl[0]->bufOutBlks = NULL; }
  if (conTbl[0]->bufCmp) { SystemFree(conTbl[0]->bufCmp); conTbl[0]->bufCmp = NULL; }
  if (conTbl[0]->dtf) { SystemFree(conTbl[0]->dtf); conTbl[0]->dtf = NULL; }
}

/**
 * \internal
 * \brief Gets or finds a free slot in the client connection table.
 *
 * Returns an existing link if the input parameters match an entry in the
 * table
 *
 * This function tries to find an inactive link in the connection
 * table which has the same context, device name, device group, property,
 * and input/ouput data formats.
 *
 * \param dout Pointer to the output data buffer
 * \param din Pointer to the input data buffer 
 * \param devcxt EqmContext name
 * \param devgrp Device group
 * \param devnam Device name
 * \param property Property
 * 
 * \todo Move static int initialized to initClient().
 * \todo nConnectionTableEntries: connect incrementation with memory allocation.
 * 
 */
int getConTblEntry(DTYPE *dout,DTYPE *din,char *devcxt,char *devgrp,char *devnam,char *devprp)
{
  static int initialized = 0;
  UINT32 doutsize,dinsize;
  int srvidx = -1,fecidx = -non_existent_fec,slot = 0,boundToId = 0;
  int cc,i,nb,fmtsizein,devnamlen,devnampad;
  struct structStruct *s;
  BYTE *dataPtr;
  char fecName[FEC_NAME_SIZE],eqmName[EQM_NAME_SIZE];
  char ldevgrp[EXPORT_NAME_SIZE+4],ldevprp[PROPERTY_NAME_SIZE+4];
  char ldevnam[DEVICE_NAME_MAXSIZE+4],ldevcxt[CONTEXT_NAME_SIZE];
  time_t t = time(NULL);
  UINT16 datamtu;

  if (tineDebug) dbglog("connect to %s %s %s",devgrp,devnam,devprp);
  if (!initialized)
  {
    initialized = 1;
    if ((cc=initClient()) != 0) return -cc;
  }
  if (NameServerLoaded && !strcmp(devgrp,CSHOSTS)) /* Name Server request */
  { /* NOTE: direct CSHOSTS requests do not handle input data sets (no swapping, etc) */
    prepNameServerEntry(dout,din,devcxt,devgrp,devnam,devprp);
    return 0;
  }  
  /* has this been redirected ? */
  ldevcxt[0] = 0;
  strncpy(ldevgrp,devgrp,EXPORT_NAME_SIZE);
  strncpy(ldevprp,devprp,PROPERTY_NAME_SIZE);
  strncpy(ldevnam,devnam,DEVICE_NAME_MAXSIZE);
  if (devcxt != NULL) strncpy(ldevcxt,devcxt,CONTEXT_NAME_SIZE);
  redirectLink(ldevcxt,ldevgrp,ldevprp,ldevnam);
  /* resolve FEC name and EQM name: */
  if ((fecidx=locateExport(ldevgrp,ldevcxt,&srvidx)) < 0) return -address_unknown;
  if (srvidx < 0) return -non_existent;
  strncpy(fecName,FecTbl[fecidx].Name,FEC_NAME_SIZE);
  strncpy(eqmName,SrvTbl[srvidx].EqmName,EQM_NAME_SHORTSIZE);
  
  if (WaitForSystemMutex(hLinkTblMutex,gSystemTick) != 0) return -semaphore_busy;
  /* check to see if in connection Table */
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (conTbl[i] == NULL) continue; /* should be impossible ! */
    if (i > 0 && slot == 0 &&
        conTbl[i]->heartbeat < SUB_PENDING &&
        conTbl[i]->needsNotification == FALSE &&
        (time_t)conTbl[i]->starttime < (t-2) &&
        conTbl[i]->mode == CM_CANCEL) slot = i;     /* slot is available */
    if (strnicmp(ldevnam,conTbl[i]->devName,DEVICE_NAME_SIZE)) 
    { /* not equal at normal length */
      if (conTbl[i]->devName[DEVICE_NAME_SIZE-1] == '&' && conTbl[i]->extStringspace != NULL)
      { /* are the long names equal ? */
        if (strnicmp(&ldevnam[DEVICE_NAME_SIZE-1],conTbl[i]->extStringspace,conTbl[i]->extStringsize))
        { /* not equal at long length -> keep looking */
          continue;
        }
        /* else they're equal -> so fall thru */
      }
      else 
      { /* connection entry is not a long name -> keep looking */
        continue;
      }
    }
    if (ldevcxt[0] != 0 && strnicmp(ldevcxt,conTbl[i]->context,CONTEXT_NAME_SIZE)) continue;
    if ((!strnicmp(fecName,FecTbl[conTbl[i]->fecIdx].Name,FEC_NAME_SIZE) &&
         !strncmp(eqmName,SrvTbl[conTbl[i]->srvIdx].EqmName,EQM_NAME_SHORTSIZE) &&
         !strnicmp(ldevprp,conTbl[i]->prpName,PROPERTY_NAME_SIZE)) ||
         !ConTblRdrCmp(conTbl[i],ldevcxt,devgrp,devnam,devprp))
    {
      if (conTbl[i]->boundToId > 0) continue; /* already bound to a parent link */
      if (dout && dout->dArrayLength > 0)
      {
        if (dout->dFormat != conTbl[i]->formatOut) continue;
        if (isTaggableFormat(dout->dFormat) && strncmp(dout->dTag,conTbl[i]->strTagOut,TAG_NAME_SIZE)) continue;
        if (dout->dFormat == CF_STRUCT && (s=getStructFromTag(dout->dTag)) != NULL)
        {
          if (dout->dArrayLength*s->rawlen != conTbl[i]->sizeOut) continue;
        }
        else
        {
          if (dout->dArrayLength != conTbl[i]->sizeOut) continue;
        }
      }
      else if (conTbl[i]->sizeOut) continue;
      if (din && din->dArrayLength > 0)
      {
        if (conTbl[i]->dataIn == NULL) continue;
        if (din->dFormat != conTbl[i]->formatIn) continue;
        if (isTaggableFormat(din->dFormat) && strncmp(din->dTag,conTbl[i]->strTagIn,TAG_NAME_SIZE)) continue;
        if (din->dFormat == CF_STRUCT && (s=getStructFromTag(din->dTag)) != NULL)
        {
          if (din->dArrayLength*s->rawlen != conTbl[i]->sizeIn) continue;
          AssignClnStructDataPtr(dataPtr,conTbl[i]->dataIn,RTOL);
          if (memcmp(din->data.vptr,dataPtr,(size_t)din->dArrayLength*s->size)) continue;
        }
        else
        {
          if (din->dArrayLength != conTbl[i]->sizeIn) continue;
          dinsize = conTbl[i]->sizeIn*fmtsizeof(conTbl[i]->formatIn);
          if (memcmp(din->data.vptr,conTbl[i]->dataIn,(size_t)dinsize)) continue;
        }
      }
      else if (conTbl[i]->sizeIn) continue;
      if (dout != NULL)
      {
        if (BASEMODE(conTbl[i]->mode) != CM_CANCEL)  
        { /* active link */
          if (tineDebug) dbglog("latch onto active %s link",modetoa[BASEMODE(conTbl[i]->mode)]);
          boundToId = i;
          lastOutputDataLink = i;
          if (allowCommonLinks) 
          { /* add to dependency list (below) */
            i = nConnectionTableEntries;
            continue;               
          }
          ReleaseSystemMutex(hLinkTblMutex);
          return -link_exists; /* redirect output data */
        }
        else                                            
        { /* not active link */
          if (dout->data.vptr != conTbl[i]->dataOut) 
          { /* different stored buffer */
            continue;
          }
        }
      }
      if (tineDebug) dbglog("re-use inactive link");
      conTbl[i]->starttime = (UINT32)t;
      conTbl[i]->dtimestamp = conTbl[i]->dtimestampMSEC = 0;
      conTbl[i]->mcastGrp = 0;
      conTbl[i]->rdrLink = NULL;
      conTbl[i]->rdrID = 0;
      conTbl[i]->callbackID = 0;
      conTbl[i]->linkStatus = notification_pending;
      conTbl[i]->hasNotifiedOnce = FALSE;
      conTbl[i]->cancelPending = FALSE;
      conTbl[i]->notifyPending = 0;
      conTbl[i]->needsNotification = FALSE;
      conTbl[i]->srvID = 0;
      conTbl[i]->blksin = 0;
      conTbl[i]->bytesin = 0;
      conTbl[i]->queueLen = 0;
      conTbl[i]->xferReason = CX_NULL;
      memset(conTbl[i]->field,0,FIELD_NAME_SIZE);
      if (conTbl[i]->bufOutBlks) memset(conTbl[i]->bufOutBlks,0,conTbl[i]->numblks);
      ReleaseSystemMutex(hLinkTblMutex);
      return i;  /* re-use inactive link */
    } /* if contract is in table */
  }  /* for all connection table entries */
  
  /* brace ourselves for a new entry: */
  if (redirectedEntrySlot >= 0 && redirectedEntrySlot < nConnectionTableEntries)
  {
    slot = redirectedEntrySlot;
    markTCPConnectionAsIdle(slot);
    conTbl[slot]->rdrLink = NULL;
  }
  else if (slot == 0)
  {
    slot = nConnectionTableEntries;
    if ((cc=assignConTblMemory(nConnectionTableEntries)) != 0) 
    {
      ReleaseSystemMutex(hLinkTblMutex);
      return -cc;
    }
  }
  else
  {
    markTCPConnectionAsIdle(slot);
    if (conTbl[slot]->allocated && conTbl[slot]->dataOut)
      SystemFree(conTbl[slot]->dataOut);
    if (conTbl[slot]->dataIn) SystemFree(conTbl[slot]->dataIn);
    if (conTbl[slot]->bufOut) SystemFree(conTbl[slot]->bufOut);
    if (conTbl[slot]->bufOutBlks) SystemFree(conTbl[slot]->bufOutBlks);
    if (conTbl[slot]->bufCmp) SystemFree(conTbl[slot]->bufCmp);
    if (conTbl[slot]->dtf) SystemFree(conTbl[slot]->dtf);
    if (conTbl[slot]->extStringspace) SystemFree(conTbl[slot]->extStringspace);
    memset(conTbl[slot],0,sizeof(ConnTableEntry));
  }
  
  /* assign data indices and other parameters in connection Table: */
  if (ldevcxt[0] != 0)
    strncpy(conTbl[slot]->context,(char *)ldevcxt,CONTEXT_NAME_SIZE);
  else conTbl[slot]->context[0] = 0;
  strncpy(conTbl[slot]->expName,(char *)ldevgrp,EXPORT_NAME_SIZE);
  strncpy(conTbl[slot]->devName,(char *)ldevnam,DEVICE_NAME_SIZE);
  if ((devnamlen=strlen(ldevnam)) > DEVICE_NAME_SIZE)
  { /* long device name */
    conTbl[slot]->devName[DEVICE_NAME_SIZE-1] = '&';
    devnamlen -= DEVICE_NAME_SIZE-1;
    devnampad = devnamlen % 2; /* make sure it's an even number of extra bytes */
    if ((conTbl[slot]->extStringspace=(char *)SystemCalloc(1,devnamlen+devnampad)) == NULL)
    {
      ReleaseSystemMutex(hLinkTblMutex);
      return -out_of_client_memory;
    }
    conTbl[slot]->extStringsize = devnamlen + devnampad;
    strncpy(conTbl[slot]->extStringspace,(char *)&ldevnam[DEVICE_NAME_SIZE-1],devnamlen);
    if (devnampad > 0) conTbl[slot]->extStringspace[devnamlen] = 0;
  }
  strncpy(conTbl[slot]->prpName,(char *)ldevprp,PROPERTY_NAME_SIZE);
  conTbl[slot]->formatIn = din ? din->dFormat : CF_NULL;
  conTbl[slot]->formatOut = dout ? dout->dFormat : CF_NULL;
  if (din && isTaggableFormat(din->dFormat))
    strncpy(conTbl[slot]->strTagIn,din->dTag,TAG_NAME_SIZE);
  if (dout && isTaggableFormat(dout->dFormat))
    strncpy(conTbl[slot]->strTagOut,dout->dTag,TAG_NAME_SIZE);
  conTbl[slot]->srvIdx = srvidx;
  conTbl[slot]->fecIdx = fecidx;
  conTbl[slot]->sizeIn  = din ? din->dArrayLength : 0;
  conTbl[slot]->sizeOut = dout ? dout->dArrayLength : 0;
  conTbl[slot]->tineProtocol = FecTbl[fecidx].TineProtocol;
  conTbl[slot]->mode = CM_REGISTER;
  conTbl[slot]->blknumIn  = 1;
  conTbl[slot]->numblksIn = 1;
  conTbl[slot]->nextDataPtr = 0;
  conTbl[slot]->blkidIn = 0;
  conTbl[slot]->datamtu = SUBMTU(FecTbl[fecidx].TransportProtocol,DEFAULT_TRANSPORT_PROTOCOL_LEVEL); /* input data mtu */
  conTbl[slot]->inetProtocol = (FecTbl[fecidx].TransportProtocol & IPX) ? IPX : UDP ;
  if ((time_t)conTbl[slot]->starttime == t) conTbl[slot]->starttime++;
  else conTbl[slot]->starttime = (UINT32)t;
  conTbl[slot]->dtimestamp = 0;
  conTbl[slot]->dtimestampMSEC = 0;
  conTbl[slot]->dataStatus = not_posted;
  conTbl[slot]->linkStatus = notification_pending;
  conTbl[slot]->hasNotifiedOnce = FALSE;
  conTbl[slot]->cancelPending = FALSE;
  conTbl[slot]->notifyPending = 0;
  conTbl[slot]->needsNotification = FALSE;
  conTbl[slot]->queueDepth = gClnRecvQueueDepth;
  conTbl[slot]->srvID = 0;
  conTbl[slot]->blksin = 0;
  conTbl[slot]->bytesin = 0;
  conTbl[slot]->queueLen = 0;
  conTbl[slot]->xferReason = CX_NULL;
  if (redirectedEntrySlot >= 0 && redirectedEntrySlot < nConnectionTableEntries)
  {
    redirectedEntrySlot = -1;
    ReleaseSystemMutex(hLinkTblMutex);
    return slot;  /* return here and don't mess with the memory ...*/
  }
  if (din && din->data.vptr != NULL)
  {
    if (din->dFormat == CF_STRUCT && (s=getStructFromTag(din->dTag)) != NULL)
    {
      conTbl[slot]->sizeIn = din->dArrayLength*s->rawlen;
    }
    fmtsizein = fmtsizeof(din->dFormat);
    if ((dinsize=conTbl[slot]->sizeIn*fmtsizein) == 0)
    {
      conTbl[slot]->dataIn = NULL;
    }
    else 
    {
      if ((conTbl[slot]->dataIn=(char *)SystemCalloc(1,(size_t)dinsize)) == NULL)
      {
        ReleaseSystemMutex(hLinkTblMutex);
        return -out_of_client_memory;
      }
      if (fmtsizein > 0) conTbl[slot]->datamtu -= conTbl[slot]->datamtu % fmtsizein;
      cpyClnStructData(conTbl[slot]->dataIn,din->data.vptr,(size_t)dinsize);

      conTbl[slot]->bytesTotal = dinsize;
      conTbl[slot]->numblksIn = (UINT16)(dinsize/conTbl[slot]->datamtu +
                                            (dinsize%conTbl[slot]->datamtu ? 1 : 0));
      conTbl[slot]->bytesRemaining = conTbl[slot]->bytesTotal;
      conTbl[slot]->nextDataSize = MIN(conTbl[slot]->datamtu,conTbl[slot]->bytesRemaining);
      if (conTbl[slot]->extStringsize > 0)
      { /* need to adjust the above since the extended string space will eat up the initial data payload space */
        conTbl[slot]->nextDataSize -= conTbl[slot]->extStringsize; 
      }
    }
  }
  else conTbl[slot]->dataIn = NULL;
  if (dout)
  {
    if (dout->dFormat == CF_STRUCT && (s=getStructFromTag(dout->dTag)) != NULL)
    {
      conTbl[slot]->sizeOut = dout->dArrayLength*s->rawlen;
      doutsize = dout->dArrayLength*s->size;
    }
    else
    {
      doutsize = conTbl[slot]->sizeOut*fmtsizeof(dout->dFormat);
      if (dout->dFormat == CF_SPECTRUM) doutsize += 96;
    }
    if (dout->data.vptr == NULL) /* unassigned pointer */
    {
      if ((conTbl[slot]->dataOut=(char *)SystemCalloc(1,(size_t)doutsize)) == NULL)
      {
        ReleaseSystemMutex(hLinkTblMutex);
        return -out_of_client_memory;
      }
      dout->data.vptr = conTbl[slot]->dataOut;
      conTbl[slot]->allocated = TRUE;
    }
    else conTbl[slot]->dataOut = (char *)dout->data.vptr;
    datamtu = PMTU(FecTbl[fecidx].TransportProtocol,6); /* output data mtu */
    conTbl[slot]->numblks = (UINT16)(doutsize/datamtu + (doutsize%datamtu ? 1 : 0));
    if (conTbl[slot]->numblks)
    {
      nb = conTbl[slot]->numblks; nb += nb % 8;
      conTbl[slot]->bufOut = (char *)SystemCalloc(1,(size_t)doutsize);
      conTbl[slot]->bufOutBlks = (BYTE *)SystemCalloc(nb,1);
    }
  }
  else conTbl[slot]->dataOut = NULL;

  /* better together with allocation */
  if (slot == nConnectionTableEntries) IncrementConnections(nConnectionTableEntries);
# ifndef SMALL_TINE_LIB
  if (boundToId > 0)
  {
    conTbl[slot]->boundToId = boundToId;
    if ((cc=addDependentLink(slot)) != 0) 
    {
      ReleaseSystemMutex(hLinkTblMutex);
      return -cc;
    }
  }
# endif /* SMALL_TINE_LIB */
  ReleaseSystemMutex(hLinkTblMutex);
  return slot; /* everythings OK, return list No. */
}

static int gSyncLnkCount = 0; /* no synchronous activity */
void SyncAsync(int id,int cc)
{
  gSyncLnkCount--;
}
#define tvDeltaMs(a,b) (((a).tv_sec-(b).tv_sec)*1000 + ((a).tv_usec-(b).tv_usec)/1000)
int WaitForSyncLink(int to)
{
  struct timeval tv0,tv;
  int td = 0;
  if (to < 0) return 0;
  gettimeofday(&tv0,(struct timezone *)NULL);
  tv = tv0;
  while (gSyncLnkCount>0 && (td=tvDeltaMs(tv,tv0)) < to)
  {
    clientCycle();
    gettimeofday(&tv,(struct timezone *)NULL);
  }
  if (tineDebug > 3) dbglog("synchronous link : %d msec",td);
  return gSyncLnkCount > 0 ? -1 : 0;
}
TINE_EXPORT int PutLnkErrorValue(int i, void *errValue)
{
  if (i >= nConnectionTableEntries || i < 0) return argument_list_error;
  if (conTbl[i] == NULL || conTbl[i]->mode == CM_CANCEL) return invalid_link;
  memcpy(conTbl[i]->errValue,errValue,64);
  conTbl[i]->useErrValue = TRUE;
  return 0;
}

int prepIncomingSubscription(int protocol,ProducerHeaderStruct *pPHdr,BYTE *pHdrstream)
{
  ProducerHeaderStruct5 PHdr5, *pPHdr5 = &PHdr5;
  switch (protocol)
  {
    case 5:
#     ifdef INTEL
      memcpy(pPHdr5,pHdrstream,sizeof(ProducerHeaderStruct5));
#     else
      memset(pPHdr5,0,sizeof(ProducerHeaderStruct5));
      _StructSwap(ProducerHeaderStruct5Struct,pHdrstream,(unsigned char *)pPHdr5,RTOL);
#     endif
      pPHdr->msgsize = pPHdr5->msgsize;
      pPHdr->SubscriptionID = pPHdr5->SubscriptionID;
      pPHdr->CompletionCode = pPHdr5->CompletionCode;
      pPHdr->stscode = pPHdr5->CompletionCode;
      pPHdr->numblks = pPHdr5->numblks;
      pPHdr->blknum = pPHdr5->blknum;
      pPHdr->mtu = pPHdr5->mtu;
      pPHdr->timestampUSec = pPHdr5->timestampMSEC * 1000;
      pPHdr->EqmFormat = pPHdr5->EqmFormat;
      pPHdr->counter = pPHdr5->counter;
      pPHdr->timestamp = pPHdr5->timestamp;
      pPHdr->ClnStarttime = pPHdr5->ClnStarttime;
      pPHdr->stssize = pPHdr->CompletionCode == 0 ? 0 : STATUS_SHORTSIZE;
      pPHdr->SrvProtocol = 5;
      break;
    case 6:
      _StructSwap(ProducerHeaderStructStruct,pHdrstream,(unsigned char *)pPHdr,RTOL);
      break;
    default:
       return illegal_protocol;
  }
  return 0;
}

int prepOutgoingSubscription(int protocol,ConnTableEntry *c,char *eqm,int datasize,int subid,BYTE *pSubstream)
{
  SUBSCRIPTION sub, *pSub = &sub;
  SUBSCRIPTION5 sub5, *pSub5 = &sub5;
  
  switch (protocol)
  {
    case 5:
      memset(pSub5,0,sizeof(SUBSCRIPTION5));
      pSub5->msgsize = SUBSCRIPTION5_SIZE + datasize;
      pSub5->mtu = c->datamtu;
      pSub5->numblks = (BYTE)c->numblksIn;
      pSub5->blknum = (BYTE)c->blknumIn;
      pSub5->id = subid;
      pSub5->mode = (short)TARGETMODE(c->mode);
      pSub5->blkid = c->blkidIn;
      pSub5->PollingRate = c->pollRate;
      pSub5->starttime = c->starttime;
      strncpy(pSub5->contract.EqmName,eqm,EQM_NAME_SIZE);
      strncpy(pSub5->contract.EqmProperty,c->prpName,PROPERTY_NAME_SHORTSIZE);
      strncpy(pSub5->contract.EqmDeviceName,c->devName,DEVICE_NAME_SHORTSIZE);
      if (strlen(c->devName) > DEVICE_NAME_SHORTSIZE && strlen(c->prpName) <= DEVICE_NAME_SHORTSIZE) 
      {
       pSub5->contract.EqmDeviceName[DEVICE_NAME_SHORTSIZE-1] = '&';
       pSub5->contract.EqmProperty[DEVICE_NAME_SHORTSIZE] = 0;
       strncpy(&pSub5->contract.EqmProperty[DEVICE_NAME_SHORTSIZE+1],&c->devName[DEVICE_NAME_SHORTSIZE-1],DEVICE_NAME_SHORTSIZE-1);
      }
      pSub5->contract.EqmSizeOut = c->sizeOut;
      pSub5->contract.EqmSizeIn = c->sizeIn;
      pSub5->contract.EqmAccess = (BYTE)c->access;
      pSub5->contract.hEqmName = 0;
      pSub5->contract.EqmFormatOut = (BYTE)(BFMT(c->formatOut));
      pSub5->contract.EqmFormatIn = (BYTE)(BFMT(c->formatIn));
      strncpy(pSub5->contract.strTagIn,c->strTagIn,TAG_NAME_SHORTSIZE);
      strncpy(pSub5->contract.strTagOut,c->strTagOut,TAG_NAME_SHORTSIZE);
#     ifdef INTEL
      memcpy(pSubstream,pSub5,sizeof(SUBSCRIPTION5));
#     else
      memset(pSub5,0,sizeof(SUBSCRIPTION5));
      _StructSwap(SUBSCRIPTION5Struct,pSubstream,(unsigned char *)pSub5,LTOR);
#     endif
      break;
    case 6:
      memset(pSub,0,sizeof(SUBSCRIPTION));
      pSub->msgsize = SUBSCRIPTION_SIZE + datasize;
      pSub->PollingRate = c->pollRate;
      pSub->id = subid;
      pSub->blknum = c->blknumIn;
      pSub->numblks = c->numblksIn;
      pSub->blkid = c->blkidIn;
      pSub->mtu = c->datamtu;
      pSub->mode = (short)TARGETMODE(c->mode);
      pSub->starttime = c->starttime;
      strncpy(pSub->contract.EqmName,eqm,EQM_NAME_SIZE);
      strncpy(pSub->contract.EqmProperty,c->prpName,PROPERTY_NAME_SIZE);
      strncpy(pSub->contract.EqmDeviceName,c->devName,DEVICE_NAME_SIZE);
      pSub->contract.EqmSizeOut = c->sizeOut;
      pSub->contract.EqmSizeIn = c->sizeIn;
      pSub->contract.EqmAccess = (BYTE)c->access;
      pSub->contract.hEqmName = 0;
      pSub->contract.EqmFormatOut = (BYTE)(BFMT(c->formatOut));
      pSub->contract.EqmFormatIn = (BYTE)(BFMT(c->formatIn));
      strncpy(pSub->contract.strTagIn,c->strTagIn,TAG_NAME_SIZE);
      strncpy(pSub->contract.strTagOut,c->strTagOut,TAG_NAME_SIZE);
      pSub->extsize = c->extStringspace == NULL  ? 0 : c->extStringsize;
      pSub->msgsize += pSub->extsize;
#     ifdef INTEL
      if ((void *)pSubstream != (void *)pSub) memcpy(pSubstream,pSub,sizeof(SUBSCRIPTION));
#     else
      memset(pSubstream,0,sizeof(SUBSCRIPTION));
      _StructSwap(SUBSCRIPTIONStruct,pSubstream,(BYTE *)pSub,LTOR);
#     endif
      break;
    default:
       return illegal_protocol;
  }
  return 0;
}

static UINT32 assignHeartbeat(ConnTableEntry *c)
{ /* not exported : don't check null pointers */
  if (BASEMODE(c->mode) != (UINT16)CM_REFRESH) return (UINT32)(2+c->pollRate/1000);
  if (c->linkStatus == (short)link_not_open ||
      c->linkStatus == (short)notification_pending) return (UINT32)REPLY_PENDING;
  return (UINT32)(2+HEARTBEAT);
}

int prepSubRequest(BYTE *data,int isENS)
{
  HeaderStruct hTmp,*h = &hTmp;
  short fecIdx = -1,i,j,k,pktlen=0,dsize,dpadding,extsize,totalsize = HEADERSTRUCT_SIZE;
  short stepsize=0,prot=0;
  ConnTableEntry *c=NULL;
  time_t timestamp = time(NULL);
  char tin[TAG_NAME_SIZE+4],tout[TAG_NAME_SIZE+4];
  TCPBCKT *bckt;
  IDLE_CONNECTION *ic=NULL;
  int cc,sck=0,transport=0;
# if defined(UNIX) && !defined(NO_TINE_PIPE)
  char ipcpipe[64];
# endif

  if (data == NULL) return -1;
  AssignOutgoingHeader(h,data);
  for (i=0,j=0; i<nConnectionTableEntries; i++) 
  {
    if (i > 0 && isENS) break;
    if ((c=conTbl[i]) == NULL) continue;
    if (c->tineProtocol < 5) continue;
    if (c->boundToId > 0) continue;
    if (c->heartbeat < SUB_PENDING) continue;
    if (!j)
    {
      prot = c->tineProtocol;
      fecIdx = c->fecIdx;
      pktlen = FecTbl[fecIdx].TransportProtocol == IPX ? IPX_BUFFER_SIZE : UDP_BUFFER_SIZE;
      memset(data,0,pktlen);
      stepsize = SUB_STEPSIZE(prot);
    }
    else 
    {
      if (fecIdx != c->fecIdx) continue;
    }

    if (c->mode & CM_CONNECT && FecTbl[fecIdx].TransportProtocol != PIPE)
    {
      if (transport == UDP || transport == IPX) continue;
      if (c->TCPsocket == 0)
      { /* not yet connected */
        for (k=0; k<nConnectionTableEntries; k++)
          if (conTbl[k]->fecIdx == fecIdx &&
              conTbl[k]->TCPsocket != 0)
          { /* some one else is ! */
            c->TCPsocket = conTbl[k]->TCPsocket;
            break;
          }
      }
      if (c->TCPsocket == 0)
      { /* still not connected */
        if ((ic=getIdleConnection(c->fecIdx)) != NULL)
        { /* no longer idle */
          c->TCPsocket = ic->sck;
          RemoveIdleConnection(ic);
        }
      }
      if (c->TCPsocket == 0)
      { /* still not connected */
        if (connectTCP(gtTCPPort,i) == 0)
        { /* success */
          if ((bckt=getBucket(c->TCPsocket)) != NULL)
          {
            bckt->idx = fecIdx;
            FecTbl[fecIdx].TransportProtocol |= TCP;
            c->inetProtocol = TCP;
          }
          else CloseConnectionSocket(c->TCPsocket);
        }
      }
      sck = c->TCPsocket;
      transport = TCP;
    }
#   if defined(UNIX) && !defined(NO_TINE_PIPE)
    else if (FecTbl[fecIdx].TransportProtocol == PIPE) 
    {
      memcpy(&sck,FecTbl[fecIdx].IPXNetwork,4);
      if (sck == 0)
      {
        sprintf(ipcpipe,"%s%s-server.ipc",IPC_PATH,FecTbl[fecIdx].Name);
        if ((sck=CMDconnect(ipcpipe)) < 0) 
        {
          perror("IPC Socket");
          sck = 0;
        }
        else
        {
          memcpy(FecTbl[fecIdx].IPXNetwork,&sck,4);
          if (tineDebug) dbglog("FEC %s: attach pipe",FecTbl[fecIdx].Name);
          if ((bckt=getBucket(sck)) != NULL) bckt->idx = fecIdx;
          else close(sck);
        }
      }
      transport = PIPE;
    }
#   endif
    else
    {
      if (transport == TCP || transport == PIPE) continue;
      sck = 0;  /* let sendSubRequest figure it out */
      transport = FecTbl[fecIdx].TransportProtocol;
    }

    if (c->extStringspace != NULL && c->extStringsize > 0 && c->nextDataPtr == 0)
    { /* initial packet carries the extended string space (if any) */
      extsize = c->extStringsize;
      memcpy(&data[totalsize + stepsize],c->extStringspace,extsize);
    }
    else
    {
      extsize = 0;
    }

    if (c->sizeIn && c->dataIn)
    { /* input data size to be sent to the server */
      dsize = (short)c->nextDataSize;
      if ((totalsize + stepsize + dsize) > pktlen) break;
      dpadding = dsize % 2;                         /* make sure of even boundaries */
      PrepOutgoingData((char *)&data[totalsize+stepsize+extsize],&c->dataIn[(int)c->nextDataPtr],(size_t)dsize,c->formatIn);
    }
    else dpadding = dsize = 0;

    prepOutgoingSubscription(prot,c,SrvTbl[c->srvIdx].EqmName,dsize+dpadding,i,&data[totalsize]);

    j++;
    totalsize += stepsize + extsize + dsize + dpadding;
    if (tineDebug)
    {
      memset(tin,0,TAG_NAME_SIZE+4);
      if (c->strTagIn[0] != 0) strncpy(tin,c->strTagIn,TAG_NAME_SIZE);
      memset(tout,0,TAG_NAME_SIZE+4);
      if (c->strTagOut[0] != 0) strncpy(tout,c->strTagOut,TAG_NAME_SIZE);
      dbglog("subs %s %s %s (%d %d %s %d %d %s %d of %d) [%s] (%d ms): %s",
              c->expName,c->prpName,c->devName,
              (int)c->sizeOut,c->formatOut,tout,(int)c->sizeIn,c->formatIn,tin,
              c->blknumIn,c->numblksIn,
              FecTbl[fecIdx].Name,(int)c->pollRate,modetoa[BASEMODE(c->mode)]);
    }
    if (tineDebug > 3 && c->dataIn)
      DataDump("data sent: ",c->dataIn,(int)c->sizeIn,c->formatIn,LTOR);
    ReturnToClientCycle = 0;
    if (c->blknumIn == c->numblksIn)
    {
      c->blknumIn = 1;
      c->nextDataPtr = 0;
      c->blkidIn = (++(c->blkidIn))%1000;
      c->bytesRemaining = c->bytesTotal;
      c->heartbeat = assignHeartbeat(c);
      c->counter = -1;
      c->time = (UINT32)timestamp;
    }
    else if (c->dataIn && c->numblksIn > c->blknumIn)
    {
      ReturnToClientCycle = c->numblksIn - c->blknumIn;
      c->blknumIn++;
      c->nextDataPtr += c->nextDataSize;
      c->bytesRemaining = c->bytesTotal - c->nextDataPtr;
    }
    c->nextDataSize = MIN(c->datamtu,c->bytesRemaining);

    if ((totalsize + stepsize) > pktlen) break;  /* no more packing possible */
  }

  h->TotalSize = totalsize;
  memcpy(h->UserName,gUserName,USERNAME_SIZE);
  h->Protocol = prot;
  h->EqmSize = j;
  PrepOutgoingHeader(h,data);
  if (fecIdx >= 0) 
  {
    if (!strncmp(FecTbl[fecIdx].IP,gtNetSrvAddr,ADDR_SIZE)) sck = NETSRV_SOCKET;
    if ((cc=sendSubRequest(fecIdx,sck,transport,data)) != 0)
    { /* send unsuccessful ! */
      for (i=0; i<nConnectionTableEntries; i++) 
      { /* try, try, again ... */
        if ((c=conTbl[i]) == NULL) continue;
        if (c->fecIdx != fecIdx) continue;
        if (cc != not_implemented && transport != PIPE && BASEMODE(c->mode) > CM_SINGLE) 
          c->heartbeat = SUB_PENDING;  /* set it back */
        if (BASEMODE(c->mode) == CM_CANCEL) c->mode = CM_CANCEL; /* strip off CM_NETWORK */
      }
      fecIdx = -1;  /* get out of the loop ! */
    }
  }
  return fecIdx;
}
int sendSubRequest(int fecIdx,int sck,int transport,BYTE *data)
{
  int cc;
  UINT16 poff;
  static CLN cln;
# if defined(IPXNW) || defined(IPXBSD)
  static BYTE reg_sock[2];
# endif
# if defined(UNIX) && !defined(NO_TINE_PIPE)
  static UINT32 addr = 0;
# endif

  memset(&cln,0,sizeof(CLN));
  strcpy(cln.userName,FecTbl[fecIdx].Name);
  cln.transport = transport;
  poff = FecTbl[fecIdx].PortOffset;
  switch (transport)
  {
   case IPX:
     if (!IPXloaded) return ipx_not_supported;
#    if defined(IPXNW) || defined(IPXBSD)
     reg_sock[0] = REG_SOCKET + (BYTE)poff;
     reg_sock[1] = REG_SOCKET + (BYTE)poff;
     memcpy(&cln.IPXaddress,FecTbl[fecIdx].IPXNetwork,10);
     memcpy(cln.IPXaddress.socket,reg_sock,2);
     memcpy(cln.node,FecTbl[fecIdx].IPXImmediateAddress,6);
     cln.sck = IPXConsumerSocket;
#    endif
     break;
   case UDP:
     if (!TCPloaded) return tcp_not_supported;
     memset(&prod_name,0,sizeof(prod_name));
     if (sck == NETSRV_SOCKET)
     {
       prod_name.sin_addr.s_addr = tNetSrvAddr.s_addr;
       prod_name.sin_port = htons((UINT16)(gtNetSrvPort));
     }
     else
     {
       memcpy(&prod_name.sin_addr,&FecTbl[fecIdx].IPh_addr,sizeof(struct in_addr));
       prod_name.sin_port = htons((UINT16)(gtUDPPort + poff));
     }
     prod_name.sin_family=AF_INET;
     memcpy(&cln.IPaddress,&prod_name,sizeof(prod_name));
     cln.sck = ConsumerSocket;
     break;
   case TCP:
     if (!TCPloaded) return tcp_not_supported;
     memset(&prod_name,0,sizeof(prod_name));
     memcpy(&prod_name.sin_addr,&FecTbl[fecIdx].IPh_addr,sizeof(struct in_addr));
     prod_name.sin_family=AF_INET;
     prod_name.sin_port = htons((UINT16)(gtTCPPort + poff));
     memcpy(&cln.IPaddress,&prod_name,sizeof(prod_name));
     cln.sck = sck;
     break;
#   if defined(MAKE_TINE32_DLL) && !defined(NO_REMOTE_PIPE)
    case MMF:
     memcpy(&cln.sck,FecTbl[fecIdx].IPXImmediateAddress,4);
     break;
#  endif
#  if defined(UNIX) && !defined(NO_TINE_PIPE)
   case PIPE:
     if (addr == 0) addr = getmyipaddr();
     cln.sck = sck;
     memcpy(&cln.IPaddress.sin_addr,&addr,4);
     break;
#  endif
   case   0:
   default:
     if (tineDebug > 1) dbglog("FEC %s: acquire protocol",FecTbl[fecIdx].Name);
     return not_implemented;
   }
   if ((cc=sendToPeer(&cln,data,CLN_REQ_)) != 0)
   {
     feclog("Send Subscription: %s",erlst[cc]);
#    if defined(UNIX) && !defined(NO_TINE_PIPE)
     close(sck); freeBucket(sck);
     memset(FecTbl[fecIdx].IPXNetwork,0,4);
     FecTbl[fecIdx].TransportProtocol = 0;
     gUseLoopback = TRUE;
#    endif
     return cc;
   }
   return 0;
}
extern struct sockaddr_in IncomingIPfec;
extern IPXAddress IncomingIPXfec;
UINT32 IncomingNetworkProtocol;
int getIncomingFecIndex(void)
{
  int i,idx;
  UINT16 poff;

  if (IncomingNetworkProtocol == UDP || IncomingNetworkProtocol == TCP)
  {
    for (i=0; i<nConnectionTableEntries; i++) if (conTbl[i] != NULL && conTbl[i]->inetProtocol == UDP)
    { /* scan thru my connection table until the incoming address matches */
      idx = conTbl[i]->fecIdx;
      poff = FecTbl[idx].PortOffset;
      if (IncomingIPfec.sin_port == htons((UINT16)(gtUDPPort + poff))
          && !memcmp(&IncomingIPfec.sin_addr.s_addr,&FecTbl[idx].IPh_addr,4) )
        return idx;
    }
    return -1;
  }
# if defined(MAKE_TINE32_DLL) && !defined(NO_REMOTE_PIPE)
  if (IncomingNetworkProtocol == MMF)
  {
    for (i=0; i<nConnectionTableEntries; i++) if (conTbl[i] != NULL && conTbl[i]->inetProtocol == UDP)
    {
      idx = conTbl[i]->fecIdx;
      if (!memcmp(&IncomingLocalFec,FecTbl[idx].IPXImmediateAddress,4)) return idx;
    }
    return -1;
  }
# endif
  if (IncomingNetworkProtocol == IPX)
  {
    for (i=0; i<nConnectionTableEntries; i++) if (conTbl[i] != NULL && conTbl[i]->inetProtocol == IPX)
    {
      idx = conTbl[i]->fecIdx;
      if (!memcmp(FecTbl[idx].IPXNetwork,&IncomingIPXfec,10)) return idx;
    }
    return -1;
  }
  return -1;
}
extern int fecNameId;
#define TO_THRESHOLD 10
#define RETRY_THRESHOLD 3
void resetLinkTOcounters(int fecIdx)
{
  int i;
  ConnTableEntry *c;
  for (i=0; i<nConnectionTableEntries; i++)
    if ((c=conTbl[i]) != NULL && c->fecIdx == fecIdx) c->tocounter = 0;
}
int allLinksAboveThreshold(int fecIdx)
{
  int i;
  ConnTableEntry *c;
  for (i=0; i<nConnectionTableEntries; i++)
    if ((c=conTbl[i]) != NULL && c->fecIdx == fecIdx && 
         BASEMODE(c->mode) > CM_SINGLE &&
         c->tocounter < TO_THRESHOLD) return 0;
  return 1;
}
UINT32 nPartialTransfers = 0;
void CheckConnections(void)
{
  int i;
  time_t timestamp = time(NULL);
  static time_t lasttime = 0, lastaddrreq = 0;
  ConnTableEntry *c;
  IDLE_CONNECTION *ic=NULL;
# if defined(UNIX) && !defined(NO_TINE_PIPE)
  int pipe;
# endif

  for (i=0; i<nConnectionTableEntries; i++) 
  {
    if ((c=conTbl[i]) == NULL) continue;
    if (c->tineProtocol < 5) continue;
    if (BASEMODE(c->mode) && c->heartbeat &&  /* heartbeat = 0 -> REGISTER */
        timestamp > (time_t)(c->time + c->heartbeat))
    {
      if (tineDebug > 1)
#       ifdef OS_16BIT
        dbglog("%s %s %s : %ld > %ld + %d",
#       else
        dbglog("%s %s %s : %d > %d + %d",
#       endif
          c->expName,c->devName,c->prpName,timestamp,c->time,(int)c->heartbeat);
      lastLnkErrSrc = LNK_ERROR_LCL;
      lastLnkErr = link_not_open;
      memcpy(gLastStatusString,erlst[link_not_open],STATUS_SHORTSIZE);
      nConnectionTimeouts++;
      c->tocounter++;
      if (c->access & CA_RETRY && (c->tocounter % RETRY_THRESHOLD))
      { /* RETRY bit is set and retried fewer than 3 times (N.B. tocounter will increase to 30 before being reset!) */
        c->heartbeat = SUB_PENDING;
        if (c->mode & CM_SINGLE) c->mode |= CM_RETRY;
        if (tineDebug > 1) dbglog("retry link : %d timeouts so far",c->tocounter);
        return;
      }
      c->dataStatus = not_posted;
      c->linkStatus = link_not_open;
      c->xferReason = CX_NULL;
      NotifyClient(i,link_not_open);
      if (c->bytesin > 0) nPartialTransfers++;
      c->dtimestamp = 0L; c->dtimestampMSEC = 0;
      c->blksin = 0; c->bytesin = 0;
      if (c->boundToId > 0) continue;
      if (BASEMODE(c->mode) > CM_SINGLE) 
      { /* is a persistent link */
        c->heartbeat = SUB_PENDING;
      }
      else
      { /* a single link has timed out => cancel it ! */
        c->heartbeat = 0;
        c->mode = CM_CANCEL;
        msglog(link_not_open,"/%s/%s/%s[%s] : timed out",
          SrvTbl[c->srvIdx].EqmContext,SrvTbl[c->srvIdx].ExportName,c->devName,c->prpName);
      }
      c->time = (UINT32)timestamp;
#     if defined(UNIX) && !defined(NO_TINE_PIPE)
      if (FecTbl[c->fecIdx].TransportProtocol == PIPE) 
      {
        memcpy(&pipe,FecTbl[c->fecIdx].IPXNetwork,4);
        memset(FecTbl[c->fecIdx].IPXNetwork,0,4);
        if (pipe > 0)
        {
          close(pipe); freeBucket(pipe);
        }
        if (tineDebug)
        {
          feclog("FEC %s : reset pipe due to timeout",FecTbl[c->fecIdx].Name);
        }
      }
#     endif
      if (strncmp(c->expName,CSHOSTS,EXPORT_NAME_SIZE)) /* not the name server */
      { /* locateExport will call CheckConnections too */
        if (c->tocounter > TO_THRESHOLD*3 &&
            allLinksAboveThreshold(c->fecIdx) &&
            fecNameId == -1 && c->rdrID == 0 && timestamp != lastaddrreq)
        {
          resetLinkTOcounters(c->fecIdx);
          feclog("Re-acquire (%d) %s",i,FecTbl[c->fecIdx].Name);
          FecTbl[c->fecIdx].TransportProtocol = 0; /* forces address re-acquisition */
          FecTbl[c->fecIdx].TineProtocol = 0;
          if (NameServerLoaded) asyncLocateExport(SrvTbl[c->srvIdx].ExportName,SrvTbl[c->srvIdx].EqmContext);
          else locateExport(SrvTbl[c->srvIdx].ExportName,SrvTbl[c->srvIdx].EqmContext,NULL);
          lastaddrreq = timestamp;
          msglog(link_not_open,"/%s/%s/%s[%s] : force new address resolution",
            SrvTbl[c->srvIdx].EqmContext,SrvTbl[c->srvIdx].ExportName,c->devName,c->prpName);
        }
      }
      else
      { /* a link to the ENS ! */
        feclog("%s : timed out!",CSHOSTS);
        if (c->tocounter >= RETRY_THRESHOLD) c->tocounter = 0;
      }
    }
  }
  for (ic=tcpIdleConnectionList; ic != NULL; )
  {
    ic->ttl -= (int)(timestamp - lasttime)*1000;
    if (ic->ttl < 0) 
    {
      closeIPConnectionToFec(ic->idx);
      ic = RemoveIdleConnection(ic); 
    }
    else ic = ic->nxt;
  }
  for (i=0; i<nglobals; i++)             
  { /* raise alarm for missing heartbeat */
    if (glbTbl[i]->disabled) continue;
    if ((glbTbl[i]->fcn != NULL) &&
        timestamp > glbTbl[i]->time + glbTbl[i]->maxPeriod*GLOBALGRACE)
    {
      NotifyGlobalClient(i,link_not_open);
      glbTbl[i]->time = timestamp;
    }
  }
  if (timestamp != lasttime)
  {
    lasttime = timestamp;
  }
  return;
}
void checkRemotePipe(void)
{ /* TODO: do a better job of this */
# if defined(WIN32) && !defined(NO_REMOTE_PIPE)
  int len;
  char cmdbuf[CMDSIZE];
  if (IPCSocket == 0) return;
# if !defined(NO_REMOTE_PIPE)
  if ((len=GetCommandEx(IPCSocket,cmdbuf,CMDSIZE)) > 0) InterpretCommand(cmdbuf);
# endif
# endif
}
void clientCycle(void)
{
  int saveClientCycle;

  if (WaitForSystemMutex(hSystemClientMutex,0) == 0)
  {
    CheckConsumerData(FALSE);                  /* flush input on the socket */
    CheckSubscriptions(FALSE);
    CheckConsumerData(TRUE);
    CheckConnections();                       /* missing heartbeats ?   */
    ReleaseSystemMutex(hSystemClientMutex);
  }
# if defined(MAKE_TINE32_DLL) && !defined(NO_REMOTE_PIPE)
  DoClientEvents(1000);
# endif
  checkClientNotification(); /* always check, irrespective of Semaphore */
  checkRemotePipe();
  return;
}
void checkClientNotification(void)
{
  int i;
  ConnTableEntry *c;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if ((c=conTbl[i]) == NULL) continue;
    if (c->needsNotification) NotifyClient(i,c->linkStatus);
  }
}
void applyErrorValue(ConnTableEntry *c,void *errValue)
{
  int i,fmt = c->formatOut,len = fmtsizeof(fmt);
  BYTE *ptr;
  if (errValue == NULL) return;
  if (len < 1 || len > 64) return;
  for (i=0; i<(int)c->sizeOut; i++) 
  {
    ptr = (BYTE *)(c->dataOut + i*len);
    memcpy(ptr,errValue,len);
  }
}
void NotifyGlobalClient(int i,int cc)
{
  int k;
  GlobalListStruct *g = NULL;
  GroupTableEntry *gt = NULL;

  if (!(g=glbTbl[i])->isPending) return;
  g->isPending = FALSE;
  if (g->isGrouped)
  {
    for (gt=grpTable; gt && gt->GrpID != (size_t)g->fcn; gt = gt->next);
    if (gt == NULL) g->isGrouped = 0;
    else
    {
      gt->GrpCc |= cc;
      if (--(gt->numPending) > 0) return;
      gt->numPending = gt->numInGrp;
      cc = gt->GrpCc;
      gt->GrpCc = 0;
    }
  }
  gDataTimeStamp = makeDataTimeStamp();
  if (g->fcn != NULL) (g->fcn)((int)g->callbackID,cc);
  for (k=0; k<nglobals; k++) if (glbTbl[k]->fcn == g->fcn)
    glbTbl[k]->isPending = TRUE;
  return;
}
void NotifyClient(int i,int cc)
{
  ConnTableEntry *c = conTbl[i];
  GroupTableEntry *g = NULL;
  ConnTableQueue *q = NULL;
  REDIRECTED_LINK *r;
  DEPENDENT_LINK *lnk;
  int isCancelled = FALSE;

  if (c == NULL) return;
  if (tineDebug > 2)
    dbglog("%s %s %s <%d>: %s",
		  c->expName,c->devName,c->prpName,cc,c->notifyPending ? "notification" : "not pending");
  if (!c->notifyPending) return;
  c->notifyPending = 0;
  c->needsNotification = FALSE;
  if (c->queueLen > 0) c->queueLen--; /* reduce the queue length -> usually sets it back to 0 */

  if (cc != 0)
  {
    if ((cc == illegal_property || cc == non_existent || 
         cc == non_existent_elem || cc == non_existent_fec) &&
        (r=c->rdrLink) != NULL)
    {
      c->rdrLink = NULL;
      removeRedirectedLink(r,cc);
    }
    if (!(cc & CE_SENDDATA) && c->useErrValue)
    {
      applyErrorValue(c,c->errValue);
    }
  }
  if (c->grouped)
  {
    for (g=grpTable; g && g->GrpID != (size_t)c->fcn; g = g->next);
    if (g == NULL) c->grouped = 0;
    else
    {
      g->GrpCc |= cc;
      if (--(g->numPending) > 0) return;
      if (BASEMODE(c->mode) <= CM_SINGLE) /* will be marked for cancellation */
           g->numInGrp = 0;               /* so reduce the group size to 0 */
      g->numPending = g->numInGrp;
      cc = g->GrpCc;
      g->GrpCc = 0;
    }
  }
  LastCompletionDataType = c->formatOut;
  LastCompletionDataSize = c->lastDataSize;
  gDataTimeStamp = PutDataTimeStamp(0,c->dtimestamp,c->dtimestampMSEC);
  if (c->fcn != NULL && !c->cancelPending && !c->supressCallback) (c->fcn)((int)c->callbackID,cc);
  if (cc == 0 || cc & CE_SENDDATA) c->hasNotifiedOnce = TRUE;
# ifndef SMALL_TINE_LIB
  if (WaitForSystemMutex(hLinkTblMutex,gSystemTick) == 0)
  {
    for (lnk=c->dpdLink; lnk != NULL; lnk=lnk->nxt)
    { /* there are dependent links : notify them as well */
      if (conTbl[lnk->linkId]->dataOut != NULL) 
        memcpy(conTbl[lnk->linkId]->dataOut,c->dataOut,c->sizeOut*fmtsizeof(c->formatOut));
      conTbl[lnk->linkId]->lastDataSize = c->lastDataSize;
      conTbl[lnk->linkId]->dataStatus = c->dataStatus;
      conTbl[lnk->linkId]->linkStatus = c->linkStatus;
      conTbl[lnk->linkId]->xferReason = c->xferReason;
      conTbl[lnk->linkId]->dtimestamp = c->dtimestamp;
      conTbl[lnk->linkId]->dtimestampMSEC = c->dtimestampMSEC;
      if (!c->supressCallback && conTbl[lnk->linkId]->fcn != NULL) 
        conTbl[lnk->linkId]->fcn(conTbl[lnk->linkId]->callbackID,cc);
    }
    ReleaseSystemMutex(hLinkTblMutex);
  }
# endif
  for (i=0; i<nConnectionTableEntries; i++) if (conTbl[i]->fcn == c->fcn)
  { /* in case of grouped calls, make sure they're all in a wait state */
    if (c != conTbl[i] && /* different connection entries with same callback */
        (!c->grouped || !conTbl[i]->grouped)) /* not grouped ! */
      continue; /* so skip it */
    conTbl[i]->notifyPending = 1;
    while (conTbl[i]->queueLen > 0 && WaitForSystemMutex(hLinkQueueMutex,gSystemTick) == 0)
    { /* there's a queue -> work thru it */
      if ((q=conTbl[i]->queue) == NULL) 
      { /* shouldn't be possible */
        conTbl[i]->queueLen = 0;
        continue; 
      }
      if (conTbl[i]->fcn == NULL) continue;
      memcpy(conTbl[i]->dataOut,q->dataOut,q->dataSize);
      conTbl[i]->lastDataSize = q->dataSize;
      conTbl[i]->dataStatus = q->dataStatus;
      conTbl[i]->linkStatus = q->statusCode;
      conTbl[i]->dtimestamp = q->timeStamp;
      conTbl[i]->dtimestampMSEC = q->timeStampMSEC;
      if (tineDebug > 1) dbglog("fire queued callback");
      if (!conTbl[i]->supressCallback) conTbl[i]->fcn(conTbl[i]->callbackID,q->statusCode);
      conTbl[i]->queueLen--;
      q = q->next;
      SystemFree(conTbl[i]->queue->dataOut);
      SystemFree(conTbl[i]->queue);
      conTbl[i]->queue = q;
      ReleaseSystemMutex(hLinkQueueMutex);
    }
  }
}
int _datcmp(char *d1,char *d2,int dsz,short fmt,DataToleranceFactor *dtf)
{
  int i,n;
  float delta = 0.0, epsilon = 0.0;

  if (dtf == NULL) return memcmp(d1,d2,dsz);
  if (fmt < CF_DOUBLE || fmt > CF_FLOAT || fmt == CF_TEXT) return memcmp(d1,d2,dsz);
  n = dsz/fmtsizeof(fmt);
  switch (fmt)
  {
    case CF_DOUBLE:
      for (i=0; i<n; i++) 
      {
        delta = (float)fabs(((double *)d1)[i] - ((double *)d2)[i]);
        epsilon = (float)fabs(dtf->percent * ((double *)d2)[i]) + dtf->absolute;
        if (delta > epsilon) return TRUE;
      }
      break;
    case CF_FLOAT:
      for (i=0; i<n; i++) 
      {
        delta = (float)fabs(((float *)d1)[i] - ((float *)d2)[i]);
        epsilon = (float)fabs(dtf->percent * ((float *)d2)[i]) + dtf->absolute;
        if (delta > epsilon) return TRUE;
      }
      break;
    case CF_SHORT:
      for (i=0; i<n; i++) 
      {
        delta = (float)abs(((short *)d1)[i] - ((short *)d2)[i]);
        epsilon = (float)abs((int)(dtf->percent * (float)((short *)d2)[i])) + dtf->absolute;
        if (delta > epsilon) return TRUE;
      }
      break;
    case CF_LONG:
      for (i=0; i<n; i++) 
      {
        delta = (float)abs(((SINT32 *)d1)[i] - ((SINT32 *)d2)[i]);
        epsilon = (float)abs((int)(dtf->percent * (float)((SINT32 *)d2)[i])) + dtf->absolute;
        if (delta > epsilon) return TRUE;
      }
      break;
    case CF_BYTE:
    default:
      return memcmp(d1,d2,dsz);
  }
  return FALSE;
}
int gClnRecvQueueDepth = DEFAULT_RECV_QUEUE_DEPTH;
#ifdef OS_16BIT /* 16-bit OS needs the %ld, 64-bit OS can get confused with %ld ! */
#define bowout(txt) {\
   if (tineDebug) { sprintf(dbgbuf,"subscription %d %ld",i,pPHdr->timestamp);\
     if (c != NULL) sprintf(&dbgbuf[strlen(dbgbuf)],"(%s %s %s %ld)",c->expName,c->devName,c->prpName,c->dtimestamp);\
     sprintf(&dbgbuf[strlen(dbgbuf)],": %s",txt);\
     dbglog(dbgbuf);\
   }\
   goto bow;\
}
#else
#define bowout(txt) {\
   if (tineDebug) { sprintf(dbgbuf,"subscription %d %d",i,pPHdr->timestamp);\
     if (c != NULL) sprintf(&dbgbuf[strlen(dbgbuf)],"(%s %s %s %d)",c->expName,c->devName,c->prpName,c->dtimestamp);\
     sprintf(&dbgbuf[strlen(dbgbuf)],": %s",txt);\
     dbglog(dbgbuf);\
   }\
   goto bow;\
}
#endif /* OS_16BIT */
#define qbowout(txt) { ReleaseSystemMutex(hLinkQueueMutex); c->queueLen--; c->blksin = 0; c->bytesin = 0; bowout(txt); }
int prepRedirString(ConnTableEntry *c,char *rdr,char *str)
{
  char *prpstr, *bpstr, *mpstr;
  switch (c->tineProtocol)
  {
    case 5:
      memset(rdr,0,STATUS_SIZE);
      strncpy(&rdr[PRP_REDIR_SRV_OFFSET],str,16);
      strncpy(&rdr[PRP_REDIR_PRP_OFFSET],&str[16],16);
      if (strlen(c->context) > 0) strncpy(&rdr[PRP_REDIR_CTX_OFFSET],c->context,CONTEXT_NAME_SIZE);
      strncpy(&rdr[PRP_REDIR_DEV_OFFSET],c->devName,DEVICE_NAME_SIZE);
      break;
    case 6:
      memcpy(rdr,str,STATUS_SIZE);
      if (rdr[PRP_REDIR_CTX_OFFSET] == 0) strncpy(&rdr[PRP_REDIR_CTX_OFFSET],c->context,CONTEXT_NAME_SIZE);
      if (rdr[PRP_REDIR_DEV_OFFSET] == 0) strncpy(&rdr[PRP_REDIR_DEV_OFFSET],c->devName,DEVICE_NAME_SIZE);
      break;
    default:
      return illegal_protocol;
  }
  prpstr = &rdr[PRP_REDIR_PRP_OFFSET];
  if (strlen(prpstr) == 0) 
  { /* use the calling property name as is */
    strncpy(prpstr,c->prpName,PROPERTY_NAME_SIZE);
  }
  else
  { /* add the meta modifier back on if necessary */
    if (!IsMetaProperty(prpstr,NULL,NULL) && 
         IsMetaProperty(c->prpName,&bpstr,&mpstr) &&
         strlen(prpstr) < PROPERTY_NAME_SIZE - 16)
    {
      strcat(prpstr,".");
      strncat(prpstr,mpstr,15);
    }
  }
  return 0;
}
void consumeData(BYTE *buf)
{
  short i,m,nb,cc=0,rc=0,fmtsize,totalsize,bytesread = sizeof(totalsize);
  short idx,prot,stepsize,baseMode,renewalMultiplier,syncNotify = 0,supressNotify = 0;
  int fecidx, srvidx;
  UINT32 extsize, extptr, stssize;
  UINT32 datasize,dcmpsize,dataPtr,doutsize,dretsize, mcaLength, ipeek;
  ProducerHeaderStruct PHdr, *pPHdr = &PHdr;
  ConnTableEntry *c = NULL;
  time_t timestamp = time(NULL);
  BYTE *srcbuf;
  char namstr[196],devstr[128],*pctx,*psrv,*pdev,*pprp;
  DTYPE dout, din;
  struct structStruct *s=NULL;
  ConnTableQueue *q=NULL,**nxt=NULL;

  nArrivals++;
  if ((totalsize = PeekWord(*(short *)buf)) <= 0) return;

  switch (cc=PeekWord(*(short *)&buf[2+PHDR_CC_OFFSET]))
  {
    case call_redirection:
    case server_redirection:
      if ((i=PeekWord(*(short *)&buf[2+PHDR_SUBID_OFFSET])) < 0 || i >= nConnectionTableEntries) ccerr(non_existent);
      /* TODO:  need to know the protocol -> p6 will carry more redirect info */
      if ((c=conTbl[i]) == NULL) ccerr(non_existent);
      if (c->mode == CM_CANCEL) ccerr(non_existent);
      stepsize = PHDR_STEPSIZE(c->tineProtocol);
      stssize = STSSIZE(c->tineProtocol);
      prepRedirString(c,namstr,(char *)&buf[bytesread+stepsize]);
      pctx = &namstr[PRP_REDIR_CTX_OFFSET]; psrv = &namstr[PRP_REDIR_SRV_OFFSET]; 
      pdev = &namstr[PRP_REDIR_DEV_OFFSET]; pprp = &namstr[PRP_REDIR_PRP_OFFSET];
      if (strlen(pctx) > 0) sprintf(devstr,"/%.32s/%.32s/%.64s",pctx,psrv,pdev);
      else sprintf(devstr,"%.32s/%.64s",psrv,pdev);
      dout.dArrayLength = c->sizeOut;
      dout.dFormat = c->formatOut;
      memcpy(dout.dTag,c->strTagOut,TAG_NAME_SIZE);
      dout.data.cptr = c->dataOut;
      din.dArrayLength = c->sizeIn;
      din.dFormat = c->formatIn;
      memcpy(din.dTag,c->strTagIn,TAG_NAME_SIZE);
      din.data.cptr = c->dataIn;
      if (tineDebug) dbglog("redirect %s/%s %s to %s %s",c->expName,c->devName,c->prpName,devstr,pprp);
      c->rdrID = redirectedEntrySlot = i;
      c->notifyPending = 0;
      markTCPConnectionAsIdle(i);
      c->TCPsocket = 0;
      m = c->mode;
      c->rdrLink = 
        appendRedirectedLinkList(c->context,c->expName,c->prpName,c->devName,pctx,psrv,pdev,pprp);
      if ((i=AttachLinkEx(devstr,pprp,&dout,&din,c->access,(int)c->pollRate,c->fcn,m,c->callbackID)) < 0)
      {
        if (cc == server_redirection) c->rdrLink = NULL;
      }
      return;
    case illegal_protocol:
      if ((idx=getIncomingFecIndex()) < 0) ccerr(invalid_index);
      for (i=0; i<nConnectionTableEntries; i++) if ((c=conTbl[i]) != NULL && c->fecIdx == idx && c->tineProtocol == 6)
      {
        FecTbl[c->fecIdx].TineProtocol = 5;
        c->tineProtocol = 5;
        c->heartbeat = SUB_PENDING;
        if (tineDebug) dbglog("%s %s %s : switch to protocol 5",c->expName,c->devName,c->prpName);
        if (c->mode & CM_MCAST)
        { /* join legacy mcast group */
          attachPortToMulticastGroup(MCastConsumerSocket,(unsigned short)gtMCastPort,tMCastGenAddr,"server common");
        }
        break;
      }
      return;
    case get_subscription_id:
      if ((idx=getIncomingFecIndex()) < 0) ccerr(invalid_index);
      if ((i=PeekWord(*(short *)&buf[2+PHDR_SUBID_OFFSET])) < 0 || i >= nConnectionTableEntries) ccerr(non_existent);
      if ((c=conTbl[i]) == NULL) ccerr(non_existent);
      stepsize = PHDR_STEPSIZE(c->tineProtocol);
      c->srvID = PeekWord(*(short *)&buf[bytesread+stepsize]);
      if (tineDebug) dbglog("BCAST sub id %d from %s",c->srvID,FecTbl[idx].Name);
      return;
#   ifndef SMALL_TINE_LIB
    case property_is_mca:
      /* TODO: finish this */
      if ((i=PeekWord(*(short *)&buf[2+PHDR_SUBID_OFFSET])) < 0 || i >= nConnectionTableEntries) ccerr(non_existent);
      if ((c=conTbl[i]) == NULL) ccerr(non_existent);
      if (c->mode == CM_CANCEL) ccerr(non_existent);
      memcpy(&ipeek,&buf[2+PDHR_SYST_OFFSET],4);
      mcaLength = PeekDWord(ipeek);
      memcpy(&ipeek,&buf[2+PDHR_URST_OFFSET],4);
      c->mcaIndex = PeekDWord(ipeek);

      dout.dArrayLength = mcaLength;
      dout.dFormat = c->formatOut;
      memcpy(dout.dTag,c->strTagOut,TAG_NAME_SIZE);
      /* TODO: free and realloc >> */
      dout.data.cptr = c->dataOut;

      din.dArrayLength = c->sizeIn;
      din.dFormat = c->formatIn;
      memcpy(din.dTag,c->strTagIn,TAG_NAME_SIZE);
      din.data.cptr = c->dataIn;
      if (tineDebug) dbglog("reassign %s/%s %s to multi-channel array",c->devName,c->prpName);
      c->notifyPending = 0;
      markTCPConnectionAsIdle(i);
      c->TCPsocket = 0;
      m = c->mode;
      if ((i=AttachLinkEx(devstr,c->prpName,&dout,&din,c->access,(int)c->pollRate,c->fcn,m,c->callbackID)) >= 0)
      { /* bind the original link to the mca link */
        c->boundToId = i;
        c->mode = CM_REGISTER;
        c->heartbeat = 0;
        addDependentLink(i);
        /* when it comes in -> only memcpy what the call wanted (the original sizeOut @ the mcaIndex) */
      }
#   endif
    default:
      break;
  }

  memcpy(&ipeek,&buf[2+PHDR_STTM_OFFSET],4); /* avoid alignment problems */
  if (PeekDWord(ipeek) == BCAST_ID) 
  { /* PHDR_PROT_OFFSET only avail for prot > 5 */
    prot = PeekWord(*(UINT16 *)&buf[2+PHDR_PROT_OFFSET]) == 6 ? 6 : 5;
  }
  else if ((i=PeekWord(*(short *)&buf[2+PHDR_SUBID_OFFSET])) >= 0 && i < nConnectionTableEntries)
  {
    if ((c=conTbl[i]) == NULL) ccerr(non_existent);
    if ((prot=c->tineProtocol) < 5 || prot > 6) ccerr(illegal_protocol);
  }
  else ccerr(un_allocated);

  stepsize = PHDR_STEPSIZE(prot);

  while (bytesread < totalsize)
  {
    prepIncomingSubscription(prot,pPHdr,&buf[bytesread]);
    stssize = pPHdr->stssize;
    if (pPHdr->ClnStarttime == BCAST_ID)   
    { /* BCAST or MCAST contract */
      if ((idx=getIncomingFecIndex()) < 0) 
      { /* is this multi/broad-cast coming from someone I know ? */
        return;         
      }
      for (i=0; i<nConnectionTableEntries; i++)        
      { /* is this multi/broad-cast something I've asked for ? */
        if (conTbl[i] == NULL) continue;
        if (BASEMODE(conTbl[i]->mode) == CM_CANCEL) continue;
        if (conTbl[i]->cancelPending) continue;
        if (conTbl[i]->fecIdx != idx) continue;
        if (conTbl[i]->srvID != pPHdr->SubscriptionID) continue;
        /* YES: sync with local contract */
        pPHdr->SubscriptionID = i; 
        pPHdr->ClnStarttime = conTbl[i]->starttime;
        break;
      }
      if (i == nConnectionTableEntries) return;/* not interested in this one */
    }
    i = pPHdr->SubscriptionID;
    if (i < 0 || i >= nConnectionTableEntries) bowout("invalid id");
    if ((c=conTbl[i]) == NULL) bowout("id not ready");
    if (c->starttime != pPHdr->ClnStarttime) bowout("invalid starttime");
    if ((baseMode=BASEMODE(c->mode)) == CM_CANCEL) bowout("cancelled");
    if (pPHdr->timestamp < c->dtimestamp) bowout("stale");
    c->numblks = pPHdr->numblks;
    fmtsize = fmtsizeof(c->formatOut);
    doutsize = pPHdr->msgsize - stepsize;
    datasize = fmtsize*c->sizeOut;
    if (c->formatOut == CF_SPECTRUM) datasize += 96;
    dcmpsize = datasize;
    if (dcmpsize%2) dcmpsize++;                         /* even boundaries */
    if (tineDebug > 2 || (tineDebug && baseMode <= CM_SINGLE && pPHdr->numblks <= 4))
    {
      dbglog("%s (p%d %d of %d)-> %s %s [%s] (+%d) %d vals (%d)<%d>",
	      fecprot,prot,pPHdr->blknum,pPHdr->numblks,
	      c->expName,c->prpName,gFecAddrBuffer,pPHdr->counter,
	      (int)c->sizeOut,c->formatOut,
	      (int)(pPHdr->CompletionCode & 0xfff));
    }
    memset(gLastStatusString,0,STATUS_SIZE);
    c->lastDataSize = 0; extsize = 0;
    switch (cc=pPHdr->CompletionCode)
    {
      case non_existent_elem:
        if (strncmp(c->expName,CSHOSTS,EXPORT_NAME_SIZE))
        {
          c->tocounter++;
          if (c->tocounter > TO_THRESHOLD*3 &&
              allLinksAboveThreshold(c->fecIdx) &&
              fecNameId == -1)
          {
            resetLinkTOcounters(c->fecIdx);
            c->tocounter = 0;
            FecTbl[c->fecIdx].TransportProtocol = 0; /* forces address re-aquisition */
            if (NameServerLoaded) asyncLocateExport(SrvTbl[c->srvIdx].ExportName,SrvTbl[c->srvIdx].EqmContext);
            else locateExport(SrvTbl[c->srvIdx].ExportName,SrvTbl[c->srvIdx].EqmContext,NULL);
          }
        }
        if (c->tocounter == 1)
        {
          feclog("%s/%s %s -> %s [%s] %s [%s]: %s",
            c->expName,c->devName,c->prpName,
            SrvTbl[c->srvIdx].ExportName,SrvTbl[c->srvIdx].EqmName,
            FecTbl[c->fecIdx].Name,gFecAddrBuffer,erlst[cc]);
          if (strnicmp(c->expName,SrvTbl[c->srvIdx].ExportName,EXPORT_NAME_SIZE))
          { /* not equal! connection table is corrupted, try to recover */
            fecidx = locateExport(c->expName,c->context,&srvidx);
            if (srvidx >= 0 && fecidx >= 0)
            {
              c->fecIdx = fecidx; c->srvIdx = srvidx;
            }
          }
        }
      default: /* non-zero */
        if (doutsize < stssize && c->blksin == 0)
        {
          c->bytesin = 0;
          bowout("partial error block");
        }
        if (!(pPHdr->CompletionCode & CE_SENDDATA)) datasize = 0;
        extsize = 0;
        if (pPHdr->blknum == pPHdr->numblks || datasize == 0)
        {
		      memset(c->lastStatusString,0,STATUS_SIZE);
          extsize = stssize;
          if (doutsize < stssize)
          {
            if (pPHdr->numblks == 1) bowout("phantom block");
            extsize = doutsize;
          }
          extptr = stssize - extsize;
          memcpy(&c->lastStatusString[(int)extptr],&buf[bytesread+pPHdr->msgsize-(int)extsize],(size_t)extsize);
          doutsize -= (UINT16)extsize;
          c->bytesin += (UINT16)extsize;
        }
        memcpy(gLastStatusString,&c->lastStatusString,STATUS_SIZE);
        lastLnkErr = cc;
        lastLnkErrSrc = LNK_ERROR_RMT;
        if (!(pPHdr->CompletionCode & CE_SENDDATA)) break;
      case 0:
        if (fmtsize == 0) break;
        if (dcmpsize < doutsize)
        {
          cc = illegal_format;
          if (tineDebug) dbglog("data size mismatch: %lu vs %lu",datasize,doutsize);
        }
        else
        {
          if (pPHdr->numblks == 1 && c->counter == pPHdr->counter) bowout("phantom block");
          if (!c->blksin)
          { /* new set */
            c->counter = pPHdr->counter;
            if (c->bufOutBlks) memset(c->bufOutBlks,0,c->numblks);
          }
          if (c->counter != pPHdr->counter)   
          { /* we're being passed */
            if (tineDebug) dbglog("transfer block out of sequence: %d vs %d",pPHdr->counter,c->counter);
            nPartialTransfers++;
            if (pPHdr->counter < c->counter || pPHdr->counter > c->counter + 30)
            { /* more recent set */
              if (tineDebug) dbglog("reset sequence counter");
              c->blksin = 0; c->bytesin = 0;
              c->counter = pPHdr->counter;       
              if (c->bufOutBlks) memset(c->bufOutBlks,0,c->numblks);
            }
            else bowout("old data block"); /* old stuff coming in */
          }
          if (pPHdr->numblks > 1)               
          { /* multiple blocks !*/
            if (!c->bufOut && !(c->bufOut=(char *)SystemCalloc(1,(size_t)(datasize+STATUS_SIZE)))) bowout("no local memory");
            nb = c->numblks; nb += nb % 8;
            if (!c->bufOutBlks && !(c->bufOutBlks=(BYTE *)SystemCalloc(nb,1))) bowout("no local memory");
            if (pPHdr->blknum > c->numblks) bowout("unexpected block");
            if (c->bufOutBlks[pPHdr->blknum-1]) bowout("phantom block");
            c->bufOutBlks[pPHdr->blknum-1] = 1;
            dataPtr = (pPHdr->blknum-1)*pPHdr->mtu;
            if (doutsize > 0) memcpy(&c->bufOut[(int)dataPtr],&buf[bytesread+stepsize],(size_t)doutsize);
            srcbuf = (BYTE *)c->bufOut;
            ReturnToClientCycle = pPHdr->numblks - c->blksin;
          }
          else 
          { 
            dataPtr = 0; 
            srcbuf = &buf[bytesread+stepsize]; 
          }
          c->bytesin += doutsize;
          if (++(c->blksin) < pPHdr->numblks) goto bow;
          ReturnToClientCycle = 0;
          dretsize = c->bytesin - (cc == 0 ? 0 : stssize);
          if (dretsize <= 0) break;
          if (c->formatOut == CF_DEFAULT)
          { /* record the default format and size */
            c->formatOut = LFMT(pPHdr->EqmFormat); 
            if ((fmtsize=fmtsizeof(c->formatOut)) == 0) bowout("illegal format");
            c->sizeOut = dretsize/fmtsize;
          }
          if (c->dataOut == NULL || srcbuf == NULL) 
          {
            cc = code_failure;
            break;
          }
          if (c->queueLen++ > 0)
          { /* we've got another one and have not yet finished with this one -> queue it */
            if (WaitForSystemMutex(hLinkQueueMutex,gSystemTick) != 0) bowout("can't queue entry");
#           ifndef SMALL_TINE_LIB
            if (c->queueLen >= c->queueDepth) qbowout("queue full");
            if (c->fcn == wcSingleCallback || c->fcn == wcGroupCallback) qbowout("queue not allowed");
            if ((q=(ConnTableQueue *)SystemCalloc(1,sizeof(ConnTableQueue))) == NULL) qbowout("no memory for queue");
            q->dataSize = dretsize;
            if ((q->dataOut=(BYTE *)SystemCalloc(1,q->dataSize)) == NULL) qbowout("no memory for queue data");
            q->dataStatus = pPHdr->stscode;
            q->statusCode = cc;
            q->sysStamp = pPHdr->sysstamp;
            q->usrStamp = pPHdr->userstamp;
            q->xfrReason = pPHdr->xferReason;
            q->timeStamp = pPHdr->timestamp;
            q->timeStampMSEC = pPHdr->timestampUSec / 1000;
            PrepIncomingData(q->dataOut,srcbuf,(size_t)datasize,(size_t)dretsize,c->formatOut);
            for (nxt=&(c->queue); *nxt != NULL; nxt = &((*nxt)->next));
            *nxt = q;
#           endif
            qbowout("entry queued");
          }
          PrepIncomingData(c->dataOut,srcbuf,(size_t)datasize,(size_t)dretsize,c->formatOut);
          c->lastDataSize = dretsize/fmtsize;
          if (c->formatOut == CF_STRUCT && s != NULL && s->rawlen > 0) c->lastDataSize /= s->rawlen;
          supressNotify = CompIncomingData(c->dataOut,c->bufCmp,datasize,c->formatOut,c->dtf);
          if (!c->hasNotifiedOnce) supressNotify = FALSE; /* hasn't yet called in */ 
          if (c->linkStatus != 0) supressNotify = FALSE; /* pre-existing link status not 0 */
          if (!supressNotify && c->bufCmp != NULL) memcpy(c->bufCmp,c->dataOut,datasize);
          if (tineDebug > 1)
          {
#           ifdef OS_16BIT
            dbglog("%s %s [%s] (+%d) data set received (%lu values)",
#           else
            dbglog("%s %s [%s] (+%d) data set received (%u values)",
#           endif
	            c->expName,c->prpName,gFecAddrBuffer,pPHdr->counter,c->lastDataSize);
            if (tineDebug > 3) DataDump("data : ",c->dataOut,(int)c->sizeOut,c->formatOut,RTOL);
          }
        }
        break;
      case use_stream_transport:
	      bowout("use stream transport");
        break;
    }
    syncNotify = (c->mode & CM_SYNCNOTIFY) ? TRUE : FALSE;
    switch (baseMode)
    {
      default:
      case CM_SINGLE : 
        if (c->wcLink == NULL) c->mode = CM_CANCEL;  /* got reply */
		    break;
      case CM_EVENT:
        supressNotify = (pPHdr->xferReason & (CX_EVENT|CX_MESSAGE|CX_ERROR)) ? FALSE : TRUE;
      case CM_REFRESH: 
        c->heartbeat = 2 + HEARTBEAT;
		    break;
      case CM_POLL   : 
        c->heartbeat = 2 + c->pollRate/1000;
		    break;
    }

    if (cc == non_existent_elem)
    {
      if (baseMode > CM_SINGLE && (time_t)c->time != timestamp) c->heartbeat = SUB_PENDING; 
    }
    else 
    {
      c->tocounter = 0;
    }
    c->xferReason = pPHdr->xferReason;
    c->dataStatus = pPHdr->stscode;
    c->linkStatus = cc;
    c->counter = pPHdr->counter;
    c->dtimestamp = pPHdr->timestamp;
    c->dtimestampMSEC = pPHdr->timestampUSec / 1000;
    c->time = (UINT32)timestamp;
    c->datastamp = pPHdr->userstamp;
    c->systemstamp = pPHdr->sysstamp;
    c->blksin = 0; c->bytesin = 0;
    if (c->bufOutBlks) memset(c->bufOutBlks,0,c->numblks);
    if (SystemLocalDataTimeSynchronization) /* don't let the incoming data be too far in the future ! */
    {
      gDataTimeStamp = makeDataTimeStamp();
      if (c->dtimestamp > (UINT32)(gDataTimeStamp) + 1) 
      {
#       ifdef OS_16BIT
        feclog("adjusting incoming data timestamp for %s/%s[%s] from %ld to %ld",
#       else
        feclog("adjusting incoming data timestamp for %s/%s[%s] from %d to %d",
#       endif
          c->expName,c->devName,c->prpName,c->dtimestamp,(SINT32)gDataTimeStamp);
        c->dtimestamp = (SINT32)gDataTimeStamp;
        c->dtimestampMSEC = 0;
      }
    }
    if (c->linkStatus != 0) supressNotify = FALSE; /* this link status not 0 */
    c->supressCallback = (UINT8) supressNotify;
    c->needsNotification = TRUE;
    if (syncNotify) NotifyClient(i,cc);

    if (baseMode > CM_SINGLE && pPHdr->counter > 0)
    {
      renewalMultiplier = 1;
      if (c->pollRate < 1000 && c->pollRate >= 20) renewalMultiplier = (short)(1000/c->pollRate);
      if (pPHdr->counter < REMINDER*renewalMultiplier) c->heartbeat = SUB_PENDING;
      else if (pPHdr->counter <= ACK_REQUEST*renewalMultiplier) c->heartbeat = ACK_PENDING;
    }
bow:
    if ((bytesread += pPHdr->msgsize) <= 0 || (bytesread%2 != 0)) return;
  }
  return;
err:
  if (cc && tineDebug)
  {
    rc = cc & 0xfff;
    if (strlen(gLastStatusString) == 0 && rc > 0 && rc < numErr)
      strncpy(gLastStatusString,erlst[rc],STATUS_SHORTSIZE);
    dbglog("ConsumeData : %s",gLastStatusString);
  }
  return;
}
#ifndef GLOBAL_SYNCHRONIZATION_TOLERANCE
# define GLOBAL_SYNCHRONIZATION_TOLERANCE 0.2
#endif
double gSyncTolerance = GLOBAL_SYNCHRONIZATION_TOLERANCE;
double gSyncTimestamp = 0;
#define TIMESYNC_DEPTH 6
void cbGlobalSynchronization(int id, int cc)
{
  int applyCorrection = FALSE;
  int i,k;
  double correction = 0, delta_t = 0, last_dt = 0, last_rb;
  static int last_cc = -1, last_ptr = 0, ncorrections = 0;
  static double last_logged = 0;
  static double last_dataTimeOffset = 0, largest_correction = 0, last_correction = 0;
  static double last_dataTimeRbDeltas[TIMESYNC_DEPTH] = { 0, 0, 0, 0, 0, 0 };
  static double last_dataTimeDtDeltas[TIMESYNC_DEPTH] = { 0, 0, 0, 0, 0, 0 };
  static double last_dataTimeStamps[TIMESYNC_DEPTH] = { 0, 0, 0, 0, 0, 0 };
  static double last_dataTimeRbTimes[TIMESYNC_DEPTH] = { 0, 0, 0, 0, 0, 0 };

  if (cc != last_cc)
  {
    last_cc = cc;
    feclog("global synchronization : %s",erlst[cc]);
  }
  if (cc) return;

  last_rb = last_dataTimeRbTimes[last_ptr];
  last_dt = last_dataTimeStamps[last_ptr];
  last_ptr = (last_ptr + 1) % TIMESYNC_DEPTH;
  last_dataTimeRbTimes[last_ptr] = gDataTimeStamp;            /* my clock's current timestamp */
  last_dataTimeRbDeltas[last_ptr] = gDataTimeStamp - last_rb; /* time difference based on my clock */
  last_dataTimeStamps[last_ptr] = gSyncTimestamp;             /* remote clock's current timestamp */
  last_dataTimeDtDeltas[last_ptr] = gSyncTimestamp - last_dt; /* time difference based on remote clock */

  for (i=0; i<TIMESYNC_DEPTH; i++)
  {
    k = (last_ptr + i) % TIMESYNC_DEPTH;
    if (last_dataTimeStamps[k] < 1000) return;
    delta_t = last_dataTimeRbDeltas[k] - last_dataTimeDtDeltas[k];
    if (fabs(delta_t) > 0.01)
    { /* these deltas should be really close if the data is coming in and being read at the correct rate */
      return;
    }
    correction += last_dataTimeStamps[k] - last_dataTimeRbTimes[k];
  }
  correction /= TIMESYNC_DEPTH;
  if (correction > gSyncTimestamp)
  {
    feclog("invalid time correction %g greater than time server clock %g",correction,gSyncTimestamp);
    return;
  }

  if (correction > gSyncTolerance || correction < -gSyncTolerance) applyCorrection = TRUE;
  else if (last_correction < gSyncTolerance && correction > gSyncTolerance/2) applyCorrection = TRUE;
  else if (last_correction > gSyncTolerance && correction < -gSyncTolerance/2) applyCorrection = TRUE;
  if (!applyCorrection) return;
  gDataTimeStampOffset += correction;
  last_correction = correction;
  if (tineDebug) dbglog("global synchronization : %g sec (current offset %g sec)",correction,gDataTimeStampOffset);
  ncorrections++;
  if (fabs(correction) > largest_correction) largest_correction = fabs(correction);
  if (gDataTimeStamp < last_logged + 3600) return;
  last_logged = gDataTimeStamp;
  if (fabs(gDataTimeStampOffset-last_dataTimeOffset) < 0.1) return;
  feclog("global synchronization offset : %g sec",gDataTimeStampOffset);
  feclog("largest correction : %g sec  (nr. corrections %d)",largest_correction,ncorrections);
  last_logged = gDataTimeStamp; ncorrections = 0; largest_correction = 0;
  last_dataTimeOffset = gDataTimeStampOffset;
}
int SystemStartGlobalSynchronization(void)
{
  static int glbSyncStarted = FALSE;
  int cc = 0;
  DTYPE dout;

  if (glbSyncStarted) return 0;
  dout.dArrayLength = 1;
  dout.dFormat = CF_DOUBLE;
  dout.data.dptr = &gSyncTimestamp;
  dout.dTag[0] = 0;
  cc = recvNetGlobalEx("SYSTIME",&dout,cbGlobalSynchronization,UNASSIGNED_CALLBACKID);
  if (cc < 0) 
  {
    cc = -cc;
    feclog("could not start global synchronization : %s",erlst[cc]);
  }
  else 
  {
    cc = 0;
    glbSyncStarted = TRUE;
  }
  return cc;
}
void cbTimeSynchronization(int id,int cc)
{
  gDataTimeStamp = makeDataTimeStamp();
  cbGlobalSynchronization(id,cc);
}
int SystemStartTimeSynchronization(char *timesrv)
{
  int cc = 0;
  DTYPE dout;

  if (timesrv == NULL) ccerr(-argument_list_error);
  dout.dArrayLength = 1;
  dout.dFormat = CF_DOUBLE;
  dout.data.dptr = &gSyncTimestamp;
  dout.dTag[0] = 0;
  cc = AttachLink(timesrv,"SYSTIME",&dout,NULL,CA_READ,5000,cbTimeSynchronization,CM_POLL);
err:
  if (cc < 0) feclog("could not start time synchronization : %s",erlst[-cc]);
  return cc < 0 ? -cc : 0;
}
void getGlobalUpdate(BYTE *buf)
{
  GlobalDataHdr tmpGDH, *gd = &tmpGDH; /* what is coming in over the net */
  HeaderStruct hTmp, *GlobalHdr = &hTmp;
  int num,size=HEADERSTRUCT_SIZE;
  int cc=0,i,j,fmtsize,dsize,datasize,dretsize;
  GlobalListStruct *g = NULL; /* what this client wants to see */
  ConnTableEntry *c = NULL;
  struct structStruct *s=NULL;

  BRDcount++; /* increment statistics counter */
  PrepIncomingHeader(GlobalHdr,buf);
  switch (GlobalHdr->Protocol)
  {
    case 3: break;
    default: ccerr(illegal_protocol);
  }
  num = GlobalHdr->EqmSize;
  for (i=0; i<num; i++)
  {
    PrepIncomingGCastData(gd,&buf[size])
    size += sizeof(GlobalDataHdr);
    if ((fmtsize=fmtsizeof(gd->format)) < 0) ccerr(illegal_format);
    for (j=0; j<nglobals; j++)
    {
      if ((g=glbTbl[j])->disabled) continue;
      if (!strcmp(gd->name,g->keyword))
      {
        if (gd->time < (SINT32)g->time) continue;
        g->time = gd->time;
        gLastGlobalDataSize = g->lastsize = gd->size;
        dsize = MIN(g->size,gd->size);
        dretsize = fmtsize*dsize; 
        datasize = fmtsize*g->size;
        if (g->format != gd->format) 
        {
          if (g->refdata == NULL)
          {
            if ((g->refdata=(BYTE *)SystemCalloc(gd->size,fmtsize)) == NULL) 
            {
              if (g->fcn != NULL) (g->fcn)((int)g->callbackID,out_of_client_memory);
              continue;
            }
          }
          PrepIncomingData(g->refdata,&buf[size],datasize,dretsize,gd->format);
          cc = reformat(g->data,g->format,g->refdata,gd->format,gd->size);
        }
        else
        {
          PrepIncomingData(g->data,&buf[size],datasize,dretsize,gd->format);
          cc = 0;
        }
        NotifyGlobalClient(j,cc);
      }
    }
    size += gd->size*fmtsize;
  }
err:
  if (tineDebug > 2)
  {
    sprintf(dbgbuf,"Received %s globals packet",fecprot);
    if (cc) sprintf(&dbgbuf[strlen(dbgbuf)],"ERR: %s",erlst[cc]);
    dbglog(dbgbuf);
  }
  return;
}
void dumpGlobals(void)
{
  int i,j,dsize;
  char Buffer[256];

  for (i=0; i<nglobals; i++)
  {
    sprintf(Buffer,"\t%s : ",glbTbl[i]->keyword);
    dsize = MIN(glbTbl[i]->size,10);
    switch (glbTbl[i]->format)
    {
      case CF_SHORT:
        for (j=0; j<dsize; j++) 
          sprintf(&Buffer[strlen(Buffer)],"%d ",((short *)glbTbl[i]->data)[j]);
        break;
      case CF_LONG:
        for (j=0; j<dsize; j++) 
#         ifdef OS_16BIT
          sprintf(&Buffer[strlen(Buffer)],"%ld ",((SINT32 *)glbTbl[i]->data)[j]);
#         else
          sprintf(&Buffer[strlen(Buffer)],"%d ",((SINT32 *)glbTbl[i]->data)[j]);
#         endif
          break;
      case CF_FLOAT:
        for (j=0; j<dsize; j++) 
          sprintf(&Buffer[strlen(Buffer)],"%g ",((float *)glbTbl[i]->data)[j]);
        break;
      case CF_DOUBLE:
        for (j=0; j<dsize; j++) 
          sprintf(&Buffer[strlen(Buffer)],"%g ",((double *)glbTbl[i]->data)[j]);
        break;
      case CF_NAME16:
        for (j=0; j<dsize; j++) 
          sprintf(&Buffer[strlen(Buffer)],"%s ",((NAME16 *)glbTbl[i]->data)[j].name);
        break;
      case CF_NAME32:
        for (j=0; j<dsize; j++) 
          sprintf(&Buffer[strlen(Buffer)],"%s ",((NAME32 *)glbTbl[i]->data)[j].name);
        break;
      case CF_TEXT:
        strncat(Buffer,(char *)glbTbl[i]->data,MIN(glbTbl[i]->size,200));
        break;
    }
    ttyoutput(Buffer);
  }
  return;
}
/*
 * \internal
 * Try to find a reasonable match from the link table.
 * The caller has to manage his callback IDs effectively
 * for this routine to return something sensible.
 * TODO: make all of the GetCompletion<> routines take
 * the callback function address as a parameter so as to
 * be able to dereference it properly!
 */
int getLinkIdFromCallbackId(int cbId)
{
  int i;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (conTbl[i] == NULL || conTbl[i]->mode == CM_CANCEL) continue;
    if (conTbl[i]->wcLink != NULL && 
        conTbl[i]->wcLink->callbackId == cbId) return i;
    if (conTbl[i]->hasUserCallbackId &&
        conTbl[i]->callbackID == (UINT32)cbId) return i;
  }
  if (cbId >= 0 && cbId < nConnectionTableEntries && !conTbl[cbId]->hasUserCallbackId)
    return cbId;
  return -1;
}
TINE_EXPORT int GetGlobalLinkId(char *keyword)
{
  char *kwd,*ctx,kwdbuf[68];
  int cc;
  strncpy(kwdbuf,keyword,64); keyword[64] = 0;
  if ((cc=parseGlobalsKeyword(kwdbuf,&ctx,&kwd)) != 0) return -cc;
  return getGlobalTableIndex(ctx,kwd);
}
TINE_EXPORT int GetGlobalLinkIdFromCallbackId(int cbId)
{
  int i;
  for (i=0; i<nglobals; i++)
  {
    if (glbTbl[i]->callbackID == (UINT32)cbId) return i;
  }
  return -1;
}
#ifndef SMALL_TINE_LIB
int addDependentLink(int linkId)
{
  ConnTableEntry *c;
  DEPENDENT_LINK *lnk;
  int parentId;
  if (linkId < 0 || linkId > nConnectionTableEntries) return argument_list_error; /* not a valid link table entry ! */
  parentId = conTbl[linkId]->boundToId;
  if (parentId < 1 || parentId > nConnectionTableEntries) return argument_list_error; /* not a valid link table entry ! */
  c = conTbl[parentId];
  if ((lnk=(DEPENDENT_LINK *)SystemCalloc(1,sizeof(DEPENDENT_LINK))) == NULL) return out_of_client_memory;
  lnk->linkId = linkId;
  lnk->nxt = c->dpdLink;
  if (c->dpdLink != NULL) c->dpdLink->prv = lnk;
  c->dpdLink = lnk;
  return 0;
}
void removeDependentLink(int linkId)
{
  ConnTableEntry *c;
  DEPENDENT_LINK *lnk;
  int parentId;
  if (linkId < 0 || linkId > nConnectionTableEntries) return; /* not a valid link table entry ! */
  parentId = conTbl[linkId]->boundToId;
  if (parentId < 1 || parentId > nConnectionTableEntries) return; /* not a valid link table entry ! */
  c = conTbl[parentId];
  if ((lnk=c->dpdLink) == NULL) return; /* no dependency list ! */
  conTbl[linkId]->mode = CM_CANCEL;
  if (WaitForSystemMutex(hLinkTblMutex,gSystemTick) != 0) return;
  for (lnk=c->dpdLink; lnk != NULL; lnk=lnk->nxt)
  {
    if (lnk->linkId == linkId)
    { /* got it */
      if (lnk == c->dpdLink) 
      { /* remove the head ! */
        c->dpdLink = lnk->nxt;
        if (c->dpdLink != NULL) c->dpdLink->prv = NULL;
      }
      else 
      {
        lnk->prv = lnk->nxt;
      }
      free(lnk);
      break;
    }
  }
  ReleaseSystemMutex(hLinkTblMutex);
  return;
}
void wcGroupCallback(int id,int cc)
{
  int wccc, wcid, lstid;
  int i, k, fmtlen, nmlen, noff, voff, soff;
  wildCardTableEntry *wc;
  ConnTableEntry *c = NULL;

  wcid = WILDCARD_LINKID(id);
  lstid = WILDCARD_LISTID(id);

  wc = wcTable[wcid];
  if (wc->linkId < 0) return;
  if (wc->status[lstid] == not_ready) wc->numPending--;
  wc->status[lstid] = cc;
  if (cc == illegal_property) CloseLink(wc->listId[lstid]);
  /* TODO: the illegal_format check is here only because DOOCS returns it ! */
  if (cc == illegal_format) CloseLink(wc->listId[lstid]);

  if (wc->numPending > 0) return;

  c = conTbl[wc->linkId]; /* the original link entry */
  k = 0; /* count the legal links */
  fmtlen = fmtsizeof(wc->format);
  switch (wc->format)
  {
    case CF_NAME16DBLDBL:
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL:
      nmlen = fmtlen - 16;
      for (i=0; i<wc->length && k<(int)c->sizeOut; i++)
      { 
        if (wc->status[i] == illegal_format) continue;
        if (wc->status[i] == illegal_property) continue;
        *(double *)(c->dataOut + k*fmtlen + nmlen) = (double)((float *)wc->data)[i];
        *(double *)(c->dataOut + k*fmtlen + nmlen + 8) = (double)wc->status[i];
        strncpy(c->dataOut + k*fmtlen,wc->list[i].name,16);
        k++;
      }
      break;
    case CF_USTRING:
    case CF_NAME16II:
    case CF_NAME16FI:
      if (wc->format == CF_USTRING)
      {
        nmlen = 80;
        noff = 16; voff = 4; soff = 0;
      }
      else
      {
        nmlen = fmtlen - 8;
        noff = 0; voff = nmlen; soff = nmlen + 4;
      }
      for (i=0; i<wc->length && k<(int)c->sizeOut; i++)
      { 
        if (wc->status[i] == illegal_format) continue;
        if (wc->status[i] == illegal_property) continue;
        strncpy(c->dataOut + k*fmtlen + noff,wc->list[i].name,16);
        memcpy(c->dataOut + k*fmtlen + voff,&((UINT32 *)wc->data)[i],4);
        *(UINT32 *)(c->dataOut + k*fmtlen + soff) = wc->status[i];
        k++;
      }
      break;
    default:
      if (wc->format <= CF_NAME32 && strcmp(c->devName,"*") == 0)
      {
        for (i=0; i<wc->length && k<(int)c->sizeOut; i++)
        {
          if (wc->status[i] == illegal_format) continue;
          if (wc->status[i] == illegal_property) continue;
          memcpy((BYTE *)c->dataOut+k*fmtlen,wc->data+i*fmtlen,fmtlen);
          k++;
        }
      }
      else
      {
        for (i=0; i<wc->length; i++)
        {
          wc->status[i] = illegal_format;
          CloseLink(wc->listId[i]);
        }
      }
      break;
  }
  wc->numActive = k;
  wccc = cc;
  wc->numPending = wc->numActive;
  for (i=0; i<wc->length; i++)
  {
    if (wc->status[i] == illegal_format) continue;
    if (wc->status[i] == illegal_property) continue;
    if (wccc != 0 && wc->status[i] == 0) wccc = 0; /* at least one success */
    wc->status[i] = not_ready;
  }
  c->lastDataSize = k;
  LastCompletionDataSize = c->lastDataSize;
  GetErrorString(wccc,c->lastStatusString);
  wc->callback(wc->callbackId,wccc);
  c->tocounter = 0;
  return;
}
void wcSingleCallback(int id,int cc)
{
  ConnTableEntry *c = NULL;
  wildCardTableEntry *wc = wcTable[id];
  char srv[64],dev[128];
  int fmt, fmtsiz, cbId, mode, i, n, rc;
  DTYPE dout, din;

  switch (cc)
  {
    case illegal_property:
      /* will be returned by the group server */
    case data_not_local:
      /* mark this entry as CM_REGISTER */
      c = conTbl[wc->linkId]; /* the original link entry */
      mode = c->mode;
      c->heartbeat = 0;
      c->mode = CM_REGISTER;
      /* get the device list and length and allocate the relevant arrays */
      if (strlen(c->context) > 0) sprintf(srv,"/%s/",c->context); else srv[0] = 0;
      strcat(srv,c->expName);
      sprintf(dev,"%s/%s",srv,c->devName);
      /* first find out how many */
      dout.dArrayLength = 1; dout.dFormat = CF_LONG; 
      dout.data.lptr = (SINT32 *)&n;
      rc = ExecLinkEx(dev,"NDEVICES",&dout,NULL,CA_READ|CA_RETRY,800);
      gSyncLnkCount = 1;
      if (rc != 0) ccerr(rc);
      /* now allocate the device tables */
      wc->length = n;
      /* NDEVICES returns the total number, so this is a worst case allocation ... */
      if ((wc->list=(NAME16 *)SystemCalloc(n,sizeof(NAME16))) == NULL) ccerr(out_of_client_memory);
      if ((wc->status=(int *)SystemCalloc(n,sizeof(int))) == NULL) ccerr(out_of_client_memory);
      if ((wc->listId=(int *)SystemCalloc(n,sizeof(int))) == NULL) ccerr(out_of_client_memory);
      if (wc->format == CF_DEFAULT)
      { /* a wild card call with unknown format + propertyQueryFunction returning illegal_property */
        c->formatOut = wc->format = CF_FLOAT; 
      }
      switch (wc->format)
      {
        case CF_NAME16DBLDBL:
        case CF_NAME32DBLDBL:
        case CF_NAME64DBLDBL:
        case CF_USTRING:
        case CF_NAME16FI:
          fmt = CF_FLOAT;
          break;
        case CF_NAME16II:
          fmt = CF_LONG;
          break;
        default:
          if (wc->format > CF_NAME32 && strcmp(c->devName,"*") != 0) ccerr(illegal_format);
          fmt = wc->format;
          break;
      }
      fmtsiz = fmtsizeof(fmt);
      if ((wc->data=(BYTE *)SystemCalloc(n,fmtsiz)) == NULL) ccerr(out_of_client_memory);   
      /* fill in the device list */
      rc = GetDeviceNames(dev,wc->list,&n);
      gSyncLnkCount = 1;
      if (rc != 0) ccerr(rc);
      /* if the wildcard contains characters, then n returned here is likely to be smaller than above */
      wc->length = n;
      wc->numPending = wc->length;
      dout.dArrayLength = 1;
      dout.dFormat = fmt;
      din.dArrayLength = c->sizeIn;
      din.dFormat = c->formatIn;
      din.data.vptr = c->dataIn;
      for (i=0; i<wc->length; i++)
      { /* start the individual links (these will be redirected by the source) */
        cbId = MAKE_WILDCARD_CALLBACKID(id,i);
        sprintf(dev,"%s/%s",srv,wc->list[i].name);
        dout.data.vptr = wc->data + i*fmtsiz;
        wc->status[i] = not_ready;
        wc->listId[i] = AttachLinkEx(dev,c->prpName,&dout,&din,c->access,c->pollRate,wcGroupCallback,mode,cbId);
      }
      /* cross-reference back into client connection table */
      c->wcLink = wc;
      c->tocounter = -1; /* reset */
      break;
    default:
      wc->callback(wc->callbackId,cc);
      break;
  }
  return;
err:
  wc->callback(wc->callbackId,cc);
  CloseLink(wc->linkId);
  feclog("wildcard call cannot get device information : %s",erlst[cc]);
  return;
}
int getWildCardTableId(void)
{
  int i, cc = resources_exhausted;

  for (i=0; i<MaxNumConnections; i++)
  {
    if (wcTable[i] == NULL)
    {
      if ((wcTable[i]=(wildCardTableEntry *)SystemCalloc(1,sizeof(wildCardTableEntry))) == NULL) 
        ccerr(out_of_client_memory);
      nWildCardTableEntries++;
      return i;
    }
    if (wcTable[i]->linkId == -1) return i; /* available slot */
  }
err:
  return -cc;
}
#endif /* not MS_DOS */

int isENSCall(char *devName)
{
  if (strstr(devName,CSHOSTS) != NULL) return TRUE;
  if (!stricmp(devName,"ENS")) return TRUE;
  if (!strnicmp(devName,"/SERVICE/ENS/",13)) return TRUE;
  return FALSE;
}
int parseGlobalsKeyword(char *key,char **context,char **keyword)
{ /* not exported, so we know that the args are always okay */
  char *c, *ctx, *kwd;
  if (key[0] == '/')
  { /* context given */
    ctx = &key[1];
    if ((c=strchr(ctx,'/')) == NULL) return invalid_property; /* badly formed -> context but no real keyword given */
    *c++ = 0; 
    kwd = c;
  }
  else
  { /* naked keyword */
    ctx = "SERVICE";
    kwd = key;
  }
  *context = ctx;
  *keyword =kwd;
  return 0;
}
int getGlobalTableIndex(char *context,char *keyword)
{
  int i;
  for (i=0; i<nglobals; i++) 
  {
    if (strcmp(glbTbl[i]->context,context)) continue;
    if (strcmp(glbTbl[i]->keyword,keyword)) continue;
    return i;
  }
  return -1;
}
TINE_EXPORT int ListenForMulticasts(void) 
{
  return ListenForIPMulticasts();
}
TINE_EXPORT int ListenForNetGlobals(void) 
{
  return ListenForIPNetGlobals();
}
TINE_EXPORT char *GetRPCLastError(short cc)
{
  static char s[56];

  if (cc == -1) cc = lastLnkErr;
  if (lastLnkErr == cc && strlen(gLastStatusString) > 0) 
  {
    sprintf(s,"%s: %s",gStatusSourceString[lastLnkErrSrc],gLastStatusString);
  }
  else
  {
    if (cc < 0) cc = -cc;
    if (cc > numErr) sprintf(s,"LCL: (ERR %d undefined)",cc);
    else sprintf(s,"LCL: %s",erlst[cc]);
  }
  return s;
}
#endif /* NO_DOXYGEN */

UINT32 LastCompletionDataSize = 0;
int LastCompletionDataType = CF_NULL;
TINE_EXPORT int MaxNumConnections = MAX_ENTRIES;
/* TINE Client API Calls follow: */
TINE_EXPORT int AttachLinkEx(const char *devNam,const char *devProp,DTYPE *dout,DTYPE *din,short access,
	       int pollingRate,void (*callback)(int,int),int mode,UINT32 callbackID)
{
  char fulname[FULL_NAME_MAXSIZE+4],devProperty[PROPERTY_NAME_SIZE+4],devName[FULL_NAME_MAXSIZE+4],devField[FIELD_NAME_SIZE+4];
  int i=-1,baseMode,dinsize=0,doutsize=0,cc=0;
  int isENS,grouped,supressDuplication,wait,useOnError,wcid,isBfPrp=FALSE;
  GroupTableEntry *gt = NULL;
  struct structStruct *s;
  char *c = NULL, *cn = NULL,*en = NULL,*dn = NULL,*ap = NULL,devNameBuffer[256];
  wildCardTableEntry *wc;
  char haddr[4],srv[2*CONTEXT_NAME_SIZE+4];
  struct in_addr inaddr;

  if (devNam == NULL || devProp == NULL) ccerr(argument_list_error);
  if (pollingRate < 0) ccerr(invalid_interval);
  strncpy(devName,devNam,FULL_NAME_MAXSIZE);
  strncpy(devProperty,devProp,PROPERTY_NAME_SIZE);
  if (isBitfieldAccess(dout)) 
  {
    acquireAndRegisterBitfield(devName,dout->dTag);
    if ((cc=splitFullName(devName,srv,NULL)) != 0) ccerr(cc);
    if (isBitfieldProperty(srv,dout->dTag,devProperty,devField)) isBfPrp = TRUE;
  }
  strncpy(fulname,devName,FULL_NAME_MAXSIZE); fulname[FULL_NAME_MAXSIZE] = 0;
  if (getCdiLinkNameHook != NULL && (dn=getCdiLinkNameHook(devName,devNameBuffer)) != NULL)
  {
    cc = -not_implemented;
    if (cdiAttachLinkHook != NULL) cc = cdiAttachLinkHook(dn,devProperty,dout,din,access,pollingRate,callback,mode,callbackID);
    if (cc != -server_redirection) return cc;
    strncpy(fulname,dn,64);
    dn = NULL; cc = 0;
  }
  lastLnkErrSrc = LNK_ERROR_LCL;
  isENS = isENSCall(devName);
  if (!isENS && AttachLinkSemaphore++) 
  {
    lastLnkErrSrc = LNK_ERROR_LCL;
    lastLnkErr = semaphore_busy;
    memcpy(gLastStatusString,erlst[semaphore_busy],STATUS_SHORTSIZE);
    msglog(semaphore_busy,"AttachLinkEx %s %s : semaphore busy",devName,devProperty);
    return -semaphore_busy;
  }

  if (gUseSynchronousNotification) mode |= CM_SYNCNOTIFY;
  supressDuplication = mode & CM_NODUPLICATES; 
  grouped = mode & CM_GROUPED; 
  wait = mode & CM_WAIT;
  useOnError = mode & CM_USEONERROR; 
  mode &= ~(CM_REMOVEVOLATILE); /* strip off these bits */
  if ((mode & CM_BCAST || mode & CM_MCAST) && !ListenForMCasts)
  {
    if ((cc=ListenForMulticasts()) != 0) ccerr(cc);
  }
  baseMode = BASEMODE(mode);
  if (baseMode == CM_RECEIVE) 
  {
    AttachLinkSemaphore = 0;
    return recvNetGlobalEx(devProperty,dout,callback,grouped ? CM_GROUPED : callbackID);
  }
  if (baseMode > CM_EVENT  || baseMode < CM_CANCEL) ccerr(illegal_mode);
  if (baseMode == CM_REFRESH || baseMode == CM_EVENT) grouped = 0;
  pollingRate = pollingRate ? pollingRate : 1000;
  if (SystemSupressHeartbeatNotification && (baseMode == CM_REFRESH)) supressDuplication = TRUE;

  if (din)
  {
    din->dTimeStamp = 0;
    dinsize = (int)(fmtsizeof(din->dFormat)*din->dArrayLength);
    if (din->dFormat == CF_STRUCT && (s=getStructFromTag(din->dTag)) != NULL) dinsize *= s->size;
    if (din->dFormat == CF_NULL) memset(din->dTag,0,TAG_NAME_SIZE);
  }
  if (dout)
  {
    dout->dTimeStamp = 0;
    doutsize = (int)(fmtsizeof(dout->dFormat)*dout->dArrayLength);
    if (dout->dFormat == CF_STRUCT && (s=getStructFromTag(dout->dTag)) != NULL) doutsize *= s->size;
    if (dout->dFormat == CF_SPECTRUM) doutsize += 96;
    if (dout->dFormat == CF_NULL) memset(dout->dTag,0,TAG_NAME_SIZE);
  }
  if (tineDebug)
  {
    feclog("Attach %s %s (%d msec): %s",devName,devProperty,pollingRate,modetoa[baseMode]);
  }
  if ((cn=strtok_r(fulname,DELIMITER,&c)) == NULL) ccerr(argument_list_error);
  if ((en=strtok_r(NULL,DELIMITER,&c)) == NULL) /* only one given */
  {
    en = cn; cn = NULL; dn = &fulname[FULL_NAME_MAXSIZE];
  }
  else if ((dn=strtok_r(NULL,DELIMITER,&c)) == NULL) /* only two given */
  {
     if (cn == fulname) /* no leading delimiter */
     {
       dn = en; en = cn; cn = NULL;
     }
     else dn = &fulname[FULL_NAME_MAXSIZE];
  }
  else if ((ap=strtok_r(NULL,"\0",&c)) != NULL) /* a '/' in the device name */
  {
    dn[strlen(dn)] = '/';  /* put it back */
  }
  if ((i=getConTblEntry(dout,din,cn,en,dn,devProperty)) < 0)
  {
    if (i != -link_exists)
    {
      feclog("Unable to subscribe to %s %s: %s",devName,devProperty,erlst[-i]);
      if (callback != NULL && callbackID != UNASSIGNED_CALLBACKID) callback(callbackID,-i); /* in case of server_redirect */
      ccerr(-i);
    }
    if ((i=lastOutputDataLink) < 0) ccerr(link_exists);
    if (BASEMODE(conTbl[i]->mode) == CM_SINGLE)  /* was a SINGLE link */
    {
      if (tineDebug) dbglog("attaching %s %s to link %d",devName,devProperty,i);
      if (dout->data.vptr != NULL) conTbl[i]->dataOut = (char *)dout->data.vptr;
      else dout->data.vptr = (void *)conTbl[i]->dataOut;
    }
    ccerr(link_exists);
  }
  if (callbackID == UNASSIGNED_CALLBACKID) callbackID = i; else conTbl[i]->hasUserCallbackId = TRUE;
  if (conTbl[i]->tineProtocol == 0) conTbl[i]->tineProtocol = DEFAULT_TRANSPORT_PROTOCOL_LEVEL;
  if (useConnectedSockets) mode |= CM_CONNECT;
  conTbl[i]->pollRate  = pollingRate;
  conTbl[i]->mode      = mode;
  conTbl[i]->fcn       = callback;
  conTbl[i]->callbackID= callbackID;
  conTbl[i]->time      = (UINT32)time(NULL);
  conTbl[i]->heartbeat = mode == CM_REGISTER ? 0 : SUB_PENDING;
  conTbl[i]->access    = access;
  conTbl[i]->notifyPending   = 1;
  conTbl[i]->needsNotification = FALSE;
  if (isBfPrp) strncpy(conTbl[i]->field,devField,FIELD_NAME_SIZE); 
  if (supressDuplication) conTbl[i]->bufCmp = (char *)SystemCalloc(1,(size_t)doutsize);
  if (useOnError && dout->dArrayLength > 0) putLnkErrorValue(i,dout->data.vptr);
  if (mode & CM_MCAST && baseMode > CM_SINGLE)
  { /* join the server's specific multicast group */
    memcpy(haddr,FecTbl[conTbl[i]->fecIdx].IPh_addr,4);
    haddr[0] = (BYTE)238; haddr[1] = (BYTE)1; /* "a.b.c.d" -> "238.1.c.d" */
    memcpy(&inaddr,&haddr,4);
    attachPortToMulticastGroup(MCastConsumerSocket,(unsigned short)gtMCastPort,inaddr,"server specific");
    conTbl[i]->mcastGrp = inaddr.s_addr;
  }
# ifndef SMALL_TINE_LIB
  /* check this is a dependent link */
  if (conTbl[i]->boundToId > 0)
  {
    conTbl[i]->mode = CM_REGISTER;
    conTbl[i]->heartbeat = 0;
  }
  /* check for wildcard call */
  else if (strchr(devName,'*') != NULL  && !IsStockProperty(devProperty))
  {
    /* don't allow the CM_GROUPED flag ! */
    if (grouped) ccerr(illegal_mode);
    if ((wcid=getWildCardTableId()) < 0) ccerr(-wcid);
    wc = wcTable[wcid];
    wc->linkId = i;
    wc->callbackId = conTbl[i]->callbackID;
    wc->callback = conTbl[i]->fcn;
    wc->format = conTbl[i]->formatOut;
    /* reset these to the wildcard callback  */
    conTbl[i]->fcn = wcSingleCallback;
    conTbl[i]->callbackID = wcid;
    if (access & CA_SYNC) wait = TRUE; /* have to let possible queries complete ! */
    /* have to reset these here for wild card collection */
    AttachLinkSemaphore = 0;
    ExecLinkSemaphore = 0;
  }
# endif /* SMALL_TINE_LIB */
  if (grouped && (gt=findGroup((size_t)callback)) == NULL) grouped = 0;
  if (gt)
  {
    gt->numPending++;
    gt->numInGrp++;
    conTbl[i]->grouped = 1;
  }
  if (wait)
  {
    conTbl[i]->tocounter = -1;
    while (conTbl[i]->tocounter == -1) WaitForSyncCount(gSyncLnkCount);
  }
  ReturnToClientCycle = 1;
err:
  if (cc || mode != CM_REGISTER)
  {
    msglog(cc,"AttachLink [%d] %s %s (%d bytes read %d bytes sent) %s %s %d ms",
      i,devName,devProperty,doutsize,dinsize, access & CA_WRITE ? "WRITE" : "READ",
      modetoa[BASEMODE(mode)],pollingRate);
  }
  if (!isENS ) AttachLinkSemaphore = 0;
  return cc ? -cc : i;
}
TINE_EXPORT int AttachLink(const char *devName,const char *devProperty,DTYPE *dout,DTYPE *din,short access,
	       int pollingRate,void (*callback)(int,int),int mode)
{
  return AttachLinkEx(devName,devProperty,dout,din,access,pollingRate,callback,mode,UNASSIGNED_CALLBACKID);
}
TINE_EXPORT int CloseGlobalLink(int linkId)
{
  struct in_addr inip;
  if (linkId < 0 || linkId >= nglobals) return invalid_link;
  memcpy(&inip,&glbTbl[linkId]->mcastGrp,4);
  detachPortFromMulticastGroup(GlobalSocket,(unsigned short)gtGCastPort,inip,"globals specific");
  glbTbl[linkId]->disabled = TRUE;
  return 0;
}
TINE_EXPORT void CloseLink(int i)
{
  ConnTableEntry *c;
  wildCardTableEntry *wc;
  int k;
  struct in_addr inip;

  if (cdiCloseLinkHook != NULL && cdiCloseLinkHook(i) == 0) return;
  if (i < 0 || i >= nConnectionTableEntries) return; /* not a valid link table entry ! */
  if ((c=conTbl[i]) == NULL) return;
  msglog(0,"/%s/%s/%s[%s] : closing link [%d]",
    SrvTbl[c->srvIdx].EqmContext,SrvTbl[c->srvIdx].ExportName,c->devName,c->prpName,i);
# ifndef SMALL_TINE_LIB
  if ((k=c->boundToId) > 0)
  { /* this one is just bound to another link (don't close for real) */
    removeDependentLink(i);
    if (conTbl[k]->dpdLink == NULL)
    { /* removed last dependency; is the parent still active ? */
      if (!conTbl[k]->cancelPending) return;
      /* can now cancel the parent link ! */
      c = conTbl[k]; 
    }
    else
    { /* there are still dependencies, don't do anything */
      return;
    }
  }
  if (c->dpdLink != NULL)
  { /* parent link is cancelled with dependencies ! */
    c->cancelPending = TRUE;
    c->needsNotification = FALSE;
    return;
  }
  if ((wc=c->wcLink) != NULL)
  { /* closing a wildcard link */
    if (wc->length > 0)
    { /* not a single-server link */
      for (k=0; k<wc->length; k++)
      { /* re-entrancy shouldn't be a problem here ...*/
        CloseLink(wc->listId[k]);
      }
      c->mode = CM_CANCEL; 
      c->heartbeat = 0;
    }
    c->wcLink = NULL;
    /* mark as available and cleanup ... */
    wc->linkId = -1; 
    if (wc->list != NULL) free(wc->list); wc->list = NULL;
    if (wc->data != NULL) free(wc->list); wc->data = NULL;
    if (wc->status != NULL) free(wc->list); wc->status = NULL;
    if (wc->listId != NULL) free(wc->list); wc->listId = NULL;
    wc->length = 0;
  }
# endif
  if (BASEMODE(c->mode) > CM_SINGLE)
  {
    if (c->mode & CM_MCAST && c->mcastGrp != 0) 
    {
      memcpy(&inip,&c->mcastGrp,4);
      detachPortFromMulticastGroup(MCastConsumerSocket,(unsigned short)gtMCastPort,inip,"server specific");
      c->mcastGrp = 0;
    }
    if (c->mode & (CM_BCAST|CM_MCAST)) c->mode = CM_CANCEL | CM_MCAST;
    else c->mode = CM_CANCEL;
    if (c->TCPsocket > 0) c->mode |= CM_CONNECT;
    c->heartbeat = SUB_PENDING;
  }
  c->needsNotification = FALSE;
  return;
}
TINE_EXPORT int CloseNetGlobal(const char *keyword)
{
  int i, cc;
  char *kwd,*ctx,kwdbuf[68];

  strncpy(kwdbuf,keyword,64); kwdbuf[64] = 0;
  if ((cc=parseGlobalsKeyword(kwdbuf,&ctx,&kwd)) != 0) return cc;
  if ((i=getGlobalTableIndex(ctx,kwd)) == -1) return invalid_keyword;
  return closeGlobalLink(i);
}
TINE_EXPORT int ExecLinkEx(const char *devName,const char *devProperty,DTYPE *dout,DTYPE *din,short access,UINT16 timeout)
{
  int cc=0,id,pid=0;
  int mode,to,cnt=2,canRetry=0,isRetry=0,isAdjustable=0,isENS=0,keepLink=FALSE,wait=TRUE;
  char *dn=NULL,devNameBuffer[256];

  if (devName == NULL || devProperty == NULL) return argument_list_error;
  if (isBitfieldAccess(dout)) acquireAndRegisterBitfield(devName,dout->dTag);
  isAdjustable = access & CA_MUTABLE;
  canRetry = access & CA_RETRY;
  access &= ~(CA_MUTABLE|CA_RETRY);
  if (getCdiLinkNameHook != NULL && (dn=getCdiLinkNameHook((char *)devName,devNameBuffer)) != NULL)
  {
    cc = not_implemented;
    if (cdiExecLinkHook != NULL) cc = cdiExecLinkHook(dn,(char *)devProperty,dout,din,access,timeout);
    if (cc != server_redirection) return cc;
    cc = 0;
  }
  dn = (char *)devName; /* not a CDI hook ! */
  lastLnkErrSrc = LNK_ERROR_LCL;

  isENS = isENSCall((char *)devName);
  if (!isENS && ExecLinkSemaphore++) 
  {
    lastLnkErrSrc = LNK_ERROR_LCL;
    lastLnkErr = semaphore_busy;
    memcpy(gLastStatusString,erlst[semaphore_busy],STATUS_SHORTSIZE);
    msglog(semaphore_busy,"ExecLinkEx %s %s : semaphore busy",devName,devProperty);
    return semaphore_busy;
  }
retry:
  mode = CM_SINGLE;
  if (access & CA_CONNECT) mode |= CM_CONNECT;
  mode |= CM_SYNCNOTIFY; /* always use synchronous notification on synchronous links */
  if (isRetry) mode |= CM_RETRY;
  gSyncLnkCount = 1; /* to be reset by sync callback */
  if ((id=AttachLink(dn,devProperty,dout,din,access,timeout,SyncAsync,mode)) < 0 ||
      (pid=conTbl[id]->boundToId) > 0)
  {
    if (pid == 0 && id == -link_exists) 
    {
      pid = lastOutputDataLink; /* this isn't exactly thread-safe ! */
      keepLink = TRUE;
    }
    if (pid > 0) /* get output from parent */
    {
      if (tineDebug) dbglog("redirecting output for %s %s to link %d",dn,devProperty,pid);
      if (dout->data.vptr == NULL)
	      dout->data.vptr = (void *)conTbl[pid]->dataOut;
      else if (dout->data.vptr != conTbl[pid]->dataOut)
	      memcpy(dout->data.vptr,conTbl[pid]->dataOut,(size_t)conTbl[pid]->sizeOut*fmtsizeof(conTbl[pid]->formatOut));
      if (isAdjustable) dout->dArrayLength = conTbl[pid]->lastDataSize;
#     ifndef SMALL_TINE_LIB
      if (!keepLink)
      {
        conTbl[id]->heartbeat = 0;
        conTbl[id]->mode = CM_CANCEL; /* make sure it is CANCELled ! */
        removeDependentLink(id);
      }
#     endif
      ccerr(0);
    }
    ccerr(-id);
  }
  if (conTbl[id]->wcLink != NULL) ExecLinkSemaphore = 0;
  to = (int)conTbl[id]->pollRate;
  if (to >= 1000) canRetry = TRUE;
# ifdef THIS_IS_HERE_FOR_A_REASON
  if (strchr(devName,'*') != NULL  && !IsStockProperty((char *)devProperty)) wait = FALSE;
# endif
  if (wait && WaitForSyncLink(to) != 0)
  { /* if the caller wants this to be quick, don't retry! */
    if (cnt-- > 0 && canRetry)
    {
      conTbl[id]->heartbeat = 0;
      conTbl[id]->mode = CM_CANCEL;  /* CANCEL it in the link table ! */
      isRetry = TRUE;
      if (isENS && cnt == 0) ToggleNameServer();
      goto retry;
    }
    conTbl[id]->lastLnkErrSrc = LNK_ERROR_LCL;
    conTbl[id]->dataStatus = not_posted;
    conTbl[id]->linkStatus = connection_timeout;
    strncpy(conTbl[id]->lastStatusString,erlst[connection_timeout],32);
    strncpy(gLastStatusString,conTbl[id]->lastStatusString,32);
    FecTbl[conTbl[id]->fecIdx].TransportProtocol = 0;/* forces address re-aquisition */
    FecTbl[conTbl[id]->fecIdx].TineProtocol = 0;
  }
  conTbl[id]->heartbeat = 0;
  conTbl[id]->mode = CM_CANCEL;        /* make sure it is CANCELled ! */
  cc = lastLnkErr = conTbl[id]->linkStatus;
  if (cc == 0) strcpy(conTbl[id]->lastStatusString,"success");
  if (cc != 0 && conTbl[id]->rdrID > 0) cc |= CE_REDIRECTED;
  if (dout != NULL)
  {
    dout->dTimeStamp = PutDataTimeStamp(0,conTbl[id]->dtimestamp,conTbl[id]->dtimestampMSEC);
    if (dout->dFormat == CF_DEFAULT)
    {
      dout->dFormat = conTbl[id]->formatOut;
      dout->dArrayLength = conTbl[id]->sizeOut;
    }
    dout->dStamp = conTbl[id]->datastamp;
    dout->sysStamp = conTbl[id]->systemstamp;
    if (isAdjustable) dout->dArrayLength = conTbl[id]->lastDataSize;
    dout->xferReason = conTbl[id]->xferReason;
  }
  if (tineDebug && cc) dbglog("ExecLink: %s",gLastStatusString);
err:
  if (cc == connection_timeout) 
    msglog(cc,"ExecLink to %s[%s] timed out (%d retries)",devName,devProperty,2-cnt);
  if (!isENS) ExecLinkSemaphore = 0;
  return cc;
}
TINE_EXPORT int ExecLink(const char *devName,const char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  char lclDevName[64];
  if (!strnicmp((char *)devName,"/LOCAL/",7) || !strnicmp((char *)devName,"\\LOCAL\\",7))
  {
    if (strlen(devName) > 64) return non_existent_elem;
    strcpy(lclDevName,&devName[7]);
    return ExecLocalLink(lclDevName,devProperty,dout,din,access);
  }
  return ExecLinkEx(devName,devProperty,dout,din,access,DEFAULT_CLIENT_POLLRATE);
}
TINE_EXPORT int GetClnRecvQueueDepth(void)
{
  return gClnRecvQueueDepth;
}
TINE_EXPORT int GetLinkQueueDepth(int linkId)
{
  if (linkId < 0 || linkId > nConnectionTableEntries-1) return -invalid_link;
  if (conTbl[linkId] == NULL) return -non_existent_elem;
  return conTbl[linkId]->queueDepth;
}
TINE_EXPORT int GetCompletionDataSize(int i)
{
  if (i == -1) return (int)LastCompletionDataSize;
  if (i < 0 || i >= nConnectionTableEntries) return -out_of_range;
  if (conTbl[i] == NULL) return -un_allocated;
  return (int)conTbl[i]->lastDataSize;
}
TINE_EXPORT int GetCompletionDataSizeFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? -link_not_open : (int)conTbl[i]->lastDataSize;
}
TINE_EXPORT int GetCompletionDataType(int i)
{
  if (i == -1) return (int)LastCompletionDataType;
  if (i < 0 || i >= nConnectionTableEntries) return -out_of_range;
  if (conTbl[i] == NULL) return -un_allocated;
  return (int)conTbl[i]->formatOut;
}
TINE_EXPORT int GetCompletionDataTypeFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? -link_not_open : (int)conTbl[i]->formatOut;
}
TINE_EXPORT char *GetCompletionStatus(int i)
{
  if (i < 0 || i >= nConnectionTableEntries) return NULL;
  if (conTbl[i] == NULL) return NULL;
  return conTbl[i]->lastStatusString;
}
TINE_EXPORT short GetCompletionSource(int i)
{
  if (i < 0) return lastLnkErrSrc;
  if (i >= nConnectionTableEntries) return -1;
  if (conTbl[i] == NULL) return -un_allocated;
  return conTbl[i]->lastLnkErrSrc;
}
TINE_EXPORT char *GetCompletionStatusFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? NULL : conTbl[i]->lastStatusString;
}
TINE_EXPORT short GetCompletionSourceFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? -link_not_open : conTbl[i]->lastLnkErrSrc;
}
TINE_EXPORT double GetCurrentDataTimeStamp(int i)
{
  if (i < 0 || i >= nConnectionTableEntries) return out_of_range;
  if (conTbl[i] == NULL) return un_allocated;
  return PutDataTimeStamp(0,conTbl[i]->dtimestamp,conTbl[i]->dtimestampMSEC);
}
TINE_EXPORT double GetCurrentDataTimeStampFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? 0 : PutDataTimeStamp(0,conTbl[i]->dtimestamp,conTbl[i]->dtimestampMSEC);
}
TINE_EXPORT int GetCurrentLinkStatus(int i)
{
  if (i < 0 || i >= nConnectionTableEntries) return out_of_range;
  if (conTbl[i] == NULL) return un_allocated;
  return conTbl[i]->linkStatus;
}
TINE_EXPORT int GetCurrentLinkStatusFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? -link_not_open : (int)conTbl[i]->linkStatus;
}
TINE_EXPORT int GetCurrentDataStatus(int i)
{
  if (i < 0 || i >= nConnectionTableEntries) return out_of_range;
  if (conTbl[i] == NULL) return un_allocated;
  return conTbl[i]->dataStatus;
}
TINE_EXPORT int GetCurrentDataStatusFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? -link_not_open : (int)conTbl[i]->dataStatus;
}
TINE_EXPORT int GetDataFromLinkId(int linkId,DTYPE *dout,UINT16 *status)
{
  ConnTableEntry *c;
  if (dout == NULL) return argument_list_error;
  if (linkId < 0 || linkId >= nConnectionTableEntries) return invalid_index; /* not a valid link table entry ! */
  c = conTbl[linkId];
  dout->dFormat = c->formatOut;
  dout->dArrayLength = c->lastDataSize;
  strncpy(dout->dTag,c->strTagOut,8);
  dout->dTimeStamp = (double)c->dtimestamp + ((double)c->dtimestampMSEC)/1000.0;
  dout->dStamp = conTbl[linkId]->datastamp;
  dout->sysStamp = conTbl[linkId]->systemstamp;
  dout->data.cptr = c->dataOut;
  dout->xferReason = conTbl[linkId]->xferReason;
  if (status != NULL) *status = (UINT16)c->linkStatus;
  return 0;
}
TINE_EXPORT int GetDataFromCallbackId(int id,DTYPE *dout,UINT16 *status)
{
  int i = getLinkIdFromCallbackId(id);
  return GetDataFromLinkId(i,dout,status);
}
TINE_EXPORT double GetGlobalsTimeStamp(int linkId)
{
  if (linkId < 0 || linkId >= nglobals) return 0;
  return (double)glbTbl[linkId]->time;
}
TINE_EXPORT int GetLastGlobalDataSize(int id)
{
  if (id == -1) return gLastGlobalDataSize;
  if (id < 0 || id >= nglobals) return -out_of_range;
  return glbTbl[id]->lastsize;
}

TINE_EXPORT int GetLastGlobalDataSizeFromCallbackId(int id)
{
  int i = getGlobalLinkIdFromCallbackId(id);
  if (i == -1) return gLastGlobalDataSize;
  return glbTbl[i]->lastsize;
}

TINE_EXPORT char *GetLinkName(int i)
{
  static char linkname[128];
  strcpy(linkname,"link not found");
  if (i >= 0 && i < nConnectionTableEntries)
  {
    sprintf(linkname,"/%s/%s/%s[%s]",conTbl[i]->context,conTbl[i]->expName,
      conTbl[i]->devName,conTbl[i]->prpName);
  }
  return linkname;
}
TINE_EXPORT char *GetLinkNameFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  static char linkname[128];
  strcpy(linkname,"link not found");
  if (i >= 0) 
  {
    sprintf(linkname,"/%s/%s/%s[%s]",conTbl[i]->context,conTbl[i]->expName,
      conTbl[i]->devName,conTbl[i]->prpName);
  }
  return linkname;
}
TINE_EXPORT short GetLastLinkError(short cc, char  *errstr)
{
  if (errstr == NULL) return argument_list_error;
  strncpy(errstr,GetRPCLastError(cc),56);
  return 0;
}
TINE_EXPORT double GetSynchronizationTolerance(void) { return gSyncTolerance; }
TINE_EXPORT int GetSystemDataStamp(int linkId)
{
  if (linkId < 0) return gSystemDataStamp;
  if (linkId > nConnectionTableEntries-1) return -out_of_range;
  return conTbl[linkId]->systemstamp;
}
TINE_EXPORT int GetSystemDataStampFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? -link_not_open : conTbl[i]->systemstamp;
}
TINE_EXPORT int GetDataStamp(int linkId)
{
  if (linkId < 0 || linkId > nConnectionTableEntries-1) return -out_of_range;
  return conTbl[linkId]->systemstamp;
}
TINE_EXPORT int GetDataStampFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? -link_not_open : conTbl[i]->systemstamp;
}
TINE_EXPORT UINT16 GetTransferFlag(int linkId)
{
  if (linkId < 0 || linkId > nConnectionTableEntries-1) return CX_NULL;
  return conTbl[linkId]->xferReason;
}
TINE_EXPORT UINT16 GetTransferFlagFromCallbackId(int id)
{
  int i = getLinkIdFromCallbackId(id);
  return i < 0 ? CX_NULL : conTbl[i]->xferReason;
}
TINE_EXPORT void SetSupressHeartbeatNotification(int value)
{
  SystemSupressHeartbeatNotification = value;
}
TINE_EXPORT int recvNetGlobalEx(const char *keyword,DTYPE *dout,void (*callback)(int,int),UINT32 callbackID)
{
  static int initialized = 0;
  GroupTableEntry *gt = NULL;
  int cc = 0,i = 0,isGrouped = 0,rc;
  UINT16 dsize;
  char *kwd,*ctx,glbSrv[64],kwdbuf[68];
  char haddr[4];
  struct in_addr inaddr;
  FecDataStruct *fec;

  if (keyword == NULL || dout == NULL) return -argument_list_error;
  if (recvNetGlobalSemaphore) return -semaphore_busy; else recvNetGlobalSemaphore = -1;

  if (!ListenForGlobals)
  {
    if ((cc=ListenForNetGlobals()) != 0) ccerr(cc);
  }
  if (!initialized)
  {
    initialized = TRUE;
    if ((glbTbl=(GlobalListStruct **)SystemCalloc(MaxNumGlobals,sizeof(GlobalListStruct *))) == NULL)
      ccerr(out_of_client_memory);
  }

  strncpy(kwdbuf,keyword,64); kwdbuf[64] = 0;
  if ((cc=parseGlobalsKeyword(kwdbuf,&ctx,&kwd)) != 0) return cc;
  sprintf(glbSrv,"/%s/%s",ctx,GLOBALSRV);
  if ((i=getGlobalTableIndex(ctx,kwd)) != -1) 
  {
    if (!glbTbl[i]->disabled) 
    { /* is in the list and is active */
      if (dout->dFormat != glbTbl[i]->format) return -illegal_format;
      if (dout->dArrayLength > (UINT32)glbTbl[i]->size) dout->dArrayLength = (UINT32)glbTbl[i]->size;
      dsize = (UINT16)(dout->dArrayLength*fmtsizeof(dout->dFormat));
      if (dout->data.vptr != NULL) memcpy(dout->data.vptr,glbTbl[i]->data,dsize);
      dout->dTimeStamp = (double)glbTbl[i]->time;
      ccerr(0);
    }
    glbTbl[i]->disabled = FALSE;
  }
  else 
  {
    i = nglobals;
  }
  if (nglobals >= MaxNumGlobals) ccerr(resources_exhausted);
  /* request for new global ... */
  if (glbTbl[i] == NULL && (glbTbl[i]=(GlobalListStruct *)SystemCalloc(1,sizeof(GlobalListStruct))) == NULL)
    ccerr(out_of_client_memory);
  if (callbackID == UNASSIGNED_CALLBACKID) callbackID = i; else glbTbl[i]->hasUserCallbackId = TRUE;
  strncpy(glbTbl[i]->context,ctx,CONTEXT_NAME_SIZE);
  strncpy(glbTbl[i]->keyword,kwd,KEYWORD_NAME_SIZE);
  glbTbl[i]->size = (short)dout->dArrayLength;
  glbTbl[i]->format = dout->dFormat;
  glbTbl[i]->fcn = callback;
  glbTbl[i]->callbackID = callbackID;
  glbTbl[i]->maxPeriod = GlobalHeartbeat;
  dsize = (UINT16)(dout->dArrayLength*fmtsizeof(dout->dFormat));
  if (dout->data.vptr == NULL && (dout->data.vptr=(BYTE *)SystemCalloc(1,dsize)) == NULL)
      ccerr(out_of_client_memory);
  glbTbl[i]->data = (BYTE *)dout->data.vptr;

  rc = ExecLinkEx(glbSrv,glbTbl[i]->keyword,dout,NULL,CA_READ|CA_RETRY,100);
  glbTbl[i]->time = rc == 0 ? time(NULL) : 0;

  fec = GetFecDataFromCache(ctx,GLOBALSRV);
  if (fec != NULL)
  { /* It's a known address */
    if (fec->TineProtocol < 6)
    { /* one general globals server */
      attachPortToMulticastGroup(GlobalSocket,(unsigned short)gtGCastPort,tGCastGenAddr,"globals common");
      glbTbl[i]->mcastGrp = tGCastGenAddr.s_addr;
    }
    else
    { /* context specific globals */
      memcpy(haddr,fec->IPh_addr,4);
      haddr[0] = (BYTE)238; haddr[1] = (BYTE)1; /* "a.b.c.d" -> "238.1.c.d" */
      memcpy(&inaddr,&haddr,4);
      attachPortToMulticastGroup(GlobalSocket,(unsigned short)gtGCastPort,inaddr,"globals specific");
      glbTbl[i]->mcastGrp = inaddr.s_addr;
    }  
  }
  glbTbl[i]->isPending = TRUE;
  isGrouped = (callbackID == CM_GROUPED) ? TRUE : FALSE;
  if (isGrouped && (gt=findGroup((size_t)callback)) == NULL) isGrouped = 0;
  glbTbl[i]->isGrouped = isGrouped;
  feclog("Listening for global parameter : %s",glbTbl[i]->keyword);
  nglobals++;
err:
  recvNetGlobalSemaphore = 0;
  return cc ? -cc : i;
}
TINE_EXPORT int recvNetGlobal(const char *keyword,DTYPE *dout,void (*callback)(int,int))
{
  return recvNetGlobalEx(keyword,dout,callback,UNASSIGNED_CALLBACKID);
}

TINE_EXPORT void SetClnRecvQueueDepth(int depth)
{
  gClnRecvQueueDepth = depth > 0 ? depth : 0;
}
TINE_EXPORT void SetLinkQueueDepth(int linkId,int depth)
{
  if (linkId < 0 || linkId > nConnectionTableEntries-1) return;
  if (conTbl[linkId] == NULL) return;
  conTbl[linkId]->queueDepth = depth > 0 ? depth : 0;
}
#ifndef SMALL_TINE_LIB /* for DOS, etc. ... */
TINE_EXPORT int ModifyLinkAttributes(int i,short access,int pollingRate,void (*callback)(int,int),int mode,UINT32 callbackID)
{
  int isGrouped,wasGrouped,baseMode,cc = 0;
  GroupTableEntry *gt = NULL;

  if (i < 0 || i >= nConnectionTableEntries) return invalid_index;
  if (conTbl[i] == NULL || conTbl[i]->mode == CM_CANCEL) return un_allocated;
  if ((wasGrouped=conTbl[i]->grouped) != 0)
  {
    if ((gt=findGroup((size_t)conTbl[i]->fcn)) == NULL) wasGrouped = 0;
  }
  if (mode == 0)
  {
    mode = conTbl[i]->mode;                                     /* keep it */
    isGrouped = conTbl[i]->grouped;                             /* keep it */
  }
  else
  {
    isGrouped = mode & CM_GROUPED; mode &= ~(CM_GROUPED);
    if ((mode & CM_BCAST || mode & CM_MCAST) && !ListenForMCasts)
    {
      if ((cc=ListenForMulticasts()) != 0) return cc;
    }
    baseMode = BASEMODE(mode);
    if (baseMode > CM_EVENT  || baseMode < CM_CANCEL || baseMode == CM_RECEIVE) return illegal_mode;
    if (baseMode == CM_REFRESH || baseMode == CM_EVENT) isGrouped = 0;
  }
  if (pollingRate == 0) pollingRate = (int) conTbl[i]->pollRate; /* keep it */
  if (access == 0) access = conTbl[i]->access;                   /* keep it */
  if (callbackID == UNASSIGNED_CALLBACKID) callbackID = conTbl[i]->callbackID;       /* keep it */
  if (callback == NULL) callback = conTbl[i]->fcn;               /* keep it */

  conTbl[i]->pollRate  = pollingRate;
  conTbl[i]->mode      = mode;
  conTbl[i]->fcn       = callback;
  conTbl[i]->callbackID= callbackID;
  conTbl[i]->heartbeat = mode == CM_REGISTER ? 0 : SUB_PENDING;
  conTbl[i]->access    = access;
  conTbl[i]->notifyPending   = 1;

  if (!wasGrouped && isGrouped)
  {
    if ((gt=findGroup((size_t)callback)) != NULL)
    {
      gt->numPending++;
      gt->numInGrp++;
      conTbl[i]->grouped = 1;
    }
  }
  else if (wasGrouped && !isGrouped)
  {
    if (gt != NULL)
    {
      gt->numPending--;
      gt->numInGrp--;
    }
    conTbl[i]->grouped = 0;
  }
  return 0;
}
TINE_EXPORT int SetNotificationTolerance(int linkId, float toleranceAbsolute, float tolerancePercent)
{
  if (toleranceAbsolute <= 0 && tolerancePercent <= 0) return 0; /* no reason to allocate a dtf */
  if (linkId < 0 || linkId > nConnectionTableEntries-1) return out_of_range;
  conTbl[linkId]->dtf = (DataToleranceFactor *)SystemCalloc(1,sizeof(DataToleranceFactor));
  if (conTbl[linkId]->dtf == NULL) return out_of_local_memory;
  conTbl[linkId]->dtf->absolute = toleranceAbsolute;
  conTbl[linkId]->dtf->percent = tolerancePercent;
  return 0;
}
TINE_EXPORT void SetSynchronizationTolerance(double toleranceInSeconds)
{
  if (toleranceInSeconds < 0.1) return; /* this would be ridiculous */
  gSyncTolerance = toleranceInSeconds;
}
#endif /* SMALL_TINE_LIB */
