/** \file clnip.c
 *
 * some client-side IP routines ...
 *
 */
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "srvcore.h"
#include "srvip.h"
#include "globals.h"
#include "dbtools.h"
#include "ttycmd.h"
#ifdef UNIX
#  ifdef NOTTY
int foregroundTTY = 0;
#  else
int foregroundTTY = 1;
#  endif
#endif

int ConsumerSocket = 0;
int MCastConsumerSocket = 0;
int ListenForMCasts = FALSE;
#ifndef LISTEN_FOR_GLOBALS
# define LISTEN_FOR_GLOBALS TRUE
#endif
#ifndef CLIENT_BURST_LIMIT
# define CLIENT_BURST_LIMIT 1000
#endif
int ListenForGlobals = LISTEN_FOR_GLOBALS;
struct timeval conto;
struct sockaddr_in srv_name;
int GlobalSocket = 0;
int defaultClientTimeout = SYSTEM_TICK;
int clientTimeout = 0;
int clientTimeoutMsec = 0;
int ReturnToClientCycle = 0;
int gInitIPClientDone = FALSE;
int gClientBurstLimit = CLIENT_BURST_LIMIT;
struct sockaddr_in IncomingIPfec;
extern IPXAddress IncomingIPXfec;
extern int gInitIPServicesDone;
IDLE_CONNECTION *tcpIdleConnectionList = NULL;

TINE_EXPORT void SetClientBurstLimit(int npackets)
{
  if (npackets >= 1) 
  {
    gClientBurstLimit = npackets;
  }
}
TINE_EXPORT void SetClientTimeout(int timeoutMilliseconds)
{
  if (defaultClientTimeout >= 0) 
  {
    clientTimeout = timeoutMilliseconds/1000;
    clientTimeoutMsec = timeoutMilliseconds%1000;
  }
}
void setupGlobalsSocket(int *psck,UINT16 p)
{
  if (listenUDP(psck,p,FALSE)) 
  {
    *psck = 0;
    feclog("cannot open %s socket",p == GCAST_PORT ? "GLOBALS" : "MCAST");
  }
}
/**
 * \internal
 * \brief Initialize file descriptor sets and bind sockets for an IP client.
 *
 *
 * Binds sockets for the consumer, MCast, and GCast,
 * and adds these sockets to conset.
 * 
 * Initialize the file descriptor set ext-set.
 *
 * Also adds stdin to conset if stdin is a tty.
 * 
 * \param port consumer port number.
 * \todo describe MCast and GCast.
 * \todo document global fd sets.
 */
int initIPClient(unsigned short port)
{
  int cc = 0;
  int optval = 1, optlen = sizeof(int);
# ifndef NOT_MULTICASTING
# ifdef WIN32
  int ttl = gtMCastTTL;
# else
  BYTE ttl = gtMCastTTL;
# endif
# endif
  if (defaultClientTimeout <= 0) defaultClientTimeout = 1000;
  if (gInitIPClientDone) return 0; gInitIPClientDone = TRUE;
# ifndef DOS
  TCPloaded = 1;
  if (ProducerSocket != 0) defaultClientTimeout = 0;
# ifdef WIN32
  if (!WinsockLoaded()) TCPloaded = 0;
  MaxPollingRate = 10; MinPollingRate = 10;
# endif
# else
  if (!TCPloaded) ccerr(tcp_not_supported);
  defaultClientTimeout = 0;
# endif
  clientTimeout = defaultClientTimeout/1000;
  clientTimeoutMsec = defaultClientTimeout%1000;
  if (MaxPollingRate > 1000) MaxPollingRate = 1000;
  if (SysPoll > 1000) SysPoll = 1000;
  /* set up UDP Consumer Sockets */
  conto.tv_sec = clientTimeout; conto.tv_usec = clientTimeoutMsec*1000;
  if (listenUDP(&ConsumerSocket,port,FALSE) != 0) ccerr(udp_socket_error);
# ifndef DOS
  if (setsockopt(ConsumerSocket,SOL_SOCKET,SO_BROADCAST,(char *)&optval,optlen))
  {
    feclog("cannot set broadcast option on consumer socket !");
  }
# endif /* DOS */
# ifndef NOT_MULTICASTING
  if (setsockopt(ConsumerSocket,IPPROTO_IP,IP_MULTICAST_TTL,(char *)&ttl,sizeof(ttl)) < 0)
  {
    feclog("cannot set multicast time to live on consumer socket!");
  }
# endif
  if (ListenForMCasts && MCastConsumerSocket == 0)
  {
    setupGlobalsSocket(&MCastConsumerSocket,(UINT16)gtMCastPort);
    if (MCastConsumerSocket == 0) ListenForMCasts = 0;
  }
  /* does it make sense to listen for globals ? */
  if (ListenForGlobals && GlobalSocket == 0) 
  {
    setupGlobalsSocket(&GlobalSocket,(UINT16)gtGCastPort);
    if (GlobalSocket == 0) ListenForGlobals = 0;
  }
err:
  if (cc)
  {
    feclog("Consumer socket : %s",erlst[cc]);
  }
  return cc;
}

int ListenForIPMulticasts(void)
{
  int cc = 0;
  if (ListenForMCasts) ccerr(0); /* is already listening for multicasts */
  ListenForMCasts = TRUE;
  initClient();
  if (!TCPloaded) ccerr(tcp_not_supported);
  setupGlobalsSocket(&MCastConsumerSocket,(UINT16)gtMCastPort);
  if (MCastConsumerSocket == 0) ccerr(tcp_socket_error);
# if defined(UNIX)
  gUseLoopback = TRUE;
# endif
err:
  if (cc) ListenForMCasts = FALSE;
  return cc;
}

int ListenForIPNetGlobals(void)
{
  int cc = 0;
  if (ListenForGlobals) ccerr(0)
  if (!TCPloaded) ccerr(tcp_not_supported);
  setupGlobalsSocket(&GlobalSocket,(UINT16)gtGCastPort);
  if (GlobalSocket == 0) ccerr(tcp_socket_error);
  ListenForGlobals = TRUE;
err:
  if (cc) ListenForGlobals = FALSE;
  return cc;
}

void closeIPClients(void)
{
  int i,done = 0;
  char str[256];
  if (!TCPloaded) return;
  if (soclose(ConsumerSocket)) soperror("UDP soclose");
  if (ListenForMCasts)
  {
    if (MCastConsumerSocket != 0 && soclose(MCastConsumerSocket)) soperror("UDP soclose");
  }
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (conTbl[i] == NULL) continue;
    if (FecTbl[conTbl[i]->fecIdx].TransportProtocol == TCP)
    {
      if (!done)
      {
        done = -1;
        shutdown(conTbl[i]->TCPsocket,2);
        soclose(conTbl[i]->TCPsocket);
      }
      conTbl[i]->TCPsocket = 0;
      sprintf(str,"TCP link to %s %s terminated",conTbl[i]->expName, conTbl[i]->prpName);
      feclog(str);
    }
  }
  return;
}

void closeIPConnectionToFec(int index)
{
  int i,done = 0;
  if (!TCPloaded) return;
  if (FecTbl[index].TransportProtocol != UDP) return;
  FecTbl[index].TransportProtocol = UDP;
  for (i=0; i<nConnectionTableEntries; i++)
  {
    if (conTbl[i] == NULL) continue;
    if (index != conTbl[i]->fecIdx) continue;
    if (!done)
    {
      done = -1;
      shutdown(conTbl[i]->TCPsocket,2);
      soclose(conTbl[i]->TCPsocket);
    }
    conTbl[i]->TCPsocket = 0;         /* signal for new connection */
    feclog("TCP link to %s %s terminated",conTbl[i]->expName, conTbl[i]->prpName);
  }
  return;
}
int connectTCP(int port,int i)  /* TCP client for connection table entry #i */
{
  ConnTableEntry *c;
  int fec,cc=0;
  int optval = 1, mode = 1;
  struct timeval tv0,tv;
  UINT16 poff;

  if ( i < 0 || i >= nConnectionTableEntries) return un_allocated;
  if ((c=conTbl[i]) == NULL) return un_allocated;
  fec = c->fecIdx;
  memset(&srv_name,0,sizeof(srv_name));
  memcpy(&srv_name.sin_addr,&FecTbl[fec].IPh_addr,sizeof(struct in_addr));
  srv_name.sin_family=AF_INET;
  poff = FecTbl[fec].PortOffset;
  srv_name.sin_port = htons((UINT16)(port + poff));
  if ((c->TCPsocket=(int)socket(AF_INET,SOCK_STREAM,0))<0)
  {
    soperror("socket");
    return tcp_socket_error;
  }
  if (setsockopt(c->TCPsocket,SOL_SOCKET,SO_KEEPALIVE,(char *)&optval,sizeof(optval)))
    ccerr(tcp_socket_error);
  if (soioctl(c->TCPsocket,FIONBIO,(char *)&mode)) ccerr(tcp_connect_error);

  gettimeofday(&tv0,(struct timezone *)NULL);  tv = tv0;

  while (connect(c->TCPsocket,(struct sockaddr *)&srv_name,sizeof(srv_name)) < 0)
  {
    if (soerrno == EISCONN) break;
    gettimeofday(&tv,(struct timezone *)NULL);
    if (MSECS(tv,tv0) > DEFAULT_TCP_CONNECTION_TIMEOUT) ccerr(connection_timeout);
  }
err:
  if (cc) 
  {
    soclose(c->TCPsocket);
    c->TCPsocket = 0;
  }
  return cc;
}
void addClientSocketsToSet(fd_set *fdset)
{
  int i;
  ConnTableEntry *c;
# ifdef UNIX
  int sck=0;
# endif

  if (ConsumerSocket) FD_SET(ConsumerSocket,fdset);
  if (ListenForMCasts) 
  {
    if (MCastConsumerSocket != 0) FD_SET(MCastConsumerSocket,fdset);
  }
  if (GlobalSocket) FD_SET(GlobalSocket,fdset);
  for (i=0; i<nConnectionTableEntries; i++)    /* possible TCP connections */
  {
    if ((c=conTbl[i]) != NULL && c->TCPsocket != 0) FD_SET(c->TCPsocket,fdset);
  }
# ifdef UNIX
# ifndef NO_TINE_PIPE
  for (i=0; !gUseLoopback && i<nConnectionTableEntries; i++)    /* PIPE connections */
  {
    if ((c=conTbl[i]) != NULL && c->fecIdx >= 0 && c->fecIdx < numFecTblEntries)
    {
      if (FecTbl[c->fecIdx].TransportProtocol != PIPE) continue;
      memcpy(&sck,FecTbl[c->fecIdx].IPXNetwork,4);
      if (sck > 0) FD_SET(sck,fdset);
    }
  }
# else
  sck = sck;
# endif
# endif
# ifdef IPXBSD
  FD_SET(IPXConsumerSocket,&fdset);
  if (ListenForMCasts && IPXMCastConsumerSocket) FD_SET(IPXMCastConsumerSocket,fdset);
# endif
}
void doUDPConsumer(void)
{
  short i,k,n=0,nread,nleft;
  int retry=2,nready;
  unsigned int len = sizeof(IncomingIPfec);
  ConnTableEntry *c;
  TCPBCKT *bckt;
  BYTE *bufptr=clnWorkArea;
  fd_set fdset;
# ifdef UNIX
  char cmdbuf[CMDSIZE];
  int sck;
# endif
  
  if (!TCPloaded) return;
  FD_ZERO(&fdset);
  addServerSocketsToSet(&fdset);
  addClientSocketsToSet(&fdset);
  if (ReturnToClientCycle || ReturnToServerCycle)
  {
    conto.tv_sec = conto.tv_usec = 0;
  }
  else
  {
    conto.tv_sec =  clientTimeout;
    conto.tv_usec = clientTimeoutMsec*1000;
  }

retry_select:
  for (k=0; (nready=select(maxFdSets,&fdset,NULL,NULL,&conto)) > 0 && k<gClientBurstLimit; k++)
  {
    conto.tv_usec = 0; conto.tv_sec = 0;
    IncomingNetworkProtocol = TCP;
    if (tineDebug) sprintf(fecprot,"TCP");
    for (i=0; i<nConnectionTableEntries; i++)    
    { /* possible TCP connections */
      if ((c=conTbl[i]) != NULL && c->TCPsocket != 0) recvIncomingIPData(c->TCPsocket);
    }
    IncomingNetworkProtocol = UDP;
    if (tineDebug) sprintf(fecprot,"UDP");
    /* check the bread-and-butter UDP socket */
    recvIncomingIPData(ConsumerSocket); n = 0;
    if (ListenForMCasts) 
    { /* something on the multicast socket ? */
      if (MCastConsumerSocket != 0) recvIncomingIPData(MCastConsumerSocket);
    }
    if (GlobalSocket > 0 && FD_ISSET(GlobalSocket,&fdset))
    { /* something on the globals socket ? */
      if ((nread=recvfrom(GlobalSocket,(char *)clnWorkArea,BRDCST_BUFFER_SIZE,0,(struct sockaddr *)&IncomingIPfec,(int *)&len)) > 0)
      {
        if (tineDebug) sprintf(fecprot,"GLB");
        if (nread == PeekWord(*(short *)clnWorkArea)) getGlobalUpdate(clnWorkArea);
      }
    }
#   if defined(UNIX) && !defined(NO_TINE_PIPE)
    for (i=0; !gUseLoopback && i<nConnectionTableEntries; i++)    
    { /* are there PIPE connections ? */
      if ((c=conTbl[i]) != NULL && c->fecIdx >= 0 && c->fecIdx < numFecTblEntries)
      {
        if (FecTbl[c->fecIdx].TransportProtocol != PIPE) continue;
        IncomingNetworkProtocol = PIPE;
        if (tineDebug) sprintf(fecprot,"LCL");
        memcpy(&sck,FecTbl[c->fecIdx].IPXNetwork,4);
        if (sck > 0) 
        { /* there is a pipe connection ! */
          recvIncomingIPData(sck);
          if (sck == 0) 
          { /* was there an error, socket reset to 0 ? */
            FecTbl[c->fecIdx].TransportProtocol = 0; 
            memset(FecTbl[c->fecIdx].IPXNetwork,0,4); 
            gUseLoopback = TRUE;
            feclog("pipe error : resort to loopback");
          }
        }
      }
    }
#   endif
#   ifdef UNIX
    if (foregroundTTY && FD_ISSET(STDIN_FILENO,&fdset) &&
        IPCSocket == 0 && GetCommandEx(STDIN_FILENO,cmdbuf,CMDSIZE) > 0)
    {
      InterpretCommand(cmdbuf);
    }
#   ifndef NO_TINE_PIPE
    for (i=0; !gUseLoopback && i<nConnectionTableEntries; i++)    
    { /* PIPE connections */
      if ((c=conTbl[i]) != NULL && c->fecIdx >= 0 && c->fecIdx < numFecTblEntries)
      {
        if (FecTbl[c->fecIdx].TransportProtocol != PIPE) continue;
        memcpy(&sck,FecTbl[c->fecIdx].IPXNetwork,4);
        if (sck > 0) FD_SET(sck,&fdset);
      }
    }
#   else
    sck = sck; /* aviod a warning */
#   endif /* !NO_TINE_PIPE */
#   endif /* UNIX */
#   ifdef IPXBSD
    if (IPXConsumerSocket && FD_ISSET(IPXConsumerSocket,&fdset)) break;
#   endif
    conto.tv_sec = conto.tv_usec = 0;
    FD_ZERO(&fdset);
    addClientSocketsToSet(&fdset);
  }
  if (k == gClientBurstLimit && tineDebug) dbglog("client burst limit reached");
  if (nready == -1)
  {
    if ((i=soerrno) != EWOULDBLOCK)
    {
      if (tineDebug > 2) soperror("UDP select");
      if (retry-- > 0) goto retry_select;
      if (tineDebug) dbglog("select socket error %d",i);
    }
  }
  ReturnToClientCycle = (k > 0) ? 1 : 0;
}
IDLE_CONNECTION *RemoveIdleConnection(IDLE_CONNECTION *ic)
{
  IDLE_CONNECTION *nxt=NULL;
  if (ic == NULL) return ic;
  nxt = ic->nxt;
  if (ic->prv != NULL) ic->prv->nxt = ic->nxt;
  else if (ic == tcpIdleConnectionList) tcpIdleConnectionList = ic->nxt;
  if (tineDebug) dbglog("TCP connection to %s (sck %d) no longer idle",FecTbl[ic->idx].Name,ic->sck);
  SystemFree(ic);
  return nxt;
}
IDLE_CONNECTION *AddIdleConnection(int sck,int idx)
{
  IDLE_CONNECTION *ic=NULL;

  if ((ic=getIdleConnection(idx)) != NULL) return ic;
  if ((ic=(IDLE_CONNECTION *)SystemCalloc(1,sizeof(IDLE_CONNECTION))) == NULL) return NULL;
  ic->sck = sck;
  ic->idx = idx;
  ic->ttl = 300000;
  ic->nxt = tcpIdleConnectionList;
  if (tcpIdleConnectionList != NULL) tcpIdleConnectionList->prv = ic;
  tcpIdleConnectionList = ic;
  if (tineDebug) dbglog("mark TCP connection to %s (sck %d) as idle",FecTbl[idx].Name,ic->sck);
  return ic;
}
IDLE_CONNECTION *getIdleConnection(int idx)
{
  IDLE_CONNECTION *ic=NULL;

  for (ic=tcpIdleConnectionList; ic != NULL; ic=ic->nxt)
  {
    if (ic->idx == idx) return ic;
  }
  return NULL;
}
IDLE_CONNECTION *getIdleConnectionBySocket(int sck)
{
  IDLE_CONNECTION *ic=NULL;

  for (ic=tcpIdleConnectionList; ic != NULL; ic=ic->nxt)
  {
    if (ic->sck == sck) return ic;
  }
  return NULL;
}
