/** \file srvdbase.c
 * \brief TINE server engine and registration API routines
 *
 * The API routines presented here deal almost exclusively with server-side registration calls.
 * All TINE servers must export their behavior in the form of properties, devices, server names,
 * etc.  The routines offered here provide ways to do this.  In most cases there are also 
 * alternatives based on startup configuration files such as fecid.csv, exports.csv, etc.  On
 * front-end systems where a local disk is available, this is the suggested registration 
 * method, since it avoids the hard coding of 'names' inside of the server programs themselves.
 *
 */
#ifndef NO_DOXYGEN
#include "prolog.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "srvcore.h"
#include "threader.h"
#include "srvip.h"
#include "srvipx.h"
#include "dbtools.h"
#include "errors.h"
#include "hstlib.h"
#include "almlib.h"
#include "globals.h"
#include "ttycmd.h"
#if defined(MAKE_TINE32_DLL) && !defined(NO_REMOTE_PIPE)
# include "shmalloc.h"
#endif

time_t appmktime = 0;
char appmkversion[64] = "1.00.0000";
TINE_EXPORT char gFecName[20] =  FECNAME_UNKNOWN;
TINE_EXPORT char gDeviceContext[20] = "";
TINE_EXPORT char gDeviceSubSystem[20] = "";
TINE_EXPORT int gScanForNameLists = SCAN_FOR_NAME_LISTS;
TINE_EXPORT int gScanForDeviceLists = SCAN_FOR_DEVICE_LISTS;
extern char *vFeclogBuffer;
extern NAME32 *gPrpList;
extern int gNeedToStartEqmThreads;
int queryDeviceNamesFromProperties = FALSE;
int acquireDeviceNamesFromProperties = FALSE;
int isUsingXMLFecDatabase = FALSE;
extern struct sockaddr_in IncomingIPfec;
char gFecDBSubDir[20] = "";
int seekAlternateExportsFile = FALSE; /* remove this at Release 4 time */
int gEqmListSize = 0;
int populateFIDS(DBLNKLIST *fecList,FIDS **fids);
int populateXIS(char *fecname,char *eqmname,DBLNKLIST *fecList,XIS **xis);
int populateDEVIS(char *fecname,char *eqmname,DBLNKLIST *fecList,DEVIS **devis);
int populatePRPDEVIS(char *fecname,char *eqmname,char *prpname,DBLNKLIST *fecList,DEVIS **devis);
int populateDEVPRPIS(char *fecname,char *eqmname,char *lstname,DBLNKLIST *fecList,NAME64 **prps);
int populateFIDNameList(DBLNKLIST *fecList,char *fecName,char *listName,NAME16 **lst);
int populateAIS(char *fecname,DBLNKLIST *fecList,ALIAS_TABLE **ais);
/**
 * \internal
 */
int updateFecManifest(ExportListStruct *el);

static CLN qrycln;
static CONSUMER qrycon;
static ContractListStruct qrycl;

TINE_EXPORT char *GetExportedContext(char *eqmName)
{
  static char contextBuffer[32];
  eqmName = eqmName; /* avoid warning until this is used */
  strncpy(contextBuffer,gDeviceContext,CONTEXT_NAME_SIZE);
  return contextBuffer;
}

TINE_EXPORT char *GetExportedSubSystem(char *eqmName)
{
  static char strBuffer[32];
  eqmName = eqmName; /* avoid warning until this is used */
  strncpy(strBuffer,gDeviceSubSystem,SUBSYSTEM_NAME_SIZE);
  return strBuffer;
}

TINE_EXPORT char *GetExportedFecName(void)
{
  static char fecNameBuffer[32];
  strncpy(fecNameBuffer,gFecName,FEC_NAME_SIZE);
  return fecNameBuffer;
}

TINE_EXPORT void _SystemFreeResources(int exitOnFree)
{
  closeServers(TRUE);
  ExitEQPModules();
# if !defined(DOS) && !defined(NIOS)
  freeServerMemory();
# endif
# if defined(WIN32) && !defined(NO_SERVICE_PIPE)
  if (IPCSocket != 0) CloseHandle(((PIPEINST *)IPCSocket)->hPipeInst);
# endif
  if (exitOnFree) exit(0);
}
TINE_EXPORT void haltServer(void)
{
  feclog("Stopping TINE Server ...");
  ServerExitCondition = 4;
  _SystemFreeResources(TRUE);
}

int adjustWorkAreaSize(UINT32 newSizeInBytes)
{
  int cc = 0;
  BYTE *m = (BYTE *)realloc(srvWorkArea,(size_t)newSizeInBytes);
  if (m == NULL) ccerr(out_of_server_memory);
  srvWorkArea = m;
  srvWorkAreaSize = MaxSystemTransportSize = newSizeInBytes;
  feclog("Work Space re-allocation to %u bytes",newSizeInBytes);
# ifdef INTEL
  tmpWorkArea = srvWorkArea;
  tmpWorkAreaSize = srvWorkAreaSize;
# endif
err:
  feclog("Work Space re-allocation to %u bytes: %s",newSizeInBytes,erlst[cc]);
  return cc;
}
void ParseRedirector(char *str,PrpRedirBlk *prd)
{
  char *c = str, *cstr = NULL, *sstr = NULL, *pstr = str, *dstr = NULL, *cp = NULL;

  if (str == NULL) return;
  if (str[0] == '/')
  { /* context given ? */
    *c++ = 0; cstr = c;
    if ((c=strchr(cstr,'/')) == NULL) return; /* device server must be there !*/
    *c++ = 0;
  }
  sstr = c;                                                /* device server */
  if ((c=strchr(sstr,'/')) != NULL)
  { /* device name given ? */
    *c++ = 0; dstr = c;
  }
  else
  { /* no, put the pointer back */
    c = sstr;
  }
  if ((pstr=strtok_r(c,"[",&cp)) != NULL) pstr = strtok_r(NULL,"]",&cp);
  memset(prd->prpRedirection,0,STATUS_SIZE);
  if (cstr != NULL) strncpy(&prd->prpRedirection[PRP_REDIR_CTX_OFFSET],cstr,CONTEXT_NAME_SIZE);
  if (sstr != NULL) strncpy(&prd->prpRedirection[PRP_REDIR_SRV_OFFSET],sstr,EXPORT_NAME_SIZE);
  if (dstr != NULL) strncpy(&prd->prpRedirection[PRP_REDIR_DEV_OFFSET],dstr,DEVICE_NAME_SIZE);
  if (pstr != NULL) strncpy(&prd->prpRedirection[PRP_REDIR_PRP_OFFSET],pstr,PROPERTY_NAME_SIZE);
}
void ParsePropertyDescription(char *str,ExportPropertyListStruct *prp)
{
  char *c=NULL, *c1, tmp[64], *s = str;

  if (str == NULL) return;
  if ((c=strchr(str,'[')) != NULL)
  {
    if ((c1=strchr(++c,']')) != NULL)               /* 1st token: egu */
    {
      *c1++ = 0;
      str = c1;         
      if ((c1=strstr(c,"!")) != NULL)                  /* graph type */
      {
        *c1++ = 0; prp->egu.graph = getGraphType(c1);
      }
      if ((c1=strchr(c,':')) != NULL)               /* prp egu given */
      {
        c1++;
        prp->egu.min = (float)strtod(c,&c); c = c1;
        prp->egu.max = (float)strtod(c,&c); c++;
        if (c != str)
        {
          strncpy(tmp,c,64);
          strncpy(prp->egu.units,tmp,16);
        }
      }
    }
  }
  if (str != NULL && (c=strchr(str,'[')) != NULL)
  {
    if ((c1=strchr(++c,']')) != NULL)             /* 2nd token: x egu */
    {
      *c1++ = 0;
      str = c1;         
      if ((c1=strstr(c,"!")) != NULL)                  /* graph type */
      {
        *c1++ = 0; prp->xegu.graph = getGraphType(c1);
      }
      if ((c1=strchr(c,':')) != NULL)                 /* x egu given */
      {
        c1++;
        prp->xegu.min = (float)strtod(c,&c); c = c1;
        prp->xegu.max = (float)strtod(c,&c); c++;
        if (c != str)
        {
          strncpy(tmp,c,64);
          strncpy(prp->xegu.units,tmp,16);
        }
      }
    }
  }
  if (str == NULL) str = s;
  strncpy(prp->prpDescription,str,PROPERTY_DESC_SIZE);  /* 3rd token: description part */
}
#ifdef VXWORKS
# ifndef VXWBKG_FLAGS
#   define VXWBKG_FLAGS VX_FP_TASK
# endif
# ifndef VXWBKG_STACK
#   define VXWBKG_STACK 10000
# endif
# ifndef VXWBKG_PRIORITY
#   define VXWBKG_PRIORITY 100
# endif
void VXWbkgtsk(void (*tsk)(void),int rate,int ref);
#endif
void makeEqpDevicesHashTable(ExportListStruct *el)
{
  HashXRefStruct *xref;
  char lclname[DEVICE_NAME_SIZE+4];
  int cc=0,idx,i;

  if (el == NULL || el->EqmDeviceXRef == NULL) return;
  if (el->EqmDeviceList == NULL) return;
  lclname[DEVICE_NAME_SIZE] = 0;
  for (i=0; i<el->EqmDeviceSpaceSize; i++)
  {
    strncpy(lclname,el->EqmDeviceList[i].name,DEVICE_NAME_SIZE);
    idx = (int)(ElfHash((unsigned char *)lclname) % DEVICES_HASH_SIZE);
    for (xref=el->EqmDeviceXRef[idx]; xref && xref->index != i; xref=xref->next);
    if (xref == NULL) /* it's a new one*/
    {
      if ((xref=(HashXRefStruct *)SystemCalloc(1,sizeof(HashXRefStruct))) == NULL) ccerr(out_of_server_memory);
      xref->index = i;
      xref->next = el->EqmDeviceXRef[idx];
      el->EqmDeviceXRef[idx] = xref;
    }
  }
err:
  if (cc) feclog("makeEqpDevicesHashTable: %s\n",erlst[cc]);
  return;
}

CSVCOLUMN csvFecNameColumns[] =
{
  { "FEC_NAME","",0,CF_TEXT,FEC_NAME_SIZE,(BYTE *)OFFSETIN(FIDS,fecName),NULL,0 },
  { "OS",OS_TAG,0,CF_TEXT,16,(BYTE *)OFFSETIN(FIDS,fecOs),NULL,0 },
  { "PORT_OFFSET","0",0,CF_LONG,1,(BYTE *)OFFSETIN(FIDS,fecPort),NULL,0 },
  { "CONTEXT","TEST",0,CF_TEXT,CONTEXT_NAME_SIZE,(BYTE *)OFFSETIN(FIDS,fecContext),NULL,0 },
  { "EXPORT_NAME","",0,CF_TEXT,EXPORT_NAME_SIZE,(BYTE *)OFFSETIN(FIDS,expNameTarget),NULL,0 },
  { "LOCAL_NAME","",0,CF_TEXT,EQM_NAME_SIZE,(BYTE *)OFFSETIN(FIDS,eqmNameTarget),NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,FEC_DESC_SIZE,(BYTE *)OFFSETIN(FIDS,fecDescription),NULL,0 },
  { "LOCATION","",0,CF_TEXT,FEC_LOCATION_SIZE,(BYTE *)OFFSETIN(FIDS,fecLocation),NULL,0 },
  { "HARDWARE","none",0,CF_TEXT,FEC_HDW_SIZE,(BYTE *)OFFSETIN(FIDS,fecHardware),NULL,0 },
  { "SUBSYSTEM","",0,CF_TEXT,SUBSYSTEM_NAME_SIZE,(BYTE *)OFFSETIN(FIDS,fecSubsystem),NULL,0 },
  { "RESPONSIBLE","",0,CF_TEXT,FEC_RESP_SIZE,(BYTE *)OFFSETIN(FIDS,responsible),NULL,0 },
  { "HISTORY_HOME","",0,CF_TEXT,HISTORY_HOME_SIZE,(BYTE *)OFFSETIN(FIDS,historyHome),NULL,0 },
};
CSVDB csvFecNameDb =
{
  csvFecNameColumns,
  sizeof(csvFecNameColumns)/sizeof(csvFecNameColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(FIDS)
};

int getSrvAddress(DTYPE *sa,char *eqm,char *devsrv)
{
  ExportListStruct *el=NULL;
  int len;
  char lDeviceServer[EXPORT_NAME_SIZE+4], lEqmName[EQM_NAME_SIZE], lSubName[SUBSYSTEM_NAME_SIZE], lCxtName[CONTEXT_NAME_SIZE+4];
  if (sa == NULL) return argument_list_error;
  switch (sa->dFormat)
  {
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME64:
      len = fmtsizeof(sa->dFormat);
      break;
    default:
      return illegal_format;
  }
  if (eqm != NULL && strlen(eqm) > 0)
  {
    for (el=ExportList; el != NULL; el=el->next) 
      if (!strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) break;
  }
  else if (devsrv != NULL && strlen(devsrv) > 0)
  {
    for (el=ExportList; el != NULL; el=el->next) 
      if (!strnicmp(devsrv,el->EqmExportName,EXPORT_NAME_SIZE)) break;
  }
  lSubName[0] = lDeviceServer[0] = lEqmName[0] = lCxtName[0] = 0;
  if (el != NULL)
  {
    strncpy(lDeviceServer,el->EqmExportName,EXPORT_NAME_SIZE);
    memset(lEqmName,0,EQM_NAME_SIZE);
    strncpy(lEqmName,el->EqmName,EQM_NAME_SHORTSIZE);
    strncpy(lSubName,el->EqmSubSystem,SUBSYSTEM_NAME_SIZE);
    strncpy(lCxtName,el->EqmContext,CONTEXT_NAME_SIZE);
  }
  switch (sa->dArrayLength)
  {
    case 6: strncpy(sa->data.cptr + 5*len,lSubName,SUBSYSTEM_NAME_SIZE);
    case 5: strncpy(sa->data.cptr + 4*len,lDeviceServer,MIN(EXPORT_NAME_SIZE,len));
    case 4: strncpy(sa->data.cptr + 3*len,lEqmName,EQM_NAME_SHORTSIZE);
    case 3: strncpy(sa->data.cptr + 2*len,lCxtName,MIN(CONTEXT_NAME_SIZE,len));
    case 2: strncpy(sa->data.cptr + len,gFecName,FEC_NAME_SIZE);
    case 1: sprintf(sa->data.cptr,"%d",gPortOffset);
      break;
    default: 
      return dimension_error;
  }
  return 0;
}
int getFecName(void)
{
  char *c,scratch[32];
  int i,siz=0,cc=0;
  FIDS *fids = NULL;
  ExportListStruct *el = NULL;
  BYTE *ver = SystemVersion();

  if (FecNameRegistered) return 0;
  gLastMinuteInitializationDone = FALSE;
  memset(gFecDBSubDir,0,20);

  gFecInfoList = listCreate();
  if ((cc=xmlReadFile(FecDBpath,FECXML_FILE,gFecInfoList)) == 0)
  {
    siz = populateFIDS(gFecInfoList,&fids);
    isUsingXMLFecDatabase = TRUE;
  }
  else
  {
    SystemFree(gFecInfoList);
    if ((cc=csvReadFile(FecDBpath,FECID_FILE,&csvFecNameDb,(void **)&fids)) != 0) ccerr(cc);
    siz = csvFecNameDb.siz;
  }
  for (i=0; i<siz; i++)
  {
    if (strlen(fids[i].expNameTarget) > 0)
    { /* target export name in fecid.csv file -> find a match */
      for (el=ExportList; el != NULL; el=el->next)
      {
        if (!strnicmp(fids[i].expNameTarget,el->EqmExportName,EXPORT_NAME_SIZE)) break;
      }
      if (el == NULL) continue;
      feclog("export entry %s found!",el->EqmExportName);
    }
    if (strlen(fids[i].eqmNameTarget) > 0)
    { /* target eqm (local) name in fecid.csv file -> find a match */
      for (el=ExportList; el != NULL; el=el->next)
      {
        if (!strnicmp(fids[i].eqmNameTarget,el->EqmName,EQM_NAME_SIZE)) break;
      }
      if (el == NULL) continue;
      feclog("local name entry %s found!",el->EqmName);
    }
    strncpy(gFecName,fids[i].fecName,FEC_NAME_SIZE);
    strncpy(gDeviceContext,fids[i].fecContext,CONTEXT_NAME_SIZE);
    strncpy(gDeviceSubSystem,fids[i].fecSubsystem,SUBSYSTEM_NAME_SIZE);
    strncpy(gFecInfo.os,fids[i].fecOs,FEC_OS_SIZE);
    strncpy(gFecInfo.desc,fids[i].fecDescription,FEC_DESC_SIZE);
    strncpy(gFecInfo.loc,fids[i].fecLocation,FEC_LOCATION_SIZE);
    strncpy(gFecInfo.hdw,fids[i].fecHardware,FEC_HDW_SIZE);
    strncpy(gFecInfo.resp,fids[i].responsible,FEC_RESP_SIZE);
    sprintf(scratch,"%d.%02d.%04d",ver[0],ver[1],ver[2]*256 + ver[3]);
    strncpy(gFecInfo.ver,scratch,FEC_VERSION_SIZE);
    gPortOffset = fids[i].fecPort;
    strncpy(gUserName,gFecName,USERNAME_SIZE);
    strncpy(ArchiveLocation,fids[i].historyHome,HISTORY_HOME_SIZE);
    if (strlen(ArchiveLocation) > 0) feclog("HISTORY HOME %s set from FEC configuration file",ArchiveLocation);
    FecNameRegistered = 1;
    break;
  }
  if (!FecNameRegistered)
  { /* reset these if no match found ... */
    strcpy(gFecName,"*unknown*");
    gDeviceContext[0] = 0;
    feclog("%s%s : %s",FecDBpath,FECID_FILE,erlst[cc]);
    feclog("unable to register as fec name is unknown!");
    ccerr(name_unknown);
  }
  feclog("accepting %s as FEC name",gFecName);
  if (strlen(gDeviceSubSystem) == 0)
  { /* subsystem not explicitly given; is it to be parsed from the description ? */
    if ((c=strchr(gFecInfo.desc,'[')) != 0)
    {
      strncpy(scratch,++c,20);
      if ((c=strchr(scratch,']')) != 0) *c=0;
      strncpy(gDeviceSubSystem,scratch,16);
      feclog("assigning FEC subsystem %s from description",gDeviceSubSystem);
    }
  }
err:
  if (fids != NULL) SystemFree(fids);
  return cc;
}
void checkAccessLocks(time_t t)
{
  if (currentAccessLock.lockDuration < 0) return;  /* forever */
  if (t > (time_t)(currentAccessLock.lockStart + currentAccessLock.lockDuration))
    memset(&currentAccessLock,0,sizeof(currentAccessLock));
}
int IsMemberLockSet(CLN *cln)
{
  if (!currentAccessLock.lockType) return TRUE; /* unlocked */
  if (strnicmp(currentAccessLock.cln.userName,cln->userName,USERNAME_SIZE)) return FALSE;
  if (memcmp(&currentAccessLock.cln.IPXaddress,&cln->IPXaddress,sizeof(IPXAddress)))
    return FALSE;
  if (memcmp(&currentAccessLock.cln.IPaddress,&cln->IPaddress,sizeof(struct sockaddr_in)))
    return FALSE;
  return TRUE;
}
int DENYaccess = 1;
int WRITEaccess = 0;
int NrOfUsers = 0;
int NrOfDeniedUsers = 0;
int MaxNumAllowedUsers = MAX_NUM_USERS;
NAME16 *gAllowedUsersList = NULL;
NAME16 *gDeniedUsersList = NULL;
int isDeniedUser(char *usr)
{
  int i;
  for (i=0; i<NrOfDeniedUsers && strnicmp(usr,gDeniedUsersList[i].name,USERNAME_SIZE); i++);
  return i == NrOfDeniedUsers ? 0 : -1;
}
int isAllowedUser(char *usr)
{
  int i;
  for (i=0; i<NrOfUsers; i++) 
    if (!strnicmp(usr,gAllowedUsersList[i].name,USERNAME_SIZE)) return TRUE;
  return FALSE;
}
int removeRegisteredUser(NAME16 *userlist,int listsize)
{
# if !defined(FS_RDONLY) && !defined(FS_NONE)
  FILE *fp = NULL, *tmpfp = NULL;
  int i,n=0,len,cc=0;
  char *c,str[256],userfilename[128],tmpfilename[128];

  if (userlist == NULL || listsize == 0) ccerr(argument_list_error);
  for (i=0; i<listsize; i++)
  {
    if (!isAllowedUser(userlist[i].name)) userlist[i].name[0] = 0;
    if (strlen(userlist[i].name) > 0) n++;
  }
  if (n == 0) ccerr(0); /* users weren't in the list anyway ! */

  sprintf(userfilename,"%s%s",FecDBpath,USERS_FILE);
  strncpy(tmpfilename,userfilename,128);
  if ((c=strstr(tmpfilename,".csv")) == NULL) ccerr(file_error);
  *c = 0; strcat(tmpfilename,".bak");
  unlink(tmpfilename);
  rename(userfilename,tmpfilename);

  if ((tmpfp=fopen(tmpfilename,"r")) == NULL) ccerr(no_such_file);
  if ((fp=fopen(userfilename,"w")) == NULL) ccerr(no_such_file);

  while (fgets(str,255,tmpfp) != NULL)
  {
    if ((len=(int)strlen(str)-1) <= 0) continue;
    for (i=0; i<listsize; i++) if (!strnicmp(str,userlist[i].name,len)) break;
    if (i < listsize) continue; /* skip, i.e. remove, this one */
    fwrite(str,strlen(str),1,fp);
  }
  fclose(fp); fp = NULL;
  fclose(tmpfp); tmpfp = NULL;
  getRegisteredUsers(&gAllowedUsersList,USERS_FILE,&NrOfUsers,"WRITE",&WRITEaccess);
err:
  if (fp != NULL) fclose(fp);
  if (tmpfp != NULL) fclose(tmpfp);
  return cc;
# else
  int i, n;
  if (userlist == NULL || listsize == 0) return argument_list_error;
  for (i=0; i<NrOfUsers; i++) 
  { /* clear those entries to be removed */
    for (n=0; n<listsize; n++) if (!strnicmp(gAllowedUsersList[i].name,userlist[n].name,USERNAME_SIZE)) break;
    if (n < listsize) gAllowedUsersList[i].name[0] = 0; /* clear this one */
  }
  for (i=0; i<NrOfUsers; i++) 
  { /* now collapse the list */
    if (strlen(gAllowedUsersList[i].name) == 0)
    {
      NrOfUsers--;
      if (i < NrOfUsers)
      {
        strncpy(gAllowedUsersList[i].name,gAllowedUsersList[i+1].name,USERNAME_SIZE);
        i--; 
      }
    }
  }
  return 0;
# endif
}
TINE_EXPORT int AppendRegisteredUserList(NAME16 *userlist,int listsize)
{
  int cc=0, i, n=0;
# ifndef FS_NONE /* normal case: there is a file system */
  FILE *fp = NULL;
  char userfilename[128],usr[USERNAME_SIZE+4];
  if (userlist != NULL && listsize > 0)
  { /* there's a list coming */
    for (i=0; i<listsize; i++)
    {
      if (isAllowedUser(userlist[i].name)) userlist[i].name[0] = 0;
      if (strlen(userlist[i].name) > 0) n++;
    }
    if (n == 0) return 0; /* everybody's already in the list */
    /* append the new guys to the file ... */
    sprintf(userfilename,"%s%s",FecDBpath,USERS_FILE);
    if ((fp=fopen(userfilename,"a")) == NULL) ccerr(no_such_file);
    fseek(fp,0,SEEK_END);
    if (ftell(fp) == 0)
    {
      fprintf(fp,"USERNAME\n");  /* file did not exist before */
    }
    for (i=0; i<listsize; i++)
    {
      if (strlen(userlist[i].name) == 0) continue;
      strncpy(usr,userlist[i].name,USERNAME_SIZE); usr[USERNAME_SIZE] = 0;
      fprintf(fp,"%s\n",usr);
    }
    fclose(fp);
  }
  getRegisteredUsers(&gAllowedUsersList,USERS_FILE,&NrOfUsers,"WRITE",&WRITEaccess);
# else
  if (gAllowedUsersList == NULL)
  {
    if ((gAllowedUsersList=(NAME16 *)SystemCalloc(MaxNumAllowedUsers,sizeof(NAME16))) == NULL)
      ccerr(out_of_server_memory);
    NrOfUsers = 0;
  }
  if (userlist == NULL || listsize == 0) ccerr(argument_list_error);
  for (i=0; i<listsize; i++)
  {
    if (isAllowedUser(userlist[i].name)) userlist[i].name[0] = 0;
    if (strlen(userlist[i].name) > 0) n++;
  }
  if (n == 0) return 0; /* everybody's already in the list */
  if (NrOfUsers == MaxNumAllowedUsers) ccerr(resources_exhausted);
  n = NrOfUsers;
  for (i=0; i<listsize && n<MaxNumAllowedUsers; i++)
  {
    if (strlen(userlist[i].name) == 0) continue;
    strncpy(gAllowedUsersList[n++].name,userlist[i].name,USERNAME_SIZE);
  }
  NrOfUsers = n;
# endif
err:
  return cc;
}
CSVCOLUMN csvUserFileColumns[] = 
{
  { "USER_NAME","",0,CF_TEXT,USERNAME_SIZE,NULL,NULL,0 }
};
CSVDB csvUserFileDb =
{
  csvUserFileColumns,
  sizeof(csvUserFileColumns)/sizeof(csvUserFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(NAME16)
};
int getRegisteredUsers(NAME16 **ul,char *fn,int *nr,char *accstr,int *accflg)
{
  int i,sz,fl,cc = 0;
  char *tag, usr[USERNAME_SIZE+4];

  if (isUsingXMLFecDatabase)
  {
    tag = !strcmp(accstr,"WRITE") ? USERS_ALLOWED_SETNAME : USERS_DENIED_SETNAME;
    sz = populateFIDNameList(gFecInfoList,gFecName,tag,ul);
    if (sz < 0) cc = -sz;
    fl = (sz == 0) ? 1 : 0;
  }
  else
  {
    cc = csvReadFile(FecDBpath,fn,&csvUserFileDb,(void **)ul);
    sz = csvUserFileDb.siz;
    fl = cc == 0 ? 0 : 1;
  }
  if (nr != NULL) *nr = sz;
  if (accflg != NULL) *accflg = fl;
  if (fl == 0)
  {
    for (i=0; i<sz; i++)
    {
      strncpy(usr,(*ul)[i].name,USERNAME_SIZE);
      feclog("%s has %s access",usr,accstr);
    }
    if (sz == 0)
    {
      strcpy(usr,!strcmp(accstr,"WRITE") ? "NONE" : "EVERYONE");
      feclog("%s has %s access",usr,accstr);
    }
  }
  else
  {
    strcpy(usr,!strcmp(accstr,"WRITE") ? "EVERYONE" : "NONE");
    feclog("%s has %s access",usr,accstr);
  }
  return cc;
}  
void dumpRegisteredUsers(void)
{
  int i;
  char usr[USERNAME_SIZE+4];
  if (WRITEaccess)
  {
    dbglog("Write Access open to all users\n");
    return;
  }
  dbglog("Write Access open to :\n");
  for (usr[USERNAME_SIZE]=0,i=0; i<NrOfUsers; i++)
  {
    strncpy(usr,gAllowedUsersList[i].name,USERNAME_SIZE);
    dbglog("\t%s\n",usr);
  }
}
int RegisterExport(char *expName,char *eqmName,int numdevices)
{
  int i, cc = 0;
  ExportListStruct *el;
  static int inside = 0;
  char expfilepath[64], expNameBuffer[128], *c, *exp = NULL, *ctx = NULL, *sub = NULL;

  if (inside) return -1; else inside = -1;

  if (eqmName == NULL || strlen(eqmName) == 0)
  {
    feclog("RegisterExport: missing local equipment module name");
    ccerr(argument_list_error);
  }
  feclog("RegisterExport : server (export) name -> %s, eqm (local) name -> %s, num devices -> %d",
    expName == NULL ? "(null)" : expName, eqmName, numdevices);
  if ((el=getExportListItem(eqmName)) != NULL) ccerr(0);        /* already in list */
  el = (ExportListStruct *)SystemCalloc(1,sizeof(ExportListStruct));
  if (el == NULL) ccerr(out_of_server_memory);
  strncpy(el->EqmName,eqmName,EQM_NAME_SHORTSIZE);
  if (expName == NULL || strlen(expName) == 0) 
  { /* registered without an export name -> find in config file ! */
#   ifdef FS_DELIMITER
    memset(gFecDBSubDir,0,20);
    strncpy(gFecDBSubDir,eqmName,EQM_NAME_SHORTSIZE);
    gFecDBSubDir[strlen(gFecDBSubDir)] = FS_DELIMITER;
    feclog("looking for exports for local name %s",eqmName);
    sprintf(expfilepath,"%s%cexports.csv",eqmName,FS_DELIMITER);
    GetAliasTable();
    if ((cc=getRegisteredExports(expfilepath,eqmName)) != 0) 
    { /* came up empty, so reset this */
      gFecDBSubDir[0] = 0; 
      ccerr(cc);
    }
    if (nAliasTableEntries == 0) GetAliasTable();
    ccerr(0);
#   else
    feclog("no export name for %s given !",eqmName);
    ccerr(argument_list_error);
#   endif
  }
  strncpy(expNameBuffer,expName,128);
  if (expNameBuffer[0] == '/')
  { /* context given */
    ctx = &expNameBuffer[1];
    if ((exp=strchr(ctx,'/')) == NULL) ccerr(argument_list_error);
    *exp++ = 0;
    if ((sub=strchr(ctx,'[')) != NULL)
    { /* subsystem given */
      *sub++ = 0;
      if ((c=strchr(sub,']')) != NULL) *c = 0;
    }
  }
  else
  { /* just the export */
    exp = expNameBuffer;
  }
  if (strlen(exp) == 0) ccerr(argument_list_error);
  if (ctx == NULL) ctx = gDeviceContext;
  if (sub == NULL) sub = gDeviceSubSystem;
  strncpy(el->EqmExportName,exp,EXPORT_NAME_SIZE);
  strncpy(el->EqmContext,ctx,CONTEXT_NAME_SIZE);
  strncpy(el->EqmSubSystem,sub,SUBSYSTEM_NAME_SHORTSIZE);
  if ((i=getAliasTableIndexFromName(el->EqmName)) >= 0)
    strncpy(el->EqmAlias,gAliasTable[i].alias,EQM_NAME_SHORTSIZE);
  el->EqmDeviceSpaceSize = numdevices;
  el->EqmNumDevices = numdevices;
  el->EqmDeviceList = GetEqpDeviceNames(el->EqmName,NULL,el->EqmNumDevices,&el->EqmDeviceRdrList,&el->EqmDevicePrpList);
  if (el->EqmDeviceList != NULL && strlen(el->EqmDeviceList[0].name) > 0) el->hasRegisteredDevicesNames = TRUE;
  if (el->EqmNumDevices >= DEVICES_HASH_PAYOFF)
  {
    if ((el->EqmDeviceXRef == NULL) &&
        (el->EqmDeviceXRef=(HashXRefStruct **)SystemCalloc(DEVICES_HASH_SIZE,sizeof(HashXRefStruct *))) == NULL)
      ccerr(out_of_server_memory);
    makeEqpDevicesHashTable(el);
  }
  el->EqmNumProperties = 0;
  if ((el->almLst=(ALARM **)SystemCalloc(el->EqmDeviceSpaceSize,sizeof(ALARM *))) == NULL)
    ccerr(out_of_server_memory);
  el->exportId = gEqmListSize++;
  el->next = ExportList;
  ExportList = el;

# ifdef WIN32
  if (seekAlternateExportsFile && !FecNameRegistered) 
  {
    getFecName();
    memset(gFecDBSubDir,0,20);
    strncpy(gFecDBSubDir,eqmName,16);
    gFecDBSubDir[strlen(gFecDBSubDir)] = FS_DELIMITER;
    feclog("looking for exports for local name %s",eqmName);
    sprintf(expfilepath,"%s%cexports.csv",eqmName,FS_DELIMITER);
    GetAliasTable();
    if (getRegisteredExports(expfilepath,eqmName) != 0) 
    { /* came up empty, so reset this */
      gFecDBSubDir[0] = 0; 
    }
    if (el->EqmDeviceList != NULL) SystemFree(el->EqmDeviceList);
    el->EqmDeviceList = GetEqpDeviceNames(el->EqmName,NULL,el->EqmNumDevices,&el->EqmDeviceRdrList,&el->EqmDevicePrpList);
    if (el->EqmDeviceList != NULL && strlen(el->EqmDeviceList[0].name) > 0) el->hasRegisteredDevicesNames = TRUE;
    if (strlen(el->EqmContext) == 0) strncpy(el->EqmContext,gDeviceContext,CONTEXT_NAME_SIZE);
    if (strlen(el->EqmSubSystem) == 0) strncpy(el->EqmSubSystem,gDeviceSubSystem,SUBSYSTEM_NAME_SHORTSIZE);
  }
# endif
  getAlarmInfoTable(eqmName);

err:
  if (!FecNameRegistered) getFecName();
  inside = 0;
  return cc;
}
TINE_EXPORT int GetNumberEquipmentModules(void)
{
  int n=0;
  ExportListStruct *el;
  for (el=ExportList; el != NULL; el=el->next) n++; 
  return n;
}

CSVCOLUMN csvExportsInfoColumns[] = 
{
  { "CONTEXT","",0,CF_TEXT,CONTEXT_NAME_SIZE,(BYTE *)OFFSETIN(XIS,EqmContext),NULL,0 },
  { "SUBSYSTEM","",0,CF_TEXT,SUBSYSTEM_NAME_SIZE,(BYTE *)OFFSETIN(XIS,SubSystem),NULL,0 },
  { "EXPORT_NAME","",0,CF_TEXT,EXPORT_NAME_SIZE,(BYTE *)OFFSETIN(XIS,ExpName),NULL,0 },
  { "LOCAL_NAME","",0,CF_TEXT,EQM_NAME_SIZE,(BYTE *)OFFSETIN(XIS,EqmName),NULL,0 },
  { "NUM_DEVICES","1",0,CF_LONG,1,(BYTE *)OFFSETIN(XIS,DevSpace),NULL,0 },
  { "NUM_MODULES","1",0,CF_LONG,1,(BYTE *)OFFSETIN(XIS,DevSpace),NULL,0 },
  { "PROPERTY","",0,CF_TEXT,PROPERTY_NAME_SIZE,(BYTE *)OFFSETIN(XIS,PrpName),NULL,0 },
  { "PROPERTY_SIZE","0",0,CF_LONG,1,(BYTE *)OFFSETIN(XIS,PrpSizeOut),NULL,0 },
  { "PROPERTY_INSIZE","0",0,CF_LONG,1,(BYTE *)OFFSETIN(XIS,PrpSizeIn),NULL,0 },
  { "PROPERTY_ID","-1",0,CF_LONG,1,(BYTE *)OFFSETIN(XIS,PrpId),NULL,0 },
  { "FORMAT","float",0,CF_TEXT,FORMAT_STR_SIZE,(BYTE *)OFFSETIN(XIS,fmtOut),NULL,0 },
  { "INFORMAT","",0,CF_TEXT,FORMAT_STR_SIZE,(BYTE *)OFFSETIN(XIS,fmtIn),NULL,0 },
  { "ACCESS","READ",0,CF_TEXT,ACCESS_STR_SIZE,(BYTE *)OFFSETIN(XIS,access),NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,128,(BYTE *)OFFSETIN(XIS,description),NULL,0 },
  { "REDIRECTION","",0,CF_TEXT,PROPERTY_REDIR_SIZE,(BYTE *)OFFSETIN(XIS,redirection),NULL,0 },
  { "GROUP","",0,CF_TEXT,EXPORT_NAME_SIZE,(BYTE *)OFFSETIN(XIS,EqmGroup),NULL,0 },
  { "GROUP_INDEX","",0,CF_LONG,1,(BYTE *)OFFSETIN(XIS,EqmGroupIdx),NULL,0 }
};
CSVDB csvExportsInfoDb =
{
  csvExportsInfoColumns,
  sizeof(csvExportsInfoColumns)/sizeof(csvExportsInfoColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(XIS)
};

int getRegisteredExports(char *fname,char *eqm)
{
  int i, k, idx, sz, cc=0;
  SINT32 prpId;
  UINT32 dSizeInBytes = 0, tSizeInBytes;
  char *c, *ctx, *sub;
  XIS *xis = NULL;
  ExportPropertyListStruct *prp;
  ExportListStruct *el;
  PrpRedirBlk *prd = NULL;

  if (isUsingXMLFecDatabase)
  {
    sz = populateXIS(gFecName,eqm,gFecInfoList,&xis);
    if (sz < 0) ccerr(-sz);
  }
  else
  {
    if ((cc=csvReadFile(FecDBpath,fname,&csvExportsInfoDb,(void **)&xis)) != 0) ccerr(cc);
    sz = csvExportsInfoDb.siz;
  }
  feclog("get registered exports from file (%d entries)",sz);
  for (i=0; i<sz; i++)
  { /* prepare export list table */
    if (strlen(xis[i].ExpName) == 0 || strlen(xis[i].EqmName) == 0) 
    { /* not accepted */
      feclog("%s : entry %d (property %s) was not assigned an export or local equipment name",i,xis[i].PrpName);
      continue; 
    }
    for (el=ExportList; el != NULL; el=el->next) 
    {
      if (!strncmp(xis[i].EqmName,el->EqmName,EQM_NAME_SHORTSIZE)) 
      {
        if (strnicmp(xis[i].ExpName,el->EqmExportName,EXPORT_NAME_SIZE))
        {
          feclog("configuration error: module %s exported as %s is attempting to export as %s",
            el->EqmName,el->EqmExportName,xis[i].ExpName);
        }
        break;
      }
      if (!strnicmp(xis[i].ExpName,el->EqmExportName,EXPORT_NAME_SIZE)) 
      {
        if (strncmp(xis[i].EqmName,el->EqmName,EQM_NAME_SHORTSIZE))
        {
          feclog("configuration error: module %s exported as %s is attempting to assign module name %s",
            el->EqmName,el->EqmExportName,xis[i].EqmName);
        }
        break;
      }
    }
    if (el == NULL)               
    { /* new eqm module */
      el = (ExportListStruct *)SystemCalloc(1,sizeof(ExportListStruct));
      if (el == NULL) ccerr(out_of_server_memory);
      ctx = strlen(xis[i].EqmContext) > 0 ? xis[i].EqmContext : gDeviceContext;
      sub = strlen(xis[i].SubSystem) > 0 ? xis[i].SubSystem : gDeviceSubSystem;
      strncpy(el->EqmContext,ctx,CONTEXT_NAME_SIZE);
      strncpy(el->EqmSubSystem,sub,SUBSYSTEM_NAME_SIZE);
      el->exportId = gEqmListSize++;
      el->next = ExportList;
      ExportList = el;
    }
    if (strlen(el->EqmExportName) == 0)
    {
      if (el->EqmDeviceList != NULL) SystemFree(el->EqmDeviceList);
      if (el->almLst != NULL) SystemFree(el->almLst);
      strncpy(el->EqmExportName,xis[i].ExpName,EXPORT_NAME_SIZE);
      strncpy(el->EqmName,xis[i].EqmName,EQM_NAME_SIZE);
      if (xis[i].DevSpace == 0) xis[i].DevSpace = 1;
      el->EqmDeviceSpaceSize = el->EqmNumDevices = xis[i].DevSpace;
      el->EqmDeviceList = GetEqpDeviceNames(el->EqmName,NULL,el->EqmNumDevices,&el->EqmDeviceRdrList,&el->EqmDevicePrpList);
      if (el->EqmDeviceList != NULL && strlen(el->EqmDeviceList[0].name) > 0) el->hasRegisteredDevicesNames = TRUE;
      if (el->EqmDeviceSpaceSize >= DEVICES_HASH_PAYOFF)
      {
        if ((el->EqmDeviceXRef == NULL) &&
            (el->EqmDeviceXRef=(HashXRefStruct **)SystemCalloc(DEVICES_HASH_SIZE,sizeof(HashXRefStruct *))) == NULL)
          ccerr(out_of_server_memory);
        makeEqpDevicesHashTable(el);
      }
      el->EqmNumProperties = 0;
      if ((el->almLst=(ALARM **)SystemCalloc(el->EqmDeviceSpaceSize,sizeof(ALARM *))) == NULL)
        ccerr(out_of_server_memory);
    }
    if ((prp=(ExportPropertyListStruct *)SystemCalloc(1,sizeof(ExportPropertyListStruct))) == NULL) ccerr(out_of_server_memory);
    strncpy(prp->prpName,xis[i].PrpName,PROPERTY_NAME_SIZE);
    if ((k=getAliasTableIndexFromName(prp->prpName)) >= 0)
      strncpy(prp->prpAlias,gAliasTable[k].alias,PROPERTY_NAME_SIZE);
    prp->prpSize = xis[i].PrpSizeOut;
    if (!strnicmp(xis[i].fmtOut,"STRUCT.",7)) /* tag coming ... */
    {
      if ((c=strstr(xis[i].fmtOut,".")) == NULL) ccerr(code_failure);
      *c++ = 0;
      strncpy(prp->prpFormatTag.name,c,TAG_NAME_SIZE);
    }
    if (!strnicmp(xis[i].fmtOut,"BITFIELD",8)) /* tag coming ? ... */
    {
      if ((c=strstr(xis[i].fmtOut,".")) != NULL)
      {
        *c++ = 0;
        strncpy(prp->prpFormatTag.name,c,TAG_NAME_SIZE);
      }
    }
    prp->prpFormat = ftoi(xis[i].fmtOut);
    tSizeInBytes = prp->prpSize * fmtsizeof(prp->prpFormat);
    if ( tSizeInBytes > dSizeInBytes) dSizeInBytes = tSizeInBytes;
    prp->prpArrayType = getArrayType(xis[i].fmtOut);
    if (!(prp->prpArrayType & AT_DOUBLE)) prp->prpArrayType |= prp->prpSize > 1 ? AT_SINGLE : AT_SCALAR;
    prp->rowSize = getArrayRowLength(xis[i].fmtOut);
    if (prp->rowSize > 0) prp->numRows = (UINT16)(prp->prpSize/prp->rowSize); else prp->rowSize = (short)prp->prpSize;
    prp->prpAccessMode = mtoi(xis[i].access);
    prp->prpSizeIn = xis[i].PrpSizeIn;
    prp->prpFormatIn = CF_NULL;
    if (strlen(xis[i].fmtIn) > 0)
    {
      if (!strnicmp(xis[i].fmtIn,"STRUCT.",7)) /* tag coming ... */
      {
        if ((c=strstr(xis[i].fmtIn,".")) == NULL) ccerr(code_failure);
        *c++ = 0;
        strncpy(prp->prpFormatTagIn.name,c,TAG_NAME_SIZE);
      }
      prp->prpFormatIn = ftoi(xis[i].fmtIn);
    }
    else
    {
      if (prp->prpAccessMode & CA_WRITE) prp->prpFormatIn = prp->prpFormat;
    }
    ParsePropertyDescription(xis[i].description,prp);
    if (strlen(xis[i].EqmGroup) > 0)
    {
      strncpy(el->EqmGroup,xis[i].EqmGroup,EXPORT_NAME_SIZE);
      el->groupIndex = xis[i].EqmGroupIdx;
    }
    if (strlen(xis[i].redirection) > 0)
    {
      if ((prd=(PrpRedirBlk *)SystemCalloc(1,sizeof(PrpRedirBlk))) == NULL) ccerr(out_of_server_memory);
      prd->next = prp->prd;
      prp->prd = prd;
      ParseRedirector(xis[i].redirection,prp->prd);
    }
    if (strchr(prp->prpName,'*')) idx = PROPERTY_HASH_SIZE;
    else idx = (int)(ElfHash((unsigned char *)prp->prpName) % PROPERTY_HASH_SIZE);
    prpId = xis[i].PrpId >= 0 ?  xis[i].PrpId: el->EqmNumProperties;
    prp->prpId = MAKE_PROPERTY_ID(prpId,idx);
    prp->next = el->EqmProperty[idx];
    el->EqmProperty[idx] = prp;
    if ((k=(int)strlen(prp->prpName)) > 4 && !strcmp(&prp->prpName[k-4],".NAM")) 
    {
      el->hasPropertyDeviceNames = TRUE;
      if (queryDeviceNamesFromProperties) acquireDeviceNamesFromProperties = TRUE;
    }
    prp->numDevices = xis[i].DevSpace;
    if ((prp->devNames=GetEqpDeviceNames(eqm,prp->prpName,prp->prpSize,NULL,NULL)) != NULL)
    {
      if (prp->prpArrayType & AT_CHANNEL) prp->numDevices = prp->prpSize;
      el->hasPropertyDeviceNames = TRUE;
    }
    el->EqmNumProperties++;
  }

  getAlarmInfo();

# if !defined(VXWORKS)
  if (dSizeInBytes > MaxSystemTransportSize)
  { /* need to make adjustments */
    adjustWorkAreaSize(dSizeInBytes+0x10000);
  }
# else
  feclog("maximum property size (%d bytes) larger than registered work area (%d bytes)",dSizeInBytes,MaxSystemTransportSize);
# endif

err:
  if (xis != NULL) SystemFree(xis);
  feclog("exports file %s : %s",fname, cc ? erlst[cc] : "found");
  if (cc) seekAlternateExportsFile = TRUE;
  return cc;
}
ExportListStruct *getExportListItem(char *eqm)
{
  ExportListStruct *el;

  if (eqm == NULL) return NULL;
  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) return el;
  return NULL;
}
ExportListStruct *getExportListItemFromId(int id)
{
  ExportListStruct *el;

  for (el=ExportList; el != NULL; el=el->next) if (el->exportId == id) return el;
  return NULL;
}
int GetEqpPropertyList(char *eqm,char *filename,EqmDevicePropertyList *prpList)
{
  NAME64 *prps = NULL;
  int n,i=0,cc=0,done=0,name_col=-1;
  char str[256],hdr[256],fn[128];
  FILE *fp = NULL;
  
  if (filename == NULL || prpList == NULL) ccerr(argument_list_error);

  if (isUsingXMLFecDatabase)
  {
    n = populateDEVPRPIS(gFecName,eqm,filename,gFecInfoList,&prps);
    if (n <= 0) ccerr(-n);
  }
  else
  {
    sprintf(fn,"%s%s",FecDBpath,filename);
    if ((n=GetDataBaseSize(fn)) < 0) ccerr(-n);
    if ((prps=(NAME64 *)SystemCalloc(n,sizeof(NAME64))) == NULL) ccerr(out_of_server_memory);
    if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
    for (done=0; fgets(str,255,fp) != NULL;)   /* scan file       */
    {
      strtrm(str);
      if (strlen(str) == 0) continue;          /* carriage return */
      if (strchr("%;#\n",str[0])) continue;    /* comment line(s) */
      if (!done)
      {
        strncpy(hdr,str,255);
        if ((name_col=findcol(hdr,"PROPERTY_NAME")) < 0) ccerr(no_such_column);
        done = 1;                              /* read header info */
        continue;
      }
      getColumnStr(name_col,str,prps[i++].name,64);
    }
  }
  strncpy(prpList->filename,filename,64);
  prpList->prpList = prps;
  prpList->prpListSize = n;
err:
  if (fp) fclose(fp);
  feclog("property list %s : %s",filename, cc ? erlst[cc] : "found");
  return cc;
}

CSVCOLUMN csvDevicesColumns[] =
{
  { "DEVICE_NAME","",0,CF_TEXT,DEVICE_NAME_SIZE,(BYTE *)OFFSETIN(DEVIS,devname),NULL,0 },
  { "DEVICE_NUMBER","0",0,CF_LONG,1,(BYTE *)OFFSETIN(DEVIS,devnum),NULL,0 },
  { "REDIRECTION","",0,CF_TEXT,32,(BYTE *)OFFSETIN(DEVIS,devrdr),NULL,0 },
  { "PROPERTY_LIST","",0,CF_TEXT,32,(BYTE *)OFFSETIN(DEVIS,prplst),NULL,0 },
};
CSVDB csvDevicesDb =
{
  csvDevicesColumns,
  sizeof(csvDevicesColumns)/sizeof(csvDevicesColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(DEVIS)
};

int getDevicesFromFile(char *fname,DEVIS **devis,int *size)
{
  int cc;
  if ((cc=csvReadFile(FecDBpath,fname,&csvDevicesDb,(void **)devis)) != 0) ccerr(cc);
  if (size != NULL) *size = csvDevicesDb.siz;
  feclog("file %s found",fname);
err:
  return cc;
}
char *makeDeviceFileName(char *eqm,int decorate)
{
  static char fname[64];
  char dname[32];

  if (decorate == 0)
  {
    sprintf(fname,"%sdevices.csv",gFecDBSubDir);
  }
  else
  {
    memset(dname,0,sizeof(dname));
    strncpy(dname,eqm,EQM_NAME_SHORTSIZE);
    strlwr(dname);
#   if !defined(FS_SHORTNAMES)
    strcat(dname,"-devices.csv");
#   else
    strcat(dname,"-d.csv");
#   endif
    sprintf(fname,"%s%s",gFecDBSubDir,dname);
  }
  return fname;
}
char *makePropertyFileName(char *prp)
{
  static char fname[64];
  sprintf(fname,"%s%s-names.csv",gFecDBSubDir,prp);
  return fname;
}
NAME64 *GetEqpDeviceNames(char *eqm,char *prpname,int num,EqmDeviceRdrListType **rdrnames,EqmDevicePropertyList **prplist)
{
  NAME64 *eqpDeviceList = NULL;
  int i,k,idx,cc=0,decorate=0,size=0;
  char *fname,*tgt="[unknown]";
  PrpRedirBlk prd;
  EqmDevicePropertyList *pl;
  DEVIS *devis = NULL;

  if (num <= 0) return NULL;

  if (eqm != NULL && prpname == NULL)
  { /* common device list */
    if ((eqpDeviceList=(NAME64 *)SystemCalloc(num,sizeof(NAME64))) == NULL) ccerr(out_of_server_memory);
    if (isUsingXMLFecDatabase)
    {
      size = populateDEVIS(gFecName,eqm,gFecInfoList,&devis);
      if (size <= 0) ccerr(-size);
    }
    else
    {
      if (!gScanForDeviceLists) ccerr(no_such_file);
retryEqmDeviceFile:
      fname = makeDeviceFileName(eqm,decorate);
      if ((cc=getDevicesFromFile(fname,&devis,&size)) == no_such_file && decorate++ == 0) goto retryEqmDeviceFile;
      if (cc != 0) ccerr(cc);
      if (devis == NULL) ccerr(code_failure);
    }
    tgt = eqm;
  }
  else if (prpname != NULL)
  { /* property-specific device list */
    if (isUsingXMLFecDatabase)
    {
      size = populatePRPDEVIS(gFecName,eqm,prpname,gFecInfoList,&devis);
      if (size <= 0) ccerr(-size);
    }
    else
    {
      if (!gScanForNameLists) ccerr(no_such_file);
      fname = makePropertyFileName(prpname);
      cc = getDevicesFromFile(fname,&devis,&size);
      if (cc != 0) ccerr(cc);
    }
    if (devis == NULL) ccerr(code_failure);
    if ((eqpDeviceList=(NAME64 *)SystemCalloc(num,sizeof(NAME64))) == NULL) ccerr(out_of_server_memory);
    tgt = prpname;
  }
  else
    ccerr(argument_list_error);

  for (i=0; i<size; i++)
  {
    idx = devis[i].devnum;
    if (idx < 0 || idx >= num) continue;
    strncpy(eqpDeviceList[idx].name,devis[i].devname,DEVICE_NAME_SIZE);
    if (strlen(devis[i].devrdr) > 0 && rdrnames != NULL)
    {
      if (*rdrnames == NULL)
      { /* haven't allocated space yet */
        if ((*rdrnames=(EqmDeviceRdrListType *)SystemCalloc(num,sizeof(EqmDeviceRdrListType))) == NULL) 
          ccerr(out_of_server_memory);
      }
      if (((*rdrnames)[idx].rdrName=(char *)SystemCalloc(1,CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE+DEVICE_NAME_SIZE)) == NULL)
        ccerr(out_of_server_memory);
      ParseRedirector(devis[i].devrdr,&prd); /* only the device server name is relevant */
      strncpy((*rdrnames)[idx].rdrName,&prd.prpRedirection[PRP_REDIR_CTX_OFFSET],CONTEXT_NAME_SIZE); 
      strncpy(&(*rdrnames)[idx].rdrName[CONTEXT_NAME_SIZE],&prd.prpRedirection[PRP_REDIR_SRV_OFFSET],EXPORT_NAME_SIZE); 
      strncpy(&(*rdrnames)[idx].rdrName[CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE],&prd.prpRedirection[PRP_REDIR_DEV_OFFSET],DEVICE_NAME_SIZE); 
    }
    if (strlen(devis[i].prplst) > 0 && prplist != NULL)
    { 
      if (*prplist == NULL)
      { /* haven't allocated space yet */
        if ((*prplist=(EqmDevicePropertyList *)SystemCalloc(num,sizeof(EqmDevicePropertyList))) == NULL) 
          ccerr(out_of_server_memory);
      }
      pl = *prplist;
      for (k=0; k<i; k++) if (!strncmp(pl[k].filename,devis[i].prplst,64)) break;
      if (k < i) /* already have the list for this one */
      {
        strncpy(pl[i].filename,pl[k].filename,64);
        pl[i].prpList = pl[k].prpList;
        pl[i].prpListSize = pl[k].prpListSize;
      }
      else /* don't have the list, get it ! */
      {
        GetEqpPropertyList(eqm,devis[i].prplst,&pl[i]);
      }
    }
  }
err:
  if (devis != NULL) SystemFree(devis);
  if (cc != no_such_file) feclog("device file for %s : %s",tgt, cc ? erlst[cc] : "found");
  return eqpDeviceList;
}
int lookupPropertyName(ExportListStruct *el,char *prpName,char *devName)
{
  ExportPropertyListStruct *prp;
  PrpRedirBlk *prd = NULL;
  int idx, redir = FALSE;

  if (el == NULL || prpName == NULL) return -1;
  if (el->EqmDeviceRdrList != NULL)
  { /* if not NULL there's a device-specific redirection list */
    if ((idx=GetDeviceNumberEx(el->EqmName,devName,NULL)) >= 0 && idx < el->EqmNumDevices)
    { /* given device was found */
      if (el->EqmDeviceRdrList[idx].rdrName != NULL && strlen(&el->EqmDeviceRdrList[idx].rdrName[CONTEXT_NAME_SIZE]) > 0)
      { /* redirect, regardless of property */
        SetRedirectionParameters(el->EqmName,el->EqmDeviceRdrList[idx].rdrName,
                                 &el->EqmDeviceRdrList[idx].rdrName[CONTEXT_NAME_SIZE],
                                 &el->EqmDeviceRdrList[idx].rdrName[CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE],NULL);
        return -server_redirection;
      }
    }
  }
  idx = (int)(ElfHash((unsigned char *)prpName) % PROPERTY_HASH_SIZE);
  for (prp=el->EqmProperty[idx]; prp; prp=prp->next)
    if (!strnicmp(prp->prpName,prpName,PROPERTY_NAME_SIZE)) 
    { /* property is registered */
      if (prp->prd != NULL)
      {
        for (prd=prp->prd; prd != NULL; prd=prd->next)
        { /* property is redirected, but is it true for the device name given ? */
          if (prd->prpRedirection[PRP_REDIR_DEV_OFFSET] == 0) redir = TRUE;
          if (prd->prpRedirection[PRP_REDIR_DEV_OFFSET] != 0 &&
              !strnicmp(&prd->prpRedirection[PRP_REDIR_DEV_OFFSET],devName,DEVICE_NAME_SIZE)) redir = TRUE;
          if (redir)
          { /* there's a match, redirect it */
            memcpy(el->EqmLastErrorString,prd->prpRedirection,STATUS_SIZE);
            return -server_redirection;
          }
        }
      }
      return (int)PROPERTY_HASH_INDEX(prp->prpId);
    }
  return -1;
}
char *lookupPropertyId(ExportListStruct *el,short id)
{
  ExportPropertyListStruct *prp;
  int idx;

  if (el == NULL) return (char *)NULL;
  idx = PROPERTY_HASH_INDEX(id);
  for (prp=el->EqmProperty[idx]; prp; prp=prp->next)
    if (prp->prpId == id) return prp->prpName;
  return (char *)NULL;
}
int RegisterLocalAlias(char *eqmName,char *alias)
{
  int cc=0;
  ExportListStruct *el;

  if ((el=getExportListItem(eqmName)) == NULL) ccerr(non_existent_elem);
  strncpy(el->EqmAlias,alias,EQM_NAME_SHORTSIZE);
  feclog("Register local alias %s for %s",alias,eqmName);
err:
  return cc;
}
int joinEnsGroup(ExportListStruct *el)
{
  int cc=0,pass=0;
  NAME32 n32data[5];
  NAME16 *n16data = (NAME16 *)n32data;
  DTYPE d;

  if (el == NULL) return non_existent_elem;

retryIllegalFormat:
  if (pass == 0)
  {
    strncpy(n32data[0].name,el->EqmExportName,32); 
    strncpy(n32data[1].name,el->EqmGroup,32); 
    strncpy(n32data[2].name,el->EqmContext,32); 
    sprintf(n32data[3].name,"%d",el->groupIndex); 
    strncpy(n32data[4].name,el->EqmSubSystem,32); 
    d.dArrayLength = 5;
    d.dFormat = CF_NAME32;
    memset(d.dTag,0,TAG_NAME_SIZE);
    d.data.cptr = n32data[0].name;
  }
  else
  {
    strncpy(n16data[0].name,el->EqmExportName,16); 
    strncpy(n16data[1].name,el->EqmGroup,16); 
    strncpy(n16data[2].name,el->EqmContext,16); 
    sprintf(n16data[3].name,"%d",el->groupIndex); 
    strncpy(n16data[4].name,el->EqmSubSystem,16); 
    d.dArrayLength = 5;
    d.dFormat = CF_NAME16;
    memset(d.dTag,0,TAG_NAME_SIZE);
    d.data.cptr = n16data[0].name;
  }
  cc = tExecLink("GENS","GRPEQM","#0","JOIN",&d,&d,CA_WRITE,500);
  if (cc == illegal_format && pass == 0)
  { /* legacy GENS ? */
    pass++;
    goto retryIllegalFormat;
  }
  if (cc != 0) feclog("%s joining group %s failed : %s",el->EqmExportName,el->EqmGroup,erlst[cc & 0xfff]);

  return cc;
}

int prepOutgoingAddr(int isLegacy,char *addrstream,FecDataStruct *fec,ExpDataStruct *srv,FecInfoStruct *inf)
{
  RPCFecStruct lfec;
  RPCDataStruct lsrv;
  RPCFecExStruct linf;
  if (fec == NULL || srv == NULL) return argument_list_error;
  if (isLegacy)
  {
    memset(&lfec,0,sizeof(RPCFecStruct));
    strncpy(lfec.Name,fec->Name,FEC_NAME_SIZE);
    memcpy(lfec.Network,fec->IPXNetwork,16);
    strncpy(lfec.IP,fec->IP,ADDR_SIZE);
    memcpy(lfec.IPh_addr,fec->IPh_addr,4);
    lfec.Port[0] = (BYTE)fec->PortOffset;
    lfec.Port[1] = (BYTE)(fec->PortOffset >> 8);
    lfec.PortOffset = (BYTE)fec->PortOffset;
    lfec.Prot = fec->TransportProtocol;
    memset(&lsrv,0,sizeof(RPCDataStruct));
    strncpy(lsrv.FecName,srv->FecName,FEC_NAME_SIZE);
    strncpy(lsrv.SubSystem,srv->SubSystem,SUBSYSTEM_NAME_SHORTSIZE);
    strncpy(lsrv.ExportName,srv->ExportName,EXPORT_NAME_SHORTSIZE);
    strncpy(lsrv.EqmName,srv->EqmName,EQM_NAME_SHORTSIZE+1);
    strncpy(lsrv.EqmContext,srv->EqmContext,CONTEXT_NAME_SHORTSIZE);
    memset(&linf,0,sizeof(RPCFecExStruct));
    strncpy(linf.os,inf->os,12);
    strncpy(linf.desc,inf->desc,64);
    strncpy(linf.loc,inf->loc,32);
    strncpy(linf.hdw,inf->hdw,32);
    strncpy(linf.resp,inf->resp,32);
    strncpy(linf.sub,gDeviceSubSystem,SUBSYSTEM_NAME_SIZE);
#   ifdef INTEL
    memcpy(addrstream,&lsrv,RPCDATASTRUCT_SIZE);
    memcpy(&addrstream[RPCDATASTRUCT_SIZE],&lfec,RPCFECSTRUCT_SIZE);
    memcpy(&addrstream[RPCDATASTRUCT_SIZE+RPCFECSTRUCT_SIZE],&linf,sizeof(RPCFecExStruct));
#   else
    /* avoid alignment problems on some machines: */
	  _StructSwap(RPCFecStructStruct,(unsigned char *)addrstream,(unsigned char *)&lfec,LTOR);
    memcpy(&addrstream[RPCDATASTRUCT_SIZE],addrstream,RPCFECSTRUCT_SIZE);
    _StructSwap(RPCDataStructStruct,(unsigned char *)addrstream,(unsigned char *)&lsrv,LTOR);
    memcpy(&addrstream[RPCDATASTRUCT_SIZE+RPCFECSTRUCT_SIZE],&linf,sizeof(RPCFecExStruct));
#   endif
  }
  else
  {
#   ifdef INTEL
    memcpy(&addrstream[EXPDATASTRUCT_SIZE],fec,FECDATASTRUCT_SIZE);
#   else
	  _StructSwap(FDSStructStruct,(unsigned char *)addrstream,(unsigned char *)fec,LTOR);
    memcpy(&addrstream[EXPDATASTRUCT_SIZE],addrstream,FECDATASTRUCT_SIZE);
#   endif
    memcpy(addrstream,srv,EXPDATASTRUCT_SIZE);
    memcpy(&addrstream[EXPDATASTRUCT_SIZE+FECDATASTRUCT_SIZE],inf,sizeof(FecInfoStruct));
  }
  return 0;
}
int sndRegisteredExports(struct timeval *tv)
{
  int cc=0,dlen;
  static time_t lastAttempt = 0;
  static int toCount = 0, errCount = 0;
  FecDataStruct fec;
  ExpDataStruct srv;
  FecInfoStruct inf;
  static BYTE fecAddressBuf[sizeof(FecDataStruct)+sizeof(ExpDataStruct)+sizeof(FecInfoStruct)];
  ExportListStruct *el;
  DTYPE d;
  BYTE *ver = SystemVersion();

  if (!FecNameRegistered)
  {
    /* Fec Name not available, so don't try again */
    ExportsRegistered = 1;
    return database_not_loaded;
  }
  if ((cc=GetNameServerAddress()) != 0)
  {
    /* Name Server Address not available, so don't try again */
    ExportsRegistered = 1;
    return cc;
  }
  if (toCount > 30)
  {
    /* Name Server not available, so don't try again */
    ExportsRegistered = 1;
    return link_not_open;
  }
  if (tv->tv_sec < (long)lastAttempt + 1) return 0;
  lastAttempt = tv->tv_sec;
  memset(&fec,0,sizeof(FecDataStruct));
  strncpy(fec.Name,gFecName,FEC_NAME_SIZE);
  memcpy(fec.IPXNetwork,&IPXFecAddress,12);
  fec.PortOffset = gPortOffset; 
  sprintf(gFecInfo.ver,"%d.%02d.%04d",ver[0],ver[1],ver[2]*256 + ver[3]);

  memset(&inf,0,sizeof(FecInfoStruct));
  strncpy(inf.os,gFecInfo.os,FEC_OS_SIZE);
  strncpy(inf.desc,gFecInfo.desc,FEC_DESC_SIZE);
  strncpy(inf.loc,gFecInfo.loc,FEC_LOCATION_SIZE);
  strncpy(inf.ver,gFecInfo.ver,FEC_VERSION_SIZE);
  strncpy(inf.hdw,gFecInfo.hdw,FEC_HDW_SIZE);
  strncpy(inf.resp,gFecInfo.resp,FEC_RESP_SIZE);

retrySend:
  dlen = gUsingLegacyAddrCalls ? RPCFECSTRUCT_SIZE + RPCDATASTRUCT_SIZE + sizeof(RPCFecExStruct) : 
                                 EXPDATASTRUCT_SIZE + FECDATASTRUCT_SIZE + sizeof(FecInfoStruct);

  for (el=ExportList; el != NULL; el=el->next)
  {
    if (!strcmp(el->EqmExportName,"LOCAL")) el->registered = REGISTERED_ALL;
    if (el->registered) continue;

    memset(&srv,0,sizeof(ExpDataStruct));
    strncpy(srv.FecName,gFecName,FEC_NAME_SIZE);
    strncpy(srv.ExportName,el->EqmExportName,EXPORT_NAME_SIZE);
    strncpy(srv.EqmName,el->EqmName,EQM_NAME_SIZE);

    if (strlen(el->EqmContext) == 0) strncpy(el->EqmContext,gDeviceContext,CONTEXT_NAME_SIZE);
    strncpy(srv.EqmContext,el->EqmContext,CONTEXT_NAME_SIZE);
    
    if (strlen(el->EqmSubSystem) == 0) strncpy(el->EqmSubSystem,gDeviceSubSystem,SUBSYSTEM_NAME_SIZE);
    strncpy(srv.SubSystem,el->EqmSubSystem,SUBSYSTEM_NAME_SIZE);

    prepOutgoingAddr(gUsingLegacyAddrCalls,(char *)fecAddressBuf,&fec,&srv,&inf);

    d.dArrayLength = dlen;
    d.dFormat = CF_STRUCT;
    memset(d.dTag,0,TAG_NAME_SIZE);
    d.data.bptr = fecAddressBuf;
    switch (cc=tExecLink("ENS",CSHOSTS,gDeviceContext,"EXPORT",&d,&d,CA_WRITE,500))
    {
      case 0:
        feclog("FEC %s, Server %s (%s) registered with equipment name server",
          gFecName,el->EqmExportName,el->EqmName);
        feclog("EqmContext is %s",strlen(gDeviceContext) > 0 ? gDeviceContext : "not given");
        toCount = errCount = 0;
        break;
      case access_denied:
        ToggleNameServer();
        if(errCount++ < 5) ccerr(cc);
        break;
      case out_of_range:
        if (!gUsingLegacyAddrCalls)
        {
          gUsingLegacyAddrCalls = TRUE;
          goto retrySend;
        }
        break;
      case server_name_in_use:
        toCount = 0;
        el->EqmGroup[0] = 0;
        feclog("FEC name %s or Server name %s is in use",gFecName,el->EqmExportName);
        break;
      case operation_busy: 
        toCount = 0;
        ccerr(cc);
      case link_not_open: 
        toCount++;
      default: 
        errCount++;
        if (errCount < 30) ccerr(cc);
        feclog("register FEC %s, Server %s : %s",gFecName,el->EqmExportName,erlst[cc]);
        break;
    }
    el->registered = REGISTERED_ENS;
  }
  for (el=ExportList; el != NULL; el=el->next)
  { /* anybody want to join a group ? */
    if (el->registered != REGISTERED_ENS) continue;
    if (strlen(el->EqmGroup) > 0)
    { /* keep trying for a while ... */
      if ((cc=joinEnsGroup(el)) != 0 && errCount++ < 30) ccerr(cc);
    }
    el->registered = REGISTERED_ALL;
  }
  ExportsRegistered = 1;
err:
  gettimeofday(tv,(struct timezone *)NULL);
  return cc;
}
int outputConsumerList(void)
{
  int k;
  char Buffer[256];

  memset(Buffer,' ',80);
  sprintf(&Buffer[0]," CLIENT"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[16],"ADDRESS"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[32],"PROTOCOL"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[48],"Contracts\n");
  ttyoutput(Buffer);
  for (k=0; k<nconsumers; k++)
  {
    memset(Buffer,' ',80);
    sprintf(&Buffer[0]," %s",Consumers[k]->userName); Buffer[strlen(Buffer)] = ' ';
    ntoa(Consumers[k],&Buffer[16]);
    Buffer[strlen(Buffer)] = ' ';
    sprintf(&Buffer[32],"%d",Consumers[k]->protocol);
    sprintf(&Buffer[32],"%s",getProtocolAsString(Consumers[k]->transport));
    Buffer[strlen(Buffer)] = ' ';
    sprintf(&Buffer[48],"%d\n",Consumers[k]->count);
    ttyoutput(Buffer);
  }
  return 0;
}
int outputEqmList(void)
{
  ExportListStruct *el;
  char Buffer[256];
  memset(Buffer,' ',80);
  sprintf(&Buffer[0]," MODULE"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[10],"EXPORT NAME"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[30],"CONTEXT"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[50],"SUBSYSTEM"); Buffer[strlen(Buffer)] = 0;
  ttyoutput(Buffer);
  for (el=ExportList; el != NULL; el=el->next)
  {
    memset(Buffer,' ',80);
    sprintf(&Buffer[0]," %s",el->EqmName); Buffer[strlen(Buffer)] = ' ';
    sprintf(&Buffer[10],"%s",el->EqmExportName); Buffer[strlen(Buffer)] = ' ';
    sprintf(&Buffer[30],"%s",el->EqmContext); Buffer[strlen(Buffer)] = ' ';
    sprintf(&Buffer[50],"%s",el->EqmSubSystem); Buffer[strlen(Buffer)] = 0;
    ttyoutput(Buffer);
  }
  return 0;
}
int outputEqmProperties(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  char Buffer[256];
  NAME64 *prps;
  int i, nprps = 100;
  if (el == NULL) return non_existent_elem;
  if ((prps=(NAME64 *)alloca(100*sizeof(NAME64))) == NULL) return out_of_server_memory;
  GetRegisteredPropertyList(eqm,prps,&nprps);
  memset(Buffer,' ',80);
  sprintf(&Buffer[0]," Registered Properties for %s :",eqm); Buffer[strlen(Buffer)] = 0;
  ttyoutput(Buffer);
  for (i=0; i<nprps; i++)
  {
    sprintf(&Buffer[0]," %.32s",prps[i].name);
    Buffer[strlen(Buffer)] = 0;
    ttyoutput(Buffer);
  }
  return 0;
}
int outputEqmDevices(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  char Buffer[256];
  int i;
  if (el == NULL) return non_existent_elem;
  memset(Buffer,' ',80);
  sprintf(&Buffer[0]," Registered Devices for %s :",eqm); Buffer[strlen(Buffer)] = 0;
  ttyoutput(Buffer);
  for (i=0; i<el->EqmNumDevices; i++)
  {
    sprintf(&Buffer[0]," %.32s",el->EqmDeviceList[i].name);
    Buffer[strlen(Buffer)] = 0;
    ttyoutput(Buffer);
  }
  return 0;
}
int outputContractList(void)
{
  int i;
  char Buffer[256];
  CONSUMER *c;
  CONTRACT *con;

  memset(Buffer,' ',80);
  sprintf(&Buffer[0]," CONTRACT"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[50],"POLL"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[60],"TO\n");
  ttyoutput(Buffer);
  memset(Buffer,' ',80);
  for (i=0; i<ncontracts; i++)
  {
    if (ContractList[i]->expired == CONTRACT_AVAILABLE) continue;
    con = &ContractList[i]->contract;
    c   = ContractList[i]->consumer;
#   ifdef OS_16BIT
    sprintf(&Buffer[0],"[%d] %s %s <%s> (%lu elements)",
      i,con->EqmName,con->EqmProperty,con->EqmDeviceName,con->EqmSizeOut);
#   else
    sprintf(&Buffer[0],"[%d] %s %s <%s> (%u elements)",
      i,con->EqmName,con->EqmProperty,con->EqmDeviceName,con->EqmSizeOut);
#   endif
    Buffer[strlen(Buffer)] = ' ';
    sprintf(&Buffer[50],"%d msec",(int)ContractList[i]->PollingRate);
    Buffer[strlen(Buffer)] = ' ';
    if (ContractList[i]->expired)
    {
      sprintf(&Buffer[60]," [expired]\n");
      ttyoutput(Buffer);
      memset(Buffer,' ',80);
    }
    for (c=ContractList[i]->consumer; c != NULL; c=c->next)
    {
      sprintf(&Buffer[60],"%s\n",c->client->userName);
      ttyoutput(Buffer);
      memset(Buffer,' ',80);
    }
  }
  return 0;
}
int outputContractStats(int i)
{
  char Buffer[256];
  CONSUMER *c;
  CONTRACT *con;

  if (i < 0 || i >= ncontracts) return out_of_range;

  con = &ContractList[i]->contract;
  if (ContractList[i]->expired != CONTRACT_ACTIVE)
  {
    memset(Buffer,' ',80);
    sprintf(&Buffer[0]," contract entry is no longer active\n");
    ttyoutput(Buffer);
    return 0;
  }
  memset(Buffer,' ',80);
  sprintf(&Buffer[0]," CONTRACT\n");
  ttyoutput(Buffer);
  memset(Buffer,' ',80);
# ifdef OS_16BIT
  sprintf(&Buffer[0]," %s %s <%s> (%lu elements) status <%d> returned <%d>\n\n",
# else
  sprintf(&Buffer[0]," %s %s <%s> (%u elements) status <%d> returned <%d>\n\n",
# endif
    con->EqmName,con->EqmProperty,con->EqmDeviceName,con->EqmSizeOut,
    ContractList[i]->compStatus,ContractList[i]->returnCode);
  ttyoutput(Buffer);
  memset(Buffer,' ',80);
  sprintf(&Buffer[0],"POLL"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[12],"LAST SENT"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[40],"LAST STALE"); Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[63],"MISSES DELAYS\n"); 
  ttyoutput(Buffer);
  memset(Buffer,' ',80);
  sprintf(&Buffer[0],"%d msec",(int)ContractList[i]->PollingRate);
  Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[12],"%ld sec %ld usec",
    ContractList[i]->lasttime.tv_sec,ContractList[i]->lasttime.tv_usec);
  Buffer[strlen(Buffer)] = ' ';
  sprintf(&Buffer[40],"%ld sec %ld usec",
    ContractList[i]->lastmarked.tv_sec,ContractList[i]->lastmarked.tv_usec);
  Buffer[strlen(Buffer)] = ' ';
# ifdef OS_16BIT
  sprintf(&Buffer[63],"%6ld %6ld\n\n",
# else
  sprintf(&Buffer[63],"%6d %6d\n\n",
# endif
   ContractList[i]->misses,ContractList[i]->delays);
  ttyoutput(Buffer);
  memset(Buffer,' ',80);
  sprintf(&Buffer[0],"CLIENTS\n");
  ttyoutput(Buffer);
  memset(Buffer,' ',80);
  for (c=ContractList[i]->consumer; c != NULL; c=c->next)
  {
    sprintf(&Buffer[0],c->client->userName); Buffer[strlen(Buffer)] = ' ';
#   ifdef OS_16BIT
    sprintf(&Buffer[9]," [%s] %d ms +%ld : %s %ld s %ld us %lu miss %lu rec\n",
#   else
    sprintf(&Buffer[9]," [%s] %d ms +%d : %s %ld s %ld us %u miss %u rec\n",
#   endif
            SubType[BASEMODE(c->status.mode)],
            c->status.PollingRate,
            c->status.counter,
            c->status.Stale ? "STALE" : "FRESH",
            c->status.lasttime.tv_sec,
            c->status.lasttime.tv_usec,
            c->status.misses,
            c->status.reconnects);
    ttyoutput(Buffer);
    memset(Buffer,' ',80);
  }
  return 0;
}
UINT32 CMDcount = 0;
int outputServerStats(void)
{
  ttyoutput("Running since %s",ctime(&startuptime));
# ifdef OS_16BIT
  ttyoutput("Total SPX requests  : %lu\n",SPXcount);
  ttyoutput("Total IPX requests  : %lu\n",IPXcount);
  ttyoutput("Total TCP requests  : %lu\n",TCPcount);
  ttyoutput("Total UDP requests  : %lu\n",UDPcount);
  ttyoutput("Total KBD commands  : %lu\n",CMDcount);
# else
  ttyoutput("Total SPX requests  : %u\n",SPXcount);
  ttyoutput("Total IPX requests  : %u\n",IPXcount);
  ttyoutput("Total TCP requests  : %u\n",TCPcount);
  ttyoutput("Total UDP requests  : %u\n",UDPcount);
  ttyoutput("Total KBD commands  : %u\n",CMDcount);
# endif
  ttyoutput("Socket RCV Buffers  : %d\n",max_rcvbuf);
  ttyoutput("Socket SND Buffers  : %d\n",max_sndbuf);
  if (gEqmListSize) /* non-servers must cover these variables */
  {
#   ifdef OS_16BIT
    ttyoutput("Server Work Area    : %lu bytes\n",srvWorkAreaSize);
#   else
    ttyoutput("Server Work Area    : %u bytes\n",srvWorkAreaSize);
#   endif
    ttyoutput("Registered clients  : %d\n",nconsumers);
    ttyoutput("Registered contracts: %d\n",ncontracts-nExpiredContracts);
#   ifdef OS_16BIT
    ttyoutput("Contract misses     : %lu\n",gContractMisses);
    ttyoutput("Contract delays     : %lu\n",gContractDelays);
    ttyoutput("Client misses       : %lu\n",gClientMisses);
    ttyoutput("Client reconnects   : %lu\n",gClientReconnects);
    ttyoutput("Client retries      : %lu\n",gClientRetries);
    ttyoutput("Synchronous calls   : %lu\n",gSingleLinkCount);
    ttyoutput("Bursts              : %lu\n",gBurstLimitReachedCount);
#   else
    ttyoutput("Contract misses     : %u\n",gContractMisses);
    ttyoutput("Contract delays     : %u\n",gContractDelays);
    ttyoutput("Client misses       : %u\n",gClientMisses);
    ttyoutput("Client reconnects   : %u\n",gClientReconnects);
    ttyoutput("Client retries      : %u\n",gClientRetries);
    ttyoutput("Synchronous calls   : %u\n",gSingleLinkCount);
    ttyoutput("Bursts              : %u\n",gBurstLimitReachedCount);
#   endif
  }
  if (nConnectionTableEntries)                 /* non-clients must cover these variables */
  {
#   ifdef OS_16BIT
    ttyoutput("Connection arrivals : %lu\n",nArrivals);
    ttyoutput("Connection timeouts : %lu\n",nConnectionTimeouts);
    ttyoutput("Incomplete transfers: %lu\n",nPartialTransfers);
    ttyoutput("Client Work Area    : %lu bytes\n",clnWorkAreaSize);
#   else
    ttyoutput("Connection arrivals : %u\n",nArrivals);
    ttyoutput("Connection timeouts : %u\n",nConnectionTimeouts);
    ttyoutput("Incomplete transfers: %u\n",nPartialTransfers);
    ttyoutput("Client Work Area    : %u bytes\n",clnWorkAreaSize);
#   endif
  }
  ttyoutput("System Polling rate : %d msec\n",(int)SysPoll);
  ttyoutput("CPU usage           : %d percent\n", AveBusyTime);
  ttyoutput("Average Cycles/sec  : %d Hz\n", gCycleCounts);
  ttyoutput("Max Cycles/sec      : %d Hz\n", gMaxCycleCounts);
  return 0;
}
int outputServerTime(void)
{
  char tstr[64];
  time_t t;

  t = time(NULL); sprintf(tstr,"%s",ctime(&t)); tstr[24] = 0;
  ttyoutput("current FEC time: %s (UNIX: %ld)",tstr,t);
  gDataTimeStamp = makeDataTimeStamp();
  t = UTCtimestamp; sprintf(tstr,"%s",ctime(&t)); tstr[24] = 0;
  sprintf(&tstr[strlen(tstr)]," %d msec",UTCtimestampMSEC);
  ttyoutput("current Data timestamp: %s (UNIX: %10.4f)",tstr,gDataTimeStamp);
  ttyoutput("current synchronization offset: %10.4f secs",gDataTimeStampOffset);
  return 0;
}
int dumpVersion(void)
{
  BYTE *ver = SystemVersion();
  char strbuf[256];
  ttyoutput("TINE version %d.%02d.%04d",ver[0],ver[1],ver[2]*256 + ver[3]);
  sprintf(strbuf,"Appl version %s compiled on ",appmkversion);
  strcat(strbuf,ctime(&appmktime)); strcat(strbuf,">");
  ttyoutput(strbuf);
  return 0;
}
int dumpConsumerList(void) { return outputConsumerList(); }
int dumpContractList(void) { return outputContractList(); }
int dumpEqmList(void) { return outputEqmList(); }
int dumpEqmProperties(char *eqm) { return outputEqmProperties(eqm); }
int dumpEqmDevices(char *eqm) { return outputEqmDevices(eqm); }
int dumpContractStats(int i) { return outputContractStats(i); }
int dumpServerStats(void) { return outputServerStats(); }
int dumpServerTime(void) { return outputServerTime(); }

ALIAS_TABLE *gAliasTable = NULL;
int nAliasTableEntries = 0;
HashXRefStruct **AliasXRefTable = NULL;

CSVCOLUMN csvAliasFileColumns[] = 
{
  { "NAME","",0,CF_TEXT,32,(BYTE *)OFFSETIN(ALIAS_TABLE,name),NULL,0 },
  { "ALIAS","",0,CF_TEXT,32,(BYTE *)OFFSETIN(ALIAS_TABLE,alias),NULL,0 }
};
CSVDB csvAliasFileDb =
{
  csvAliasFileColumns,
  sizeof(csvAliasFileColumns)/sizeof(csvAliasFileColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(ALIAS_TABLE)
};
int GetAliasTable(void)
{
  int cc=0,sz=0;
  char afn[64];

  if (nAliasTableEntries > 0) return already_assigned;
  if (isUsingXMLFecDatabase)
  {
    sz = populateAIS(gFecName,gFecInfoList,&gAliasTable);
    if (sz <= 0)ccerr(-sz);
  }
  else
  {  
    strncpy(afn,ALIAS_FILE,64);
    if (strlen(gFecDBSubDir) > 0) sprintf(afn,"%s%s",gFecDBSubDir,ALIAS_FILE);
    if ((cc=csvReadFile(FecDBpath,afn,&csvAliasFileDb,(void **)&gAliasTable)) != 0) ccerr(cc);
    sz = csvAliasFileDb.siz;
  }
  nAliasTableEntries = sz;
  if (nAliasTableEntries > ALIAS_HASH_PAYOFF && (cc=makeAliasXRefTable()) != 0) ccerr(cc);
err:
  feclog("alias file : %s", cc ? erlst[cc] : "found");
  return cc;
}
int makeAliasXRefTable(void)
{
  int k,idx;
  HashXRefStruct *xref;

  if ((AliasXRefTable=(HashXRefStruct **)SystemCalloc(ALIAS_HASH_SIZE,sizeof(HashXRefStruct *))) == NULL)
    return out_of_local_memory;
  for (k=0; k<nAliasTableEntries; k++)
  {
    idx = (int)(ElfHash((unsigned char *)gAliasTable[k].alias) % ALIAS_HASH_SIZE);
    for (xref=AliasXRefTable[idx]; xref && xref->index != k; xref = xref->next)
    {
      if (!strnicmp(gAliasTable[k].alias,gAliasTable[xref->index].alias,64)) break;
    }
    if (xref == NULL) 
    {
      if ((xref=(HashXRefStruct *)SystemCalloc(1,sizeof(HashXRefStruct))) == NULL) return out_of_server_memory;
      xref->index = k;
      xref->next = AliasXRefTable[idx];
      AliasXRefTable[idx] = xref;
    }
  }
  return 0;
}
int getTableIndex(char *str,HashXRefStruct **xrf,int hshsiz,char *tbl,int tblsiz,int off,int reclen)
{
  int i,idx;
  HashXRefStruct *xref;

  if (tbl == NULL) return -1;
  if (xrf != NULL && hshsiz > 0) /* this should be quick */
  {
    idx = (int)(ElfHash((unsigned char *)str) % hshsiz);
    for (xref=xrf[idx]; xref; xref = xref->next)
    {
      i = xref->index;
      if (!strnicmp(str,tbl+reclen*i+off,64)) return i;
    }
  }
  else
  {
    for (i=0; i<tblsiz; i++)
    {
      if (!strnicmp(str,tbl+reclen*i+off,64)) return i;
    }
  }
  return -1;
}
ExportPropertyListStruct *GetProperyListStruct(char *eqm,char *prpName)
{
  ExportListStruct *el = getExportListItem(eqm);
  ExportPropertyListStruct *prp = NULL;
  int idx;
  if (el == NULL || prpName == NULL) return NULL;
  idx = (int)(ElfHash((unsigned char *)prpName) % PROPERTY_HASH_SIZE);
  for (prp=el->EqmProperty[idx]; prp && strnicmp(prp->prpName,prpName,PROPERTY_NAME_SIZE); prp=prp->next);
  return prp;
}
int isSpectrumArray(char *eqm,char *prpName)
{
  ExportPropertyListStruct *prp = GetProperyListStruct(eqm,prpName);
  if (prp == NULL || !(prp->prpArrayType & AT_SPECTRUM)) return FALSE;
  if (prp->prpSize <= 1) return FALSE;
  return TRUE;
}
int isChannelArray(char *eqm,char *prpName)
{
  ExportPropertyListStruct *prp = GetProperyListStruct(eqm,prpName);
  if (prp == NULL || !(prp->prpArrayType & AT_CHANNEL)) return FALSE;
  return TRUE;
}
void getWildcardTokens(char *wildcardString,char *beginToken,char *endToken,char *midToken)
{
  char *s,*c,lclStr[128];

  if (wildcardString == NULL || beginToken == NULL || endToken == NULL || midToken == NULL) return;
  strncpy(lclStr,wildcardString,64); lclStr[64] = 0;
  s = lclStr;
  beginToken[0] = endToken[0] = midToken[0] = 0; /* start from scratch */
  if (*s == '*') 
  { /* no begin token */
    while (*s == '*') s++;
    if ((c=strchr(s,'*')) != NULL) 
    { /* there's a midToken ("*abc*") */
      *c = 0;
      strncpy(midToken,s,64);
    }
    else
    { /* there's an end token */
      strncpy(endToken,s,64);
    }
  }
  else
  {
    strncpy(beginToken,s,64);
    if ((c=strchr(beginToken,'*')) != NULL) 
    { /* found begin Token end */
      *c++ = 0;
      s = c; /* re-position (could still be an end token */
      while (*s == '*') s++; /* is it "abc*xyz" ? */
      if (strlen(s) > 0) strncpy(endToken,s,64);
    }
  }
}
int wildcardMatches(char *devName,char *beginToken,int beginLen,char *endToken,int endLen,char *midToken,int midLen)
{
  int dlen;
  if (devName == NULL || beginToken == NULL || endToken == NULL || midToken == NULL) return FALSE;
  if ((dlen=(int)strlen(devName)) == 0) return FALSE;
  if (beginLen > 0)
  { /* check against begin token */
    if (strnicmp(devName,beginToken,beginLen)) return FALSE;
  }
  if (endLen > 0)
  { /* check against end token */
    if (strnicmp(&devName[dlen-endLen],endToken,endLen)) return FALSE;
  }
  if (midLen > 0)
  { /* check against mid token */
    if (strstr(devName,midToken) == NULL) return FALSE;
  }
  return TRUE;
}
int isContiguous(char *eqm,char *devName,NAME64 *devlist,int *start,int *len)
{ /* called in the case of local multi-channel arrays */
  int i, n, inSequence = FALSE, finSequence = FALSE;
  char beginToken[64], endToken[64], midToken[64];
  int blen = 0, elen = 0, mlen = 0;
  ExportListStruct *el=getExportListItem(eqm);

  if (el == NULL || devName == NULL || devlist == NULL || start == NULL || len == NULL) return FALSE;

  if (!strcmp(devName,"*"))  
  { /* channel array with '*' is continguous */
    *start = 0;
    return TRUE;
  }
  getWildcardTokens(devName,beginToken,endToken,midToken);

  /* calculate these once outside the loop ... */
  blen = (int)strlen(beginToken);
  mlen = (int)strlen(midToken);
  elen = (int)strlen(endToken);

  for (i=0,n=0; i<*len; i++)
  {
    if (!wildcardMatches(devlist[i].name,beginToken,blen,endToken,elen,midToken,mlen))
    {
      if (inSequence) finSequence = TRUE;
      continue;
    }
    if (finSequence) return FALSE; /* found another match -> not contiguous ! */
    if (!inSequence) *start = i;
    inSequence = TRUE;
    n++;
  }
  *len = n;
  return inSequence;
}
int isLocal(char *eqm,char *devName,char *prpName)
{ /* not a registered multi-channel array */
  ExportListStruct *el=getExportListItem(eqm);
  ExportPropertyListStruct *prp = GetProperyListStruct(eqm,prpName);
  PrpRedirBlk *prd = NULL;
  char beginToken[64], endToken[64], midToken[64];
  int i, blen = 0, elen = 0, mlen = 0;

  if (el == NULL || prp == NULL) return FALSE;
  getWildcardTokens(devName,beginToken,endToken,midToken);
  /* calculate these once outside the loop ... */
  blen = (int)strlen(beginToken);
  mlen = (int)strlen(midToken);
  elen = (int)strlen(endToken);
  if (el->EqmDeviceRdrList != NULL) 
  { /* has device redirection list, check it */
    for (i=0; i<el->EqmNumDevices; i++)
    {
      if (el->EqmDeviceRdrList[i].rdrName != NULL && strlen(&el->EqmDeviceRdrList[i].rdrName[CONTEXT_NAME_SIZE]) > 0 &&
          wildcardMatches(el->EqmDeviceList[i].name,beginToken,blen,endToken,elen,midToken,mlen))
        return FALSE;
    }
  }
  if (prp->prd != NULL)
  { /* has property redirector, check it */
    for (prd=prp->prd; prd != NULL; prd=prd->next)
    {
      if (wildcardMatches(&prd->prpRedirection[PRP_REDIR_DEV_OFFSET],beginToken,blen,endToken,elen,midToken,mlen))
        return FALSE;
    }
  }
  if (el->EqmPropertyQueryFcn != NULL || el->XEqmPropertyQueryFcn != NULL)
  { /* has registered query function -> let it handle this */
  }
  return TRUE; /* appears to be local so-far */
}
TINE_EXPORT int GetRegisteredPropertyList(char *eqm,NAME64 *prpNames,int *nprps)
{
  ExportListStruct *el;
  ExportPropertyListStruct *prp;
  int n,k,idx;

  if (eqm == NULL ||
      prpNames == NULL ||
      nprps == NULL) return argument_list_error;
  n = *nprps;
  for (el=ExportList; el != NULL; el=el->next)
  {
    if (strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) continue;
    for (k=0,idx=0; idx<PROPERTY_HASH_SIZE+1 && k<n; idx++)
      for (prp=el->EqmProperty[idx]; prp && k<n; prp=prp->next)
      {
        if (k > 0 && !strnicmp(prpNames[k-1].name,prp->prpName,PROPERTY_NAME_SIZE)) continue;
        strncpy(prpNames[k].name,prp->prpName,PROPERTY_NAME_SIZE);
        k++;
      }
  }
  *nprps = k;
  return 0;
}
#if !defined(DOS) && !defined(NIOS)
void freeServerMemory(void)
{
  ExportListStruct *el, *el0;
  HashXRefStruct *xref, *xref0;
  ExportPropertyListStruct *prp, *prp0;
  ALARM *alm, *alm0;
  ADSList *lst=NULL, *nxtl=NULL;
  AWSList *aw, *aw0;
  PrpRedirBlk *prd, *prd0;
  GroupTableEntry *gt, *gt0;
  int i;

#if defined(MAKE_TINE32_DLL) && !defined(NO_REMOTE_PIPE)
  DestroyShmWindow();
# endif

  if (cdiSystemFreeHook != NULL) cdiSystemFreeHook();

  if (TCPsck != NULL) SystemFree(TCPsck); TCPsck = NULL;
  
  if (IPNetList != NULL) SystemFree(IPNetList); IPNetList = NULL;
  if (IPBcastList != NULL) SystemFree(IPBcastList); IPBcastList = NULL;

  if (vFeclogBuffer != NULL) SystemFree(vFeclogBuffer);
  vFeclogBuffer = NULL;

  if (gPrpList != NULL)
  {
    SystemFree(gPrpList);
    gPrpList = NULL;
  }
  if (gAliasTable != NULL) 
  {
    for (i=0; AliasXRefTable != NULL && i<ALIAS_HASH_SIZE; i++)
    {
      for (xref=AliasXRefTable[i]; xref != NULL; )
      {
        xref0 = xref->next;
        SystemFree(xref);
        xref = xref0;
      }
    }
    SystemFree(gAliasTable);
    gAliasTable = NULL;
  }
  if (StockPropXRefTable != NULL)
  {
    for (i=0; i<STOCKPROP_HASH_SIZE; i++)
    {
      for (xref=StockPropXRefTable[i]; xref != NULL; )
      {
        xref0 = xref->next;
        SystemFree(xref);
        xref = xref0;
      }
    }
    SystemFree(StockPropXRefTable);
  }
  if (hstTbl != NULL)
  {
    for (i=0; i<maxHistoryRecords && i<nHistoryRecords; i++)
    {
      if (hstTbl[i].timestamp != NULL) SystemFree(hstTbl[i].timestamp);
      if (hstTbl[i].data != NULL) SystemFree(hstTbl[i].data);
      if (hstTbl[i].prd != NULL) SystemFree(hstTbl[i].prd);
    }
    SystemFree(hstTbl); hstTbl = NULL;
  }
  for (aw=almWatchList; aw != NULL; )
  {
    aw0 = aw->nxt;
    SystemFree(aw);
    aw = aw0;
  }
  for (el=ExportList; el != NULL; )
  {
    for (i=0; i<PROPERTY_HASH_SIZE; i++)
    {
      for (prp=el->EqmProperty[i]; prp != NULL; )
      {
        prp0 = prp->next;
        if (prp->prd != NULL) 
        {
          for (prd=prp->prd; prd != NULL; )
          {
            prd0 = prd->next;
            SystemFree(prd);
            prd = prd0;
          }
        }
        SystemFree(prp);
        prp = prp0;
      }
    }
    if (el->EqmDeviceXRef != NULL)
    {
      for (i=0; i<DEVICES_HASH_SIZE; i++)
      {
        for (xref=el->EqmDeviceXRef[i]; xref != NULL; )
        {
          xref0 = xref->next;
          SystemFree(xref);
          xref = xref0;
        }
      }
    }
    if (el->adsTable != NULL) 
    {
      for (lst=el->adsTable; lst !=NULL; )
      {
        nxtl = lst->next;
        SystemFree(lst);
        lst = nxtl;
      }
    }
    for (i=0; i<el->EqmNumDevices; i++) if (el->almLst[i] != NULL)
    {
      for (alm=el->almLst[i]; alm != NULL; )
      {
        alm0 = alm->next;
        SystemFree(alm);
        alm = alm0;
      }
    }
    SystemFree(el->almLst);

    el0 = el->next;
    SystemFree(el);
    el = el0;
  }
  ExportList = NULL;
  for (i=0; i<ngcasts; i++)
  {
    if (gcastTbl[i]->refdata != NULL) SystemFree(gcastTbl[i]->refdata);
    if (gcastTbl[i] != NULL) SystemFree(gcastTbl[i]);
  }

  if (ContractList != NULL)
  {
    for (i=0; i<MaxNumContracts; i++)
    {
      if (ContractList[i] == NULL) continue;
      if (ContractList[i]->DataIn != NULL) SystemFree(ContractList[i]->DataIn);
      if (ContractList[i]->DataOut != NULL) SystemFree(ContractList[i]->DataOut);
      SystemFree(ContractList[i]);
    }
    SystemFree(ContractList); ContractList = NULL;
  }
  if (Consumers != NULL)
  {
    for (i=0; i<MaxNumClients; i++)
    {
      if (Consumers[i] == NULL) continue;
      SystemFree(Consumers[i]);
    }
    SystemFree(Consumers); Consumers = NULL;
  }

  if (conTbl != NULL)
  {
    for (gt=grpTable; gt != NULL; )
    {
      gt0 = gt->next;
      SystemFree(gt);
      gt = gt0;
    }
    for (i=0; i<MaxNumConnections; i++)
    {
      if (conTbl[i] != NULL && (i > 1 || !NameServerLoaded))
      {
        if (conTbl[i]->dataIn != NULL) SystemFree(conTbl[i]->dataIn);
        if (conTbl[i]->bufOut != NULL) SystemFree(conTbl[i]->bufOut);
        if (conTbl[i]->bufOutBlks != NULL) SystemFree(conTbl[i]->bufOutBlks);
        SystemFree(conTbl[i]);
      }
    }
    SystemFree(conTbl); conTbl = NULL;
  }

  if (glbTbl!= NULL)
  {
    for (i=0; i<nglobals; i++)
    {
      if (glbTbl[i] == NULL) continue;
      if (glbTbl[i]->refdata != NULL) SystemFree(glbTbl[i]->refdata);
      SystemFree(glbTbl[i]);
    }
    SystemFree(glbTbl); glbTbl = NULL;
    nglobals = 0;
  }

  if (SrvTbl != NULL) SystemFree(SrvTbl); SrvTbl = NULL;
  if (FecTbl != NULL) SystemFree(FecTbl); FecTbl = NULL;
  if (NameServer != NULL) SystemFree(NameServer); NameServer = NULL;

# ifndef INTEL
  if (tmpWorkArea != NULL && tmpWorkArea != srvWorkArea) SystemFree(tmpWorkArea); tmpWorkArea = NULL;
# endif
# ifndef srvWorkArea
  SystemFree(srvWorkArea); srvWorkArea = NULL;
# else
  SystemFree(WorkArea); WorkArea = NULL;
# endif
# ifndef clnWorkArea
  if (clnWorkArea != NULL) SystemFree(clnWorkArea); clnWorkArea = NULL;
# endif
  freeStruct(HeaderStructStruct);
  freeStruct(ProducerHeaderStructStruct);
  freeStruct(ProducerHeaderStruct5Struct);
  freeStruct(SUBSCRIPTIONStruct);
  freeStruct(SUBSCRIPTION5Struct);
  freeStruct(RPCDataStructStruct);
  freeStruct(RPCFecStructStruct);
  freeStruct(ServerStateStructStruct);
  freeStruct(GlobalDataHdrStruct);
  freeStruct(SpectrumHeaderStruct);

  freeStructList();
  structList = NULL;
  return;
}
int GetDeviceListFromProperty(char *eqm,char *prp,NAME64 *devlst,UINT32 *devlen)
{
  int cc = 0,nid, i;
  ExportListStruct *el=getExportListItem(eqm);
  ExportPropertyListStruct *pl=GetProperyListStruct(eqm,prp);
  char namProperty[64], dev[16];
  DTYPE dout,din;
  NAME16 *devlst16 = (NAME16 *)tmpWorkArea;
  static int done = 0;

  if (!done)
  {
    strncpy(qrycln.userName,"LOCAL_QUERY",USERNAME_SIZE);
    qrycon.client = &qrycln;
    qrycl.consumer = &qrycon;
    qrycl.nconsumers = 1;
    done = TRUE;
  }
  if (devlst == NULL || devlen == NULL) return argument_list_error;
  if (el == NULL || el->eqmFcn == NULL) return non_existent_elem;
  if (pl == NULL) return illegal_property;
  if (strlen(prp) > PROPERTY_NAME_SIZE - 4) return string_too_long;
  sprintf(namProperty,"%s.NAM",prp);
  if ((nid=GetPrpId(eqm,namProperty)) < 0) return illegal_property;
  din.dArrayLength = 0;
  din.dFormat = CF_NULL;
  din.dTag[0] = 0;
  din.data.vptr = NULL;
  dout.dArrayLength = MIN(pl->prpSize,*devlen);
  dout.dFormat = CF_NAME64;
  dout.data.vptr = devlst;
  qrycl.exp = el;
  qrycl.compStatus = not_signalled;
  cc = callEqm(&qrycl,"#0",namProperty,&dout,&din,CA_READ|CA_QUERY);
  if (cc == illegal_format)
  { /* legacy eqm with new lib ? */
    dout.dArrayLength = MIN(pl->prpSize,*devlen);
    dout.dFormat = CF_NAME16;
    dout.data.vptr = devlst16;
    qrycl.compStatus = not_signalled;
    cc = callEqm(&qrycl,"#0",namProperty,&dout,&din,CA_READ|CA_QUERY);
    if (cc == 0)
    { /* this worked so re-map : */
      for (i=dout.dArrayLength-1; i >= 0; i--)
      {
        strncpy(dev,devlst16[i].name,16);
        memset(devlst[i].name,0,64);
        strncpy(devlst[i].name,dev,16);
      }
    }
  }
  if (dout.dArrayLength == 0) return illegal_data_size;
  *devlen = dout.dArrayLength;
  return cc;
}
ExportPropertyListStruct *AssignDeviceListFromProperty(char *eqm,char *prp)
{
  int cc = 0,i;
  ExportListStruct *el=getExportListItem(eqm);
  ExportPropertyListStruct *pl=GetProperyListStruct(eqm,prp);
  static int inside = 0;
  NAME64 *namlst = (NAME64 *)tmpWorkArea;
  UINT32 devlen;

  if (el == NULL || el->eqmFcn == NULL) return NULL;
  if (pl == NULL) return NULL;
  if (strlen(prp) > PROPERTY_NAME_SIZE-4) return NULL;
  if (pl->devNames == NULL)
  {
    if (inside) return NULL;
    inside = TRUE;
    if ((cc=GetDeviceListFromProperty(eqm,prp,namlst,&devlen)) != 0) ccerr(cc);
    if ((pl->devNames=(NAME64 *)SystemCalloc(devlen,64)) == NULL) ccerr(out_of_server_memory);
    for (i=0; i<(int)devlen; i++) strncpy(pl->devNames[i].name,namlst[i].name,DEVICE_NAME_SIZE);
    pl->numDevices = devlen;
  }
err:
  inside = FALSE;
  return cc ? NULL : pl;
}
int callEqmInSeparateThread(char *eqm,char *prp)
{
  ExportListStruct *el=getExportListItem(eqm);
  ExportPropertyListStruct *pl=GetProperyListStruct(eqm,prp);
  return pl != NULL ? pl->runInSeparateThread : 0;
}
TINE_EXPORT int AssignDeviceListToProperty(char *eqm,char *prp,NAME16 *devlst,int lstlen)
{
  int cc = 0,i;
  ExportListStruct *el=getExportListItem(eqm);
  ExportPropertyListStruct *pl=GetProperyListStruct(eqm,prp);

  if (el == NULL || el->eqmFcn == NULL || devlst == NULL || lstlen < 1) return argument_list_error;
  if (pl == NULL) return illegal_property;
  if (strlen(prp) > PROPERTY_NAME_SIZE-4) return string_too_long; 
  if (pl->devNames == NULL)
  {
    if ((pl->devNames=(NAME64 *)SystemCalloc(lstlen,64)) == NULL) ccerr(out_of_server_memory); 
    for (i=0; i<lstlen; i++) strncpy(pl->devNames[i].name,devlst[i].name,DEVICE_NAME_SIZE);
    pl->numDevices = lstlen;
  }
err:
  return cc;
}
TINE_EXPORT int GetRegisteredPropertyListStruct(char *eqm,char *prpName,ExportPropertyListStruct *prpLstStruct)
{
  ExportPropertyListStruct *prp = GetProperyListStruct(eqm,prpName);
  if (prp == NULL) return non_existent;
  if (prpLstStruct == NULL) return argument_list_error;
  memcpy(prpLstStruct,prp,sizeof(ExportPropertyListStruct));
  return 0;
}
TINE_EXPORT int GetNumberRegisteredDevices(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  if (el == NULL) return -argument_list_error;
  return el->EqmNumDevices;
}
TINE_EXPORT char *GetLocalName(char *exportName)
{
  ExportListStruct *el;
  if (exportName == NULL || strlen(exportName) == 0) return NULL;
  for (el=ExportList; el != NULL; el=el->next) 
    if (!strnicmp(exportName,el->EqmExportName,DEVICE_NAME_SIZE)) break;
  return el == NULL ? NULL : el->EqmName;
}
#else
ExportPropertyListStruct *AssignDeviceListFromProperty(char *eqm,char *prp)
{
  eqm = eqm; prp = prp;
  return NULL;
}
#endif /* SMALL_TINE_LIB */
int (*lookupRedirectionNameStub)(char *eqm,char *prpName,char *devName) = NULL;
int GetPrpId(char *eqm,char *prpName)
{
  ExportListStruct *el = getExportListItem(eqm);
  ExportPropertyListStruct *prp = NULL;
  int idx;

  if (el == NULL || prpName == NULL) return -1;
  idx = (int)(ElfHash((unsigned char *)prpName) % PROPERTY_HASH_SIZE);

  for (prp=el->EqmProperty[idx]; prp; prp=prp->next)
  {
    if (!strnicmp(prp->prpName,prpName,PROPERTY_NAME_SIZE)) return prp->prpId;
  }
  return -1;
}
TINE_EXPORT int hasDevicePropertyList(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  if (el == NULL || el->EqmDevicePrpList == NULL) return FALSE;
  return TRUE;
}
TINE_EXPORT void SetRPCCompletion(char *errstr)
{
  printf("SetRPCCompletion has been deprecated: Please use SetEqmCompletion!\n");
  SetEqmCompletion(NULL,errstr);
}
TINE_EXPORT void SetRPCRedirection(char *rdrSrv,char *rdrPrp)
{
  printf("SetRPCRedirection has been deprecated: Please use SetRedirectionParameters!\n");
  SetRedirectionParameters(NULL,NULL,rdrSrv,NULL,rdrPrp);
}
/**
 * \internal
 */
typedef struct 
{
  char expName[EXPORT_NAME_SIZE];
  char cntName[CONTEXT_NAME_SIZE];
  char eqmName[EQM_NAME_SIZE];
  char fecName[FEC_NAME_SIZE];
  char version[16];
  char fecHome[32];
  char started[32];
  UINT32 port;
} FECMF;
#define hdrFecMf "EXPORT, CONTEXT, EQM, FEC, PORT, VERSION, FEC_HOME, LAST_STARTED\n"
CSVCOLUMN csvFecMfColumns[] =
{
  { "EXPORT","",0,CF_TEXT,EXPORT_NAME_SIZE,(BYTE *)OFFSETIN(FECMF,expName),NULL,0 },
  { "CONTEXT","0",0,CF_TEXT,CONTEXT_NAME_SIZE,(BYTE *)OFFSETIN(FECMF,cntName),NULL,0 },
  { "EQM","",0,CF_TEXT,EQM_NAME_SIZE,(BYTE *)OFFSETIN(FECMF,eqmName),NULL,0 },
  { "FEC","",0,CF_TEXT,FEC_NAME_SIZE,(BYTE *)OFFSETIN(FECMF,fecName),NULL,0 },
  { "PORT","",0,CF_LONG,1,(BYTE *)OFFSETIN(FECMF,port),NULL,0 },
  { "VERSION","",0,CF_TEXT,16,(BYTE *)OFFSETIN(FECMF,version),NULL,0 },
  { "FEC_HOME","",0,CF_TEXT,32,(BYTE *)OFFSETIN(FECMF,fecHome),NULL,0 },
  { "LAST_STARTED","",0,CF_TEXT,32,(BYTE *)OFFSETIN(FECMF,started),NULL,0 },
};
CSVDB csvFecMfDb =
{
  csvFecMfColumns,
  sizeof(csvFecMfColumns)/sizeof(csvFecMfColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(FECMF)
};
/**
 * \internal
 */
int updateFecManifest(ExportListStruct *el)
{
# ifndef FS_RDONLY
  FILE *fp=NULL;
  char str[256],mffilename[128],tmpfilename[128],verstr[16],tstr[32];
  time_t t = time(NULL);
  int cc = 0,siz,i,found = FALSE;
  BYTE *ver = SystemVersion();
  FECMF *fecmf = NULL, thisfec;

  if (el == NULL) ccerr(argument_list_error);
  establishAddrCacheFilePath();
  sprintf(verstr,"%d.%02d.%04d",ver[0],ver[1],ver[2]*256 + ver[3]);
  strncpy(tstr,ctime(&t),24);
  strncpy(thisfec.cntName,el->EqmContext,CONTEXT_NAME_SIZE);
  strncpy(thisfec.expName,el->EqmExportName,EXPORT_NAME_SIZE);
  strncpy(thisfec.eqmName,el->EqmName,EQM_NAME_SHORTSIZE);
  strncpy(thisfec.fecName,gFecName,FEC_NAME_SIZE);
  thisfec.port = gPortOffset;
  strncpy(thisfec.version,verstr,16);
  strncpy(thisfec.fecHome,FecDBpath,32);
  strncpy(thisfec.started,tstr,32);

  sprintf(mffilename,"%sfec.mf",gAddrCacheFilePath);
  sprintf(tmpfilename,"%sfec.tmp",gAddrCacheFilePath);

  if ((cc=csvReadFile(gAddrCacheFilePath,"fec.mf",&csvFecMfDb,(void **)&fecmf)) != 0)
  { /* doesn't exist yet -> create it */
    if ((fp=fopen(mffilename,"w")) == NULL) ccerr(file_error);
    fwrite(hdrFecMf,strlen(hdrFecMf),1,fp);
    sprintf(str,"%.32s, %.32s, %.6s, %.16s, %d, %.16s, %.32s, %.24s\n",
      thisfec.expName,thisfec.cntName,thisfec.eqmName,thisfec.fecName,
      thisfec.port,thisfec.version,thisfec.fecHome,thisfec.started);
    fwrite(str,strlen(str),1,fp);
    fclose(fp);
    return 0;
  }
  siz = csvFecMfDb.siz;
  if ((fp=fopen(tmpfilename,"w")) == NULL) ccerr(file_error);
  fwrite(hdrFecMf,strlen(hdrFecMf),1,fp);

  for (i=0; i<siz; i++)
  {
    if (!strnicmp(fecmf[i].expName,thisfec.expName,EXPORT_NAME_SIZE) &&
        !strnicmp(fecmf[i].cntName,thisfec.cntName,CONTEXT_NAME_SIZE))
    {
      found = TRUE;
      memcpy(&fecmf[i],&thisfec,sizeof(FECMF));
    }
    sprintf(str,"%.32s, %.32s, %.6s, %.16s, %d, %.16s, %.32s, %.24s\n",
      fecmf[i].expName,fecmf[i].cntName,fecmf[i].eqmName,fecmf[i].fecName,
      fecmf[i].port,fecmf[i].version,fecmf[i].fecHome,fecmf[i].started);
    fwrite(str,strlen(str),1,fp);
  }
  if (!found)
  {
    sprintf(str,"%.32s, %.32s, %.6s, %.16s, %d, %.16s, %.32s, %.24s\n",
      thisfec.expName,thisfec.cntName,thisfec.eqmName,thisfec.fecName,
      thisfec.port,thisfec.version,thisfec.fecHome,thisfec.started);
    fwrite(str,strlen(str),1,fp);
  }
  fclose(fp); fp = NULL;

  unlink(mffilename);
  rename(tmpfilename,mffilename);
err:
  if (cc) feclog("could not update fec manifest : %s",erlst[cc]);
  if (fp) fclose(fp);
  return cc;
# else
  el = el;
  return 0;
# endif
}
TINE_EXPORT int FindServerOnNetwork(char *context,char *eqmName,char *exportName,FecDataStruct *fec,ExpDataStruct *srv)
{
  int i,p,cc=0,pass=0;
  NAME32 n32[5];
  NAME16 n16[5];
  DTYPE dout;
  DTYPE din;

retryIllegalFormat:
  if (pass == 0)
  {
    din.dFormat = CF_NAME32;
    din.dArrayLength = 3;
    din.data.vptr = n32;
    n32[0].name[0] = 0;
    n32[1].name[0] = 0;
    n32[2].name[0] = 0;
    if (context != NULL) strcpy(n32[0].name,context);
    if (eqmName != NULL) strcpy(n32[1].name,eqmName);
    if (exportName != NULL) strcpy(n32[2].name,exportName);
    dout.dFormat = CF_NAME32;
    dout.dArrayLength = 5;
    dout.data.vptr = (void *)n32;
  }
  else
  { 
    din.dFormat = CF_NAME16;
    din.dArrayLength = 3;
    din.data.vptr = n16;
    n16[0].name[0] = 0;
    n16[1].name[0] = 0;
    n16[2].name[0] = 0;
    if (context != NULL) strcpy(n16[0].name,context);
    if (eqmName != NULL) strcpy(n16[1].name,eqmName);
    if (exportName != NULL) strcpy(n16[2].name,exportName);
    dout.dFormat = CF_NAME16;
    dout.dArrayLength = 5;
    dout.data.vptr = (void *)n16;
  }
  for (i=0,cc=link_not_open; cc != 0 && i <5; i++)
  {
    cc = ExecLinkEx(SRVEXP_NAME,"SRVADDR",&dout,&din,(short)(CA_READ|CA_SYNCNOTIFY),200);
    if (cc == illegal_format && pass == 0)
    { /* legacy server ? */
      pass++;
      goto retryIllegalFormat;
    }
  }
  if (cc == 0 && pass > 0)
  {
    for (i=0; i<5; i++) 
    {
      memcpy(n32[i].name,n16[i].name,16); n32[i].name[16] = 0;
    }
  }
  if (fec != NULL)
  {
    p = atoi(n32[0].name);
    strncpy(fec->Name,n32[1].name,FEC_NAME_SIZE);
    fec->PortOffset = p;
    memcpy(fec->IPh_addr,&IncomingIPfec.sin_addr.s_addr,4);
    inet_ntoa_b(IncomingIPfec.sin_addr,fec->IP);
  }
  if (srv != NULL)
  {
    strncpy(srv->FecName,n32[1].name,FEC_NAME_SIZE);
    strncpy(srv->EqmContext,n32[2].name,CONTEXT_NAME_SHORTSIZE);
    strncpy(srv->EqmName,n32[3].name,EQM_NAME_SHORTSIZE);
    strncpy(srv->ExportName,n32[4].name,EXPORT_NAME_SHORTSIZE);
  }
  return cc;
}
#endif /* NO_DOXYGEN */

TINE_EXPORT int MaxNumContracts = NCONTRACT;
TINE_EXPORT int MaxNumClients   = NCONSUMER;
/* Documented API calls follow: */
TINE_EXPORT int AssignNumDevices(char *eqm,int num)
{
  ExportListStruct *el;
  int cc=0;

  for (el=ExportList; el != NULL; el=el->next)
    if (!strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);

  if (num < 0 || num >= el->EqmDeviceSpaceSize) ccerr(out_of_range);
  el->EqmNumDevices = num;      /* this is a one-way street  */
err:
  return cc;
}
TINE_EXPORT int AssignPropertyList(char *eqm,char *devname,char *listname,int listsize,NAME64 *list)
{
  int i, k, cc = 0, refcnt = 0;
  ExportListStruct *el, *e;
  NAME64 *lst = NULL;

  if (eqm == NULL || devname == NULL || listname == NULL || list == NULL) ccerr(argument_list_error);

  for (el=ExportList; el != NULL; el=el->next)
    if (!strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);
  if ((i=GetDeviceNumberEx(eqm,devname,NULL)) < 0 || i >= el->EqmDeviceSpaceSize) ccerr(device_not_connected);
  if (el->EqmDevicePrpList == NULL)
  {
    if ((el->EqmDevicePrpList=(EqmDevicePropertyList *)SystemCalloc(el->EqmDeviceSpaceSize,sizeof(EqmDevicePropertyList))) == NULL) 
      ccerr(out_of_server_memory);
  }

  for (e=ExportList; e != NULL; e=e->next)
  { /* how many others are referencing this ? */
    if (e->EqmDevicePrpList == NULL) continue;
    for (k=0; k<e->EqmDeviceSpaceSize; k++)
    {
      if (e == el && i == k) continue;
      if (e->EqmDevicePrpList[k].prpList == NULL) continue;
      if (!strncmp(e->EqmDevicePrpList[k].filename,listname,64) &&
          e->EqmDevicePrpList[k].prpListSize == listsize) lst = e->EqmDevicePrpList[k].prpList;
      if (el->EqmDevicePrpList[i].prpList == NULL) continue;
      if (e->EqmDevicePrpList[k].prpList == el->EqmDevicePrpList[i].prpList) refcnt++;
    }
  }
  
  if (refcnt == 0 && el->EqmDevicePrpList[i].prpList != NULL) 
  { /* only free the reference if none are using it */
    SystemFree(el->EqmDevicePrpList[i].prpList);
    el->EqmDevicePrpList[i].prpList = NULL;
    refcnt = 0;
  }

  strncpy(el->EqmDevicePrpList[i].filename,listname,64);
  el->EqmDevicePrpList[i].prpListSize = listsize;

  /* test this 
  if (el->EqmDevicePrpList[i].prpList == NULL && lst != NULL)
  { 
    el->EqmDevicePrpList[i].prpList = lst;
  }
  */
  if ((el->EqmDevicePrpList[i].prpList=(NAME64 *)SystemCalloc(listsize,sizeof(NAME64))) == NULL) 
    ccerr(out_of_server_memory);
  memcpy(el->EqmDevicePrpList[i].prpList,list,listsize*sizeof(NAME64));

err:
  return cc;
}
TINE_EXPORT int ExecLocalLink(const char *devName,const char *devProperty,DTYPE *dout,DTYPE *din,short access)
{
  ContractListStruct cl;
  CLN cln;
  BYTE *datain = NULL,*dataout = NULL;
  char fulname[128],sts[STATUS_SIZE];
  char *en = NULL,*dn = NULL,*cp=NULL;
  int cc = 0;

  strncpy(fulname,devName,71); fulname[71] = 0;
  if ((en=strtok_r(fulname,DELIMITER,&cp)) == NULL) ccerr(argument_list_error);
  if ((dn=strtok_r(NULL,DELIMITER,&cp)) == NULL) dn = &fulname[71];

  strncpy(cln.userName,gFecName,USERNAME_SIZE);

  memset(&cl.contract,0,sizeof(CONTRACT));
  strncpy(cl.contract.EqmName,en,EQM_NAME_SHORTSIZE);
  strncpy(cl.contract.EqmDeviceName,dn,DEVICE_NAME_SIZE);
  strncpy(cl.contract.EqmProperty,devProperty,PROPERTY_NAME_SIZE);

  if (din != NULL)
  {
    cl.contract.EqmSizeIn = din->dArrayLength;
    cl.contract.EqmFormatIn = BFMT(din->dFormat);
    cl.contract.strTagIn[0] = 0;
    datain = din->data.bptr;
  }
  if (dout != NULL)
  {
    cl.contract.EqmSizeOut = dout->dArrayLength;
    cl.contract.EqmFormatOut = BFMT(dout->dFormat);
    cl.contract.strTagOut[0] = 0;
    dataout = dout->data.bptr;
  }
  cl.contract.EqmAccess = (BYTE)access;

  if ((cc=IsIllegalProperty(cl.contract.EqmName,
        cl.contract.EqmProperty,cl.contract.EqmDeviceName,
        &cln,access,&cl.exp,sts)) != 0) ccerr(cc);

  setDataTimeStamp(makeDataTimeStamp());
  cl.DataOut = dataout;
  cl.DataIn = datain;
  cl.DataBuf = srvWorkArea;
  if ((cc=doContract(&cl)) != 0) goto err;
err:
  return cc;
}
TINE_EXPORT int FindNameServerOnNetwork()
{
  int cc = 0;
  FecDataStruct fec;
  ExpDataStruct srv;
  if (gotNameServerAddress) return 0;
  if ((cc=reserveFecAddressSpace()) != 0) ccerr(cc);
  if ((cc=FindServerOnNetwork(NULL,CSHOSTS,NULL,&fec,&srv)) == 0)
  {
    if ((NameServer=(FecDataStruct *)SystemCalloc(1,sizeof(FecDataStruct))) == NULL) ccerr(out_of_server_memory);
    memcpy(NameServer,&fec,sizeof(FecDataStruct));
    numNameServers = 1;
    gotNameServerAddress = 1;
    addNameServerToTables();
    feclog("Using Name Server %s at location %s",NameServer->Name,NameServer->IP);
  }
err:
  feclog("FindNameServerOnNetwork : %s",erlst[cc]);
  return cc;
}
TINE_EXPORT char *GetCaller(char *eqm)
{
  static char name[USERNAME_SIZE+4];
  NAME16 un;
  int num = 1;

  if (GetCallerInfo(eqm,&un,NULL,NULL,NULL,&num) != 0) return NULL;
  strncpy(name,un.name,USERNAME_SIZE);
  return name;
}
TINE_EXPORT int GetCallerInfo(char *eqm,NAME16 *un,BYTE *ipx,UINT32 *ip,short *prot,int *num)
{
  ExportListStruct *el = getExportListItem(eqm);
  CONSUMER *cl;
  char scratch[48];
  int i;

  if (el == NULL)
  {
    if (eqm != NULL && !strcmp(eqm,STKEQM_NAME))
    {
      if (gStockRecourceSet.StkCls == NULL) return 0;
      cl = gStockRecourceSet.StkCls->consumer;
    }
    else if (ExportList != NULL && ExportList->next == NULL)
    {
      if (ExportList->EqmCls == NULL) return 0;
      cl = ((ContractListStruct *)ExportList->EqmCls)->consumer;
    }
    else
    {
      return argument_list_error;
    }
  }
  else
  {
    if (el->EqmCls == NULL) return 0;
    cl = ((ContractListStruct *)el->EqmCls)->consumer;
  }
  if (num == NULL || *num < 1) return buffer_too_small;
  for (i=0; cl != NULL; cl=cl->next,i++)
  {
    if (cl->client == NULL) break;
    if (i < *num)
    {
      if (un != NULL) strncpy(un[i].name,cl->client->userName,16);
      if (ipx != NULL) memcpy(&ipx[i*6],cl->client->IPXaddress.network,6);
      if (ip != NULL)
      {
        inet_ntoa_b(cl->client->IPaddress.sin_addr,scratch);
        ip[i] = inet_addr(scratch);
      }
      if (prot != NULL) prot[i] = cl->client->protocol;
    }
  }
  if (i == 0)
  { /* must be a local (internal) call to the equipment module */
    if (un != NULL) strncpy(un->name,"LOCAL",16);
    if (ip != NULL) *ip = 0;
    i = 1;
  }
  *num = i;
  return 0;
}
TINE_EXPORT char *GetExportedName(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  if (el == NULL) return NULL;
  return el->EqmExportName;
}
TINE_EXPORT char *GetDeviceName(char *eqm,int devnr)
{
  ExportListStruct *el = getExportListItem(eqm);
  static char lclname[DEVICE_NAME_SIZE+4];

  sprintf(lclname,"DEVICE %d",devnr);
  if (el == NULL || devnr < 0 || devnr >= el->EqmNumDevices) return lclname;
  return el->EqmDeviceList[devnr].name;
}
TINE_EXPORT NAME64 *GetExportedDeviceList(char *eqm)
{
  ExportListStruct *el = getExportListItem(eqm);
  if (el != NULL) return el->EqmDeviceList;
  return NULL;
}
TINE_EXPORT int GetDeviceNumberEx(char *eqm,char *devname,char *prpname)
{
  char lclname[DEVICE_NAME_SIZE+4];
  int idx,i;
  HashXRefStruct *xref;
  ExportListStruct *el = getExportListItem(eqm);
  ExportPropertyListStruct *prp;

  if (eqm == NULL || devname == NULL) return -argument_list_error;
  if (prpname != NULL && (prp=AssignDeviceListFromProperty(eqm,prpname)) != NULL)
  {
    /* TODO: replace this too with a hash table ... */
    for (idx=0; idx<prp->numDevices; idx++) if (!strncmp(devname,prp->devNames[idx].name,DEVICE_NAME_SIZE)) return idx;
  }
  if (el == NULL) return -non_existent_elem;
  memset(lclname,0,DEVICE_NAME_SIZE+4);
  strncpy(lclname,devname,DEVICE_NAME_SIZE);
  if (devname[0] == '#') return atoi(&lclname[1]);
  if (el->EqmNumDevices < DEVICES_HASH_PAYOFF)
  {
    for (i=0; i<el->EqmNumDevices; i++)
      if (!strnicmp(lclname,el->EqmDeviceList[i].name,DEVICE_NAME_SIZE)) return i;
    return -device_not_connected;
  }
  if (el->EqmDeviceXRef == NULL) return -device_not_connected;
  idx = (int)(ElfHash((unsigned char *)lclname) % DEVICES_HASH_SIZE);
  for (xref=el->EqmDeviceXRef[idx]; xref; xref=xref->next)
    if (!strnicmp(lclname,el->EqmDeviceList[xref->index].name,DEVICE_NAME_SIZE)) return xref->index;
  return -device_not_connected;
}
TINE_EXPORT int GetDeviceNumber(char *eqm,char *devname)
{
  return GetDeviceNumberEx(eqm,devname,NULL);
}
TINE_EXPORT int GetNumContracts(char *eqm)
{
  int i,n=0;
  if (eqm == NULL || strlen(eqm) == 0) return ncontracts - nExpiredContracts;
  for (i=0; i<ncontracts; i++)
   if (!strncmp(ContractList[i]->contract.EqmName,eqm,EQM_NAME_SHORTSIZE) && 
       ContractList[i]->expired == CONTRACT_ACTIVE) n++;
  return n;
}
TINE_EXPORT int GetNumCallers(char *eqm)
{
  return GetNumConsumers(eqm);
}
TINE_EXPORT int GetNumConsumers(char *eqm)
{
  int i,n=0;
  if (eqm == NULL || strlen(eqm) == 0) return nconsumers;
  for (i=0; i<ncontracts; i++)
   if (!strncmp(ContractList[i]->contract.EqmName,eqm,EQM_NAME_SHORTSIZE) &&
       ContractList[i]->nconsumers > n) n = ContractList[i]->nconsumers;
  return n;
}
TINE_EXPORT int GetPropertyId(char *eqm,char *prpName)
{
  int pid = GetPrpId(eqm,prpName);
  return pid < 0 ? -1 : (int)PROPERTY_NUM_INDEX(pid);
}
TINE_EXPORT int JoinEquipmentGroup(char *eqmName,char *groupname,int groupindex)
{
  ExportListStruct *el = getExportListItem(eqmName);

  if (el == NULL) return non_existent_elem;
  if (groupname == NULL || strlen(groupname) == 0) return argument_list_error;
  if (groupindex < 0) return out_of_range;
  strncpy(el->EqmGroup,groupname,EXPORT_NAME_SIZE);
  el->groupIndex = groupindex;
  return 0;
}
TINE_EXPORT int RegisterDeviceName(char *eqm,char *devname,int devnr)
{
  ExportListStruct *el;
  HashXRefStruct *xref;
  char lclname[DEVICE_NAME_SIZE+4];
  int cc=0,idx;

  if (eqm == NULL || devname == NULL) ccerr(argument_list_error);
  for (el=ExportList; el != NULL; el=el->next)
    if (!strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);

  if (devnr < 0 || devnr >= el->EqmDeviceSpaceSize) ccerr(out_of_range);
  memset(lclname,0,DEVICE_NAME_SIZE+4);
  strncpy(lclname,devname,DEVICE_NAME_SIZE);
  strncpy(el->EqmDeviceList[devnr].name,lclname,DEVICE_NAME_SIZE);
  if (devnr == 0 && strlen(el->EqmDeviceList[0].name) > 0) el->hasRegisteredDevicesNames = TRUE;

  if (el->EqmDeviceSpaceSize < DEVICES_HASH_PAYOFF) return 0;
  if (el->EqmDeviceXRef == NULL) return 0;
  idx = (int)(ElfHash((unsigned char *)lclname) % DEVICES_HASH_SIZE);
  for (xref=el->EqmDeviceXRef[idx]; xref && xref->index != devnr; xref=xref->next);
  if (xref == NULL) /* it's a new one*/
  {
    if ((xref=(HashXRefStruct *)SystemCalloc(1,sizeof(HashXRefStruct))) == NULL) ccerr(out_of_server_memory);
    xref->index = devnr;
    xref->next = el->EqmDeviceXRef[idx];
    el->EqmDeviceXRef[idx] = xref;
  }
err:
  return cc;
}
TINE_EXPORT int RedirectDeviceName(char *eqm,char *devname,char *rdr)
{
  int i,cc = 0;
  ExportListStruct *el;
  PrpRedirBlk prd;
  if (eqm == NULL || devname == NULL || rdr == NULL) ccerr(argument_list_error);
  for (el=ExportList; el != NULL; el=el->next)
    if (!strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);
  if ((i=GetDeviceNumberEx(eqm,devname,NULL)) < 0 || i >= el->EqmNumDevices) ccerr(device_not_connected);
  if (el->EqmDeviceRdrList == NULL)
  {
    if ((el->EqmDeviceRdrList=(EqmDeviceRdrListType *)SystemCalloc(el->EqmDeviceSpaceSize,sizeof(EqmDeviceRdrListType))) == NULL) 
      ccerr(out_of_server_memory);
  }
  if ((el->EqmDeviceRdrList[i].rdrName=(char *)(SystemCalloc(1,CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE+DEVICE_NAME_SIZE))) == NULL)
    ccerr(out_of_server_memory);
  ParseRedirector(rdr,&prd); /* only the device context/server name is relevant */
  strncpy(el->EqmDeviceRdrList[i].rdrName,&prd.prpRedirection[PRP_REDIR_CTX_OFFSET],CONTEXT_NAME_SIZE); 
  strncpy(&el->EqmDeviceRdrList[i].rdrName[CONTEXT_NAME_SIZE],&prd.prpRedirection[PRP_REDIR_SRV_OFFSET],EXPORT_NAME_SIZE); 
  strncpy(&el->EqmDeviceRdrList[i].rdrName[CONTEXT_NAME_SIZE+EXPORT_NAME_SIZE],&prd.prpRedirection[PRP_REDIR_DEV_OFFSET],DEVICE_NAME_SIZE); 
err:
  return cc;
}
TINE_EXPORT int RegisterEquipmentModule(
    char *expName,
    char *eqmName,
    int numdevices,
    int (*fcn)(char *,char *,DTYPE *,DTYPE *,short),
    void (*ini)(void),
    void (*tsk)(void),
    int rate,
    void (*exi)(void)
)
{
  int cc;
  ExportListStruct *el;
# ifdef VXWORKS
  char tasktag[16];
# endif
# ifdef MULTITHREADED
  BkgThreadFcn *btf;
# endif

  if (numdevices == 0) numdevices = 1;
  if ((cc=RegisterExport(expName,eqmName,numdevices)) != 0) return cc;
  if ((el=getExportListItem(eqmName)) == NULL) return non_existent_elem;

  el->eqmFcn  = fcn;
  el->eqmBkg  = tsk;

# if defined(VXWORKS)
  if (tsk != NULL)
  {
    sprintf(tasktag,"%s_task",eqmName);
    if (taskSpawn(tasktag,VXWBKG_PRIORITY,VXWBKG_FLAGS,VXWBKG_STACK,(FUNCPTR)VXWbkgtsk,(int)tsk,(int)rate,el->exportId,0,0,0,0,0,0,0) == ERROR)
      cc = not_running;
    feclog("task %s : %s",tasktag,cc ? "not spawned !" : "spawned");
  }
  el->eqmBkg  = NULL;        /* let VxWorks handle the task */
# elif defined(MULTITHREADED)
  if (tsk != NULL && useMultiThreadedBackgroundTasks)
  {
    feclog("FEC: module %s is using multi-threaded background tasks",eqmName);
    if ((btf=(BkgThreadFcn *)SystemCalloc(1,sizeof(BkgThreadFcn))) == NULL) return out_of_local_memory;
    btf->rate = rate;   btf->tsk = tsk; btf->ref = (void *)el;
    el->bkgbusy = TRUE;
    el->hEqmBkg = bkgCreateThread(btf);
    el->eqmBkg  = NULL;     /* let the pthread handle the task */
  }
# endif

  el->rateBkg = rate;
  if (el->eqmBkg && rate && ServerCycleMode != POLLING && rate < MaxPollingRate)
  {
    SysPoll = rate;
    MaxPollingRate = rate;
    if (rate < MinPollingRate) MinPollingRate = rate;
  }
  el->eqmIni = ini;
  el->eqmExi = exi;

  feclog("%12s task registration: %s %s %s",eqmName,
          ini ? "INIT" : "----",
          tsk ? "BCKG" : "----",
          exi ? "EXIT" : "----");

  if (SystemRunning)
  {
    if (el->eqmIni) el->eqmIni();
    historyInit(el->EqmExportName,el->EqmName);
#   ifdef MULTITHREADED
    if (useMultiThreadedEquipmentFunctions) gNeedToStartEqmThreads = TRUE;
#   endif
    el->inidone = TRUE;
  }
  updateFecManifest(el);
  ExportsRegistered = 0;
  return 0;
}
TINE_EXPORT int RegisterFecInformation(char *name,char *sub,char *cntxt,char *dsc,char *loc,char *hdw,char *resp,UINT16 poff)
{
  if (FecNameRegistered)
  {
    feclog("FEC Name %s already register for this server",gFecName);
    return unauthorised_action;
  }
  if (name == NULL) return argument_list_error;
  strncpy(gFecName,name,FEC_NAME_SIZE);
  strncpy(gFecInfo.os,OS_TAG,12);
  if (dsc != NULL) strncpy(gFecInfo.desc,dsc,64);
  if (loc != NULL) strncpy(gFecInfo.loc,loc,32);
  if (hdw != NULL) strncpy(gFecInfo.hdw,hdw,32);
  if (resp != NULL) strncpy(gFecInfo.resp,resp,32);
  strncpy(gUserName,gFecName,USERNAME_SIZE);
  gPortOffset = poff;
  if (cntxt != NULL) strncpy(gDeviceContext,cntxt,CONTEXT_NAME_SIZE);
  if (sub != NULL) strncpy(gDeviceSubSystem,sub,SUBSYSTEM_NAME_SIZE);
  FecNameRegistered = 1;
  feclog("FEC Name %s registered from API",name);
  return 0;
}
TINE_EXPORT int RegisterFecNameEx(char *name,char *desc,char *os,char *loc,char *hdw,char *resp,UINT16 poff,char *context)
{
  char *c,sub[SUBSYSTEM_NAME_SIZE] = "";
  if ((c=strchr(desc,'[')) != NULL) 
  {
    strncpy(sub,++c,SUBSYSTEM_NAME_SIZE);
    if ((c=strchr(sub,']')) != NULL) *c = 0;
  }
  return RegisterFecInformation(name,sub,context,desc,loc,hdw,resp,poff);
}
TINE_EXPORT int RegisterFecName(char *name,char *desc,char *os,char *loc,char *hdw,char *resp,UINT16 poff)
{
  return RegisterFecNameEx(name,desc,os,loc,hdw,resp,poff,NULL);
}
/* Appends property list for eqm ...*/
TINE_EXPORT int RegisterPropertyInformation(char *eqm,char *property,DTYPE *dout,DTYPE *din,
                                       short acc,short atype,UINT16 rowlength,char *dsc,int propId,char *rdr) 
{
  int i,idx,cc=0,isNew=0;
  SINT32 prpId = propId, prpId0 = -1;
  UINT32 dSizeInBytes = 0;
  ExportPropertyListStruct *prp = NULL;
  ExportListStruct *el;
  PrpRedirBlk *prd;
  char strbuf[128];

  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);
  if ((i=(int)strlen(property)) > 4 && !strcmp(&property[i-4],".NAM")) 
  {
    el->hasPropertyDeviceNames = TRUE;
    if (queryDeviceNamesFromProperties) acquireDeviceNamesFromProperties = TRUE;
  }
  if (strchr(property,'*')) idx = PROPERTY_HASH_SIZE;
  else idx = (int)(ElfHash((unsigned char *)property) % PROPERTY_HASH_SIZE);

  for (prp=el->EqmProperty[idx]; prp; prp=prp->next) if (!strnicmp(prp->prpName,property,PROPERTY_NAME_SIZE)) break;
  if (prp != NULL) 
  {
    prpId0 = prp->prpId;
    if (dout != NULL)
    {
      if (dout->dArrayLength != prp->prpSize || dout->dFormat != prp->prpFormat) isNew = -1;
      if (dout->dFormat == CF_STRUCT && strncmp(dout->dTag,prp->prpFormatTag.name,TAG_NAME_SIZE)) isNew = -1;
    }
    else
    {
      if (prp->prpSize > 0) isNew = -1;
    }
    if (din != NULL)
    {
      if (din->dArrayLength != prp->prpSizeIn || din->dFormat != prp->prpFormatIn) isNew = -1;
      if (din->dFormat == CF_STRUCT && strncmp(din->dTag,prp->prpFormatTagIn.name,TAG_NAME_SIZE)) isNew = -1;
    }
    else
    {
      if (prp->prpSizeIn > 0) isNew = -1;
    }
    if (acc != prp->prpAccessMode) isNew = -1;
  }
  else
  {
    isNew = -1;
  }

  if (!isNew) ccerr(0);

  if ((prp=(ExportPropertyListStruct *)SystemCalloc(1,sizeof(ExportPropertyListStruct))) == NULL) ccerr(out_of_server_memory);
  strncpy(prp->prpName,property,PROPERTY_NAME_SIZE);
  if ((i=getAliasTableIndexFromName(property)) >= 0)
    strncpy(prp->prpAlias,gAliasTable[i].alias,PROPERTY_NAME_SIZE);
  prp->next = el->EqmProperty[idx];
  el->EqmProperty[idx] = prp;
  if (prpId < 0) prpId = el->EqmNumProperties;
  prp->prpId = prpId0 >= 0 ? prpId0 : MAKE_PROPERTY_ID(prpId,idx);
  el->EqmNumProperties++;

  if (dout != NULL)
  {
    prp->prpSize = dout->dArrayLength;
    prp->prpFormat = dout->dFormat; 
    prp->prpArrayType = atype;
    if (!(prp->prpArrayType & AT_DOUBLE)) prp->prpArrayType |= prp->prpSize > 1 ? AT_SINGLE : AT_SCALAR;
    if (strlen(dout->dTag)) strncpy(prp->prpFormatTag.name,dout->dTag,TAG_NAME_SIZE);
    if (rowlength > 0)
    {
      prp->rowSize = rowlength;
      prp->numRows = (UINT16)(prp->prpSize/rowlength);
    }
    else
    {
      prp->rowSize = (UINT16)prp->prpSize;
      prp->numRows = 1;
    }
    dSizeInBytes = prp->prpSize * fmtsizeof(prp->prpFormat);
  }
  if (din != NULL)
  {
    prp->prpSizeIn = din->dArrayLength;
    prp->prpFormatIn = din->dFormat;
    if (strlen(din->dTag)) strncpy(prp->prpFormatTagIn.name,din->dTag,TAG_NAME_SIZE);
  }
  prp->prpAccessMode = acc;
  strncpy(strbuf,dsc,127); strbuf[127] = 0;
  ParsePropertyDescription(strbuf,prp);
  if (rdr != NULL && strlen(rdr) != 0)
  {
    if ((prd=(PrpRedirBlk *)SystemCalloc(1,sizeof(PrpRedirBlk))) == NULL) ccerr(out_of_server_memory);
    prd->next = prp->prd;
    prp->prd = prd;
    strncpy(strbuf,rdr,127);
    ParseRedirector(strbuf,prp->prd);
  }
  if ((prp->devNames=GetEqpDeviceNames(eqm,prp->prpName,prp->prpSize,NULL,NULL)) != NULL)
  {
    prp->numDevices = prp->prpSize;
    el->hasPropertyDeviceNames = TRUE;
  }
# if !defined(VXWORKS)
  if (dSizeInBytes > MaxSystemTransportSize)
  { /* need to make adjustments */
    adjustWorkAreaSize(dSizeInBytes+0x10000);
  }
# else
  feclog("Property %s size (%d bytes) larger than registered work area (%d bytes)",
    property,dSizeInBytes,MaxSystemTransportSize);
# endif
err:
  return cc ? -cc : (int)PROPERTY_NUM_INDEX(prp->prpId);
}
TINE_EXPORT int RedirectProperty(char *eqm,char *property,char *rdr) 
{
  char strbuf[128];
  ExportPropertyListStruct *prp;
  ExportListStruct *el;
  PrpRedirBlk *prd;
  int idx;
  if (rdr == NULL || strlen(rdr) == 0) return argument_list_error; /* must be given ! */
  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) break;
  if (el == NULL) return non_existent_elem;

  idx = (int)(ElfHash((unsigned char *)property) % PROPERTY_HASH_SIZE);
  for (prp=el->EqmProperty[idx]; prp; prp=prp->next) if (!strnicmp(prp->prpName,property,PROPERTY_NAME_SIZE)) break;
  if (prp == NULL) return illegal_property;

  if ((prd=(PrpRedirBlk *)SystemCalloc(1,sizeof(PrpRedirBlk))) == NULL) return out_of_server_memory;
  prd->next = prp->prd;
  prp->prd = prd;
  strncpy(strbuf,rdr,127);
  ParseRedirector(strbuf,prp->prd);
  return 0;
}
TINE_EXPORT int RegisterPropertyEx(char *eqm,char *property,int siz,short fmt,short acc,char *dsc,int propId) /* Appends property list for eqm */
{
  DTYPE dout,din,*dinPtr = NULL;
  dout.dArrayLength = siz;
  dout.dFormat = fmt;
  dout.dTag[0] = 0;
  if (acc & CA_WRITE)
  {
    din.dArrayLength = siz;
    din.dFormat = fmt;
    din.dTag[0] = 0;
    dinPtr = &din;
  }
  return RegisterPropertyInformation(eqm,property,&dout,dinPtr,acc,AT_UNKNOWN,0,dsc,propId,NULL);
}
TINE_EXPORT int RegisterProperty(char *eqm,char *property,int siz,short fmt,short acc,char *dsc) /* Appends property list for eqm */
{
  return RegisterPropertyEx(eqm,property,siz,fmt,acc,dsc,-1);
}
TINE_EXPORT short RegisterPropertyQueryFunction(char *eqm,int (*fcn)(char *devName,PrpQueryStruct **prpqs),int numprops)
{
  short cc=0;
  ExportListStruct *el;

  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);
  el->EqmPropertyQueryFcn = fcn;
  el->EqmNumProperties = numprops;
err:
  return cc;
}
TINE_EXPORT short RegisterXPropertyQueryFunction(char *eqm,int (*fcn)(char *devName,XPropertyQueryStruct **xpqs),int numprops)
{
  short cc=0;
  ExportListStruct *el;

  for (el=ExportList; el != NULL; el=el->next) if (!strncmp(eqm,el->EqmName,EQM_NAME_SHORTSIZE)) break;
  if (el == NULL) ccerr(non_existent_elem);
  el->XEqmPropertyQueryFcn = fcn;
  el->EqmNumProperties = numprops;
err:
  return cc;
}
TINE_EXPORT int sendNetGlobal(char *keyword,DTYPE *din,void (*callback)(int,int),int minPeriod,int maxPeriod,double tolerance)
{
  int i=0, cc=0, isRegistered=FALSE;
  UINT16 dsize;
  static int inside = 0;

  if (!ServerInitialized) return not_ready;
  if (inside) return -semaphore_busy; else inside = -1;
  if (!canSendGlobals)
  {
    feclog("sending globals not enabled !");
    ccerr(un_allocated);
  }
  if (din == NULL || din->data.vptr == NULL) ccerr(argument_list_error);
  for (i=0; i<ngcasts; i++) if (!strcmp(gcastTbl[i]->keyword,keyword)) break;
  if (i == ngcasts)
  {
    if (ngcasts >= MaxNumGlobals) ccerr(resources_exhausted);
    if ((gcastTbl[i]=(GlobalListStruct *)SystemCalloc(1,sizeof(GlobalListStruct))) == NULL)
      ccerr(out_of_client_memory);
    ngcasts++;
    if (minPeriod <= 20) minPeriod *= 1000; /* convert to msec */
    if (maxPeriod <= 20) maxPeriod *= 1000;
    strncpy(gcastTbl[i]->keyword,keyword,16);
    gcastTbl[i]->size = (short)din->dArrayLength;
    gcastTbl[i]->lastsize = gcastTbl[i]->size; /* note the initial registration size */
    gcastTbl[i]->format = din->dFormat;
    gcastTbl[i]->fcn = callback;
    gcastTbl[i]->maxPeriod = maxPeriod;
    gcastTbl[i]->minPeriod = minPeriod;
    gcastTbl[i]->tolerance = tolerance;
    dsize = gcastTbl[i]->size*fmtsizeof(gcastTbl[i]->format);
    if ((gcastTbl[i]->refdata=(BYTE *)SystemCalloc(1,dsize)) == NULL) ccerr(out_of_server_memory);
    gcastTbl[i]->time = time(NULL);
    feclog("GLOBAL %s (%d bytes) : %s",gcastTbl[i]->keyword,dsize,cc ? erlst[cc] : "registered");
    if (minPeriod > 0 && minPeriod < MaxPollingRate)
    {
      SysPoll = MaxPollingRate = minPeriod;
      feclog("adjust maximum polling interval to %d msec",MaxPollingRate);
    }
  }
  else
  {
    if ((short)din->dArrayLength > gcastTbl[i]->lastsize) ccerr(dimension_error);
    if (din->dFormat != gcastTbl[i]->format) ccerr(illegal_format);
    gcastTbl[i]->size = (short)din->dArrayLength;
    isRegistered = TRUE;
  }
  dsize = gcastTbl[i]->size*fmtsizeof(gcastTbl[i]->format);
  gcastTbl[i]->data = (BYTE *)din->data.vptr;
  if (isRegistered)
  {
    gcastTbl[i]->tv.tv_sec = 0;
    doGCast(&thisCycleTime);       /* send it out immediately if already registered */
  }
err:
  inside = 0;
  return cc ? -cc : i;
}
TINE_EXPORT void SetAppDate(time_t appdate)
{
  appmktime = appdate;
  return;
}
TINE_EXPORT void SetAppVersion(int major,int minor,int revision)
{
  sprintf(appmkversion,"%d.%02d.%04d",major,minor,revision);
  return;
}
/* this is a documented API call (see tinetype.h) */
TINE_EXPORT int SetNameServerAddress(char *ens)
{
  int i, n = 1, cc = 0;
  char s[256],*c, *c1, *ip, *po, *cp=NULL;
  DWORD haddr;

  if (gotNameServerAddress) return 0;
  if (ens == NULL) ccerr(argument_list_error);
  strncpy(s,ens,256);
  for (i=0; i<256; i++) if (s[i] == ',') n++;
  if ((NameServer=(FecDataStruct *)SystemCalloc(n,sizeof(FecDataStruct))) == NULL) ccerr(out_of_server_memory);
  for (i=0, c=strtok_r(s,",",&cp); i<n && c != NULL; i++, c=strtok_r(NULL,",",&cp))
  {
    ip = c; po = NULL;
    if ((c1=strchr(c,':')) != NULL)
    {
      *c1++ = 0; po = c1;
    }
    haddr = inet_addr(ip);
    sprintf(NameServer[i].Name,"ENS#%d",i);
    strncpy(NameServer[i].IP,ip,ADDR_SIZE);
    memcpy(NameServer[i].IPh_addr,&haddr,4);
    NameServer[i].TransportProtocol = UDP;
    if (po != NULL) 
    {
      NameServer[i].PortOffset = atoi(po);
    }
  }
  numNameServers = i;
  gotNameServerAddress = 1;
err:
  return cc;
}
TINE_EXPORT void SetEqmCompletion(char *eqm,char *errstr)
{
  ExportListStruct *el=getExportListItem(eqm);
  
  if (el == NULL) 
  {
    if (ExportList != NULL && ExportList->next == NULL) el = ExportList; else return;
  }
  memset(el->EqmLastErrorString,0,STATUS_SIZE);
  if (errstr != NULL && strlen(errstr) > 0)
  {
    strncpy(el->EqmLastErrorString,errstr,STATUS_SIZE-1);
  }
}
TINE_EXPORT void SetRedirectionParameters(char *eqm,char *rdrCnt,char *rdrSrv,char *rdrDev,char *rdrPrp)
{
  ExportListStruct *el=getExportListItem(eqm);
  
  if (el == NULL) 
  {
    if (ExportList != NULL && ExportList->next == NULL) el = ExportList; else return;
  }
  memset(el->EqmLastErrorString,0,STATUS_SIZE);
  if (rdrCnt != NULL) memcpy(&el->EqmLastErrorString[PRP_REDIR_CTX_OFFSET],rdrCnt,CONTEXT_NAME_SIZE);
  if (rdrSrv != NULL) memcpy(&el->EqmLastErrorString[PRP_REDIR_SRV_OFFSET],rdrSrv,EXPORT_NAME_SIZE);
  if (rdrDev != NULL) memcpy(&el->EqmLastErrorString[PRP_REDIR_DEV_OFFSET],rdrDev,DEVICE_NAME_SIZE);
  if (rdrPrp != NULL) memcpy(&el->EqmLastErrorString[PRP_REDIR_PRP_OFFSET],rdrPrp,PROPERTY_NAME_SIZE);
}

TINE_EXPORT void SetlookupRedirectionNameStub(int (*fcn)(char *eqm,char *prpName,char *devName))
{
  lookupRedirectionNameStub = fcn;
}
