/** \file toolkit.c
 * \brief TINE API tools
 *
 * Many optional routines are offered to make programming easier.  Below are API routines which
 * are useful both for client-side programming and server-side programming.  Some query calls
 * such as GetDeviceNames(), etc. are clearly suitable for information gathering.  The set of
 * of data manipulation routines such at GetValuesAsFloat() or PutValuesFromLongEx() are 
 * designed for easy integration into server code, where format conversion, array-wrapping, 
 * array segmenting, etc. are automatically provided.  Helper routines such as sendPMTrigger()
 * are also provided.
 */
#ifndef NO_DOXYGEN
#include "prolog.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "tinedefs.h"
#include "srvcore.h"
#include "errors.h"
#include "swaplib.h"
#include "dbtools.h"
#include "toolkit.h"

/**
 * \internal
 */
DBITFIELD *gBitFieldList = NULL;

char feclogfile[80] = "";

static char *months[12] = {"Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"};
TINE_EXPORT time_t parseCtimeStr(char *cTimeStr)
{
  struct tm it;
  char *c = NULL;
  int i;

  for (i=0; i<12; i++) if ((c=strstr(cTimeStr,months[i])) != NULL) break;
  if (c == NULL) return (time_t)0;
  it.tm_mon = i; c += 3;
  it.tm_mday = atoi(c); c += 4;
  it.tm_hour = atoi(c); c += 3;
  it.tm_min = atoi(c);  c += 3;
  it.tm_sec = atoi(c);
  /* if this string comes from 'date' then there's a TZ field in front of the year */
  c = &cTimeStr[strlen(cTimeStr)-5];
  it.tm_year = atoi(c) - 1900;
  it.tm_isdst = -1;
  return mktime(&it);
}
TINE_EXPORT time_t parseCDateAndTimeStr(char *cDateStr,char *cTimeStr)
{
  struct tm it;
  char *c = NULL;
  int i;

  for (i=0; i<12; i++) if ((c=strstr(cDateStr,months[i])) != NULL) break;
  if (c == NULL) return (time_t)0;
  it.tm_mon = i; c += 3;
  it.tm_mday = atoi(c); c += 4;
  it.tm_year = atoi(c) - 1900;
  if ((c=cTimeStr) == NULL) return (time_t)0;
  it.tm_hour = atoi(c); c += 3;
  it.tm_min = atoi(c);  c += 3;
  it.tm_sec = atoi(c);  c += 3;
  it.tm_isdst = -1;
  return mktime(&it);
}
#ifndef SMALL_TINE_LIB
TINE_EXPORT int PropertyDataExchange(void *data,short fmt,short siz,
                        DTYPE *dout, DTYPE *din,short access)
{
  BYTE *dst=NULL,*src=NULL;
  int nbytes=0;
  if (din->dArrayLength > 0 && data != NULL)
  {
    if (!(access & CA_WRITE)) return illegal_read_write;
    switch(fmt)
    {
      case CF_INT16: 
        return GetValuesAsShort(din,(short *)data,siz);
      case CF_INT32:
        return GetValuesAsLong(din,(SINT32 *)data,siz);
      case CF_FLOAT:
        return GetValuesAsFloat(din,(float *)data,siz);
      case CF_DOUBLE:
        return GetValuesAsDouble(din,(double *)data,siz);
      default:
        if (din->dFormat != fmt) return illegal_format;
        dst = (BYTE *)data; src = din->data.bptr; nbytes = (int)(din->dArrayLength*fmtsizeof(fmt));
        goto err;
    }
  }
  if (dout->dArrayLength > 0 && data != NULL)
  {
    switch(fmt)
    {
      case CF_INT16: 
        return PutValuesFromShort(dout,(short *)data,siz);
      case CF_INT32:
        return PutValuesFromLong(dout,(SINT32 *)data,siz);
      case CF_FLOAT:
        return PutValuesFromFloat(dout,(float *)data,siz);
      case CF_DOUBLE:
        return PutValuesFromDouble(dout,(double *)data,siz);
      default:
        if (dout->dFormat != fmt) return illegal_format;
        dst = dout->data.bptr; src = (BYTE *)data; nbytes = (int)(dout->dArrayLength*fmtsizeof(fmt));
        goto err;
    }
  }
err:
  if (dst != NULL && src != NULL && nbytes > 0) memcpy(dst,src,nbytes);
  return 0;
}
int isAbsolutePath(char *fname);
extern char gFeclogPath[80];
time_t getTimeStampFromLogEntry(char *str)
{
  struct tm tms;
  tms.tm_mday = atoi(str);
  tms.tm_mon = atoi(&str[3]) - 1;
  tms.tm_year = 100 + atoi(&str[6]);
  tms.tm_hour = atoi(&str[9]);
  tms.tm_min = atoi(&str[12]);
  tms.tm_sec = atoi(&str[15]);
  tms.tm_isdst = memcmp(&str[22],ltzname[1],3) == 0 ? 1 : 0;
  return mktime(&tms);
}
int findLogFileEntry(FILE *fp,int fpos,time_t timestamp)
{
  int pos=fpos;
  time_t t=0;
  char str[256];

  fseek(fp,pos,SEEK_SET);
  while (fgets(str,255,fp) != NULL)
  {
    t = getTimeStampFromLogEntry(str);
    if (t > timestamp) return pos;
    pos += (int)strlen(str);
  }
  return pos;
}
int readFecLogFileEx(char *fname,char *textbuf,time_t start,time_t stop,int sizeInBytes)
{
  int cc=0,done=0,useAbsolutePath=0;
  FILE *fp = NULL, *fpb = NULL;
  char str[256],fn[128],bakname[96], *c;
  int sz=0,pos=0;
  int pos1=0,pos2=0;
  static char *fmod = "r";

  if (nofeclog) return 0;
  if (fname == NULL) return no_such_file;
  if (stop <= start) stop = time(NULL);
  useAbsolutePath = isAbsolutePath(fname);
  if (useAbsolutePath) strncpy(fn,fname,128); else sprintf(fn,"%s%s",gFeclogPath,fname);
  if ((fp=fopen(fn,fmod)) == NULL) goto checkbak;
  pos1 = findLogFileEntry(fp,0,start);
  pos2 = findLogFileEntry(fp,pos1,stop);
  pos = 0;
checkbak:
  if (pos1 == 0)  /* look for back file */
  {
    strncpy(bakname,fname,95); bakname[95] = 0;
    if ((c=strstr(bakname,".log")) != NULL) *c = 0;
    strcat(bakname,".bak");
    if (useAbsolutePath) strncpy(fn,bakname,128); else sprintf(fn,"%s%s",gFeclogPath,bakname);
    if ((fpb=fopen(fn,fmod)) == NULL) goto filesread;
    pos1 = findLogFileEntry(fpb,0,start);
    pos2 = findLogFileEntry(fpb,pos1,stop);
    fseek(fpb,pos1,SEEK_SET);
    while (fgets(str,255,fp) != NULL)
    {
      if ((sz=(int)strlen(str)) + pos > sizeInBytes) 
      {
        done = TRUE;
        break;
      }
      strncpy(&textbuf[pos],str,sz);
      if ((pos += sz) > pos2) 
      {
        done = TRUE;
        break;
      }
    }
    fclose(fpb);
    pos1 = 0;
  }
filesread:
  if (fp)
  {
    pos2 = findLogFileEntry(fp,pos1,stop);
    fseek(fp,pos1,SEEK_SET);
    while (fgets(str,255,fp) != NULL && !done)
    {
      if ((sz=(int)strlen(str)) + pos > sizeInBytes) 
      {
        done = TRUE;
        break;
      }
      strncpy(&textbuf[pos],str,sz);
      if ((pos += sz) > pos2) 
      {
        done = TRUE;
        break;
      }
    }
    fclose(fp);
  }
  textbuf[pos] = 0;
  if (sz == 0) cc = no_such_file;
  if (cc) dbglog("%s : %s",fn,erlst[cc]);
  return cc;
}
int GetTagList(char *context,char *tagtype,int *tagfmt,void *taglist,int *num)
{
  int cc = 0, fmt, retryfmt;
  short n = 0;
  char *c,prop[32],srv[64],*sub = NULL;
  DTYPE dout,din, *dinptr = NULL;

  if (tagfmt == NULL) return argument_list_error;
  if (num == NULL || *num <= 0) return buffer_too_small;
  switch (*tagfmt)
  {
    case CF_NAME16:
    case CF_NAME32:
      retryfmt = CF_NAME16;
      break;
    case CF_NAME64:
      retryfmt = CF_NAME32;
      break;
    default:
      return illegal_format;
  }
  strncpy(srv,ENSTAG,16);
  /* is a context given ? */
  if (context != NULL && strlen(context) > 0) 
  { /* is a subsystem given ? */
    if ((c=strchr(context,'/')) != NULL)
    { /* has the form <context>/<subsystem> */
      *c++ = 0; sub = c;
    }
    if (strlen(context) > 0) 
    { /* might have been given as /<subsystem> ! */
      strcat(srv,"/");
      strncat(srv,context,CONTEXT_NAME_SIZE);
    }
  }
  if (sub != NULL)
  {
    din.dFormat = CF_TEXT;
    din.dArrayLength = (UINT32)strlen(sub);
    din.data.cptr = sub;
    dinptr = &din;
  }
  dout.dFormat = CF_INT16;
  dout.dArrayLength = 1;
  dout.data.vptr = &n;
  sprintf(prop,"N%s",tagtype);
  fmt = *tagfmt;
retryIllegalFmt:
  if ((cc=ExecLinkEx(srv,prop,&dout,dinptr,CA_READ|CA_RETRY,500)) != 0) return cc;
  if (n > *num) n = *num;
  if (taglist != NULL)
  { /* Get Tags from name server (synchronous call) */
    dout.dFormat = fmt;
    dout.dArrayLength = n;
    dout.data.vptr = taglist;
    if ((cc=ExecLinkEx(srv,tagtype,&dout,dinptr,CA_READ|CA_RETRY,500)) != 0) 
    {
      if (cc == illegal_format && fmt != retryfmt) 
      {
        fmt = retryfmt;
        goto retryIllegalFmt;
      }
      return cc;
    }
  }
  *tagfmt = fmt;
  *num = n;
  return 0;
}
TINE_EXPORT int GetDevicePropertyInformationEx(char *srv,char *prp,PropertyQueryStructEx *srvProps,int *num)
{
  DTYPE dout, din, *dinptr = NULL;
  int cc = 0;
  short n;
  BYTE *data = (BYTE *)srvProps;

# ifndef INTEL
  data = tmpWorkArea;
# endif
  if (num == NULL || (n = *num) <= 0) return buffer_too_small;
  if (prp != NULL)
  {
    din.dArrayLength = 1;
    din.dFormat = CF_NAME32;
    din.data.cptr = prp;
    memset(din.dTag,0,TAG_NAME_SIZE);
    dinptr = &din;
  }
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = n*PROPERTYQUERYSTRUCTEX_SIZE;
  dout.data.vptr = data;
  if ((cc=ExecLinkEx(srv,"PROPS",&dout,dinptr,CA_READ|CA_RETRY,800)) != 0) return cc;
  *num = GetCompletionDataSize(-1);
# ifndef INTEL
  if (*num > 0) PrepIncomingPQSX(srvProps,data,*num);
# endif
  return 0;
}
/**
 * \internal
 */
int parseFullDeviceName(char *fullname,char *dev,char *prp,int isHistoryCall)
{
  char *c,tmp[96];
  int isBracketed = FALSE;

  strncpy(tmp,fullname,96);
  if ((c=strchr(tmp,'[')) != 0)
  { /* fullname given as "/<context>/<server>/<device>[<property>]" */
    *c++ = 0;
    strncpy(prp,c,32);
    if ((c=strchr(prp,']')) != 0) *c = 0;
    isBracketed = TRUE;
  }
  if (!isBracketed)
  { /* fullname given as "/<context>/<server>/<device>/<property>" */
    strrev(tmp);
    if ((c=strchr(tmp,'/')) == 0) return illegal_delimiter;
    *c++ = 0;
    strncpy(prp,tmp,32); strrev(prp);
    strncpy(dev,c,64); strrev(dev);
  }
  else
  {
    strncpy(dev,tmp,64);
  }
  if (isHistoryCall)
  {
    if (strstr(dev,"HISTORY") == NULL && /* not a central archive call */
      strstr(prp,".HIST") == NULL && strstr(prp,".HST") == NULL) /* meta property not yet appended */
    {
      if (strlen(prp) < 28) strcat(prp,".HST");
    }
  }
  return 0;
}
#endif /* not SMALL_TINE_LIB */

int GetIntValue(DTYPE *d,SINT32 *ival)
{
  if (d == NULL || ival == NULL) return argument_list_error;
  if (d->dArrayLength < 1) return dimension_error;
  switch (d->dFormat)
  {
    case CF_FLTINT: 
      *ival = (SINT32)((FLTINT *)d->data.vptr)->ival;
      return 0;
    case CF_INTINT: 
      *ival = (SINT32)((INTINT *)d->data.vptr)->i2val;
      return 0;
    case CF_FLTFLTINT: 
      *ival = (SINT32)((FLTFLTINT *)d->data.vptr)->ival;
      return 0;
    case CF_INTINTINT: 
      *ival = (SINT32)((INTINTINT *)d->data.vptr)->i3val;
      return 0;
    case CF_NAME16FI:
      *ival = (SINT32)((NAME16FI *)d->data.vptr)->ival;
      return 0;
    case CF_NAME16II:
      *ival = (SINT32)((NAME16II *)d->data.vptr)->i2val;
      return 0;
    case CF_INTFLTINT:
      *ival = (SINT32)((INTFLTINT *)d->data.vptr)->i2val;
      return 0;
    case CF_FLTINTINT:
      *ival = (SINT32)((FLTINTINT *)d->data.vptr)->i2val;
      return 0;
    default:
      break;
  }
  return illegal_format;
}
/**
 * \internal
 */
void augmentNameSize(int fmt,char *list,int siz)
{
  int i,fromlen,tolen;
  char scratch[64];
  if (list == NULL || siz <= 0) return;
  switch (fmt)
  {
    case CF_NAME16:
      fromlen = 16; tolen = 32; break;
    case CF_NAME32:
      fromlen = 32; tolen = 64; break;
    default: return;
  }
  for (i=siz-1; i >= 0; i--)
  {
    memset(scratch,0,64);
    memcpy(scratch,list + i*fromlen, fromlen);
    memcpy(list + i*tolen,scratch, tolen);
  }
}
/**
 * \internal
 */
int getBitShiftFromMask(UINT32 mask,int fmt)
{
  int shf = 0, b = 1, mshf;

  mshf = fmtsizeof(fmt) * 8 - 1;
  while ((mask & b) == 0)
  {
    if (++shf == mshf) break;
    b *= 2;
  }
  return shf;
}
/**
 * \internal
 */
DBITFIELD *findBitField(char *srv,char *tag)
{
  DBITFIELD *bf;
  if (tag == NULL || strlen(tag) == 0) return NULL;
  for (bf=gBitFieldList; bf; bf=bf->next)
  {
     if (strncmp(bf->name,tag,TAG_NAME_SIZE)) continue;
     if (srv != NULL && strncmp(bf->srvName,srv,2*CONTEXT_NAME_SIZE+2)) continue;
     break;
  }
  return bf;
}
/**
 * \internal
 */
int queryBitfieldFormat(char *srv,char *tag,NAME16I *bffq,int *bffqsiz)
{
  int i,n,nq,cc=0;
  DBITFIELD *bf;
  struct bitfieldFormat *bff;
  
  if (bffqsiz == NULL) ccerr(argument_list_error);
  nq = *bffqsiz;
  if ((bf=findBitField(srv,tag)) == NULL) ccerr(invalid_structure_tag);
  for (bff=bf->bfs,n=0; bff != NULL; bff=bff->nxt,n++); /* count them first (reverse ordered!) */
  for (i=0,bff=bf->bfs; i<nq && i<n && bff != NULL; i++, bff=bff->nxt)
  {
    strncpy(bffq[n-i-1].name,bff->field,16);
    bffq[n-i-1].ival = bff->bitmask;
  }
  if (i == nq) i--;
  strncpy(bffq[i].name,tag,16);
  bffq[i].ival = bf->fmt;
  *bffqsiz = n+1;
err:
  return cc;
}
/**
 * \internal
 */
UINT32 getBitfieldValue(int fmt,BYTE *dptr)
{
  switch (fmt)
  {
    case CF_BYTE:
    case CF_BITFIELD8:   return (UINT32)(*dptr);
    case CF_INT16:
    case CF_BITFIELD16: return (UINT32)(*(UINT16 *)dptr);
    case CF_INT32:
    case CF_BITFIELD32: return *(UINT32 *)dptr;
    default: break;
  }
  return (UINT32)(*dptr);
}
/**
 * \internal
 */
void mapBitfieldToTarget(char *ctx,char *exp,char *tag,char *field,int fmt,BYTE *dptr)
{
  UINT32 v;
  char srv[CONTEXT_NAME_SIZE*2+4];
  if (tag == NULL || strlen(tag) == 0) return;
  if (field == NULL || strlen(field) == 0) return;
  if (dptr == NULL) return;
  if (ctx != NULL && exp != NULL) sprintf(srv,"/%s/%s",ctx,exp); else srv[0] = 0;
  switch (fmt)
  {
    case CF_BYTE:
    case CF_BITFIELD8:   
      v = (UINT32)(*dptr);
      v = getFieldFromBitfield(srv,tag,field,v);
      *dptr = (BYTE)v;
      break;
    case CF_INT16:
    case CF_BITFIELD16: 
      v = (UINT32)(*(UINT16 *)dptr);
      v = getFieldFromBitfield(srv,tag,field,v);
      *(UINT16 *)dptr = (UINT16)v;
      break;
    case CF_INT32:
    case CF_BITFIELD32: 
      v = *(UINT32 *)dptr;
      v = getFieldFromBitfield(srv,tag,field,v);
      *(UINT32 *)dptr = v;
      break;
    default: 
      return;
  }
  return;
}
/**
 * \internal
 */
int acquireAndRegisterBitfield(const char *devName, char *tag)
{
  int i, n, fmt, cc=0;
  DTYPE dout,din;
  NAME16I n16i[32];
  char srv[2*CONTEXT_NAME_SIZE+4];

  if (devName == NULL || tag == NULL) ccerr(argument_list_error);
  if ((cc=splitFullName(devName,srv,NULL)) != 0) ccerr(cc);
  if (findBitField(srv,tag) != NULL) return already_assigned;

  dout.dFormat = CF_NAME16I;
  dout.dArrayLength = 32;
  memset(dout.dTag,0,TAG_NAME_SIZE);
  dout.data.vptr = n16i;

  din.dFormat = CF_TEXT;
  din.dArrayLength = strlen(tag);
  memset(din.dTag,0,TAG_NAME_SIZE);
  din.data.cptr = tag;

  if ((cc = ExecLinkEx(devName,"BITFIELDFORMAT",&dout,&din,CA_READ|CA_MUTABLE|CA_RETRY,200)) != 0)
    ccerr(cc);

  n = dout.dArrayLength;
  fmt = n16i[n-1].ival;
  OpenBitField(srv,tag,fmt);
  for (i=0; i<n-1; i++)
  {
    addFieldToBitField(srv,tag,n16i[i].ival,n16i[i].name); 
  }
err:
  if (cc) feclog("acquireAndRegisterBitfield %s %s : %s",devName,tag,erlst[cc]);
  return cc;
}
int isBitfieldAccess(DTYPE *d)
{
  if (d == NULL) return FALSE;
  if (d->dFormat < CF_BITFIELD8 || d->dFormat > CF_BITFIELD32) return FALSE;
  if (d->dTag[0] == 0) return FALSE;
  return TRUE;
}
/**
 * \internal
 */
int isBitfieldProperty(char *srv,char *tag,char *prp,char *fld)
{
  DBITFIELD *bf = findBitField(srv,tag);
  struct bitfieldFormat *bff;
  int i, p;

  if (prp == NULL || bf == NULL) return FALSE;
  if ((p=strlen(prp)) == 0) return FALSE;
  p = MIN(p,PROPERTY_NAME_SIZE) - 1;
  for (i=p; i>0; i--) if (prp[i] == '.') break;
  if (i == 0 || i == p) return FALSE;
  strncpy(fld,&prp[i+1],FIELD_NAME_SIZE); prp[i] = 0;
  for (bff=bf->bfs; bff != NULL; bff=bff->nxt)
  {
    if (!strncmp(bff->field,fld,FIELD_NAME_SIZE)) return TRUE;
  }
  return FALSE;  
}
/**
 * \internal
 */
int isTaggedStructAccess(DTYPE *d)
{
  if (d == NULL) return FALSE;
  if (d->dFormat != CF_STRUCT) return FALSE;
  if (d->dTag[0] == 0) return FALSE;
  return TRUE;
}
/**
 * \internal
 */
int isStructProperty(char *tag,char *prp,char *fld)
{
  struct structStruct *ss = findstruct(tag);
  struct structFormat *sf;
  int i, p;

  if (prp == NULL || ss == NULL) return FALSE;
  if ((p=strlen(prp)) == 0) return FALSE;
  p = MIN(p,PROPERTY_NAME_SIZE) - 1;
  for (i=p; i>0; i--) if (prp[i] == '.') break;
  if (i == 0 || i == p) return FALSE;
  strncpy(fld,&prp[i+1],FIELD_NAME_SIZE); prp[i] = 0;
  sf = ss->s;
  for (sf=ss->s; sf != NULL; sf=sf->nxt)
  {
    if (!strncmp(sf->field,fld,FIELD_NAME_SIZE)) return TRUE;
  }
  return FALSE;  
}
/**
 * \internal
 */
int acquireAndRegisterStruct(char *devName, char *tag,int num)
{
  int i, n, siz, offset, cc=0;
  DTYPE dout,din;
  NAME16II n16ii[64];

  if (devName == NULL || tag == NULL) ccerr(argument_list_error);
  if (findstruct(tag) != NULL) return already_assigned;

  dout.dFormat = CF_NAME16II;
  dout.dArrayLength = 64;
  memset(dout.dTag,0,TAG_NAME_SIZE);
  dout.data.vptr = n16ii;

  din.dFormat = CF_TEXT;
  din.dArrayLength = strlen(tag);
  memset(din.dTag,0,TAG_NAME_SIZE);
  din.data.cptr = tag;

  if ((cc = ExecLinkEx(devName,"STRUCTFORMAT",&dout,&din,CA_READ|CA_MUTABLE|CA_RETRY,200)) != 0)
    ccerr(cc);

  n = dout.dArrayLength;
  siz = n16ii[n-1].i1val;
  offset = 0;
  for (i=0; i<n-1; i++)
  {
    addFieldToStruct(tag,offset,n16ii[i].i1val,n16ii[i].i2val,n16ii[i].name);
    offset += n16ii[i].i1val * fmtsizeof(n16ii[i].i2val);
  }
  if ((cc=sealTaggedStruct(tag,siz,num)) != 0) ccerr(cc);
err:
  if (cc) feclog("acquireAndRegisterStruct %s %s : %s",devName,tag,erlst[cc]);
  return cc;
}
/**
 * TODO:
 * This split logic shows up elsewhere in other forms !
 * Make everyone use the same routine !
 */
int splitFullName(const char *fullName,char *srvName,char *devName)
{
  char *c0,*c1,dev[128];
  int cc = 0;
  if (fullName == NULL) ccerr(argument_list_error);
  strncpy(dev,fullName,128); strtrm(dev);
  if ((c0=strchr(&dev[1],'/')) == NULL) 
  { /* just the device name was given */
    if (srvName != NULL) srvName[0] = 0;
    if (devName != NULL) strncpy(devName,fullName,DEVICE_NAME_SIZE);
    ccerr(0);
  }
  c0++;
  if ((c1=strchr(c0,'/')) != NULL) *c1++ = 0;
  if (dev[0] == '/') 
  { /* normal case: context given */
    if (srvName != NULL) strncpy(srvName,dev,2*CONTEXT_NAME_SIZE+2);
    if (devName != NULL) strncpy(devName,c1,DEVICE_NAME_SIZE);
  }
  else
  { /* no context given */
    *c0++ = 0;
    if (srvName != NULL) strncpy(srvName,dev,2*CONTEXT_NAME_SIZE+2);
    if (devName != NULL) strncpy(devName,c0,DEVICE_NAME_SIZE);
  }
err:
  return cc;
}
#endif /* NO_DOXYGEN */

int fecLogFileDepth = 500;
TINE_EXPORT int AddFieldToBitField(char *srv,char *tag,UINT32 mask,char *field)
{
  DBITFIELD *bf = findBitField(srv,tag);
  struct bitfieldFormat *bff;
  if (bf == NULL) return invalid_structure_tag;
  if (field == NULL || strlen(field) == 0 || mask == 0) return argument_list_error;
  for (bff=bf->bfs; bff != NULL; bff=bff->nxt)
  {
    if (!strncmp(field,bff->field,FIELD_NAME_SIZE)) return name_already_exists;
    if (mask == bff->bitmask) return already_assigned;
  }
  bff = (struct bitfieldFormat *)SystemCalloc(1,sizeof(struct bitfieldFormat));
  if (bff == NULL) return out_of_local_memory;
  bff->bitmask = mask;
  bff->bitshift = getBitShiftFromMask(mask,bf->fmt);
  strncpy(bff->field,field,FIELD_NAME_SIZE);
  bff->nxt = bf->bfs;
  bf->bfs = bff;
  return 0;
}
TINE_EXPORT int GetBitfieldAsString(char *srv,char *tag,UINT32 value,char *strbuf,int buflen)
{
  DBITFIELD *bf = findBitField(srv,tag);
  struct bitfieldFormat *bff;
  int n;
  if (bf == NULL) return invalid_structure_tag;
  if (strbuf == NULL) return argument_list_error;
  for (bff=bf->bfs, n = 0; bff != NULL; bff=bff->nxt, n++);
  if (buflen < n * 32) return buffer_too_small;
  n = 0;
  for (bff=bf->bfs; bff != NULL; bff=bff->nxt)
  {
    sprintf(&strbuf[n],"%s : %d\n",bff->field,getFieldFromBitfield(srv,tag,bff->field,value));
    n = strlen(strbuf);
  }
  return 0;
}
TINE_EXPORT int GetFieldFromBitfield(char *srv,char *tag,char *field,UINT32 value)
{
  DBITFIELD *bf = findBitField(srv,tag);
  struct bitfieldFormat *bff;
  UINT32 bfvalue;
  if (bf == NULL) return -invalid_structure_tag;
  if (field == NULL || strlen(field) == 0) return -argument_list_error;
  for (bff=bf->bfs; bff != NULL; bff=bff->nxt)
  {
    if (!strncmp(field,bff->field,FIELD_NAME_SIZE)) break;
  }
  if (bff == NULL) return -invalid_field;
  bfvalue = bff->bitmask & value;
  bfvalue = bfvalue >> bff->bitshift;
  return bfvalue;
}
TINE_EXPORT int OpenBitField(char *srv,char *tag,int fmt)
{
  DBITFIELD *bf = findBitField(srv,tag);

  if (bf != NULL) return 0;
  if (!isIntegerFormat((short)fmt))
  {
    feclog("OpenBitField: cannot register bitfield with format %s",itof(fmt));
    return illegal_format;
  }
  if (fmt == CF_BYTE) fmt = CF_BITFIELD8;
  if (fmt == CF_INT16) fmt = CF_BITFIELD16;
  if (fmt == CF_INT32) fmt = CF_BITFIELD32;
  if ((bf=(DBITFIELD *)SystemCalloc(1,sizeof(DBITFIELD))) != NULL)
  { /* didn't find it -> need to allocated it */
    strncpy(bf->name,tag,TAG_NAME_SIZE);
    if (srv != NULL) strncpy(bf->srvName,srv,2*CONTEXT_NAME_SIZE+4);
    bf->fmt = fmt;
    bf->next = gBitFieldList;
    gBitFieldList = bf;
    if (tineDebug) dbglog("OpenBitField : New bitfield tag registered: %s",tag);
  }
  else
  {
    feclog("OpenBitField: cannot allocate bit field memory!");
    return out_of_local_memory;
  }
  return 0;
}
#if !defined(SMALL_TINE_LIB)
TINE_EXPORT void AppendHistoryLog(char *text,...)
{
# ifndef FS_RDONLY
  double ts = makeDataTimeStamp();
  FILE *fp = NULL, *fpb = NULL;
  SINT32 flen;
  int k, cc = 0;
  char str[512],tmpfilename[128];
  va_list args;

  if (strlen(feclogfile) == 0) sprintf(feclogfile,"%s.log",gFecName);
  if ((fp=fopen(feclogfile,"a")) == NULL)
  {
    perror("LOG: fopen");
    ccerr(file_error);
  }
  sprintf(str,"%s[%s] ",getDataTimeString(ts,0),gFecName);
# if defined(UNIX)
  va_start(args,text);
  vsnprintf(&str[strlen(str)],256,text,args); str[256] = 0;
  va_end(args);
# elif defined(WIN32)
  va_start(args,text);
  _vsnprintf(&str[strlen(str)],256,text,args); str[256] = 0;
  va_end(args);
# else
  if (strlen(text) > 128) 
  {
    if (strchr(text,'%') == NULL)
    { /* then truncate */
      strncpy(&str[strlen(str)],text,256); str[256] = 0;
    }
    ccerr(string_too_long);
  }
  else
  {
    va_start(args,text);
    vsprintf(&str[strlen(str)],text,args);
    va_end(args);
  }
# endif

  if (str[strlen(str)-1] != '\n') strcat(str,"\n");
  fprintf(fp,str);

  fseek(fp, 0L, SEEK_END);
  flen = ftell(fp);

  if (flen > 2*fecLogFileDepth*80)
  {
    fclose(fp);
    strncpy(tmpfilename,feclogfile,79); tmpfilename[79] = 0;
    strncpy(&tmpfilename[strlen(tmpfilename)-3],"BAK",3);
    rename(feclogfile,tmpfilename);
    if ((fp=fopen(feclogfile,"w")) == NULL) ccerr(file_error);
    if ((fpb=fopen(tmpfilename,"r")) == NULL) ccerr(no_such_file);
    for (k=0; fgets(str,255,fpb) != NULL; k++)
    {
      if (k < fecLogFileDepth) continue;
      fwrite(str,strlen(str),1,fp);
    }

  }
err:
  if (fp) fclose(fp);
  if (fpb) { fclose(fpb); remove(tmpfilename); }
  if (cc) dbglog("AppendHistoryLog : %s",erlst[cc]);
# endif
  return;
}
TINE_EXPORT int GetArchivedData(char *devsrv,int index,time_t start,time_t stop,FLTINT *fiDataArray,short *num)
{
  DTYPE din,dout;
  UINT32 startstopArray[3];
  char aServer[64],prp[32];
  int cc;

  if (num == NULL || *num <= 0) return buffer_too_small;
  if ((cc=parseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;

  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)stop;
  startstopArray[2] = index;
  dout.dArrayLength = *num;
  dout.dFormat = CF_FLTINT;
  dout.data.vptr = fiDataArray;
  din.dArrayLength = 3;
  din.dFormat = CF_INT32;
  din.data.ulptr = startstopArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_RETRY,200)) != 0) *num = 0;
  else *num = (int)GetCompletionDataSize(-1);

  return cc;
}
TINE_EXPORT int GetArchivedTraceDataAsFloat(char *devsrv,time_t start,time_t stop,float *fDataArray,int num)
{
  DTYPE din,dout;
  UINT32 startstopArray[3];
  char aServer[64],prp[32];
  int cc;

  if (num <= 0) return buffer_too_small;
  if ((cc=parseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)stop;
  dout.dArrayLength = num;
  dout.dFormat = CF_FLOAT;
  dout.data.vptr = fDataArray;
  din.dArrayLength = 2;
  din.dFormat = CF_INT32;
  din.data.ulptr = startstopArray;

  return ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_RETRY,200);
}
TINE_EXPORT int GetArchivedDataAsSnapshot(char *devsrv,time_t *target,float *fDataArray,int size)
{
  DTYPE din,dout;
  UINT32 startstopArray[3];
  char aServer[64],prp[32];
  int cc;

  if (size <= 0) return buffer_too_small;
  if (target == NULL) return argument_list_error;
  if ((cc=parseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  startstopArray[0] = *target;
  startstopArray[1] = *target;
  dout.dArrayLength = size;
  dout.dFormat = CF_FLOAT;
  dout.data.vptr = fDataArray;
  din.dArrayLength = 2;
  din.dFormat = CF_INT32;
  din.data.ulptr = startstopArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_RETRY,200)) != 0) *target = 0;
  else *target = (time_t)gDataTimeStamp;

  return cc;
}
TINE_EXPORT int GetArchivedDataAsFloat(char *devsrv,time_t start,time_t stop,FLTINT *fiDataArray,int *num)
{
  DTYPE din,dout;
  UINT32 startstopArray[3];
  char aServer[64],prp[32];
  int cc;

  if (num == NULL || *num <= 0) return buffer_too_small;
  if ((cc=parseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)stop;
  dout.dArrayLength = *num;
  dout.dFormat = CF_FLTINT;
  dout.data.vptr = fiDataArray;
  din.dArrayLength = 2;
  din.dFormat = CF_INT32;
  din.data.ulptr = startstopArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_RETRY,200)) != 0) *num = 0;
  else *num = (int)GetCompletionDataSize(-1);

  return cc;
}
TINE_EXPORT int GetArchivedDataAsText(char *devsrv,time_t start,time_t stop,NAME32I *niDataArray,int *num)
{
  DTYPE din,dout;
  UINT32 startstopArray[3];
  char aServer[64],prp[32];
  int cc;

  if (num == NULL || *num <= 0) return buffer_too_small;
  if ((cc=parseFullDeviceName(devsrv,aServer,prp,TRUE)) != 0) return cc;
  startstopArray[0] = (UINT32)start;
  startstopArray[1] = (UINT32)stop;
  dout.dArrayLength = *num;
  dout.dFormat = CF_NAME32I;
  dout.data.vptr = niDataArray;
  din.dArrayLength = 2;
  din.dFormat = CF_INT32;
  din.data.ulptr = startstopArray;

  if ((cc=ExecLinkEx(aServer,prp,&dout,&din,CA_READ|CA_RETRY,200)) != 0) *num = 0;
  else *num = (int)GetCompletionDataSize(-1);

  return cc;
}
TINE_EXPORT int GetDeviceContexts(NAME16 *clist,int *num)
{ /* deprecated */
  int fmt = CF_NAME16;
  int cc = GetTagList(NULL,"CONTEXTS",&fmt,clist,num);
  return cc;
}
TINE_EXPORT int GetDeviceContextsFromFile(NAME16 *clist,int *num)
{
  FILE  *fp=NULL;
  int i=0,j,n,cc=0,done=0,con_col;
  char fname[256],str[256],hdr[256],scratch[32];

  if (num == NULL || clist == NULL) ccerr(argument_list_error);
  sprintf(fname,"%s%s",tineHomePath,EQPDB_FILE);
  if ((fp=fopen(fname,"r")) == NULL) ccerr(no_such_file);
  n = *num;
  for (done=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((con_col=findcol(hdr,"CONTEXT")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(scratch,colptr(con_col,str),16);
    for (j=0; j<i && strncmp(scratch,clist[j].name,16); j++);
    if (j < i) continue;
    strncpy(clist[i].name,scratch,16);
    if (++i > n) break;
  }
  *num = i;
  fclose(fp);
  if (cc) ccerr(cc);
err:
  return cc;
}
TINE_EXPORT int GetMyServerAddress(char *eqmName,char *expName,char *ctxName,char *fecName,int *port)
{
  int i,cc=0,pass=0;
  NAME32 n32[5];
  NAME16 n16[5];
  char *c = NULL;
  DTYPE dout, din, *dinptr=NULL;

retryIllegalFormat:
  if (pass == 0)
  {
    if (eqmName != NULL && strlen(eqmName) > 0)
    {
      strcpy(n32[0].name,eqmName);
      c = n32[0].name;
      din.dArrayLength = 1;
      din.dFormat = CF_NAME32;
      din.data.cptr = n32[0].name;
      memset(din.dTag,0,TAG_NAME_SIZE);
      dinptr = &din;
    }
    dout.dArrayLength = 5;
    dout.dFormat = CF_NAME32;
    dout.data.cptr = n32[0].name;
    memset(dout.dTag,0,TAG_NAME_SIZE);
  }
  else
  {
    if (eqmName != NULL && strlen(eqmName) > 0)
    {
      strcpy(n16[0].name,eqmName);
      c = n16[0].name;
      din.dArrayLength = 1;
      din.dFormat = CF_NAME16;
      din.data.cptr = n16[0].name;
      memset(din.dTag,0,TAG_NAME_SIZE);
      dinptr = &din;
    }
    dout.dArrayLength = 5;
    dout.dFormat = CF_NAME16;
    dout.data.cptr = n16[0].name;
    memset(dout.dTag,0,TAG_NAME_SIZE);
  }
  for (i=0,cc=link_not_open; cc != 0 && i <5; i++)
  {
    cc = tExecLink("ENS",CSHOSTS,"#0","WHOAMI",&dout,dinptr,CA_READ|CA_RETRY,200);
    if (cc == illegal_format && pass == 0)
    {
      pass++;
      goto retryIllegalFormat;
    }
  }
  if (cc == 0)
  {
    if (pass > 0)
    {
      for (i=0; i<5; i++) 
      {
        memcpy(n32[i].name,n16[i].name,16); n32[i].name[16] = 0;
      }
    }
    if (expName != NULL) strncpy(expName,n32[0].name,EXPORT_NAME_SIZE);
    if (fecName != NULL) strncpy(fecName,n32[2].name,FEC_NAME_SIZE);
    if (ctxName != NULL) strncpy(ctxName,n32[3].name,CONTEXT_NAME_SIZE);
    if (port != NULL) *port = atoi(n32[4].name);
  }
  return cc;
}
TINE_EXPORT int GetDevicePropertyEGU(char *srv,char *prp,float *max,float *min,char *egu)
{
  DTYPE dout;
  DUSTRING ustr;
  int cc = 0;
  char uprp[36];

  dout.dFormat = CF_USTRING;
  dout.dArrayLength = 1;
  dout.data.vptr = &ustr;
  strncpy(uprp,prp,28); uprp[28] = 0; strcat(uprp,".EGU");
  if ((cc=ExecLinkEx(srv,uprp,&dout,NULL,CA_READ|CA_RETRY,800)) != 0) return cc;
  if (max != NULL) *max = ustr.f1val;
  if (min != NULL) *min = ustr.f2val;
  if (egu != NULL) strncpy(egu,ustr.str,80);
  return 0;
}
TINE_EXPORT int GetDevicePropertyInformation(char *srv,PropertyQueryStruct *srvProps,int *num)
{
  DTYPE dout;
  int cc = 0;
  short n, i;

  if (num == NULL || (n = *num) <= 0) return buffer_too_small;
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = n*PROPERTYQUERYSTRUCT_SIZE;
  dout.data.vptr = srvProps;
  if ((cc=ExecLinkEx(srv,"PROPS",&dout,NULL,CA_READ|CA_RETRY,800)) != 0) return cc;
  for (i=0; i<n; i++) srvProps[i].prpSize = ISWAP(srvProps[i].prpSize);
  *num = GetCompletionDataSize(-1);
  return 0;
}
TINE_EXPORT int GetDevicePropertyInformationX(char *srv,char *prp,XPropertyQueryStruct *srvProps,int *num)
{
  DTYPE dout, din, *dinptr = NULL;
  int cc = 0;
  short n;
  BYTE *data = (BYTE *)srvProps;

# ifndef INTEL
  data = tmpWorkArea;
# endif
  if (num == NULL || (n = *num) <= 0) return buffer_too_small;
  if (prp != NULL)
  {
    din.dArrayLength = 1;
    din.dFormat = CF_NAME32;
    din.data.cptr = prp;
    memset(din.dTag,0,TAG_NAME_SIZE);
    dinptr = &din;
  }
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = n;
  strncpy(dout.dTag,"XPQS",TAG_NAME_SIZE);
  dout.data.vptr = data;
  if ((cc=ExecLinkEx(srv,"PROPS",&dout,dinptr,CA_READ|CA_RETRY,800)) != 0) return cc;
  *num = GetCompletionDataSize(-1);
  return 0;
}
TINE_EXPORT int GetDeviceServers(NAME16 *dslist,int *num)
{ /* deprecated */
  int fmt = CF_NAME16;
  int cc = GetTagList(NULL,"TAGS",&fmt,dslist,num);
  return cc;
}
TINE_EXPORT int GetDeviceServersEx(char *context,NAME16 *dslist,int *num)
{ /* deprecated */
  int fmt = CF_NAME16;
  int cc = GetTagList(context,"TAGS",&fmt,dslist,num);
  return cc;
}
TINE_EXPORT int GetDeviceServersFromFile(char *context,NAME16 *dslist,int *num)
{
  FILE  *fp=NULL;
  int i=0,n,cc=0,done=0,nam_col,con_col;
  char fname[256],str[256],hdr[256],ctx[CONTEXT_NAME_SIZE+4];
  
  if (num == NULL || dslist == NULL) ccerr(argument_list_error);
  sprintf(fname,"%s%s",tineHomePath,EQPDB_FILE);
  if ((fp=fopen(fname,"r")) == NULL) ccerr(no_such_file);
  n = *num;
  for (done=0; fgets(str,255,fp) != NULL;)   /* scan file       */
  {
    strtrm(str);
    if (strlen(str) == 0) continue;          /* carriage return */
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      strncpy(hdr,str,255);
      if ((nam_col=findcol(hdr,"NAME")) < 0) ccerr(no_such_line);
      if ((con_col=findcol(hdr,"CONTEXT")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(ctx,colptr(con_col,str),16);
    if (context != NULL && strnicmp(ctx,context,CONTEXT_NAME_SIZE)) continue;
    strncpy(dslist[i].name,colptr(nam_col,str),16);
    if (++i > n) break;
  }
  *num = i;
  fclose(fp);
  if (cc) ccerr(cc);
err:
  return cc;
}
TINE_EXPORT int GetServersEx(char *context,NAME16 *slist,int *num)
{ /* deprecated */
  int fmt = CF_NAME16;
  int cc = GetTagList(context,"FECS",&fmt,slist,num);
  return cc;
}
TINE_EXPORT int GetServers(NAME16 *slist,int *num)
{ /* deprecated */
  int fmt = CF_NAME16;
  int cc = GetTagList(NULL,"FECS",&fmt,slist,num);
  return cc;
}
TINE_EXPORT int GetSystemContexts(NAME32 *clist,int *num)
{
  int fmt = CF_NAME32;
  int cc = GetTagList(NULL,"CONTEXTS",&fmt,clist,num);
  if (fmt != CF_NAME32 && cc == 0) augmentNameSize(CF_NAME16,(char *)clist,*num);
  return cc;
}
TINE_EXPORT int GetSystemDevices(char *srv,char *prp, NAME64 *dlist,int *num)
{
  DTYPE dout;
  int i,cc = 0, fmt = CF_NAME64;
  short n;
  char tmpprp[64];
  ExportListStruct *el = getExportListItem(srv);
  NAME64 *lcldevlst;

  if ((n = *num) <= 0) return buffer_too_small;
  if (prp != NULL && strlen(prp) > 0 && strlen(prp) < PROPERTY_NAME_SIZE-4)
  {
    sprintf(tmpprp,"%s.NAM",prp);
  }
  else
  {
    strcpy(tmpprp,"DEVICES");
  }
  if (el != NULL) /* local eqm name given */
  {
    lcldevlst = GetExportedDeviceList(srv);
    for (i=0; i<el->EqmNumDevices && i<n; i++) 
      strncpy(dlist[i].name,lcldevlst[i].name,64);
    *num = el->EqmNumDevices;
  }
  else
  {
retryFormat:
    dout.dFormat = fmt;
    dout.dArrayLength = n;
    dout.data.vptr = dlist;
    if ((cc=ExecLinkEx(srv,tmpprp,&dout,NULL,CA_READ|CA_RETRY,800)) != 0 && !(cc & CE_SENDDATA)) 
    {
      if (cc == illegal_format && fmt == CF_NAME64)
      {
        fmt = CF_NAME32;
        goto retryFormat;
      }
      return cc;
    }
    if (fmt != CF_NAME64 && cc == 0) augmentNameSize(CF_NAME32,(char *)dlist,*num);
    *num = GetCompletionDataSize(-1);
  }
  return 0;
}
TINE_EXPORT int GetSystemFecs(char *context,NAME16 *slist,int *num)
{
  int fmt = CF_NAME16;
  int cc = GetTagList(context,"FECS",&fmt,slist,num);
  return cc;
}
TINE_EXPORT int GetSystemProperties(char *srv,NAME64 *plist,int *num)
{
  DTYPE dout;
  int cc = 0, fmt = CF_NAME64;
  short n;

  if ((n = *num) <= 0) return buffer_too_small;
retryFormat:
  dout.dFormat = fmt;
  dout.dArrayLength = n;
  dout.data.cptr = (char *)plist;
  if ((cc=ExecLinkEx(srv,"PROPERTIES",&dout,NULL,CA_READ|CA_RETRY,800)) != 0 && !(cc & CE_SENDDATA)) 
  {
    if (cc == illegal_format && fmt == CF_NAME64)
    {
      fmt = CF_NAME32;
      goto retryFormat;
    }
    return cc;
  }
  if (fmt != CF_NAME64 && cc == 0) augmentNameSize(CF_NAME32,(char *)plist,*num);

  *num = GetCompletionDataSize(-1);
  return 0;
}
TINE_EXPORT int GetSystemServers(char *context,NAME32 *dslist,int *num)
{
  int fmt = CF_NAME32;
  int cc = GetTagList(context,"TAGS",&fmt,dslist,num);
  if (fmt != CF_NAME32 && cc == 0) augmentNameSize(CF_NAME16,(char *)dslist,*num);
  return cc;
}
TINE_EXPORT int GetTargetPropertyInformation(char *srv,char *prp,int *fmt,int *siz,char *dsc)
{
  DTYPE dout,din;
  int cc = 0;
  PropertyQueryStruct srvProps;
  NAME32 pname;

  if (prp != NULL) strncpy(pname.name,prp,32);
  din.dFormat = CF_NAME32;
  din.dArrayLength = 1;
  din.data.vptr = pname.name;
  dout.dFormat = CF_STRUCT;
  dout.dArrayLength = PROPERTYQUERYSTRUCT_SIZE;
  dout.data.vptr = &srvProps;
  if ((cc=ExecLinkEx(srv,"PROPS",&dout,&din,CA_READ|CA_RETRY,800)) != 0) return cc;
  if (fmt != NULL) *fmt = srvProps.prpFormat;
  if (siz != NULL) *siz = ISWAP(srvProps.prpSize);
  if (dsc != NULL) strncpy(dsc,srvProps.prpDescription,32);
  return 0;
}
TINE_EXPORT int RestorePropertyValues(char *name,void *values,short format,int size)
{
  int i = 0, cc = 0, done = 0, col = -1, len = fmtsizeof(format);
  FILE *fp = NULL;
  char scratch[128], str[256], fn[128];

  if (name == NULL || values == NULL || size <= 0 || len <= 0) ccerr(argument_list_error);
  sprintf(fn,"%s-setting.csv",name);
  if ((fp=fopen(fn,"r")) == NULL) ccerr(no_such_file);
  memset(values,0,size*len);
  while (fgets(str,255,fp) != NULL)
  {
    if (i > size) break;
    strtrm(str);
    if (strlen(str) == 0) continue;               /* carraige return */
    if (strchr("%;#\n",str[0]) != NULL) continue; /* comment line */
    if (!done)
    {
      col = findcol(str,"VALUES");
      done = TRUE;
      continue;
    }
    if (col >= 0) getColumnStr(col,str,scratch,64); else strcpy(str,"0");
    scratch[64] = 0;
    switch (format)
    {
      case CF_BYTE:
        ((BYTE *)values)[i] = atoi(scratch);
        break;
      case CF_INT16:
        ((short *)values)[i] = atoi(scratch);
        break;
      case CF_INT32:
        ((UINT32 *)values)[i] = atoi(scratch);
        break;
      case CF_FLOAT:
        ((float *)values)[i] = (float)strtod(scratch,NULL);
        break;
      case CF_DOUBLE:
        ((double *)values)[i] = strtod(scratch,NULL);
        break;
      case CF_NAME8:
      case CF_NAME16:
      case CF_NAME32:
      case CF_NAME48:
      case CF_NAME64:
        strncpy(&((char *)values)[i*len],scratch,len);
        break;
      default: 
        ccerr(illegal_format);
    }
    i++;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
TINE_EXPORT int SavePropertyValues(char *name,void *values,short format,int size)
{
  int i, cc = 0, len = 0;
  FILE *fp = NULL;
  char s[128], fn[128];

  if (name == NULL || values == NULL || size <= 0) ccerr(argument_list_error);
  sprintf(fn,"%s-setting.csv",name);
  if ((fp=fopen(fn,"w")) == NULL) ccerr(file_error);
  fprintf(fp,"VALUES\n");
  /* just take simple types in Release 3.XX */
  switch (format)
  {
    case CF_BYTE:
      for (i=0; i<size; i++) fprintf(fp,"%d\n",((BYTE *)values)[i]);
      break;
    case CF_INT16:
      for (i=0; i<size; i++) fprintf(fp,"%d\n",((short *)values)[i]);
      break;
    case CF_INT32:
      for (i=0; i<size; i++) fprintf(fp,"%d\n",((UINT32 *)values)[i]);
      break;
    case CF_FLOAT:
      for (i=0; i<size; i++) fprintf(fp,"%g\n",((float *)values)[i]);
      break;
    case CF_DOUBLE:
      for (i=0; i<size; i++) fprintf(fp,"%g\n",((double *)values)[i]);
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      len = fmtsizeof(format);
      for (i=0; i<size; i++) 
      {
        strncpy(s,&((char *)values)[i*len],len);  s[len] = 0;
        fprintf(fp,"%s\n",s);
      }
      break;
    default: 
      ccerr(illegal_format);
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
#endif /* not SMALL_TINE_LIB */

TINE_EXPORT int GetDeviceProperties(char *srv,NAME32 *props,int *num)
{
  DTYPE dout;
  int cc = 0;
  short n;

  if ((n = *num) <= 0) return buffer_too_small;
  dout.dFormat = CF_NAME32;
  dout.dArrayLength = n;
  dout.data.cptr = (char *)props;
  if ((cc=ExecLinkEx(srv,"PROPS",&dout,NULL,CA_READ|CA_RETRY,800)) != 0 && !(cc & CE_SENDDATA)) return cc;
  *num = GetCompletionDataSize(-1);
  return 0;
}
TINE_EXPORT int GetDeviceNamesEx(char *srv,char *prp, NAME16 *devs,int *num)
{
  DTYPE dout;
  int i,cc = 0;
  short n;
  char tmpprp[64];
  ExportListStruct *el = getExportListItem(srv);
  NAME64 *lcldevlst;

  if ((n = *num) <= 0) return buffer_too_small;
  if (prp != NULL && strlen(prp) < 28)
  {
    sprintf(tmpprp,"%s.NAM",prp);
  }
  else
  {
    strcpy(tmpprp,"DEVICES");
  }
  if (el != NULL) /* local eqm name given */
  {
    lcldevlst = GetExportedDeviceList(srv);
    for (i=0; i<el->EqmNumDevices && i<n; i++) 
      strncpy(devs[i].name,lcldevlst[i].name,16);
    *num = el->EqmNumDevices;
  }
  else
  {
    dout.dFormat = CF_NAME16;
    dout.dArrayLength = n;
    dout.data.vptr = devs;
    if ((cc=ExecLinkEx(srv,tmpprp,&dout,NULL,CA_READ|CA_RETRY,800)) != 0 && !(cc & CE_SENDDATA)) return cc;
    *num = GetCompletionDataSize(-1);
  }
  return 0;
}
TINE_EXPORT int GetDeviceNames(char *srv,NAME16 *devs,int *num)
{
  return GetDeviceNamesEx(srv,NULL,devs,num);
}
TINE_EXPORT int GetValuesAsString(DTYPE *d,char *str,UINT32 *dsiz)
{
  int i,n,len,siz,cc=0;
  int valueOnly = FALSE;
  time_t t;

  if (d == NULL) return argument_list_error;
  if (str == NULL || dsiz == NULL) return un_allocated;
  if ((siz = (int)(*dsiz)) < 16) return dimension_error;
  if (siz <= 64) valueOnly = TRUE;

  str[0] = 0; len = 0;
  n = (int)d->dArrayLength;
  switch (d->dFormat)
  {
    case CF_TEXT:
      len = MIN(siz,n);
      strncpy(str,d->data.cptr,len);
      if (n < siz) str[n] = 0;
      ccerr(0);
    case CF_BYTE:   
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%d",d->data.bptr[i]); goto err; }
        sprintf(&str[len],"%d\n",d->data.bptr[i]);
        if ((len=(int)strlen(str)) > siz + 16) break;
      }
      ccerr(0);
    case CF_INT16:  
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%d",d->data.sptr[i]); goto err; }
        sprintf(&str[len],"%d\n",d->data.sptr[i]);
        if ((len=(int)strlen(str)) > siz + 16) break;
      }
      ccerr(0);
    case CF_INT32:
      for (i=0; i<n; i++) 
      {
#       ifdef OS_16BIT
        if (valueOnly) { sprintf(str,"%ld",d->data.lptr[i]); goto err; }
        sprintf(&str[len],"%ld\n",d->data.lptr[i]);
#       else
        if (valueOnly) { sprintf(str,"%d",d->data.lptr[i]); goto err; }
        sprintf(&str[len],"%d\n",d->data.lptr[i]);
#       endif
        if ((len=(int)strlen(str)) > siz + 16) break;
      }
      ccerr(0);
#   ifdef SINT64
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%ld",d->data.llptr[i]); goto err; }
        sprintf(&str[len],"%ld\n",d->data.llptr[i]);
        if ((len=(int)strlen(str)) > siz + 16) break;
      }
      ccerr(0);
#   endif
    case CF_FLOAT:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%g",d->data.fptr[i]); goto err; }
        sprintf(&str[len],"%g\n",d->data.fptr[i]);
        if ((len=(int)strlen(str)) > siz + 16) break;
      }
      ccerr(0);
    case CF_DOUBLE:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%g",d->data.dptr[i]); goto err; }
        sprintf(&str[len],"%g\n",d->data.dptr[i]);
        if ((len=(int)strlen(str)) > siz + 16) break;
      }
      ccerr(0);
    case CF_NAME8:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME8 *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s\n",((NAME8 *)d->data.vptr)[i].name);
        if ((len=(int)strlen(str)) > siz + 16) break;
      }
      ccerr(0);
    case CF_NAME16:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME16 *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s\n",((NAME16 *)d->data.vptr)[i].name);
        if ((len=(int)strlen(str)) > siz + 32) break;
      }
      ccerr(0);
    case CF_NAME32:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME32 *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s\n",((NAME32 *)d->data.vptr)[i].name);
        if ((len=(int)strlen(str)) > siz + 48) break;
      }
      ccerr(0);
    case CF_NAME48:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME48 *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s\n",((NAME48 *)d->data.vptr)[i].name);
        if ((len=(int)strlen(str)) > siz + 64) break;
      }
      ccerr(0);
    case CF_NAME64:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME64 *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s\n",((NAME64 *)d->data.vptr)[i].name);
        if ((len=(int)strlen(str)) > siz + 80) break;
      }
      ccerr(0);
#if !defined(SMALL_TINE_LIB)
    case CF_NAME16I:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME16I *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s %d\n",((NAME16I *)d->data.vptr)[i].name,((NAME16I *)d->data.vptr)[i].ival);
        if ((len=(int)strlen(str)) > siz + 64) break;
      }
      ccerr(0);
    case CF_NAME32I:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME32I *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s %d\n",((NAME32I *)d->data.vptr)[i].name,((NAME32I *)d->data.vptr)[i].ival);
        if ((len=(int)strlen(str)) > siz + 64) break;
      }
      ccerr(0);
    case CF_NAME48I:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME48I *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s %d\n",((NAME48I *)d->data.vptr)[i].name,((NAME48I *)d->data.vptr)[i].ival);
        if ((len=(int)strlen(str)) > siz + 64) break;
      }
      ccerr(0);
    case CF_NAME64I:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) { sprintf(str,"%s",((NAME64I *)d->data.vptr)[i].name); goto err; }
        sprintf(&str[len],"%s %d\n",((NAME64I *)d->data.vptr)[i].name,((NAME64I *)d->data.vptr)[i].ival);
        if ((len=(int)strlen(str)) > siz + 80) break;
      }
      ccerr(0);
    case CF_FLTINT:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        sprintf(&str[len],"%g %d\n",((FLTINT *)d->data.vptr)[i].fval,((FLTINT *)d->data.vptr)[i].ival);
        if ((len=(int)strlen(str)) > siz + 32) break;
      }
      ccerr(0);
    case CF_LNGINT:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        sprintf(&str[len],"%d %d\n",((LNGINT *)d->data.vptr)[i].lval,((LNGINT *)d->data.vptr)[i].ival);
        if ((len=(int)strlen(str)) > siz + 32) break;
      }
      ccerr(0);
    case CF_DBLINT:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        sprintf(&str[len],"%g %d\n",((DBLINT *)d->data.vptr)[i].dval,((DBLINT *)d->data.vptr)[i].ival);
        if ((len=(int)strlen(str)) > siz + 32) break;
      }
      ccerr(0);
    case CF_INTFLTINT:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        sprintf(&str[len],"%d %g %d\n",((INTFLTINT *)d->data.vptr)[i].i1val,
          ((INTFLTINT *)d->data.vptr)[i].fval,((INTFLTINT *)d->data.vptr)[i].i2val);
        if ((len=(int)strlen(str)) > siz + 32) break;
      }
      ccerr(0);
    case CF_USTRING:
      for (i=0; i<n; i++) 
      {
        if (valueOnly) goto err;
        t = (time_t)((DUSTRING *)d->data.vptr)[i].tm;
        sprintf(&str[len],"%g %g %d %s %s",((DUSTRING *)d->data.vptr)[i].f1val,
          ((DUSTRING *)d->data.vptr)[i].f2val,((DUSTRING *)d->data.vptr)[i].ival,
          ((DUSTRING *)d->data.vptr)[i].str,ctime(&t));
        if ((len=(int)strlen(str)) > siz + 128) break;
      }
      ccerr(0);
#   endif /* not SMALL_TINE_LIB */
    default:
      return illegal_format;
  }
err:
  *dsiz = len;
  return cc;
}
TINE_EXPORT int GetValuesAsDBLDBL(DTYPE *d,DBLDBL *ddval,int num)
{
  int i,n;

  if (d == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_FLTFLT: 
      for (i=0; i<n; i++) 
      {
        ddval[i].d1val = (double)((FLTFLT *)d->data.vptr)[i].f1val; 
        ddval[i].d2val = (double)((FLTFLT *)d->data.vptr)[i].f2val; 
      }
      return 0;
    case CF_FI : 
      for (i=0; i<n; i++) 
      {
        ddval[i].d1val = (double)((FLTINT *)d->data.vptr)[i].fval; 
        ddval[i].d2val = (double)((FLTINT *)d->data.vptr)[i].ival; 
      }
      return 0;
    case CF_II : 
      for (i=0; i<n; i++) 
      {
        ddval[i].d1val = (double)((INTINT *)d->data.vptr)[i].i1val; 
        ddval[i].d2val = (double)((INTINT *)d->data.vptr)[i].i2val; 
      }
      return 0;
    case CF_DBLDBL : 
      if (ddval == (DBLDBL *)d->data.vptr) return 0;
		  memcpy(ddval,d->data.vptr,n*sizeof(DBLDBL));
      return 0;
    default: 
      return illegal_format;
  }
}
TINE_EXPORT int GetValuesAsNAME64DBL(DTYPE *d,NAME64DBL *ndval,int num)
{
  int i,n, flen, nlen;
  char *nptr;
  SINT32 *iptr;

  if (d == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_CHAR8I  : 
    case CF_CHAR16I : 
    case CF_CHAR32I :
    case CF_CHAR48I : 
    case CF_CHAR64I : 
      flen = fmtsizeof(d->dFormat);
      nlen = flen - sizeof(SINT32);
      nptr = d->data.cptr;
      for (i=0; i<n; i++) 
      {
        nptr += flen; 
        iptr = (SINT32 *)(nptr + nlen);
        memset(ndval[i].name,0,64);
        strncpy(ndval[i].name,nptr,nlen);
        ndval[i].dval = (double)(*iptr);
      }
      return 0;
    case CF_NAME64DBL:
      if (ndval == (NAME64DBL *)d->data.vptr) return 0;
		  memcpy(ndval,d->data.vptr,n*sizeof(NAME64DBL));
      return 0;
    default:
      return illegal_format;
  }
}
TINE_EXPORT int GetValuesAsNAME64DBLDBL(DTYPE *d,NAME64DBLDBL *nddval,int num)
{
  int i,n;

  if (d == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_CHAR16FI: 
      for (i=0; i<n; i++) 
      {
        memset(nddval[i].name,0,64);
        strncpy(nddval[i].name,((NAME16FI *)d->data.vptr)[i].name,16);
        nddval[i].d1val = (double)((NAME16FI *)d->data.vptr)[i].fval; 
        nddval[i].d2val = (double)((NAME16FI *)d->data.vptr)[i].ival; 
      }
      return 0;
    case CF_NAME16II : 
      for (i=0; i<n; i++) 
      {
        memset(nddval[i].name,0,64);
        strncpy(nddval[i].name,((NAME16II *)d->data.vptr)[i].name,16);
        nddval[i].d1val = (double)((NAME16II *)d->data.vptr)[i].i1val; 
        nddval[i].d2val = (double)((NAME16II *)d->data.vptr)[i].i2val; 
      }
      return 0;
    case CF_NAME16DBLDBL : 
      for (i=0; i<n; i++) 
      {
        memset(nddval[i].name,0,64);
        strncpy(nddval[i].name,((NAME16DBLDBL *)d->data.vptr)[i].name,16);
        memcpy(&nddval[i].d1val,&((NAME16DBLDBL *)d->data.vptr)[i].d1val,2*sizeof(double));
      }
      return 0;
    case CF_NAME32DBLDBL : 
      for (i=0; i<n; i++) 
      {
        memset(nddval[i].name,0,64);
        strncpy(nddval[i].name,((NAME32DBLDBL *)d->data.vptr)[i].name,32);
        memcpy(&nddval[i].d1val,&((NAME32DBLDBL *)d->data.vptr)[i].d1val,2*sizeof(double));
      }
      return 0;
    case CF_NAME64DBLDBL : 
      if (nddval == (NAME64DBLDBL *)d->data.vptr) return 0;
		  memcpy(nddval,d->data.vptr,n*sizeof(NAME64DBLDBL));
      return 0;
    default:
      return illegal_format;
  }
}
TINE_EXPORT int GetValuesAsShort(DTYPE *d,short *sval,int num)
{
  int i,n;

  if (d == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BITFIELD8:
    case CF_BYTE:   for (i=n-1; i>=0; i--) sval[i] = (short)d->data.bptr[i];
                    return 0;
    case CF_BITFIELD16:
    case CF_INT16:  if (sval == (short *)d->data.sptr) return 0;
		                memcpy(sval,d->data.sptr,n*sizeof(UINT16));
                    return 0;
    case CF_BITFIELD32:
    case CF_INT32:  for (i=0; i<n; i++) sval[i] = (short)(d->data.lptr)[i];
                    return 0;
    case CF_FLOAT:  for (i=0; i<n; i++) sval[i] = (short)d->data.fptr[i];
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) sval[i] = (short)d->data.dptr[i];
                    return 0;
#   ifdef SINT64
    case CF_INT64:  for (i=0; i<n; i++) sval[i] = (short)(d->data.llptr)[i];
                    return 0;
#   endif
    case CF_FLTINT: for (i=0; i<n; i++) sval[i] = (short)((FLTINT *)d->data.vptr)[i].fval;
                    return 0;
    case CF_INTINT: for (i=0; i<n; i++) sval[i] = (short)((INTINT *)d->data.vptr)[i].i1val;
                    return 0;
    case CF_DBLDBL: for (i=0; i<n; i++) sval[i] = (short)((DBLDBL *)d->data.vptr)[i].d1val;
                    return 0;
    case CF_FLTFLT: for (i=0; i<n; i++) sval[i] = (short)((FLTFLT *)d->data.vptr)[i].f1val;
                    return 0;
    case CF_FLTFLTINT: for (i=0; i<n; i++) sval[i] = (short)((FLTFLTINT *)d->data.vptr)[i].f1val;
                    return 0;
    case CF_INTINTINT: for (i=0; i<n; i++) sval[i] = (short)((INTINTINT *)d->data.vptr)[i].i1val;
                    return 0;
    case CF_INTINTINTINT: for (i=0; i<n; i++) sval[i] = (short)((DADDRESS *)d->data.vptr)[i].i1val;
                    return 0;
    case CF_NAME16FI:for (i=0; i<n; i++) sval[i] = (short)((NAME16FI *)d->data.vptr)[i].fval;
                    return 0;
    case CF_NAME16II:for (i=0; i<n; i++) sval[i] = (short)((NAME16II *)d->data.vptr)[i].i1val;
                    return 0;
    case CF_INTFLTINT: for (i=0; i<n; i++) sval[i] = (short)((INTFLTINT *)d->data.vptr)[i].fval;
                    return 0;
    case CF_FLTINTINT: for (i=0; i<n; i++) sval[i] = (short)((FLTINTINT *)d->data.vptr)[i].fval;
                    return 0;
    case CF_NAME16DBLDBL:for (i=0; i<n; i++) sval[i] = (short)((NAME16DBLDBL *)d->data.vptr)[i].d1val;
                    return 0;
    case CF_NAME32DBLDBL:for (i=0; i<n; i++) sval[i] = (short)((NAME32DBLDBL *)d->data.vptr)[i].d1val;
                    return 0;
    case CF_NAME64DBLDBL:for (i=0; i<n; i++) sval[i] = (short)((NAME64DBLDBL *)d->data.vptr)[i].d1val;
                    return 0;
    default:        return illegal_format;
  }
}
TINE_EXPORT int GetValuesAsFloat(DTYPE *d,float *fval,int num)
{
  int i,n;

  if (d == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=n-1; i>=0; i--) fval[i] = (float)d->data.bptr[i];
                    return 0;
    case CF_INT16:  for (i=n-1; i>=0; i--) fval[i] = (float)((short *)d->data.sptr)[i];
                    return 0;
    case CF_INT32:  for (i=0; i<n; i++) fval[i] = (float)((SINT32 *)d->data.lptr)[i];
                    return 0;
    case CF_FLOAT:  if (fval == d->data.fptr) return 0;
		                memcpy(fval,d->data.fptr,n*sizeof(float));
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) fval[i] = (float)d->data.dptr[i];
                    return 0;
#   ifdef SINT64
    case CF_INT64:  for (i=0; i<n; i++) fval[i] = (float)(d->data.llptr)[i];
                    return 0;
#   endif
    case CF_FLTINT: for (i=0; i<n; i++) fval[i] = (float)((FLTINT *)d->data.vptr)[i].fval;
                    return 0;
    case CF_INTINT: for (i=0; i<n; i++) fval[i] = (float)((INTINT *)d->data.vptr)[i].i1val;
                    return 0;
    case CF_DBLDBL: for (i=0; i<n; i++) fval[i] = (float)((DBLDBL *)d->data.vptr)[i].d1val;
                    return 0;
    case CF_FLTFLT: for (i=0; i<n; i++) fval[i] = (float)((FLTFLT *)d->data.vptr)[i].f1val;
                    return 0;
    case CF_FLTFLTINT: for (i=0; i<n; i++) fval[i] = (float)((FLTFLTINT *)d->data.vptr)[i].f1val;
                    return 0;
    case CF_INTINTINT: for (i=0; i<n; i++) fval[i] = (float)((INTINTINT *)d->data.vptr)[i].i1val;
                    return 0;
    case CF_INTINTINTINT: for (i=0; i<n; i++) fval[i] = (float)((DADDRESS *)d->data.vptr)[i].i1val;
                    return 0;
    case CF_NAME16FI:for (i=0; i<n; i++) fval[i] = (float)((NAME16FI *)d->data.vptr)[i].fval;
                    return 0;
    case CF_NAME16II:for (i=0; i<n; i++) fval[i] = (float)((NAME16II *)d->data.vptr)[i].i1val;
                    return 0;
    case CF_INTFLTINT: for (i=0; i<n; i++) fval[i] = (float)((INTFLTINT *)d->data.vptr)[i].fval;
                    return 0;
    case CF_FLTINTINT: for (i=0; i<n; i++) fval[i] = (float)((FLTINTINT *)d->data.vptr)[i].fval;
                    return 0;
    case CF_NAME16DBLDBL:for (i=0; i<n; i++) fval[i] = (float)((NAME16DBLDBL *)d->data.vptr)[i].d1val;
                    return 0;
    case CF_NAME32DBLDBL:for (i=0; i<n; i++) fval[i] = (float)((NAME32DBLDBL *)d->data.vptr)[i].d1val;
                    return 0;
    case CF_NAME64DBLDBL:for (i=0; i<n; i++) fval[i] = (float)((NAME64DBLDBL *)d->data.vptr)[i].d1val;
                    return 0;    
    default:        return illegal_format;
  }
}
TINE_EXPORT int GetValuesAsLong(DTYPE *d,SINT32 *lval,int num)
{
  int i,n;

  if (d == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BITFIELD8:
    case CF_BYTE:   for (i=n-1; i>=0; i--) lval[i] = (SINT32)d->data.bptr[i];
                    return 0;
    case CF_BITFIELD16:
    case CF_INT16:  for (i=n-1; i>=0; i--) lval[i] = (SINT32)d->data.usptr[i];
                    return 0;
    case CF_BITFIELD32:
    case CF_INT32:  if (lval == d->data.lptr) return 0;
		                memcpy(lval,d->data.lptr,n*sizeof(SINT32));
                    return 0;
    case CF_FLOAT:  for (i=0; i<n; i++) lval[i] = (SINT32)d->data.fptr[i];
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) lval[i] = (SINT32)d->data.dptr[i];
                    return 0;
#   ifdef SINT64
    case CF_INT64:  for (i=0; i<n; i++) lval[i] = (SINT32)(d->data.llptr)[i];
                    return 0;
#   endif
    case CF_FLTINT: for (i=0; i<n; i++) lval[i] = (SINT32)((FLTINT *)d->data.vptr)[i].fval;
                    return 0;
    case CF_INTINT: for (i=0; i<n; i++) lval[i] = (SINT32)((INTINT *)d->data.vptr)[i].i1val;
                    return 0;
    case CF_DBLDBL: for (i=0; i<n; i++) lval[i] = (SINT32)((DBLDBL *)d->data.vptr)[i].d1val;
                    return 0;
    case CF_FLTFLT: for (i=0; i<n; i++) lval[i] = (SINT32)((FLTFLT *)d->data.vptr)[i].f1val;
                    return 0;
    case CF_FLTFLTINT: for (i=0; i<n; i++) lval[i] = (SINT32)((FLTFLTINT *)d->data.vptr)[i].f1val;
                    return 0;
    case CF_INTINTINT: for (i=0; i<n; i++) lval[i] = (SINT32)((INTINTINT *)d->data.vptr)[i].i1val;
                    return 0;
    case CF_INTINTINTINT: for (i=0; i<n; i++) lval[i] = (SINT32)((DADDRESS *)d->data.vptr)[i].i1val;
                    return 0;
    case CF_NAME16FI:for (i=0; i<n; i++) lval[i] = (SINT32)((NAME16FI *)d->data.vptr)[i].fval;
                    return 0;
    case CF_NAME16II:for (i=0; i<n; i++) lval[i] = (SINT32)((NAME16II *)d->data.vptr)[i].i1val;
                    return 0;
    case CF_INTFLTINT: for (i=0; i<n; i++) lval[i] = (SINT32)((INTFLTINT *)d->data.vptr)[i].fval;
                    return 0;
    case CF_FLTINTINT: for (i=0; i<n; i++) lval[i] = (SINT32)((FLTINTINT *)d->data.vptr)[i].fval;
                    return 0;
    case CF_NAME16DBLDBL:for (i=0; i<n; i++) lval[i] = (SINT32)((NAME16DBLDBL *)d->data.vptr)[i].d1val;
                    return 0;
    case CF_NAME32DBLDBL:for (i=0; i<n; i++) lval[i] = (SINT32)((NAME32DBLDBL *)d->data.vptr)[i].d1val;
                    return 0;
    case CF_NAME64DBLDBL:for (i=0; i<n; i++) lval[i] = (SINT32)((NAME64DBLDBL *)d->data.vptr)[i].d1val;
                    return 0;    
    default:        return illegal_format;
  }
}
TINE_EXPORT int GetValuesAsDouble(DTYPE *d,double *dval,int num)
{
  int i,n;

  if (d == NULL) return argument_list_error;
  if (d->dArrayLength < 1 || num < 1) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=n-1; i>=0; i--) dval[i] = (double)d->data.bptr[i];
                    return 0;
    case CF_INT16:  for (i=n-1; i>=0; i--) dval[i] = (double)((short *)d->data.sptr)[i];
                    return 0;
    case CF_INT32:  for (i=n-1; i>=0; i--) dval[i] = (double)((SINT32 *)d->data.lptr)[i];
                    return 0;
    case CF_FLOAT:  for (i=n-1; i>=0; i--) dval[i] = (double)d->data.fptr[i];
                    return 0;
    case CF_DOUBLE: if (dval == d->data.dptr) return 0;
		                memcpy(dval,d->data.dptr,n*sizeof(double));
                    return 0;
#   ifdef SINT64
    case CF_INT64:  for (i=0; i<n; i++) dval[i] = (double)(d->data.llptr)[i];
                    return 0;
#   endif
    case CF_FLTINT: for (i=0; i<n; i++) dval[i] = (double)((FLTINT *)d->data.vptr)[i].fval;
                    return 0;
    case CF_INTINT: for (i=0; i<n; i++) dval[i] = (double)((INTINT *)d->data.vptr)[i].i1val;
                    return 0;
    case CF_DBLDBL: for (i=0; i<n; i++) dval[i] = (double)((DBLDBL *)d->data.vptr)[i].d1val;
                    return 0;
    case CF_FLTFLT: for (i=0; i<n; i++) dval[i] = (double)((FLTFLT *)d->data.vptr)[i].f1val;
                    return 0;
    case CF_FLTFLTINT: for (i=0; i<n; i++) dval[i] = (double)((FLTFLTINT *)d->data.vptr)[i].f1val;
                    return 0;
    case CF_INTINTINT: for (i=0; i<n; i++) dval[i] = (double)((INTINTINT *)d->data.vptr)[i].i1val;
                    return 0;
    case CF_INTINTINTINT: for (i=0; i<n; i++) dval[i] = (double)((DADDRESS *)d->data.vptr)[i].i1val;
                    return 0;
    case CF_NAME16FI:for (i=0; i<n; i++) dval[i] = (double)((NAME16FI *)d->data.vptr)[i].fval;
                    return 0;
    case CF_NAME16II:for (i=0; i<n; i++) dval[i] = (double)((NAME16II *)d->data.vptr)[i].i1val;
                    return 0;
    case CF_INTFLTINT: for (i=0; i<n; i++) dval[i] = (double)((INTFLTINT *)d->data.vptr)[i].fval;
                    return 0;
    case CF_FLTINTINT: for (i=0; i<n; i++) dval[i] = (double)((FLTINTINT *)d->data.vptr)[i].fval;
                    return 0;
    case CF_NAME16DBLDBL:for (i=0; i<n; i++) dval[i] = (double)((NAME16DBLDBL *)d->data.vptr)[i].d1val;
                    return 0;
    case CF_NAME32DBLDBL:for (i=0; i<n; i++) dval[i] = (double)((NAME32DBLDBL *)d->data.vptr)[i].d1val;
                    return 0;
    case CF_NAME64DBLDBL:for (i=0; i<n; i++) dval[i] = (double)((NAME64DBLDBL *)d->data.vptr)[i].d1val;
                    return 0;    
    default:        return illegal_format;
  }
}
#ifdef WIN32
TINE_EXPORT int getValuesAsString(DTYPE *d,char *str,UINT32 *dsiz)
{
  return GetValuesAsString(d,str,dsiz);
}
TINE_EXPORT int getValuesAsShort(DTYPE *d,short *s,int n)
{
  return GetValuesAsShort(d,s,n);
}
TINE_EXPORT int getValuesAsFloat(DTYPE *d,float *s,int n)
{
  return GetValuesAsFloat(d,s,n);
}
TINE_EXPORT int getValuesAsDouble(DTYPE *d,double *s,int n) 
{
  return GetValuesAsDouble(d,s,n);
}
TINE_EXPORT int getValuesAsLong(DTYPE *d,SINT32 *s,int n) 
{
  return GetValuesAsLong(d,s,n);
}
#endif
TINE_EXPORT int PutValuesFromDoubleEx(DTYPE *d,double *dval,int num,int offset)
{
  int i,n,n1,n2;
  DTYPE dl;

  if (d == NULL) return argument_list_error;
  if ((d->dFormat != CF_TEXT) && (d->dArrayLength < 1 || num < 1)) return dimension_error;
  if (offset < 0 || offset >= num) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=0; i<n; i++) d->data.bptr[i] = (BYTE)dval[(i+offset)%num];
                    return 0;
    case CF_INT16:  for (i=0; i<n; i++) d->data.sptr[i] = (short)dval[(i+offset)%num];
                    return 0;
    case CF_INT32:  for (i=0; i<n; i++) d->data.lptr[i] = (SINT32)dval[(i+offset)%num];
                    return 0;
#   ifdef SINT64
    case CF_INT64:  for (i=0; i<n; i++) d->data.llptr[i] = (SINT64)dval[(i+offset)%num];
                    return 0;
#   endif
    case CF_FLOAT:  for (i=0; i<n; i++) d->data.fptr[i] = (float)dval[(i+offset)%num];
                    return 0;
    case CF_DOUBLE: n1 = MIN(num-offset,n); n2 = n - n1;
                    if (n1 > 0) memcpy(d->data.dptr,&dval[offset],n1*sizeof(double));
                    if (n2 > 0) memcpy(&d->data.dptr[n1],dval,n2*sizeof(double));
                    return 0;
    case CF_TEXT:   dl.dArrayLength = num; dl.dFormat = CF_DOUBLE; dl.data.dptr = dval;
                    return GetValuesAsString(&dl,d->data.cptr,&d->dArrayLength);
    default:        return illegal_format;
  }
}
TINE_EXPORT int PutValuesFromFloatEx(DTYPE *d,float *fval,int num,int offset)
{
  int i,n,n1,n2;
  DTYPE dl;

  if (d == NULL) return argument_list_error;
  if ((d->dFormat != CF_TEXT) && (d->dArrayLength < 1 || num < 1)) return dimension_error;
  if (offset < 0 || offset >= num) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=0; i<n; i++) d->data.bptr[i] = (BYTE)fval[(i+offset)%num];
                    return 0;
    case CF_INT16:  for (i=0; i<n; i++) d->data.sptr[i] = (short)fval[(i+offset)%num];
                    return 0;
    case CF_INT32:  for (i=0; i<n; i++) d->data.lptr[i] = (SINT32)fval[(i+offset)%num];
                    return 0;
#   ifdef SINT64
    case CF_INT64:  for (i=0; i<n; i++) d->data.llptr[i] = (SINT64)fval[(i+offset)%num];
                    return 0;
#   endif
    case CF_FLOAT:  n1 = MIN(num-offset,n); n2 = n - n1;
                    if (n1 > 0) memcpy(d->data.fptr,&fval[offset],n1*sizeof(float));
                    if (n2 > 0) memcpy(&d->data.fptr[n1],fval,n2*sizeof(float));
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) d->data.dptr[i] = (double)fval[(i+offset)%num];
                    return 0;
    case CF_TEXT:   dl.dArrayLength = num; dl.dFormat = CF_FLOAT; dl.data.fptr = fval;
                    return GetValuesAsString(&dl,d->data.cptr,&d->dArrayLength);
    default:        return illegal_format;
  }
}
TINE_EXPORT int PutValuesFromLongEx(DTYPE *d,SINT32 *lval,int num,int offset)
{
  int i,n,n1,n2;
  DTYPE dl;

  if (d == NULL) return argument_list_error;
  if ((d->dFormat != CF_TEXT) && (d->dArrayLength < 1 || num < 1)) return dimension_error;
  if (offset < 0 || offset >= num) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=0; i<n; i++) d->data.bptr[i] = (BYTE)lval[(i+offset)%num];
                    return 0;
    case CF_INT16:  for (i=0; i<n; i++) d->data.sptr[i] = (short)lval[(i+offset)%num];
                    return 0;
    case CF_INT32:   n1 = MIN(num-offset,n); n2 = n - n1;
                    if (n1 > 0) memcpy(d->data.lptr,&lval[offset],n1*sizeof(SINT32));
                    if (n2 > 0) memcpy(&d->data.lptr[n1],lval,n2*sizeof(SINT32));
                    return 0;
#   ifdef SINT64
    case CF_INT64:  for (i=0; i<n; i++) d->data.llptr[i] = (SINT64)lval[(i+offset)%num];
                    return 0;
#   endif
    case CF_FLOAT:  for (i=0; i<n; i++) d->data.fptr[i] = (float)lval[(i+offset)%num];
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) d->data.dptr[i] = (double)lval[(i+offset)%num];
                    return 0;
    case CF_TEXT:   dl.dArrayLength = num; dl.dFormat = CF_INT32; dl.data.lptr = lval;
                    return GetValuesAsString(&dl,d->data.cptr,&d->dArrayLength);
    default:        return illegal_format;
  }
}
TINE_EXPORT int PutValuesFromUnsignedLongEx(DTYPE *d,UINT32 *lval,int num,int offset)
{
  int i,n,n1,n2;
  DTYPE dl;

  if (d == NULL) return argument_list_error;
  if ((d->dFormat != CF_TEXT) && (d->dArrayLength < 1 || num < 1)) return dimension_error;
  if (offset < 0 || offset >= num) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=0; i<n; i++) d->data.bptr[i] = (BYTE)lval[(i+offset)%num];
                    return 0;
    case CF_INT16:  for (i=0; i<n; i++) d->data.sptr[i] = (short)lval[(i+offset)%num];
                    return 0;
    case CF_INT32:   n1 = MIN(num-offset,n); n2 = n - n1;
                    if (n1 > 0) memcpy(d->data.lptr,&lval[offset],n1*sizeof(UINT32));
                    if (n2 > 0) memcpy(&d->data.lptr[n1],lval,n2*sizeof(UINT32));
                    return 0;
#   ifdef SINT64
    case CF_INT64:  for (i=0; i<n; i++) d->data.llptr[i] = (SINT64)lval[(i+offset)%num];
                    return 0;
#   endif
    case CF_FLOAT:  for (i=0; i<n; i++) d->data.fptr[i] = (float)lval[(i+offset)%num];
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) d->data.dptr[i] = (double)lval[(i+offset)%num];
                    return 0;
    case CF_TEXT:   dl.dArrayLength = num; dl.dFormat = CF_INT32; dl.data.ulptr = lval;
                    return GetValuesAsString(&dl,d->data.cptr,&d->dArrayLength);
    default:        return illegal_format;
  }
}
TINE_EXPORT int PutValuesFromShortEx(DTYPE *d,short *sval,int num,int offset)
{
  int i,n,n1,n2;
  DTYPE dl;

  if (d ==NULL) return argument_list_error;
  if ((d->dFormat != CF_TEXT) && (d->dArrayLength < 1 || num < 1)) return dimension_error;
  if (offset < 0 || offset >= num) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=0; i<n; i++) d->data.bptr[i] = (BYTE)sval[(i+offset)%num];
                    return 0;
    case CF_INT16:  n1 = MIN(num-offset,n); n2 = n - n1;
                    if (n1 > 0) memcpy(d->data.sptr,&sval[offset],n1*sizeof(UINT16));
                    if (n2 > 0) memcpy(&d->data.sptr[n1],sval,n2*sizeof(UINT16));
                    return 0;
    case CF_INT32:   for (i=0; i<n; i++) d->data.lptr[i] = (SINT32)sval[(i+offset)%num];
                    return 0;
#   ifdef SINT64
    case CF_INT64:  for (i=0; i<n; i++) d->data.llptr[i] = (SINT64)sval[(i+offset)%num];
                    return 0;
#   endif
    case CF_FLOAT:  for (i=0; i<n; i++) d->data.fptr[i] = (float)sval[(i+offset)%num];
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) d->data.dptr[i] = (double)sval[(i+offset)%num];
                    return 0;
    case CF_TEXT:   dl.dArrayLength = num; dl.dFormat = CF_INT16; dl.data.usptr = (UINT16 *)sval;
                    return GetValuesAsString(&dl,d->data.cptr,&d->dArrayLength);
    default:        return illegal_format;
  }
}
TINE_EXPORT int PutValuesFromUnsignedShortEx(DTYPE *d,UINT16 *sval,int num,int offset)
{
  int i,n,n1,n2;
  DTYPE dl;

  if (d ==NULL) return argument_list_error;
  if ((d->dFormat != CF_TEXT) && (d->dArrayLength < 1 || num < 1)) return dimension_error;
  if (offset < 0 || offset >= num) return dimension_error;
  n = MIN((int)d->dArrayLength,num);
  switch (d->dFormat)
  {
    case CF_BYTE:   for (i=0; i<n; i++) d->data.bptr[i] = (BYTE)sval[(i+offset)%num];
                    return 0;
    case CF_INT16:  n1 = MIN(num-offset,n); n2 = n - n1;
                    if (n1 > 0) memcpy(d->data.sptr,&sval[offset],n1*sizeof(UINT16));
                    if (n2 > 0) memcpy(&d->data.sptr[n1],sval,n2*sizeof(UINT16));
    case CF_INT32:   for (i=0; i<n; i++) d->data.lptr[i] = (SINT32)sval[(i+offset)%num];
                    return 0;
#   ifdef SINT64
    case CF_INT64:  for (i=0; i<n; i++) d->data.llptr[i] = (SINT64)sval[(i+offset)%num];
                    return 0;
#   endif
    case CF_FLOAT:  for (i=0; i<n; i++) d->data.fptr[i] = (float)sval[(i+offset)%num];
                    return 0;
    case CF_DOUBLE: for (i=0; i<n; i++) d->data.dptr[i] = (double)sval[(i+offset)%num];
                    return 0;
    case CF_TEXT:   dl.dArrayLength = num; dl.dFormat = CF_INT16; dl.data.usptr = sval;
                    return GetValuesAsString(&dl,d->data.cptr,&d->dArrayLength);
    default:        return illegal_format;
  }
}
TINE_EXPORT int PutValuesFromString(DTYPE *d,BYTE *str,int fmt,int num,int offset)
{
  int i,n,len;
  int stlen, slen, snum = num;
  int dtlen, dlen, dnum;
  char *dst, *src;

  if (d == NULL || str == NULL) return argument_list_error;
  if ((d->dFormat != CF_TEXT) && (d->dArrayLength < 1 || num < 1)) return dimension_error;
  dst = d->data.cptr;
  src = (char *)str;
  switch (fmt)
  { /* source fmt */
    case CF_TEXT:
      slen = num;
      snum = 1;
      stlen = 0;
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      slen = fmtsizeof(fmt);
      stlen = slen;
      break;
    case CF_USTRING:
      slen = 80;
      stlen = fmtsizeof(fmt);
      break;
    case CF_UNAME:
      slen = 16;
      stlen = fmtsizeof(fmt);
      break;
    default:
      return illegal_format;      
  }
  if (offset < 0 || offset >= snum) return dimension_error;
  dnum = d->dArrayLength;
  switch (d->dFormat)
  { /* destination format */
    case CF_TEXT:
      dlen = d->dArrayLength;
      dnum = 1;
      dtlen = 0;
      break;
    case CF_NAME8:
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME48:
    case CF_NAME64:
      dlen = fmtsizeof(fmt);
      dtlen = dlen;
      break;
    case CF_USTRING:
      dlen = 80;
      dtlen = fmtsizeof(fmt);
      break;
    case CF_UNAME:
      dlen = 16;
      dtlen = fmtsizeof(fmt);
      break;
    default:
      return illegal_format;
  }
  n = MIN(dnum,(snum-offset));
  len = MIN(dtlen,stlen);
  for (i=0; i<n; i++)
  {
    memset(dst + i*dtlen,0,dtlen);
    strncpy(dst + i*dtlen, src + (i+offset)*stlen, len);
  }
  return 0;
}
#ifdef WIN32
TINE_EXPORT int putValuesFromShortEx(DTYPE *d,short *s,int n,int o)
{
  return PutValuesFromShortEx(d,s,n,o);
}
TINE_EXPORT int putValuesFromFloatEx(DTYPE *d,float *s,int n,int o)
{
  return PutValuesFromFloatEx(d,s,n,o);
}
TINE_EXPORT int putValuesFromDoubleEx(DTYPE *d,double *s,int n,int o)
{
  return PutValuesFromDoubleEx(d,s,n,o);
}
TINE_EXPORT int putValuesFromLongEx(DTYPE *d,SINT32 *s,int n,int o)
{
  return PutValuesFromLongEx(d,s,n,o);
}
TINE_EXPORT int putValuesFromUnsignedLongEx(DTYPE *d,UINT32 *s,int n,int o)
{
  return PutValuesFromUnsignedLongEx(d,s,n,o);
}
TINE_EXPORT int putValuesFromUnsignedShortEx(DTYPE *d,UINT16 *s,int n,int o)
{
  return PutValuesFromUnsignedShortEx(d,s,n,o);
}
TINE_EXPORT int putValuesFromString(DTYPE *d,BYTE *s,int f,int n,int o)
{
  return PutValuesFromString(d,s,f,n,o);
}
#endif
TINE_EXPORT int sendEventTrigger(char *dev,char *cmt,short triggerLevel)
{
  DTYPE din;
  char triggerDevName[64];
  char cmtbuf[256];
  int cc;

  if (dev == NULL || strlen(dev) == 0) return illegal_equipment_name;
  if (strchr(dev,'/') == 0)
  {
    if (strlen(dev) > DEVICE_NAME_SIZE) return illegal_equipment_name;
    sprintf(triggerDevName,"/%s/STORAGE/%s",gDeviceContext,dev);
  }
  else strncpy(triggerDevName,dev,64);

  din.dFormat = CF_INT16;
  din.dArrayLength = 1;
  din.data.usptr = (UINT16 *)&triggerLevel;
  if ((cc=ExecLinkEx(triggerDevName,"TRIGGER",NULL,&din,CA_WRITE,200)) != 0) return cc;

  /* send comment buffer */
  strncpy(cmtbuf,cmt,255);
  din.dFormat = CF_TEXT;
  din.data.cptr = cmtbuf;
  sprintf(&cmtbuf[strlen(cmtbuf)],"@%s",getDataTimeString((double)time(NULL),FALSE));
  din.dArrayLength = (UINT32)strlen(cmtbuf);
  if ((cc=ExecLinkEx(triggerDevName,"ANNOTATE",NULL,&din,CA_WRITE,200)) != 0) return cc;
  return 0;
}
